/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#ifndef __FreeBSD__
#include <alloca.h>
#endif
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <rpcsvc/nis_db.h>
#include <sys/time.h>
#include "nis_db_private.h"

#ifndef lint
__RCSID("@(#) $FreeBSD$");
#endif

static uint32_t *merge_rel(data, knum)
	nis_datum		*data;
	uint32_t		knum;
{
	uint32_t		i;
	uint32_t		size;
	uint32_t		*kbuf = NULL;
	uint32_t		*datap = (uint32_t *)data->data;

	size = data->size + ULONGSZ;

	if ((kbuf = calloc(1, size)) == NULL)
		return(NULL);

	/*
	 * XXX: What we're doing here assumes that our addresses
	 * are all aligned on word boundaries. This code will fail
	 * on a SPARC if datap is an odd number.
	 *
	 * Go to a little extra trouble keep the tags sorted.
	 */
	for (i = 0; i < (data->size / ULONGSZ) && datap[i] < knum; i++) {
		kbuf[i] = datap[i];
	}

	kbuf[i] = knum;

	for (; i < (data->size / ULONGSZ); i++) {
		kbuf[i+1] = datap[i];
	}

	return(kbuf);
}

uint32_t gethash(dptr, i, numattrs, attrs, hashes)
	struct nis_dbent	*dptr;
	uint32_t		i;
	int			numattrs;
	nis_attr		*attrs;
	uint32_t		*hashes;
{
	int			j;
	uint32_t		relnum = 0;

	for (j = 0; j < numattrs; j++) {
		if (hashes[j] != dptr->index_hashes[i])
			continue;
		if (dptr->table->ta_cols.ta_cols_val[i].tc_flags & TA_CASE)
			relnum = hash(attrs[j].zattr_val.zattr_val_val,
			    attrs[j].zattr_val.zattr_val_len, HASH_UPPER);
		else
			relnum = hash(attrs[j].zattr_val.zattr_val_val,
			   attrs[j].zattr_val.zattr_val_len, HASH_NORM);
		break;
	}

	return(relnum);
}

static db_status addition_with_attrs(dptr, entry, numattrs, attrs)
	struct nis_dbent	*dptr;
	entry_obj		*entry;
	int			numattrs;
	nis_attr		*attrs;
{
	uint32_t		i;
	uint32_t		relnum;
	uint32_t		*kbuf = NULL;
	uint32_t		keynum;
	nis_datum		data = { NULL, 0 };
	uint32_t		*hashes;

	hashes = alloca(ULONGSZ * numattrs);

	for (i = 0; i < numattrs; i++) {
		hashes[i] = hash(attrs[i].zattr_ndx,
		    strlen(attrs[i].zattr_ndx), HASH_NORM);
	}

	for (i = 0; i < dptr->table->ta_cols.ta_cols_len; i++) {
		if (!(dptr->table->ta_cols.ta_cols_val[i].tc_flags &
		   TA_SEARCHABLE))
			continue;
		relnum = gethash(dptr, i, numattrs, attrs, hashes);
		if (dptr->status == STATUS_NORMAL ||
		    dptr->status == STATUS_REPLAYING_LOG)
			keynum = dptr->hdr_log.entry_just_added;
		if (dptr->status == STATUS_COMMITTING_LOG)
			keynum = dptr->hdr.entry_just_added;

		data.data = (char *)&keynum;
		data.size = ULONGSZ;

		if (rcm_get_index(dptr, i, relnum, &data) == DB_SUCCESS) {
			kbuf = merge_rel(&data, keynum);
			if (kbuf == NULL)
				return(DB_MEMORY_LIMIT);
			data.data = kbuf;
			data.size += ULONGSZ;
		}
		rcm_put_index(dptr, i, relnum, &data);
		if (kbuf) {
			free(kbuf);
			kbuf = NULL;
		}

	}

	return(DB_SUCCESS);
}

static db_status addition_without_attrs(dptr, entry)
	struct nis_dbent	*dptr;
	entry_obj		*entry;
{
	uint32_t		i;
	uint32_t		relnum;
	uint32_t		*kbuf = NULL;
	uint32_t		keynum;
	nis_datum		data = { NULL, 0 };

	for (i = 0; i < dptr->table->ta_cols.ta_cols_len; i++) {
		if (!(dptr->table->ta_cols.ta_cols_val[i].tc_flags &
		    TA_SEARCHABLE))
			continue;

		if (dptr->table->ta_cols.ta_cols_val[i].tc_flags & TA_CASE)
			relnum = hash(
			entry->en_cols.en_cols_val[i].ec_value.ec_value_val,
			entry->en_cols.en_cols_val[i].ec_value.ec_value_len,
			HASH_UPPER);
		else
			relnum = hash(
			entry->en_cols.en_cols_val[i].ec_value.ec_value_val,
			entry->en_cols.en_cols_val[i].ec_value.ec_value_len,
			HASH_NORM);

		if (dptr->status == STATUS_NORMAL ||
			dptr->status == STATUS_REPLAYING_LOG)
			keynum = dptr->hdr_log.entry_just_added;
		if (dptr->status == STATUS_COMMITTING_LOG)
			keynum = dptr->hdr.entry_just_added;

		data.data = (char *)&keynum;
		data.size = ULONGSZ;

		if (rcm_get_index(dptr, i, relnum, &data) == DB_SUCCESS) {
			kbuf = merge_rel(&data, keynum);
			if (kbuf == NULL)
				return(DB_MEMORY_LIMIT);
			data.data = kbuf;
			data.size += ULONGSZ;
		}
		rcm_put_index(dptr, i, relnum, &data);
		if (kbuf) {
			free(kbuf);
			kbuf = NULL;
		}

	}

	return(DB_SUCCESS);
}

db_status nis_db_process_addition(dptr, entry, numattrs, attrs)
	struct nis_dbent	*dptr;
	entry_obj		*entry;
	int			numattrs;
	nis_attr		*attrs;
{
	if (numattrs == 1 && !attrs[0].zattr_val.zattr_val_len)
		return(addition_without_attrs(dptr, entry));
	else
		return(addition_with_attrs(dptr, entry, numattrs, attrs));
}

db_result *db_add_entry(table_name, numattrs, attrs, entry)
	char			*table_name;
	int			numattrs;
	nis_attr		*attrs;
	entry_obj		*entry;
{
	struct timeval		t_start, t_end;
	db_result		*res, *lres, *rres;
	struct nis_dbent	*dptr;

	gettimeofday(&t_start, NULL);

	if ((res = calloc(1, sizeof(db_result))) == NULL)
		return(NULL);

	/*
	 * The Sun implementation requires that at least one
	 * attribute be specified even if we don't want to use
	 * any.
	 */

	if (!numattrs || attrs == NULL) {
		res->status = DB_BADQUERY;
		goto bail;
	}

	/*
	 * Pull the table handle out of the cache (will open
	 * the table if closed).
	 */

	if ((dptr = find_table(table_name)) == NULL) {
		res->status = DB_BADTABLE;
		goto bail;
	}

	if ((res->status = validate_attrs(dptr, numattrs, attrs)) ==
	    DB_BADQUERY)
		goto bail;

	/* extra sanity check */
#ifdef foo
	if (entry->en_cols.en_cols_len != dptr->table->ta_cols.ta_cols_len) {
		res->status = DB_BADQUERY;
		goto bail;
	}
#endif

	if (dptr->fp == NULL && dptr->fp_log == NULL) {
		if (unstandby_table(dptr) != DB_SUCCESS) {
			res->status = DB_INTERNAL_ERROR;
			goto bail;
		}
	}

	/*
	 * If the user specified an attributes, he wants to do
	 * a replacement. Find the entry to be replaced (can't
	 * replace more than one).
	 */

	if (res->status != DB_NOTFOUND) {
		lres = db_list_entries(table_name, numattrs, attrs);
		if (lres == NULL) {
			res->status = DB_MEMORY_LIMIT;
			goto bail;
		}
		if (lres->status != DB_SUCCESS) {
			if (lres->status != DB_NOTFOUND) {
				res->status = lres->status;
				db_free_result(lres);
				goto bail;
			}
			db_free_result(lres);
		} else {
			if (lres->objects.objects_len > 1) {
				res->status = DB_NOTUNIQUE;
				db_free_result(lres);
				goto bail;
			}
			db_free_result(lres);
			rres = db_remove_entry(table_name, numattrs, attrs);
			if (rres == NULL) {
				res->status = DB_MEMORY_LIMIT;
				goto bail;
			}
			if (rres->status != DB_SUCCESS) {
				res->status = DB_INTERNAL_ERROR;
				db_free_result(rres);
				goto bail;
			}
			db_free_result(rres);
		}
	}

	if (dptr->status == STATUS_NORMAL) {
		switch(xff_add_log_entry(dptr, entry, numattrs,
		    attrs, XLE_ADD)) {
		case DB_SUCCESS:
			res->status = DB_SUCCESS;
			break;
		case DB_INTERNAL_ERROR:
		default:
			if (errno == ENOSPC)
				res->status = DB_STORAGE_LIMIT;
			else
				res->status = DB_INTERNAL_ERROR;
			goto bail;
			break;
		}
	}

	res->status = nis_db_process_addition(dptr, entry, numattrs, attrs);

bail:
	gettimeofday(&t_end, NULL);
	res->ticks +=
		(t_end.tv_usec - t_start.tv_usec) +
		((t_end.tv_sec - t_start.tv_sec) * 1000000);

	return(res);
}
