/*
 * Copyright (c) 1996
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$FreeBSD$
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <errno.h>
#include <sys/stat.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nis_db.h>
#include "nis_db_struct.h"
#include "nis_db_private.h"

#ifndef lint
__RCSID("@(#) $FreeBSD$");
#endif

/*
 * This file contains the code that creates NIS+ tables using the
 * XDR File Format (XFF). XFF files are esentially nothing more than
 * XDR streams that have been saved to a file. At the very end of
 * an XFF file, there is a large chunk of data which represents
 * the relation cache that can be used to access any record in the
 * file based on supplied relation information. At the start of the
 * file is a header which can be used to determine where the
 * relation cache is stored, how large the file is, and where to
 * start looking for entries.
 */

/*
 * If a log exists, we 'replay' it into memory so that we can build
 * up a new relation cache. This should really be done using internal
 * routines; what we're doing now (calling the top-level db functions)
 * leads to a lot of interdependency between modules.
 */
db_status xff_replay_log(dptr, commit)
	struct nis_dbent	*dptr;
	int			commit;
{
	XDR			xdrs;
	xff_log_entry		log_entry;
	uint32_t		i, *w = NULL;

	if (dptr->hdr_log.type != XFF_LOG)
		return(DB_BADQUERY);

	if (dptr->fp_log == NULL)
		return(DB_INTERNAL_ERROR);

	fseek(dptr->fp_log, dptr->hdr_log.start_of_entries, SEEK_SET);
	xdrstdio_create(&xdrs, dptr->fp_log, XDR_DECODE);

	dptr->status = commit ? STATUS_COMMITTING_LOG : STATUS_REPLAYING_LOG;

	dptr->hdr_log.entry_just_added =
	    ftell(dptr->fp_log) + dptr->hdr.table_size;
	bzero((char *)&log_entry, sizeof(log_entry));

	while(xdr_xff_log_entry(&xdrs, &log_entry)) {
		if (log_entry.xle_status != XLE_PENDING) {
			xdr_free((xdrproc_t)xdr_xff_log_entry, (char *)&log_entry);
			bzero((char *)&log_entry, sizeof(log_entry));
			dptr->hdr_log.entry_just_added =
			    ftell(dptr->fp_log) + dptr->hdr.table_size;
			continue;
		}
		switch(log_entry.xle_action) {
		case (XLE_ADD):
			xff_add_entry(dptr, log_entry.xle_entry);
			nis_db_process_addition(dptr, log_entry.xle_entry,
			    log_entry.xle_attrs.xle_attrs_len,
			    log_entry.xle_attrs.xle_attrs_val);
			break;
		case (XLE_DELETE):
			if (nis_db_parse_rels(log_entry.xle_attrs.xle_attrs_len,
			    log_entry.xle_attrs.xle_attrs_val,
			    dptr, &w) != DB_SUCCESS)
				break;
			if (w == NULL)
				break;
			for (i = 0; i < w[0]; i++) {
				xff_delete_entry(dptr,
				    log_entry.xle_entry, w[i+1]);
				nis_db_unmerge_rel(dptr, log_entry.xle_entry, 
				    log_entry.xle_attrs.xle_attrs_len,
				    log_entry.xle_attrs.xle_attrs_val,
				    w[i + 1]);
			}
			free(w);
			w = NULL;
			dptr->deletions++;
			break;
		default:
			syslog(LOG_WARNING, "unknown log action: %d",
			    log_entry.xle_action);
			break;
		}
		xdr_free((xdrproc_t)xdr_xff_log_entry, (char *)&log_entry);
		bzero((char *)&log_entry, sizeof(log_entry));
		dptr->hdr_log.entry_just_added =
		    ftell(dptr->fp_log) + dptr->hdr.table_size;
	}

	xdr_destroy(&xdrs);
	dptr->status = STATUS_NORMAL;

	return(DB_SUCCESS);
}

/*
 * We need to do magic things for the 'catalog' table, thus we need
 * a seperate routine.
 */
db_status xff_replay_dict_log(dptr, commit)
	struct nis_dbent	*dptr;
	int			commit;
{
	XDR			xdrs;
	xff_log_entry		log_entry;
	table_desc		*tdesc;

	if (dptr->hdr_log.type != XFF_LOG)
		return(DB_BADQUERY);

	if (dptr->fp_log == NULL)
		return(DB_INTERNAL_ERROR);

	fseek(dptr->fp_log, dptr->hdr_log.start_of_entries, SEEK_SET);

	xdrstdio_create(&xdrs, dptr->fp_log, XDR_DECODE);

	dptr->status = commit ? STATUS_COMMITTING_LOG : STATUS_REPLAYING_LOG;

	dptr->hdr_log.entry_just_added =
	    ftell(dptr->fp_log) + dptr->hdr.table_size;
	bzero((char *)&log_entry, sizeof(log_entry));

	while(xdr_xff_log_entry(&xdrs, &log_entry)) {
		if (log_entry.xle_status != XLE_PENDING ||
		    log_entry.xle_status == XLE_OBSOLETE) {
			xdr_free((xdrproc_t)xdr_xff_log_entry, (char *)&log_entry);
			bzero((char *)&log_entry, sizeof(log_entry));
			dptr->hdr_log.entry_just_added =
			    ftell(dptr->fp_log) + dptr->hdr.table_size;
			continue;
		}

		if ((tdesc = decode_table_desc(
	log_entry.xle_entry->en_cols.en_cols_val[0].ec_value.ec_value_val,
	log_entry.xle_entry->en_cols.en_cols_val[0].ec_value.ec_value_len)) == NULL)
			return(DB_INTERNAL_ERROR);

		switch(log_entry.xle_action) {
		case (XLE_ADD):
			put_dictionary(tdesc->table_name, tdesc);
			break;
		case (XLE_DELETE):
			del_dictionary(tdesc->table_name);
			dptr->deletions++;
			break;
		default:
			syslog(LOG_WARNING, "unknown log action: %d",
			    log_entry.xle_action);
			break;
		}
		xdr_free((xdrproc_t)xdr_xff_log_entry, (char *)&log_entry);
		bzero((char *)&log_entry, sizeof(log_entry));
		xdr_free((xdrproc_t)xdr_table_desc, (char *)tdesc);
		free(tdesc);
		dptr->hdr_log.entry_just_added =
		    ftell(dptr->fp_log) + dptr->hdr.table_size;
	}

	xdr_destroy(&xdrs);
	dptr->status = STATUS_NORMAL;

	return(DB_SUCCESS);
}

db_status xff_commit_log(dptr)
	struct nis_dbent	*dptr;
{
	char			log_path[NIS_MAXNAMELEN];
	XDR			xdrs;

	if (dptr->hdr.log_present == 0 && dptr->fp_log == NULL)
		return (DB_SUCCESS);

	snprintf(log_path, sizeof(log_path),
	    "%s.log", dptr->tdesc->table_path);
	if (access(dptr->tdesc->table_path, R_OK|W_OK) == -1)
		if (errno == ENOENT)
			xff_create(dptr, dptr->tdesc->table_path, 0);
		else {
			syslog(LOG_ERR, "error creating file: %m");
			return(DB_INTERNAL_ERROR);
		}
	else
		xff_open(dptr, dptr->tdesc->table_path);

	rcm_free(dptr);
	rcm_load(dptr);
	xff_replay_log(dptr, 1);	/* replay for real this time */
	dptr->hdr.log_present = 0;
	if (dptr->deletions)
		xff_compact_file(dptr);	/* compact resulting file */
	else {
		fseek(dptr->fp, 0L, SEEK_SET);
		xdrstdio_create(&xdrs, dptr->fp, XDR_ENCODE);
		xdr_xff_header(&xdrs, &dptr->hdr);
		xdr_destroy(&xdrs);
		rcm_dump(dptr);
	}
	dptr->deletions = 0;
	fclose(dptr->fp_log);
	dptr->fp_log = NULL;
	bzero((char *)&dptr->hdr_log, sizeof(xff_header));
	unlink(log_path);

	return(DB_SUCCESS);
}

db_status xff_commit_dict_log(dptr)
	struct nis_dbent	*dptr;
{
	char			log_path[NIS_MAXNAMELEN];
	XDR			xdrs;

	if (dptr->hdr.log_present == 0 && dptr->fp_log == NULL)
		return (DB_SUCCESS);

	snprintf(log_path, sizeof(log_path),
	    "%s.log", dptr->tdesc->table_path);
	if (access(dptr->tdesc->table_path, R_OK|W_OK) == -1)
		if (errno == ENOENT)
			xff_create(dptr, dptr->tdesc->table_path, 0);
		else {
			syslog(LOG_ERR, "error creating file: %m");
			return(DB_INTERNAL_ERROR);
		}
	else
		xff_open(dptr, dptr->tdesc->table_path);

	rcm_free(dptr);
	rcm_load(dptr);
	xff_replay_dict_log(dptr, 1);
	dptr->hdr.log_present = 0;
	if (dptr->deletions)
		xff_compact_file(dptr);
	else {
		fseek(dptr->fp, 0L, SEEK_SET);
		xdrstdio_create(&xdrs, dptr->fp, XDR_ENCODE);
		xdr_xff_header(&xdrs, &dptr->hdr);
		xdr_destroy(&xdrs);
		rcm_dump(dptr);
	}
	dptr->deletions = 0;
	fclose(dptr->fp_log);
	dptr->fp_log = NULL;
	bzero((char *)&dptr->hdr_log, sizeof(xff_header));
	unlink(log_path);

	return(DB_SUCCESS);
}

db_status xff_add_log_entry(dptr, entry, numattrs, attrs, action)
	struct nis_dbent	*dptr;
	entry_obj		*entry;
	uint32_t		numattrs;
	nis_attr		*attrs;
	xle_action		action;
{
	XDR			xdrs;
	uint32_t		curpos;
	xff_log_entry		*log_entry;
	uint32_t		*w = NULL;

	if (dptr->fp_log == NULL)
		xff_create(dptr, dptr->tdesc->table_path, 1);

	log_entry = calloc(1, sizeof(xff_log_entry));

	log_entry->xle_time = time(NULL);
	log_entry->xle_action = action;
	log_entry->xle_status = XLE_PENDING;
	log_entry->xle_attrs.xle_attrs_len = numattrs;
	log_entry->xle_attrs.xle_attrs_val = attrs;
	log_entry->xle_entry = entry;
	
	dptr->hdr_log.entry_just_added = dptr->hdr_log.end_of_entries;
	fseek(dptr->fp_log, dptr->hdr_log.end_of_entries, SEEK_SET);

	xdrstdio_create(&xdrs, dptr->fp_log, XDR_ENCODE);

	if (!xdr_xff_log_entry(&xdrs, log_entry)) {
		xdr_destroy(&xdrs);
		return(DB_INTERNAL_ERROR);
	}

	curpos = ftell(dptr->fp_log);

	dptr->hdr_log.table_size += (curpos - dptr->hdr_log.start_of_relcache);
	dptr->hdr_log.start_of_relcache +=
	    (curpos - dptr->hdr_log.end_of_entries);
	dptr->hdr_log.end_of_entries = curpos;

	/*
	 * If the action is XLE_DELETE, we need to go back and find the
	 * log entry (entries?) where the record(s) we just marked as deleted
	 * (were?) originally added and mark it (them?) as XLE_OBSOLETE so
	 * that xff_first_entry() and xff_next_entry() can skip it (them?)
	 * if necessary necessaary.
	 */

	bzero((char *)log_entry, sizeof(xff_log_entry));

	if (action == XLE_DELETE &&
	    validate_attrs(dptr, numattrs, attrs) == DB_SUCCESS) {
		if (nis_db_parse_rels(numattrs,
		    attrs, dptr, &w) == DB_NOTFOUND)
			return(DB_SUCCESS); /* Eh? */
		/* Just do the first one */
		if (dptr->fp != NULL && w[1] < dptr->hdr.table_size) {
			xff_delete_entry(dptr, entry, w[1]);
		} else {
			w[1] -= dptr->hdr.table_size;
			bzero((char *)log_entry, sizeof(log_entry));
			fseek(dptr->fp_log, w[1], SEEK_SET);
			xdrstdio_create(&xdrs, dptr->fp_log, XDR_DECODE);
			if (!xdr_xff_log_entry(&xdrs, log_entry)) {
				xdr_destroy(&xdrs);
				return(DB_INTERNAL_ERROR);
			}
			xdr_destroy(&xdrs);
			log_entry->xle_status = XLE_OBSOLETE;
			fseek(dptr->fp_log, w[1], SEEK_SET);
			xdrstdio_create(&xdrs, dptr->fp_log, XDR_ENCODE);
			if (!xdr_xff_log_entry(&xdrs, log_entry)) {
				xdr_destroy(&xdrs);
				return(DB_INTERNAL_ERROR);
			}
			xdr_destroy(&xdrs);
			xdr_free((xdrproc_t)xdr_xff_log_entry, (char *)log_entry);
		}
		if (w != NULL)
			free(w);
	}

	fflush(dptr->fp_log);
	free(log_entry);

	/*
	 * We increase the entry_just_added counter by the size of
	 * the standard table file (or 0, if it doesn't exist).
	 * This biases all the addresses in the relation index so
	 * that we can distinguish them from addresses in the
	 * standard file (since we may have both a standard file and
	 * a log open at the same time).
	 */
	dptr->hdr_log.entry_just_added += dptr->hdr.table_size;

	if (dptr->fp != NULL)
		dptr->hdr.log_present = 1;

	return(DB_SUCCESS);
}

db_status xff_open(dptr, path)
	struct nis_dbent	*dptr;
	char			*path;
{
	XDR			xdrs;
	char			log_path[NIS_MAXNAMELEN];

	snprintf(log_path, sizeof(log_path), "%s.log", path);

	if (access(path, R_OK|W_OK) != -1) {
		if (dptr->fp != NULL)
			return(DB_NOTUNIQUE);
		if ((dptr->fp = fopen(path, "r+")) == NULL) {
			return(DB_INTERNAL_ERROR);
		}
		xdrstdio_create(&xdrs, dptr->fp, XDR_DECODE);

		if (!xdr_xff_header(&xdrs, &dptr->hdr)) {
			xdr_destroy(&xdrs);
			fclose(dptr->fp);
			return(DB_INTERNAL_ERROR);
		}
		xdr_destroy(&xdrs);
	}

	if (access(log_path, R_OK|W_OK) != -1) {
		if (dptr->fp_log != NULL)
			return(DB_NOTUNIQUE);
		if ((dptr->fp_log = fopen(log_path, "r+")) == NULL) {
			return(DB_INTERNAL_ERROR);
		}

		xdrstdio_create(&xdrs, dptr->fp_log, XDR_DECODE);

		if (!xdr_xff_header(&xdrs, &dptr->hdr_log)) {
			xdr_destroy(&xdrs);
			fclose(dptr->fp_log);
			return(DB_INTERNAL_ERROR);
		}
		xdr_destroy(&xdrs);
		dptr->hdr.log_present = 1;
		fseek(dptr->fp_log, 0L, SEEK_END);
		dptr->hdr_log.end_of_entries = ftell(dptr->fp_log);
	}

	if (dptr->fp == NULL && dptr->fp_log == NULL)
		return(DB_NOTFOUND);

	return(DB_SUCCESS);
}

db_status xff_close(dptr)
	struct nis_dbent	*dptr;
{
	XDR			xdrs;

	if (dptr->fp != NULL) {
		fseek(dptr->fp, 0L, SEEK_SET);
		xdrstdio_create(&xdrs, dptr->fp, XDR_ENCODE);

		if (!xdr_xff_header(&xdrs, &dptr->hdr)) {
			xdr_destroy(&xdrs);
			fclose(dptr->fp);
			return(DB_INTERNAL_ERROR);
		}

		xdr_destroy(&xdrs);

		fflush(dptr->fp);
		fclose(dptr->fp);
		dptr->fp = NULL;
		xdr_free((xdrproc_t)xdr_xff_header, (char *)&dptr->hdr);
	}

	if (dptr->fp_log != NULL) {
		fseek(dptr->fp_log, 0L, SEEK_SET);
		xdrstdio_create(&xdrs, dptr->fp_log, XDR_ENCODE);

		if (!xdr_xff_header(&xdrs, &dptr->hdr_log)) {
			xdr_destroy(&xdrs);
			fclose(dptr->fp_log);
			return(DB_INTERNAL_ERROR);
		}

		xdr_destroy(&xdrs);

		fflush(dptr->fp_log);
		fclose(dptr->fp_log);
		dptr->fp_log = NULL;
		xdr_free((xdrproc_t)xdr_xff_header, (char *)&dptr->hdr_log);
	}

	return(DB_SUCCESS);
}

/*
 * Initially, we create only a log. It's only until after we've
 * checkpointed the table that a real XFF file will be created.
 */
db_status xff_create(dptr, path, log)
	struct nis_dbent	*dptr;
	char			*path;
	int			log;
{
	XDR			xdrs;
	char			log_path[NIS_MAXNAMELEN];

	if (log) {
		dptr->hdr_log.vers = known_versions[0].val;
		dptr->hdr_log.log_present = 1;
		dptr->hdr.log_present = 1;
		dptr->hdr_log.type = XFF_LOG;
		dptr->hdr_log.table_size = sizeof(xff_header);
		dptr->hdr_log.start_of_entries = sizeof(xff_header);
		dptr->hdr_log.end_of_entries = sizeof(xff_header);
		dptr->hdr_log.entry_just_added = sizeof(xff_header);
		dptr->hdr_log.start_of_relcache = sizeof(xff_header);
	} else {
		dptr->hdr.vers = known_versions[0].val;
		dptr->hdr.log_present = 1;
		dptr->hdr.type = XFF_FILE;
		dptr->hdr.table_size = sizeof(xff_header);
		dptr->hdr.start_of_entries = sizeof(xff_header);
		dptr->hdr.end_of_entries = sizeof(xff_header);
		dptr->hdr.entry_just_added = sizeof(xff_header);
		dptr->hdr.start_of_relcache = sizeof(xff_header);
	}

	if (log) {
		snprintf(log_path, sizeof(log_path), "%s.log", path);
		path = (char *)&log_path;
	}

	if (log) {
		if ((dptr->fp_log = fopen(path, "w+")) == NULL)
			return(DB_INTERNAL_ERROR);
	} else {
		if ((dptr->fp = fopen(path, "w+")) == NULL)
			return(DB_INTERNAL_ERROR);
	}

	xdrstdio_create(&xdrs, log ? dptr->fp_log : dptr->fp, XDR_ENCODE);

	if (!xdr_xff_header(&xdrs, log ? &dptr->hdr_log : &dptr->hdr)) {
		xdr_destroy(&xdrs);
		fclose(log ? dptr->fp_log : dptr->fp);
		return(DB_INTERNAL_ERROR);
	}

	xdr_destroy(&xdrs);

	return(DB_SUCCESS);
}

db_status xff_add_entry(dptr, entry)
	struct nis_dbent	*dptr;
	entry_obj		*entry;
{
	XDR			xdrs;
	uint32_t		curpos;

	if (dptr->status == STATUS_REPLAYING_LOG)
		return (DB_SUCCESS);

	dptr->hdr.entry_just_added = dptr->hdr.end_of_entries;
	fseek(dptr->fp, dptr->hdr.end_of_entries, SEEK_SET);

	xdrstdio_create(&xdrs, dptr->fp, XDR_ENCODE);

	if (!xdr_entry_obj(&xdrs, entry)) {
		xdr_destroy(&xdrs);
		return(DB_INTERNAL_ERROR);
	}

	curpos = ftell(dptr->fp);

	dptr->hdr.table_size += (curpos - dptr->hdr.start_of_relcache);
	dptr->hdr.start_of_relcache +=
	    (curpos - dptr->hdr.end_of_entries);
	dptr->hdr.end_of_entries = curpos;

	xdr_destroy(&xdrs);

	return(DB_SUCCESS);
}

/*
 * Mark entry as deleted.
 * XXX Wastes space. (Only compact_file() can reclaim it.)
 */
db_status xff_delete_entry(dptr, entry, offset)
	struct nis_dbent	*dptr;
	entry_obj		*entry;
	uint32_t		offset;
{
	XDR			xdrs;
	int			i;

	if (dptr->status == STATUS_REPLAYING_LOG ||
		dptr->status == STATUS_COMMITTING_LOG)
		return (DB_SUCCESS);

	for (i = 0; i < strlen(entry->en_type) - 1; i++) {
		/* XXX */
		entry->en_type[i] = 255;
	}

	fseek(dptr->fp, offset, SEEK_SET);
	xdrstdio_create(&xdrs, dptr->fp, XDR_ENCODE);

	if (!xdr_entry_obj(&xdrs, entry)) {
		xdr_destroy(&xdrs);
		return(DB_INTERNAL_ERROR);
	}

	xdr_destroy(&xdrs);

	return(DB_SUCCESS);
}

db_status xff_get_entry(dptr, entry, offset)
	struct nis_dbent	*dptr;
	entry_obj		*entry;
	uint32_t		offset;
{
	XDR			xdrs;
	xff_log_entry		log_entry;
	int			use_log = 0;

	if (offset > dptr->hdr.table_size) {
		if (offset - dptr->hdr.table_size >
		    dptr->hdr_log.end_of_entries)
			return(DB_NOTFOUND);
		fseek(dptr->fp_log, offset - dptr->hdr.table_size, SEEK_SET);
		xdrstdio_create(&xdrs, dptr->fp_log, XDR_DECODE);
		use_log++;
	} else {
		fseek(dptr->fp, offset, SEEK_SET);
		xdrstdio_create(&xdrs, dptr->fp, XDR_DECODE);
	}

	if (use_log) {
		bzero((char *)&log_entry, sizeof(log_entry));
		log_entry.xle_entry = entry;
		if (!xdr_xff_log_entry(&xdrs, &log_entry)) {
			xdr_destroy(&xdrs);
			return(DB_INTERNAL_ERROR);
		}
	} else {
		if (!xdr_entry_obj(&xdrs, entry)) {
			xdr_destroy(&xdrs);
			return(DB_INTERNAL_ERROR);
		}
	}

	xdr_destroy(&xdrs);

	if (use_log) {
		xle_status			xst;

		xst = log_entry.xle_status;
		log_entry.xle_entry = NULL;
		xdr_free((xdrproc_t)xdr_xff_log_entry, (char *)&log_entry);
		if (xst == XLE_OBSOLETE) {
			xdr_free((xdrproc_t)xdr_entry_obj, (char *)entry);
			return(DB_NOTFOUND);
		}
	}

	if ((unsigned char)entry->en_type[0] == 255) {
		xdr_free((xdrproc_t)xdr_entry_obj, (char *)entry);
		return(DB_NOTFOUND);
	}

	return(DB_SUCCESS);
}

db_status xff_first_entry(dptr, entry)
	struct nis_dbent	*dptr;
	entry_obj		*entry;
{
	XDR			xdrs;
	xff_log_entry		log_entry;
	int			use_log = 0;

	if (dptr->fp == NULL)
		use_log++;

	if (dptr->fp_log == NULL && use_log)
		return(DB_INTERNAL_ERROR);

roll_over_to_log:

	if (use_log) {
		if (dptr->hdr_log.start_of_entries ==
		    dptr->hdr_log.end_of_entries)
			return(DB_NOTFOUND);

		fseek(dptr->fp_log, dptr->hdr_log.start_of_entries, SEEK_SET);

		xdrstdio_create(&xdrs, dptr->fp_log, XDR_DECODE);
	} else {
		if (dptr->hdr.start_of_entries ==
		    dptr->hdr.start_of_relcache) {
			if (dptr->hdr.log_present) {
				use_log++;
				goto roll_over_to_log;
			}
			return(DB_NOTFOUND);
		}

		fseek(dptr->fp, dptr->hdr.start_of_entries, SEEK_SET);

		xdrstdio_create(&xdrs, dptr->fp, XDR_DECODE);
		if (dptr->hdr.log_present) {
			if (dptr->fp_log == NULL)
				return(DB_INTERNAL_ERROR);
			else
				fseek(dptr->fp_log,
				    dptr->hdr_log.start_of_entries,
				    SEEK_SET);
		}
	}

again:
	if (use_log) {
		bzero((char *)&log_entry, sizeof(log_entry));
		log_entry.xle_entry = entry;
		if (!xdr_xff_log_entry(&xdrs, &log_entry)) {
			xdr_destroy(&xdrs);
			return(DB_INTERNAL_ERROR);
		}
		log_entry.xle_entry = NULL;
		xdr_free((xdrproc_t)xdr_xff_log_entry, (char *)&log_entry);
	} else
		if (!xdr_entry_obj(&xdrs, entry)) {
			xdr_destroy(&xdrs);
			return(DB_INTERNAL_ERROR);
		}

	if (use_log) {
		if (log_entry.xle_status == XLE_OBSOLETE ||
	    		log_entry.xle_action == XLE_DELETE) {
			xdr_free((xdrproc_t)xdr_entry_obj, (char *)entry);
			if (ftell(dptr->fp_log) !=
			    dptr->hdr_log.end_of_entries)
				goto again;
			else {
				xdr_destroy(&xdrs);
				return(DB_NOTFOUND);
			}
		}
	}

	if ((unsigned char)entry->en_type[0] == 255) {
		xdr_free((xdrproc_t)xdr_entry_obj, (char *)entry);
		if (ftell(dptr->fp) != dptr->hdr.end_of_entries)
			goto again;
		else {
			xdr_destroy(&xdrs);
			use_log++;
			goto roll_over_to_log;
		}
	}

	if (use_log)
		dptr->pos = ftell(dptr->fp_log) + dptr->hdr.table_size;
	else
		dptr->pos = ftell(dptr->fp);

	xdr_destroy(&xdrs);
	return(DB_SUCCESS);
}

db_status xff_next_entry(dptr, entry)
	struct nis_dbent	*dptr;
	entry_obj		*entry;
{
	XDR			xdrs;
	uint32_t		curpos;
	xff_log_entry		log_entry;
	int			use_log = 0;

	if (dptr->fp == NULL) {
		if (dptr->fp_log == NULL)
			return(DB_INTERNAL_ERROR);
		use_log++;
		curpos = ftell(dptr->fp_log);
		if (curpos == dptr->hdr_log.end_of_entries)
			return(DB_NOTFOUND);
	} else {
		curpos = ftell(dptr->fp);
		if (curpos == dptr->hdr.start_of_relcache ||
		    dptr->hdr.start_of_entries ==
		    dptr->hdr.start_of_relcache) {
			if (dptr->fp_log == NULL)
				return(DB_NOTFOUND);
			else {
				use_log++;
				curpos = ftell(dptr->fp_log);
				if (curpos == dptr->hdr_log.end_of_entries)
					return(DB_NOTFOUND);
			}
		}
	}

roll_over_to_log:

	xdrstdio_create(&xdrs, use_log ? dptr->fp_log : dptr->fp, XDR_DECODE);
again:
	if (use_log) {
		bzero((char *)&log_entry, sizeof(log_entry));
		log_entry.xle_entry = entry;
		if (!xdr_xff_log_entry(&xdrs, &log_entry)) {
			xdr_destroy(&xdrs);
			return(DB_INTERNAL_ERROR);
		}
		log_entry.xle_entry = NULL;
		xdr_free((xdrproc_t)xdr_xff_log_entry, (char *)&log_entry);
	} else {
		if (!xdr_entry_obj(&xdrs, entry)) {
			xdr_destroy(&xdrs);
			return(DB_INTERNAL_ERROR);
		}
	}

	if (use_log) {
		if (log_entry.xle_status == XLE_OBSOLETE ||
	    		log_entry.xle_action == XLE_DELETE) {
			xdr_free((xdrproc_t)xdr_entry_obj, (char *)entry);
			if (ftell(dptr->fp_log) !=
			    dptr->hdr_log.end_of_entries)
				goto again;
			else {
				xdr_destroy(&xdrs);
				return(DB_NOTFOUND);
			}
		}
	}

	if ((unsigned char)entry->en_type[0] == 255) {
		xdr_free((xdrproc_t)xdr_entry_obj, (char *)entry);
		if (ftell(dptr->fp) != dptr->hdr.end_of_entries)
			goto again;
		else {
			xdr_destroy(&xdrs);
			use_log++;
			goto roll_over_to_log;
		}
	}

	if (use_log)
		dptr->pos = ftell(dptr->fp_log) + dptr->hdr.table_size;
	else
		dptr->pos = ftell(dptr->fp);

	xdr_destroy(&xdrs);
	return(DB_SUCCESS);
}

db_status xff_compact_file(dptr)
	struct nis_dbent	*dptr;
{
	XDR			xdrs_in;
	XDR			xdrs_out;
	struct loserlist	*losers = NULL, *tmp = NULL;
	FILE			*fp_out;
	xff_header		header_out;
	char			path[NIS_MAXNAMELEN];
	entry_obj		*entry;
	uint32_t		pos, curpos;

	header_out.vers = known_versions[0].val;
	header_out.log_present = 0;
	header_out.type = XFF_FILE;
	header_out.table_size = dptr->hdr.table_size;
	header_out.start_of_entries = sizeof(xff_header);
	header_out.end_of_entries = sizeof(xff_header);
	header_out.entry_just_added = sizeof(xff_header);
	header_out.start_of_relcache = sizeof(xff_header);

	snprintf (path, sizeof(path), "%s.tmp", dptr->tdesc->table_path);

	if ((fp_out = fopen(path, "w+")) == NULL)
		return(DB_INTERNAL_ERROR);

	fseek(dptr->fp, dptr->hdr.start_of_entries, SEEK_SET);

	xdrstdio_create(&xdrs_out, fp_out, XDR_ENCODE);
	xdrstdio_create(&xdrs_in, dptr->fp, XDR_DECODE);

	xdr_xff_header(&xdrs_out, &header_out);

	entry = calloc(1, sizeof(entry_obj));
	while ((pos = ftell(dptr->fp)) != dptr->hdr.start_of_relcache) {
		if (!xdr_entry_obj(&xdrs_in, entry)) {
			syslog(LOG_WARNING,
			    "reading entry from input failed!");
			break;
		}
		if ((unsigned char)entry->en_type[0] == 255) {
			tmp = malloc(sizeof(struct loserlist));
			tmp->loser = pos;
			tmp->size = ftell(dptr->fp) - pos;
			tmp->next = losers;
			losers = tmp;
		} else {
			curpos = ftell(fp_out);
			header_out.entry_just_added = curpos;
			xdr_entry_obj(&xdrs_out, entry);
			curpos = ftell(fp_out);
			header_out.table_size +=
			    (curpos - header_out.start_of_relcache);
			header_out.start_of_relcache +=
			    (curpos - header_out.end_of_entries);
			header_out.end_of_entries = curpos;
		}
		xdr_free((xdrproc_t)xdr_entry_obj, (char *)entry);
	}

	fseek(fp_out, 0, SEEK_SET);

	xdr_xff_header(&xdrs_out, &header_out);

	xdr_destroy(&xdrs_in);
	xdr_destroy(&xdrs_out);
	free(entry);

	if (losers) {
		rcm_compact(losers, dptr);
		adjust_pos(dptr->table_name, losers);
	}

	while(losers) {
		tmp = losers->next;
		free(losers);
		losers = tmp;
	}

	fclose(dptr->fp);
	dptr->fp = fp_out;

	dptr->hdr.vers = header_out.vers;
	dptr->hdr.table_size = header_out.table_size;
	dptr->hdr.start_of_entries = header_out.start_of_entries;
	dptr->hdr.end_of_entries = header_out.end_of_entries;
	dptr->hdr.entry_just_added = header_out.entry_just_added;
	dptr->hdr.start_of_relcache = header_out.start_of_relcache;

	rcm_dump(dptr);
	fflush(dptr->fp);

	rename(path, dptr->tdesc->table_path);

	return(DB_SUCCESS);
}
