/*
 * Copyright (c) 1996
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <limits.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nis_db.h>
#include "nis_db_private.h"

#ifndef lint
__RCSID("@(#) $FreeBSD$");
#endif

/*
 * This function can behave in one of three ways, depending on the
 * arguments passed by the caller:
 *
 * 1) If the caller specifies no attributes, it returns the first
 *    entry in the specified table no matter what it is and sets
 *    the table's 'cursor' to the top of the table. (Caller can
 *    use db_next_entry() to enumerate the entire table.)
 *
 * 2) If the caller specifies attributes and they match exactly one
 *    entry, it returns the entry and sets the 'cursor' at the point
 *    where the entry was found. (Caller can enumerate all entries in
 *    table starting from the one that matches the specified attributes.)
 *
 * 3) If the caller specifies attributes and they match more than one
 *    entry, it returns the first matching entry and then saves the
 *    remaining entries so that they can be listed with db_next_entry().
 *    (Caller can enumerate only the list of entries that match the
 *    attributes.
 */

db_result *db_first_entry(table_name, numattrs, attrs)
	char			*table_name;
	int			numattrs;
	nis_attr		*attrs;
{
	db_result		*res;
	db_status		lstat;
	struct timeval		t_start, t_end;
	struct nis_dbent	*dptr = NULL;
	char			*p = NULL;
	struct nis_nextinfo	*nextp = NULL;
	entry_obj		**entry;
	uint32_t		rval = DB_SUCCESS;
	uint32_t		*w = NULL;
	int			i;

	gettimeofday(&t_start, NULL);

	if ((res = calloc(1, sizeof(db_result))) == NULL)
		return(NULL);

	/*
	 * Pull the table handle out of the cache (will open
	 * the table if closed).
	 */
	if ((dptr = find_table(table_name)) == NULL) {
		res->status = DB_BADTABLE;
		goto bail;
	}

	/*
	 * User has specified some attributes: see if they
	 * match anything in the table.
	 */
	if (numattrs) {
		if ((res->status = validate_attrs(dptr,
		    numattrs, attrs)) != DB_SUCCESS)
			goto bail;

		/*
		 * Try to get a list of tags that satisfy the
		 * attributes.
		 */
		if (dptr->relarray == NULL)
			rcm_load(dptr);
		lstat = nis_db_parse_rels(numattrs, attrs, dptr, &w);

		/*
		 * Attributes failed to match any entries:
		 * request is bogus.
		 */
		if (lstat != DB_SUCCESS) {
			res->status = DB_NOTFOUND;
			free(w);
			goto bail;
		}

		/*
		 * Query matched something. In any case, we'll
		 * only be able to return the first entry found
		 * by the match, so set up our result structure
		 * to point to it.
		 */

		if ((entry = calloc(1, sizeof(entry_obj *))) == NULL) {
			res->status = DB_MEMORY_LIMIT;
			free(w);
			goto bail;
		}

		if ((*entry = calloc(1, sizeof(entry_obj))) == NULL) {
			res->status = DB_MEMORY_LIMIT;
			free(w);
			free(entry);
			goto bail;
		}

		/*
		 * xff_get_entry() may return DB_NOTFOUND if we happen
		 * to stumble onto an entry that's been deleted but not
		 * reclaimed by file compaction yet. If so, we have to
		 * skip past it and try the next one.
		 */
		for (i = 0; i < w[0]; i++) {
			rval = xff_get_entry(dptr, *entry, w[i + 1]);
			if (rval == DB_NOTFOUND)
				continue;
			else
				break;
		}
		if (rval != DB_SUCCESS) {
			free(*entry);
			free(entry);
			free(w);
			if (rval == DB_NOTFOUND)
				res->status = DB_NOTFOUND;
			else
				res->status = DB_INTERNAL_ERROR;
			goto bail;
		}

		res->objects.objects_len = 1;
		res->objects.objects_val = entry;

		/*
		 * Decide on proper behavior. If the user-specified
		 * attributes match more than one entry, then what we
		 * do is allow the caller to use db_next_entry() to
		 * step through the list of matches, _not_ through
		 * the entire table. If the attributes match only one
		 * entry, then we set our table's 'cursor' to point
		 * at the entry and allow the user to step through the
		 * entire table starting at the location of the matched
		 * entry.
		 */
		/* Set up the nextinfo handle */
		p = calloc(1,sizeof(struct nis_nextinfo) +
		    ((*w + 1) * ULONGSZ));
		nextp = (struct nis_nextinfo *)p;
		nextp->etype = ETYPE_BATCH;
		nextp->pos = 0;
		if (*w > 1) {
			nextp->nextcount = i;
			bcopy((char *)w, p + sizeof(struct nis_nextinfo),
					((*w + 1) * ULONGSZ));
		} else {
			free(w);
			nextp->nextcount = 0;
		}
		res->nextinfo.db_next_desc_val = (char *)nextp;
		res->nextinfo.db_next_desc_len =
		    sizeof(struct nis_nextinfo) + ((*w + 1) * ULONGSZ);

		res->status = DB_SUCCESS;
		goto bail;
	}

	if ((entry = calloc(1, sizeof(entry_obj *))) == NULL) {
		res->status = DB_MEMORY_LIMIT;
		goto bail;
	}

	if ((*entry = calloc(1, sizeof(entry_obj))) == NULL) {
		res->status = DB_MEMORY_LIMIT;
		free(w);
		free(entry);
		goto bail;
	}

	if ((rval = xff_first_entry(dptr, *entry)) != DB_SUCCESS) {
		free(*entry);
		free(entry);
		free(w);
		if (rval == DB_NOTFOUND)
			res->status = DB_NOTFOUND;
		else
			res->status = DB_INTERNAL_ERROR;
		goto bail;
	}

	res->objects.objects_len = 1;
	res->objects.objects_val = entry;

	nextp = calloc(1, sizeof(struct nis_nextinfo));
	nextp->etype = ETYPE_ENUM;
	nextp->nextcount = 0;
	res->nextinfo.db_next_desc_val = (char *)nextp;
	res->nextinfo.db_next_desc_len = sizeof(struct nis_nextinfo);

bail:
	if (res->status == DB_SUCCESS) {
		if (nextp->etype == ETYPE_ENUM)
			 nextp->pos = dptr->pos;
		nextp->cksum = NEXTINFO_MAGIC;
		nextp->cksum = __nis_db_cksum((unsigned char *)nextp,
		    res->nextinfo.db_next_desc_len);
	}

	gettimeofday(&t_end, NULL);
	res->ticks += 
	    (t_end.tv_usec - t_start.tv_usec) +
	    ((t_end.tv_sec - t_start.tv_sec) * 1000000);

	return(res);
}
