/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <errno.h>
#include <syslog.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <rpc/rpc.h>
#include <rpc/xdr.h> 
#include <rpcsvc/nis.h>
#include <rpcsvc/nis_db.h>
#include "nis_db_struct.h"
#include "nis_db_private.h"
#include "nis_db_version.h"

#ifndef lint
__RCSID("@(#) $FreeBSD$");
#endif

#define TABLESIZE 256
#define HASH_MASK 0x000000FF

struct t_entry {
	uint32_t hashval;		/* Actual 32-bit hash value */
	struct nis_dbent *dbptr;	/* Table descriptor */
	struct nis_dbent *link;		/* Link to another table */
	struct t_entry *next;		/* Next entry in list at this slot */
};

static struct t_entry *tables[TABLESIZE];	/* open tables */

static u_int32_t hashkey(key)
	char			*key;
{
	if (key == NULL)
		return(-1);
	return(hash((void *)key, strlen(key), HASH_NORM));
}

static void t_insert(table, key, data, link)
	struct t_entry		*table[];
	char			*key;
	struct nis_dbent	*data,
				*link;
{
	struct t_entry		*new;
	u_int32_t		i;

	i = hashkey(key);

	new = malloc(sizeof(struct t_entry));
	new->hashval = i;
	new->dbptr = data;
	new->link = link;
	new->next = NULL;

	new->next = table[i & HASH_MASK];
	table[i & HASH_MASK] = new;

	return;
}

static struct nis_dbent *t_lookup(table, key)
	struct t_entry		*table[];
	char			*key;
{
	struct t_entry		*cur;
	u_int32_t		i;

	i = hashkey(key);
	cur = table[i & HASH_MASK];

	while (cur) {
		if (cur->hashval == i)
			return(cur->dbptr);
		cur = cur->next;
	}

	return(NULL);
}

static void t_delete(table, key)
	struct t_entry		*table[];
	char			*key;
{
	struct t_entry		*cur, *prev;
	u_int32_t		i;

	i = hashkey(key);
	cur = table[i & HASH_MASK];
	prev = NULL;

	while (cur) {
		if (cur->hashval == i)
			break;
		prev = cur;
		cur = cur->next;
	}

	if (cur == NULL)
		return;

	if (cur == table[i & HASH_MASK]) {
		table[i & HASH_MASK] = cur->next;
		free(cur);
		return;
	}

	if (prev == table[i & HASH_MASK])
		table[i & HASH_MASK] = cur->next;
	else
		prev = cur->next;

	free(cur);
	return;
}

static int numdbs = 0;
static char nisdir[MAXPATHLEN];

#ifndef MAXDBS
#define MAXDBS 256
#endif

static char *get_path_from_catalog __P(( char * ));

/*
 * Description of a 'dictionary' table. This is used to
 * describe the format of the catalog file, which has only
 * one searchable column. We fake this up here because we need
 * it to statisfy the RCM and XFF functions. We're
 * actually using XFF and NIS+ structures (tables and entries)
 * to store data in the catalog: a catalog is really a table
 * with just one column. However, to make it work, we need
 * to predefine certain things in order to properly bootstrap
 * ourselves.
 *
 * This is almost like kernel programming. :)
 */

static table_col dict_cols = {
	"table_descriptor",
	TA_SEARCHABLE /*| TA_CASE*/,
	DEFAULT_RIGHTS
};

static table_obj catalog_table = {
	"dictionary_tbl",
	1,
	':',
	{ 1, &dict_cols },
	"foo:bar:baz",
};

static struct table_desc catalog_desc  = {
	&catalog_table,
	"dictionary_tbl",
	NULL,
	{ 0 , 0 },
	0
};

static struct nis_dbent catalog_ptr = {
	NULL,
	{ 0, 0, 0, 0, 0, 0, 0, 0 },
	NULL,
	{ 0, 0, 0, 0, 0, 0, 0, 0 },

	&catalog_desc,
	"dictionary_tbl",
	&catalog_table,

	1,
	0,
	0,
	0,
	NULL,
	NULL
};

/*
 * Catalog access convenience functions.
 */
static db_status get_dictionary(table_name, tdesc)
	char *table_name;
	table_desc **tdesc;
{
	uint32_t hkey;
	nis_datum data = { NULL, 0 };
	entry_obj entry;

	bzero((char *)&entry, sizeof(entry_obj));

	hkey = hash(table_name, strlen(table_name), HASH_NORM);

	if (rcm_get_index(&catalog_ptr, 0, hkey, &data)!= DB_SUCCESS)
		return(DB_NOTFOUND);

	if (xff_get_entry(&catalog_ptr, &entry,
	    *(uint32_t *)data.data) != DB_SUCCESS)
		return(DB_INTERNAL_ERROR);

	if ((*(caddr_t *)tdesc = (caddr_t)decode_table_desc(
	    entry.en_cols.en_cols_val[0].ec_value.ec_value_val,
	    entry.en_cols.en_cols_val[0].ec_value.ec_value_len)) == NULL)
		return(DB_INTERNAL_ERROR);

	return(DB_SUCCESS);
}

db_status del_dictionary(table_name)
	char *table_name;
{
	uint32_t hkey;
	nis_datum data = { NULL, 0 };
	nis_attr attrs = { "table_descriptor", { 0 , "" } };
	entry_obj entry;

	bzero((char *)&entry, sizeof(entry_obj));
	attrs.zattr_val.zattr_val_val = table_name;
	attrs.zattr_val.zattr_val_len = strlen(table_name);

	hkey = hash(table_name, strlen(table_name), HASH_NORM);

	if (rcm_get_index(&catalog_ptr, 0, hkey, &data) == DB_SUCCESS) {
		if (xff_get_entry(&catalog_ptr, &entry,
		    *(uint32_t *)data.data) != DB_SUCCESS)
			return(DB_BADTABLE);
		if (catalog_ptr.status == STATUS_NORMAL)
			xff_add_log_entry(&catalog_ptr, &entry,
			    1, &attrs, XLE_DELETE);
		if (catalog_ptr.status == STATUS_COMMITTING_LOG) {
			if (xff_delete_entry(&catalog_ptr, &entry,
			    (uint32_t)data.data) != DB_SUCCESS)
			return(DB_INTERNAL_ERROR);
		}
		rcm_delete_index(&catalog_ptr, 0, hkey);
		xdr_free((xdrproc_t)xdr_entry_obj, (char *)&entry);
	} else
		return(DB_NOTFOUND);

	return(DB_SUCCESS);
}

db_status put_dictionary(table_name, tdesc)
	char *table_name;
	table_desc *tdesc;
{
	uint32_t hkey;
	nis_datum data = { NULL, 0 };
	nis_attr attrs = { "table_descriptor", { 0 , "" } };
	entry_obj entry;
	entry_col col;

	bzero((char *)&entry, sizeof(entry_obj));
	bzero((char *)&col, sizeof(entry_col));

	attrs.zattr_val.zattr_val_val = table_name;
	attrs.zattr_val.zattr_val_len = strlen(table_name);
	hkey = hash(table_name, strlen(table_name), HASH_NORM);

	if (rcm_get_index(&catalog_ptr, 0, hkey, &data) == DB_SUCCESS)
		del_dictionary(table_name);

	col.ec_flags = 0;
	col.ec_value.ec_value_val = encode_table_desc(tdesc,
	    (uint32_t *)&col.ec_value.ec_value_len);

	entry.en_cols.en_cols_val = &col;
	entry.en_cols.en_cols_len = 1;
	entry.en_type = "NIS+ catalog";

	if (catalog_ptr.status == STATUS_NORMAL)
		if (xff_add_log_entry(&catalog_ptr, &entry, 1,
		    &attrs, XLE_ADD) != DB_SUCCESS)
			return(DB_INTERNAL_ERROR);

	if (catalog_ptr.status == STATUS_COMMITTING_LOG)
		if (xff_add_entry(&catalog_ptr, &entry) != DB_SUCCESS)
			return(DB_INTERNAL_ERROR);

	free(col.ec_value.ec_value_val);

	data.size = ULONGSZ;
	if (catalog_ptr.status == STATUS_NORMAL ||
	    catalog_ptr.status == STATUS_REPLAYING_LOG)
		data.data = (char *)&catalog_ptr.hdr_log.entry_just_added;
	if (catalog_ptr.status == STATUS_COMMITTING_LOG)
		data.data = (char *)&catalog_ptr.hdr.entry_just_added;

	if (rcm_put_index(&catalog_ptr, 0, hkey, &data) != DB_SUCCESS)
		return(DB_INTERNAL_ERROR);

	return(DB_SUCCESS);
}

static db_status enum_dictionary(tdesc)
	table_desc		**tdesc;
{
	entry_obj		entry;
	db_status		rval;

	bzero((char *)&entry, sizeof(entry_obj));

	if (catalog_ptr.pos)
		rval = xff_next_entry(&catalog_ptr, &entry);
	else {
		rval = xff_first_entry(&catalog_ptr, &entry);
		catalog_ptr.pos = 1;
	}

	if (rval != DB_SUCCESS)
		return(DB_INTERNAL_ERROR);

	if ((*(caddr_t *)tdesc = (caddr_t)decode_table_desc(
	    entry.en_cols.en_cols_val[0].ec_value.ec_value_val,
	    entry.en_cols.en_cols_val[0].ec_value.ec_value_len)) == NULL)
		return(DB_INTERNAL_ERROR);

	xdr_free((xdrproc_t)xdr_entry_obj, (char *)&entry);
	return(DB_SUCCESS);
}

static void nis_free_qent(dptr)
	struct nis_dbent	*dptr;
{
	register table_desc	*tptr;


	if (dptr->relarray != NULL)
		rcm_free(dptr);

	if (dptr->fp != NULL || dptr->fp_log != NULL)
		xff_close(dptr);

#ifdef foo
	if (dptr->index_hashes)
		free(dptr->index_hashes);
#endif
	tptr = dptr->tdesc;

	/*
	 * Abuse the XDR library to make free()ing complex
	 * structures easier.
	 */
	if (tptr != NULL) {
		xdr_free((xdrproc_t)xdr_table_desc, (char *)tptr);
		free(tptr);
	}
	free(dptr);

	return;
}

static struct nis_dbent *nis_malloc_qent(tdesc)
	table_desc		*tdesc;
{
	table_desc		*tptr;
	struct nis_dbent	 *dptr;
	char			*enc;
	uint32_t		size;

	if ((dptr = calloc(1, sizeof(struct nis_dbent))) == NULL)
		return(NULL);

	/*
	 * No table was specified -- this means it'll be retrieved
	 * from the table file on disk, so we don't need to copy
	 * an existing one.
	 */
	if (tdesc == NULL)
		return(dptr);

	if ((enc = encode_table_desc(tdesc, &size)) == NULL) {
		free(dptr);
		return(NULL);
	}

	if ((tptr = decode_table_desc(enc, size)) == NULL) {
		free(enc);
		free(dptr);
		return(NULL);
	}

	free(enc);

	dptr->tdesc = tptr;
	dptr->table_name = dptr->tdesc->table_name;
	dptr->table = dptr->tdesc->table;

	return(dptr);
}

db_status remove_table(table_name)
	char			*table_name;
{
	char			*path;
	char			lpath[NIS_MAXNAMELEN];


	if ((path = get_path_from_catalog(table_name)) == NULL)
		return(DB_NOTFOUND);

	snprintf(lpath, sizeof(lpath), "%s.log", path);

	if (update_catalog(table_name, NULL, CAT_DEL) != DB_SUCCESS)
		return(DB_INTERNAL_ERROR);

	unlink(path);
	unlink(lpath);

	return(DB_SUCCESS);
}

/*
 * Cache a table handle in memory. Note that table_obj points to
 * externally allocated memory, thus we need to make our own copy of
 * it for the cache.
 */

db_status cache_table(table_name, tdesc)
	char			*table_name;
	table_desc		*tdesc;
{
	struct nis_dbent	*dbptr;
	unsigned int		i;
	char			*path;


	if ((dbptr = nis_malloc_qent(tdesc)) == NULL) {
		return (DB_INTERNAL_ERROR);
	}

	if ((path = get_path_from_catalog(table_name)) == NULL) {
		nis_free_qent(dbptr);
		return(DB_NOTFOUND);
	}

	if (xff_open(dbptr, path) == DB_NOTFOUND) {
		if (tdesc == NULL) {
			nis_free_qent(dbptr);
			return(DB_NOTFOUND);
		} else
			xff_create(dbptr, path, 1);
	}

	if (tdesc == NULL) {
		if (get_dictionary(table_name, &tdesc) != DB_SUCCESS) {
			nis_free_qent(dbptr);
			return(DB_NOTFOUND);
		}
		dbptr->table = tdesc->table;
		dbptr->table_name = tdesc->table_name;
		dbptr->tdesc = tdesc;
		dbptr->index_hashes =
		    (uint32_t *)tdesc->index_hashes.index_hashes_val;
	}

	for (i = 0; i < tdesc->table->ta_cols.ta_cols_len; i++)
		if (tdesc->table->ta_cols.ta_cols_val[i].tc_flags
		    & TA_SEARCHABLE)
			dbptr->searchable++;
	
	t_insert(tables, table_name, dbptr, NULL);
	numdbs++;

	rcm_load(dbptr);
	if (dbptr->hdr.log_present)
		xff_replay_log(dbptr, 0);

	return(DB_SUCCESS);
}

db_status validate_attrs(dptr, numattrs, attrs)
	struct nis_dbent	*dptr;
	uint32_t		numattrs;
	nis_attr		*attrs;
{
	uint32_t		i, j, hits;
	register table_col	*tcol = dptr->table->ta_cols.ta_cols_val;
	uint32_t		*hashes;

	if ((numattrs && attrs == NULL) || numattrs > dptr->searchable)
		return(DB_BADQUERY);

	hashes = calloc(ULONGSZ, numattrs);

	for (i = 0; i < numattrs; i++) {
		if (!attrs[i].zattr_val.zattr_val_len) {
			free(hashes);
			return(DB_NOTFOUND);
		}
		hashes[i] = hash(attrs[i].zattr_ndx,
		    strlen(attrs[i].zattr_ndx), HASH_NORM);
	}

	hits = 0;
	for (i = 0; i < dptr->table->ta_cols.ta_cols_len; i++) {
		if (tcol[i].tc_flags & TA_SEARCHABLE) {
			for(j = 0; j < numattrs; j++) {
				if (hashes[j] == dptr->index_hashes[i]) {
					hits++;
					break;
				}
			}
		}
	}

	free(hashes);

	if (hits != numattrs)
		return(DB_BADQUERY);

	return(DB_SUCCESS);
}

unsigned char __nis_db_cksum(data, len)
	unsigned char		*data;
	unsigned int		len;
{
	register int		i;
	unsigned char		cksum = 0;

	for (i = 0; i < len; i++)
		cksum ^= data[i];

	return(cksum);
}

db_status find_table_in_catalog(table_name)
	char			*table_name;
{
	nis_datum		data = { NULL, 0 };
	uint32_t		hkey;

	hkey = hash(table_name, strlen(table_name), HASH_NORM);

	if (rcm_get_index(&catalog_ptr, 0, hkey, &data)!= DB_SUCCESS)
		return(DB_NOTFOUND);

	return(DB_SUCCESS);
}

struct nis_dbent *find_table(table_name)
	char			*table_name;
{
	struct nis_dbent	*dptr;

	if (!strcmp(table_name, "dictionary_tbl"))
		return(&catalog_ptr);

	if ((dptr = t_lookup(tables, table_name)) == NULL) {
		/* Maybe it isn't loaded yet. */
		if (cache_table(table_name, NULL) != DB_SUCCESS)
			return(NULL);
		else
			dptr = t_lookup(tables, table_name);
	}

	return(dptr);
}

db_status flush_table(table_name)
	char			*table_name;
{
	struct nis_dbent	*dptr = NULL;

	if ((dptr = t_lookup(tables, table_name)) == NULL)
		return(DB_BADTABLE);

	t_delete(tables, table_name);
	nis_free_qent(dptr);
	numdbs--;

	return(DB_SUCCESS);
}

db_status sync_all_tables()
{
	table_desc		*tdesc;
	db_status		stat, final_stat = DB_SUCCESS;
	struct nis_dbent	*dptr;

	catalog_ptr.pos = 0;
	while(enum_dictionary(&tdesc) == DB_SUCCESS) {
		if (!strcmp(tdesc->table_name, "nisdir"))
			continue;
		if ((dptr = t_lookup(tables, tdesc->table_name)) == NULL) {
			if (cache_table(tdesc->table_name, NULL) != DB_SUCCESS)
				return(DB_NOTFOUND);
			else
				dptr = t_lookup(tables, tdesc->table_name);
		}
		stat = xff_commit_log(dptr);

		flush_table(tdesc->table_name);
		xdr_free((xdrproc_t)xdr_table_desc, (char *)tdesc);
		free(tdesc);
		if (stat != DB_SUCCESS)
			final_stat = stat;
		if (dptr->fp != NULL)
			fflush(dptr->fp);

	}

	if (xff_commit_dict_log(&catalog_ptr) != DB_SUCCESS)
		return(DB_INTERNAL_ERROR);

	return(final_stat);
}

db_status sync_table(table_name)
	char			*table_name;
{
	struct nis_dbent	*dptr;

	if ((dptr = t_lookup(tables, table_name)) == NULL) {
		if (cache_table(table_name, NULL) != DB_SUCCESS)
			return(DB_NOTFOUND);
		else
			dptr = t_lookup(tables, table_name);
	}

	if (xff_commit_log(dptr) != DB_SUCCESS)
		return(DB_INTERNAL_ERROR);

	if (xff_commit_dict_log(&catalog_ptr) != DB_SUCCESS)
		return(DB_INTERNAL_ERROR);

	if (dptr->fp != NULL)
		fflush(dptr->fp);

	flush_table(table_name);

	return(DB_SUCCESS);
}

/*
 * Close DB handles but keep table_obj in memory
 */

db_status standby_table(table_name)
	char			*table_name;
{
	struct nis_dbent	*dptr;

	if ((dptr = t_lookup(tables, table_name)) == NULL)
		return(DB_BADTABLE);
	if (dptr->fp || dptr->fp_log) {
		xff_close(dptr);
		dptr->fp = NULL;
	}

	return(DB_SUCCESS);
}

/*
 * Re-open handles on a table that's been told to stand by.
 */
db_status unstandby_table(dptr)
	struct nis_dbent	*dptr;
{
	char			*path;

	if (dptr == NULL)
		return(DB_BADQUERY);

	if (dptr->fp == NULL && dptr->fp_log == NULL) {
		if ((path = get_path_from_catalog(dptr->table_name)) == NULL)
			return(DB_BADTABLE);
		if (xff_open(dptr, path) != DB_SUCCESS)
			return(DB_INTERNAL_ERROR);
	}

	return(DB_SUCCESS);
}

/* XXX */
db_status adjust_pos(table_name, losers)
	char			*table_name;
	struct loserlist	*losers;
{
	struct nis_dbent *dptr = NULL;
	struct loserlist *tmp;

	dptr = t_lookup(tables, table_name);

	if (dptr == NULL)
		return(DB_INTERNAL_ERROR);

	tmp = losers;
	while(tmp) {
		if (dptr->pos) {
			if (dptr->pos >= tmp->loser)
				dptr->pos -= tmp->size;
			fseek(dptr->fp, dptr->pos, SEEK_SET);
		}
		tmp = tmp->next;
	}

	return(DB_SUCCESS);
}

static db_status reload_from_catalog(path)
	char			*path;
{
	table_desc		*tdesc;

	if (catalog_ptr.hdr.vers != known_versions[0].val &&
	    catalog_ptr.hdr_log.vers != known_versions[0].val)
		return(DB_INTERNAL_ERROR);

	catalog_ptr.pos = 0;
	while(enum_dictionary(&tdesc) == DB_SUCCESS) {
		if (!strcmp(tdesc->table_name, "nisdir")) {
			snprintf(nisdir, sizeof(nisdir), "%s",
			    tdesc->table_path);
		}
#ifdef foo
		} else {
#ifdef tmp
			if (access(tdesc->table_path, R_OK|W_OK))
				return(DB_INTERNAL_ERROR);
#endif
			cache_table(tdesc->table_name, NULL);
		}
#endif
		xdr_free((xdrproc_t)xdr_table_desc, (char *)tdesc);
		free(tdesc);
	}

	return(DB_SUCCESS);
}

db_status init_cache(catalog)
	char			*catalog;
{
	struct rlimit		rlp;
	char			*curdir;
	int			cold_start = 0;
	struct versions		*verp = (struct versions *)&known_versions;
	table_desc		tdesc, *tptr = NULL;
	static uint32_t		keynum;
	static char		cbuf[NIS_MAXNAMELEN];

	if (getrlimit(RLIMIT_NOFILE, &rlp) == -1)
		return(DB_INTERNAL_ERROR);

	rlp.rlim_cur = rlp.rlim_max;

	if (setrlimit(RLIMIT_NOFILE, &rlp) == -1)
		return(DB_INTERNAL_ERROR);

	/* See if catalog exists */
	if (access(catalog, R_OK|W_OK) == -1) {
		if (errno != ENOENT) {
			syslog(LOG_ERR, "init: couldn't open %s: %m", catalog);
			return(DB_INTERNAL_ERROR);
		} else
			snprintf(cbuf, sizeof(cbuf), "%s.log", catalog);
			if (access(cbuf, R_OK|W_OK) == -1) {
				if (errno != ENOENT) {
					syslog(LOG_ERR,
					    "init: couldn't open %s: %m",
					    catalog);
					return(DB_INTERNAL_ERROR);
				} else
					cold_start = 1;
			}
	}

	snprintf(cbuf, sizeof(cbuf), "%s", catalog);

	if (cold_start) {
		if (xff_create(&catalog_ptr, catalog, 1) != DB_SUCCESS) {
			syslog(LOG_ERR, "couldn't create %s: %m", catalog);
			return(DB_INTERNAL_ERROR);
		}
		rcm_load(&catalog_ptr);
	} else {
		if (xff_open(&catalog_ptr, catalog) != DB_SUCCESS) {
			syslog(LOG_ERR, "unknown catalog format -- "
			    "expected %s", known_versions[0].string);
			return(DB_INTERNAL_ERROR);
		}
		rcm_load(&catalog_ptr);
		if (catalog_ptr.hdr.log_present)
			xff_replay_dict_log(&catalog_ptr, 0);
	}

	keynum = hash("table_descriptor",
	    strlen("table_descriptor"), HASH_NORM);

	catalog_ptr.index_hashes = &keynum;
	catalog_ptr.tdesc->table_path = (char *)&cbuf;

	if (catalog_ptr.hdr.vers != known_versions[0].val &&
	    catalog_ptr.hdr_log.vers != known_versions[0].val) {
		while(verp->val) {
			if (verp->val == catalog_ptr.hdr.vers) {
				syslog(LOG_ERR, "version mismatch -- "
				    "expected %s got %s\n",
				    known_versions[0].string, verp->string);
				return(1);
			}
			verp++;
		}
		syslog(LOG_ERR, "unknown version tag %08X -- expecting: %s",
		    (u_int)catalog_ptr.hdr.vers, known_versions[0].string);
	}

	if (cold_start) {
		bzero((char *)&tdesc, sizeof(table_desc));
		if ((curdir = getcwd((char *)&nisdir, MAXPATHLEN)) == NULL)
			return(DB_INTERNAL_ERROR);
		tdesc.table_path = (char *)&nisdir;
		tdesc.table_name = "nisdir";
		tdesc.table = &catalog_table;
		tdesc.vers = known_versions[0].val;
		tdesc.index_hashes.index_hashes_val = (char *)&keynum;
		tdesc.index_hashes.index_hashes_len = 1;
		if (put_dictionary("nisdir", &tdesc) != DB_SUCCESS)
			return(DB_INTERNAL_ERROR);
	} else {
		if (get_dictionary("nisdir", &tptr) != DB_SUCCESS)
			return(DB_INTERNAL_ERROR);
		snprintf(nisdir, sizeof(nisdir), "%s", tptr->table_path);
		xdr_free((xdrproc_t)xdr_table_desc, (char *)tptr);
		free(tptr);
	}

	bzero((char *)&tables, sizeof(tables));
	numdbs = 0;

	if (reload_from_catalog(catalog) != DB_SUCCESS)
		return(DB_INTERNAL_ERROR);
	return(DB_SUCCESS);
}

static char *get_path_from_catalog(table_name)
	char			*table_name;
{
	table_desc		*tdesc;
	static char		path[NIS_MAXNAMELEN];

	if (get_dictionary(table_name, &tdesc) != DB_SUCCESS)
		return(NULL);

	snprintf(path, sizeof(path), "%s", tdesc->table_path);

	xdr_free((xdrproc_t)xdr_table_desc, (char *)tdesc);
	free(tdesc);

	return((char *)&path);
}

db_status update_catalog(table_name, tdesc, action)
	char			*table_name;
	table_desc		*tdesc;
	int			action;
{
	nis_datum		key = { NULL, 0 };
	char			tbuf[NIS_MAXNAMELEN];
	struct stat		statbuf;
	int			rval;

	snprintf(tbuf, sizeof(tbuf), "%s/%s", nisdir, table_name);
	key.data = table_name;
	key.size = strlen(key.data);

	rval = stat(tbuf, &statbuf);
	if (rval == -1 && action == CAT_DEL) {
		snprintf(tbuf, sizeof(tbuf), "%s/%s.log", nisdir, table_name);
		rval = stat(tbuf, &statbuf);
	}
	if (action == CAT_DEL) {
		if (rval == -1 || del_dictionary(table_name) != DB_SUCCESS)
			return(DB_NOTFOUND);
	} else {
		if (rval == 0)
			return(DB_BADQUERY);

		tdesc->table_path = strdup(tbuf);
		put_dictionary(table_name, tdesc);
/*		if (catalog_ptr.hdr.type != XFF_LOG)
			xff_compact_file(&catalog_ptr); */
	}

	return(DB_SUCCESS);
}
