/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <rpcsvc/nis_db.h>
#include "nis_db_private.h"

#ifndef lint
__RCSID("@(#) $FreeBSD$");
#endif

/*
 * This module actually illustrates most of the magic used in this
 * library. We retrieve a list of entries from the database in the
 * following manner:
 *
 * - We take each attribute and query the database for it. This means
 *   we actually have records called [name=foo], [uid=0] and so on, so
 *   we can actually figure out a) if any entries with matching attributes
 *   actually exist, and b) if they do exist, we know what entries are
 *   associated with which attributes.
 *
 * - If for example we perform a search for [uid=0], we get back a set
 *   of hash values (uint32_ts) that refer to a list of entries. If
 *   we also ask for [gid=0], we'll get two lists.
 *
 * - Once we have the list of hash values that satisfy the query, we
 *   actually go and retrieve the entries associated with them and return
 *   them to the caller.
 *
 */

struct objects {
	uint32_t objects_len;
	entry_object_p *objects_val;
};

static uint32_t col;

static int setcase(hval, table, hashes)
	uint32_t		hval;
	table_obj		*table;
	uint32_t		*hashes;
{
	for (col = 0; col < table->ta_cols.ta_cols_len; col++) {
		if (!(table->ta_cols.ta_cols_val[col].tc_flags &
		    TA_SEARCHABLE))
			continue;
		if (hashes[col] != hval)
			continue;
		if (table->ta_cols.ta_cols_val[col].tc_flags & TA_CASE)
			return(HASH_UPPER);
		else
			return(HASH_NORM);
	}
	return(HASH_UPPER);
}

/*
 * The arrays used by this function look like this:
 *
 * [ index 0 ] -> [ addr of tag list ]
 * [ index 1 ] -> [ addr of tag list ]
 * 
 * A tag list looks like this:
 *
 * [ # of tags ] [ tag 1 ] [ tag 2 ] [ tag n ]
 */

db_status nis_db_parse_rels(numattrs, attrs, dptr, w)
	uint32_t		numattrs;
	nis_attr		*attrs;
	struct nis_dbent	*dptr;
	uint32_t		**w;
{
	uint32_t		i, j = 0, ind = 0;
	nis_datum		data = { NULL, 0 };
	uint32_t		*winners;
	uint32_t		*b = NULL;
	uint32_t		smallest = 0, size = 0;
	uint32_t		relbuf[2];
	uint32_t		*datap;
	uint32_t		*keys[NIS_MAXCOLUMNS], *count[NIS_MAXCOLUMNS];
	uint32_t		mods;
	int			done = 0;

	/*
	 * Read the relations out of the database and create tag lists
	 * from them. (We can't go slinging around pointers to the memory
	 * returned by the RCM package: we don't own it and it could
	 * vanish out from under us later.)
	 */
	for (i = 0; i < numattrs; i++) {

		relbuf[0] = hash(attrs[i].zattr_ndx,
		    strlen(attrs[i].zattr_ndx), HASH_NORM);
		relbuf[1] = hash(attrs[i].zattr_val.zattr_val_val,
		    attrs[i].zattr_val.zattr_val_len,
		    setcase(relbuf[0], dptr->table, dptr->index_hashes));

		if (rcm_get_index(dptr, col, relbuf[1], &data) != DB_SUCCESS) {
			/* Carefully free arrays we may have allocated. */
			for(j = 0; j < i; j++) {
				if (count[j] != NULL)
					free(count[j]);
			}
			return(DB_NOTFOUND);
		}
		if ((b = calloc(1, data.size + ULONGSZ)) == NULL) {
			/* Carefully free arrays we may have allocated. */
			for(j = 0; j < i; j++) {
				if (count[j] != NULL)
					free(count[j]);
			}
			return(DB_MEMORY_LIMIT);
		}

		*b = data.size / ULONGSZ;
		if (*b < size) {
			size = *b;
			smallest = i;
		}
		/*
		 * Magic! 'b' is of type uint32_t, so b+1 puts us four
		 * bytes ahead in memory! But data.data is char *,
		 * so data.data + 1 puts us only one byte ahead.
		 */
		/*
		 * This is much like a bcopy() except we're loading
		 * and storing 32-bit quantities instead of just bytes.
		 */
		datap = (uint32_t *)data.data;
		for(j = 0; j < *b; j++) {
			b[j+1] = datap[j];
		}
		count[i] = keys[i] = b;
		keys[i]++;
	}

	/*
	 * Special case: only one attribute was specified. We are liable
	 * to encounter many of these, so handle it quick and bail out.
	 */
	if (numattrs == 1) {
		*(caddr_t *)w = (caddr_t)b;
		return(DB_SUCCESS);
	}

	/*
	 * By definition, there can be no more than count[smallest][0]
	 * possible valid entries.
	 */

	if ((winners = calloc(count[smallest][0] + 1, ULONGSZ)) == NULL) {
		return(DB_MEMORY_LIMIT);
	}

	ind = 1;

#ifdef brute_force
	/*
	 * This is a straight brute-force search. We use the smallest tag
	 * list as a base. We check to see if the tags in the smallest list
	 * appear in all the other tag lists. If we find that a tag doesn't
	 * appear in one of the other relations' lists, then it gets
	 * thrown out.
	 */
	for (aptr = 0; aptr < *a[smallest]; aptr++) {
		hits = 0;
		for (j = 0; j < numattrs; j++) {
			if (j == smallest)
				continue;
			for (k = 0; k < *a[j]; k++) {
				if (a[j][k+1] == a[smallest][aptr+1]) {
					hits++;
					break;
				}
			}
		}
		if (hits == (numattrs - 1)) {
			winners[ind] = a[smallest][aptr+1];
			ind++;
		}
	}
#endif

	/*
	 * This is (I hope) a faster algorithm that the brute force one
	 * above.
	 */

	while(!done) {
		mods = 0;
		for (i = 0; i < numattrs - 1; i++) {
			if (keys[i][0] < keys[i+1][0]) {
				if (count[i][0] > 1) {
					count[i][0]--;
					keys[i]++;
					mods++;
					continue;
				} else {
					done++;
					break;
				}
			}
			if (keys[i+1][0] < keys[i][0]) {
				if (count[i+1][0] > 1) {
					count[i+1][0]--;
					keys[i+1]++;
					mods++;
					continue;
				} else {
					done++;
					break;
				}
			}
		}
		if (done)
			break;
		if (!mods) {
			winners[ind] = keys[0][0];
			ind++;
			for (i = 0; i < numattrs; i++) {
				if (count[i][0] > 1) {
					keys[i]++;
					count[i][0]--;
				} else {
					done++;
					break;
				}
			}
		}
	}

	/* Carefully free arrays we may have allocated. */
	for(j = 0; j < numattrs; j++) {
		if (count[j] != NULL)
			free(count[j]);
	}

	/*
	 * If there were no winners, the query has failed.
	 */
	if (ind == 1) {
		free(winners);
		return(DB_NOTFOUND);
	}

	*winners = (ind - 1);

	if (w != NULL)
		*(caddr_t *)w = (caddr_t)winners;

	return(DB_SUCCESS);
}

static db_status retrieve_entries(winners, ind, dptr, obj)
	uint32_t		*winners;
	uint32_t		ind;
	struct nis_dbent	*dptr;
	struct objects		*obj;
{
	uint32_t		i, j, k;
	entry_obj		**entry;
	db_status		r;

	/*
	 * Do the actual retrievals.
	 */
	if ((entry = calloc(ind, sizeof (entry_obj *))) == NULL) {
		free(winners);
		return(DB_MEMORY_LIMIT);
	}

	obj->objects_val = entry;
	obj->objects_len = ind;
	/*
	 * Magic! a[x]+1 puts us 4 bytes ahead in memory,
	 * not one.
	 */
	k = 0;
	for (i = 0; i < ind ; i ++) {
		if (!*(winners + i + 1))
			continue;
		if ((entry[k] = calloc(1, sizeof (entry_obj))) == NULL) {
			for (j = 0; j < k; j++)
				xdr_free((xdrproc_t)xdr_entry_obj, (char *)entry[j]);
			free(entry);
			free(winners);
			return(DB_MEMORY_LIMIT);
		}
		r = xff_get_entry(dptr, entry[k], winners[i + 1]);
		/*
		 * If xff_get_entry() returns DB_NOTFOUND, it means
		 * we're trying to look up a stale index that points
		 * do a deleted entry that hasn't been flushed by a
		 * file compaction yet. just skip it.
		 */
		if (r == DB_NOTFOUND) {
			obj->objects_len -= 1;
			free(entry[k]);
			if (!obj->objects_len) {
				free(entry);
				obj->objects_val = NULL;
				return(DB_NOTFOUND);
			}
			entry = realloc(entry, sizeof(entry_obj *) *
			    obj->objects_len);
			obj->objects_val = entry;
			continue;
		}
		if (xff_get_entry(dptr, entry[k],
		    winners[i + 1]) != DB_SUCCESS) {
			free(winners);
			for (j = 0; j < k; j++)
				xdr_free((xdrproc_t)xdr_entry_obj, (char *)entry[j]);
			free(entry);
			return(DB_INTERNAL_ERROR);
		}
		k++;
	}

	free(winners);
	return(DB_SUCCESS);
}


db_result *db_list_entries(table_name, numattrs, attrs)
	char			*table_name;
	int			numattrs;
	nis_attr		*attrs;
{
	db_result		*res;
	struct timeval		t_start, t_end;
	struct nis_dbent	*dptr;
	uint32_t		*w = NULL;
	struct objects		*obj;

	gettimeofday(&t_start, NULL);

	if ((res = calloc(1, sizeof(db_result))) == NULL)
		return(NULL);

	if (!numattrs || !attrs) {
		res->status = DB_BADQUERY;
		goto bail;
	}

	/*
	 * Pull the table handle out of the cache (will open
	 * the table if closed).
	 */
	if ((dptr = find_table(table_name)) == NULL) {
		res->status = DB_BADTABLE;
		goto bail;
	}


	/* Sanity check. */
	if ((res->status = validate_attrs(dptr,
	    numattrs, attrs)) != DB_SUCCESS)
		goto bail;

	if (dptr->fp == NULL && dptr->fp_log == NULL) {
		if (unstandby_table(dptr) != DB_SUCCESS) {
			res->status = DB_INTERNAL_ERROR;
			goto bail;
		}
	}

	if ((obj = calloc(1, sizeof(struct objects))) == NULL) {
		res->status = DB_MEMORY_LIMIT;
		goto bail;
	}

	if ((res->status = nis_db_parse_rels(numattrs,
	    attrs, dptr, &w)) == DB_SUCCESS &&
	    (res->status = retrieve_entries(w, *w,
	    dptr, obj)) == DB_SUCCESS) {
		res->objects.objects_len = obj->objects_len;
		res->objects.objects_val = obj->objects_val;

	}

	if (obj != NULL)
		free(obj);
bail:
	gettimeofday(&t_end, NULL);
	res->ticks += 
	    (t_end.tv_usec - t_start.tv_usec) +
	    ((t_end.tv_sec - t_start.tv_sec) * 1000000);

	return(res);
}
