/*
 * Copyright (c) 1996
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nis_db.h>
#include "nis_db_private.h"
#include "nis_db_struct.h"

#ifndef lint
__RCSID("@(#) $FreeBSD$");
#endif

/*
 * The Solaris nis_db(3n) man page says that the Structured Storage Manager
 * (SSM) is 'memory based.' This is another way of saying they weren't
 * clever enough to make a disk-based database work fast enough. Still,
 * if we're going to clone NIS+, we may as well be bug-for-bug compatible,
 * even on a design level.
 *
 * What's implemented here is the Relation Cache Manager (RCM), which
 * is a collection of functions that load, store, search and update the
 * relation index for a given table. There are also functions for
 * freeing up an index's memory and fixing up its various addresses
 * as part of an XDR File Format (XFF) file compaction.
 */

/*
 * What we really should do here is save each 'relation' structure (which
 * is small) instead of saving the entire populated array (which is big).
 */

/*
 * Could be handled in nis_free_qent(), but this keeps the
 * separation of modules cleaner.
 */
db_status rcm_free(dptr)
	struct nis_dbent	*dptr;
{
	uint32_t		i;

	for (i = 0; i < dptr->table->ta_cols.ta_cols_len; i++) {
		if (!(dptr->table->ta_cols.ta_cols_val[i].tc_flags &
		    TA_SEARCHABLE))
			continue;
		xdr_free((xdrproc_t)xdr_relation_array, (char *)dptr->relarray[i]);
		free(dptr->relarray[i]);
		dptr->relarray[i] = NULL;
	}


	free(dptr->relarray);
	dptr->relarray = NULL;

	return(DB_SUCCESS);
}

db_status rcm_dump(dptr)
	struct nis_dbent *dptr;
{
	register uint32_t i;
	XDR xdrs;

	if (dptr->relarray == NULL)
		return(DB_SUCCESS);

	fseek(dptr->fp, dptr->hdr.start_of_relcache, SEEK_SET);

	xdrstdio_create(&xdrs, dptr->fp, XDR_ENCODE);

	for (i = 0; i < dptr->table->ta_cols.ta_cols_len; i++) {
		if (!(dptr->table->ta_cols.ta_cols_val[i].tc_flags &
		    TA_SEARCHABLE))
			continue;
		/* Only dump the array if we dirtied it. */
/*		if (!dptr->relarray[i]->dirty)
			continue; */
			
		dptr->relarray[i]->dirty = 0;
		xdr_relation_array(&xdrs, dptr->relarray[i]);
	}

	fflush(dptr->fp);
	xdr_destroy(&xdrs);

	return(DB_SUCCESS);
}

db_status rcm_load(dptr)
	struct nis_dbent	*dptr;
{
	uint32_t		i;
	XDR			xdrs;

        if ((dptr->relarray = calloc(dptr->table->ta_cols.ta_cols_len,
	     sizeof(relation_array *))) == NULL)
                return(DB_MEMORY_LIMIT);

	if (dptr->fp != NULL) {
		fseek(dptr->fp, dptr->hdr.start_of_relcache, SEEK_SET);
		xdrstdio_create(&xdrs, dptr->fp, XDR_DECODE);
	}

	for (i = 0; i < dptr->table->ta_cols.ta_cols_len; i++) {
		if (!(dptr->table->ta_cols.ta_cols_val[i].tc_flags &
		    TA_SEARCHABLE))
			continue;
		dptr->relarray[i] = calloc(1, sizeof(relation_array));
		if (dptr->fp != NULL)
			xdr_relation_array(&xdrs, dptr->relarray[i]);
		else {
			dptr->relarray[i]->pagecount = PAGESPERARRAY;
			dptr->relarray[i]->dirty = 0;
		}
	}

	if (dptr->fp != NULL)
		xdr_destroy(&xdrs);

	return(DB_SUCCESS);
}

#define MASK 0x000000FF
#define KEYINDEX (keynum >> ( 8 * i )) & MASK
#define SLOT dptr->relarray[col]->page[i].relptr[KEYINDEX]
#define SMALLKEYINDEX (keynum >> ( 8 * smallslot )) & MASK
#define SMALLSLOT dptr->relarray[col]->page[smallslot].relptr[SMALLKEYINDEX]

db_status rcm_delete_index(dptr, col, keynum)
	struct nis_dbent	*dptr;
	uint32_t		col, keynum;
{
	uint32_t		i;
	relation		*rptr;
	relation		*prev, *next;

	for (i = 0; i < PAGESPERARRAY; i++) {
		if (SLOT) {
			rptr = prev = SLOT;
			while(rptr) {
				next = rptr->next;
				if (rptr->hashkey == keynum) {
					/* save the chaincount */
					if (next)
						next->chaincount =
						    rptr->chaincount;
					free(rptr->hashvals.hashvals_val);
					rptr->hashvals.hashvals_val = NULL;
					free(rptr);
					if (prev == SLOT)
						SLOT = next;
					else
						prev->next = next;
					dptr->relarray[col]->dirty++;
					return(DB_SUCCESS);
				}
				prev = rptr;
				rptr = rptr->next;
			}
		}
	}

	return(DB_NOTFOUND);
}
/*
 * See: isn't this much nicer than going through
 * Berkeley DB?
 */
db_status rcm_get_index(dptr, col, keynum, data)
	struct nis_dbent	*dptr;
	uint32_t		col, keynum;
	nis_datum		*data;
{
	register uint32_t i;
	relation *rptr;

	for (i = 0; i < PAGESPERARRAY; i++) {
		if (SLOT) {
			rptr = SLOT;
			while(rptr) {
				if (rptr->hashkey == keynum) {
					data->data =
					    rptr->hashvals.hashvals_val;
					data->size =
					    rptr->hashvals.hashvals_len *
					    ULONGSZ;
					return(DB_SUCCESS);
				}
				rptr = rptr->next;
			}
		}
	}

	return(DB_NOTFOUND);
}

#define CSLOT dptr->relarray[col]->page[i].relptr[j]

db_status rcm_compact(losers, dptr)
	struct loserlist	*losers;
	struct nis_dbent	*dptr;
{
	uint32_t		col, i, j, k;
	relation		*rptr;
	struct loserlist	*tmp;

	for (col = 0; col < dptr->table->ta_cols.ta_cols_len; col++) {
		if (!(dptr->table->ta_cols.ta_cols_val[col].tc_flags &
		    TA_SEARCHABLE))
			continue;
		for (i = 0; i < PAGESPERARRAY; i++) {
			for (j = 0; j < RELSPERPAGE; j++) {
				rptr = CSLOT;
				while(rptr) {
					for(k = 0;
					    k < (rptr->hashvals.hashvals_len);
					    k++) {
						tmp = losers;
						while(tmp) {
			if (rptr->hashvals.hashvals_val[k] >= tmp->loser)
				rptr->hashvals.hashvals_val[k] -= tmp->size;
							tmp = tmp->next;
						}
					}
					rptr = rptr->next;
				}		
			}
		}
	}

	return(DB_SUCCESS);
}

/*
 * Three things can happen here:
 *
 * 1) The relation we're given is new and there's a slot open on
 *    one of the pages to accomodate it.
 *
 * 2) The relation is a replacement, either for the entry in the slot
 *    or for one of the entries chained onto it.
 *
 * 3) The relation is new, and all the slots are filled. We need to
 *    chain.
 */
db_status rcm_put_index(dptr, col, keynum, data)
	struct nis_dbent	*dptr;
	uint32_t		col, keynum;
	nis_datum		*data;
{
	uint32_t		*w = NULL,
				*winners = (uint32_t *)data->data;
	relation		*rptr, *cptr;
	int			i;
	uint32_t		smallest = UINT_MAX, smallslot = 0;

	if ((w = calloc(1, data->size)) == NULL)
		return(DB_MEMORY_LIMIT);

	/* Faster than bcopy(). */
	for (i = 0; i < (data->size / ULONGSZ); i++)
		w[i] = winners[i];

	dptr->relarray[col]->dirty++;
	for (i = 0; i < PAGESPERARRAY; i++) {
		/*
		 * First possibility: there's a free slot in one of the
		 * pages. Just drop the relation right in. (This is an
		 * addition of a new relation).
	 	 */
		if (SLOT == NULL) {
			rptr = calloc(1, sizeof(relation));
			rptr->hashkey = keynum;
			rptr->hashvals.hashvals_len = data->size / ULONGSZ;
			rptr->hashvals.hashvals_val = w;
			SLOT = rptr;
			return(DB_SUCCESS);
		} else {
			if (SLOT->chaincount < smallest) {
				smallslot = i;
				smallest = SLOT->chaincount;
			}
		/*
		 * Second possibility: the slot we want is full -- this could
		 * be a replace of an existing relation: scan down the
		 * chain to make sure and do a replace if so.
	 	 */
			rptr = SLOT;
			while(rptr) {
				if (rptr->hashkey == keynum) {
					/* do a replace */
					free(rptr->hashvals.hashvals_val);

					rptr->hashvals.hashvals_val = w;
					rptr->hashvals.hashvals_len =
					     data->size / ULONGSZ;
					return(DB_SUCCESS);
				}
				rptr = rptr->next;
			}
		}
	}

	/*
	 * If we get here, then the slots in all four pages were
	 * occupied. Possibility three: this is an add and we have
	 * to chain from somewhere. Pick the page with the smallest
	 * chain.
	 */

	cptr = SMALLSLOT;

	if ((rptr = calloc(1, sizeof(relation))) == NULL) {
		free(w);
		return(DB_MEMORY_LIMIT);
	}

	rptr->hashkey = keynum;
	rptr->hashvals.hashvals_len = data->size / ULONGSZ;
	rptr->hashvals.hashvals_val = w;
	rptr->next = cptr;
	rptr->chaincount = cptr->chaincount;

	SMALLSLOT = rptr;
	SMALLSLOT->chaincount++;

	return(DB_SUCCESS);
}
