/*
 * Copyright (c) 1996
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <rpcsvc/nis_db.h>
#include "nis_db_private.h"

#ifndef lint
__RCSID("@(#) $FreeBSD$");
#endif

void nis_db_unmerge_rel(dptr, entry, numattrs, attrs, keynum)
	struct nis_dbent	*dptr;
	entry_obj		*entry;
	int			numattrs;
	nis_attr		*attrs;
	uint32_t		keynum;
{
	nis_datum		data = { NULL, 0 };
	uint32_t		i, j;
	uint32_t		*new, *nptr;
	uint32_t		hval;
	register uint32_t	*datap;
	uint32_t		ncnt = 0;

	for (i = 0; i < dptr->table->ta_cols.ta_cols_len &&
	    ncnt < numattrs; i++) {
		if (!(dptr->table->ta_cols.ta_cols_val[i].tc_flags &
		    TA_SEARCHABLE)) 
			continue;

		if (dptr->table->ta_cols.ta_cols_val[i].tc_flags & TA_CASE)
			hval = hash(attrs[ncnt].zattr_val.zattr_val_val,
			    attrs[ncnt].zattr_val.zattr_val_len, HASH_UPPER);
		else
			hval = hash(attrs[ncnt].zattr_val.zattr_val_val,
			    attrs[ncnt].zattr_val.zattr_val_len, HASH_NORM);

		if (rcm_get_index(dptr, i, hval, &data) != DB_SUCCESS)
			return;

		datap = (uint32_t *)data.data;

		/*
		 * Relation has only one tag -- nuke the whole
		 * thing and move on.
		 */
		if (data.size == sizeof(uint32_t)) {
			rcm_delete_index(dptr, i, hval);
			continue;
		}

		nptr = new = calloc(1,data.size - ULONGSZ);
		/* Relation has several tags -- nuke just the one tag. */
		for (j = 0; j < (data.size / ULONGSZ); j++) {
			if (*datap == keynum)
				datap++;
			else {
				*nptr = *datap;
				datap++; nptr++;
			}
		}

		data.size -= ULONGSZ;
		data.data = (char *)new;
		rcm_put_index(dptr, i, hval, &data);
		free(new);
		ncnt++;	
	}
	return;
}

db_result *db_remove_entry(table_name, numattrs, attrs)
	char			*table_name;
	int			numattrs;
	nis_attr		*attrs;
{
	db_result		*res;
	db_status		lstat;
	struct timeval		t_start, t_end;
	struct nis_dbent	*dptr;
	uint32_t		i;
	entry_obj		entry;
	uint32_t		*w;
	db_status		rval = DB_SUCCESS;

	gettimeofday(&t_start, NULL);

	if ((res = calloc(1, sizeof (db_result))) == NULL)
		return(NULL);

	/*
	 * Pull the table handle out of the cache (will open
	 * the table if closed).
	 */

	if ((dptr = find_table(table_name)) == NULL) {
		res->status = DB_BADTABLE;
		goto bail;
	}

	if ((res->status = validate_attrs(dptr, numattrs,
	    attrs)) != DB_SUCCESS)
		goto bail;

	if (dptr->fp == NULL && dptr->fp == NULL) {
		if (unstandby_table(dptr) != DB_SUCCESS) {
			res->status = DB_INTERNAL_ERROR;
			goto bail;
		}
	}

	lstat = nis_db_parse_rels(numattrs, attrs, dptr, &w);

	if (lstat != DB_SUCCESS) {
		res->status = lstat;
		goto bail;
	}

	for (i = 0; i < w[0]; i++) {
		bzero((char *)&entry, sizeof(entry_obj));
		rval = xff_get_entry(dptr, &entry, w[i+1]);
		if (rval == DB_NOTFOUND)
			continue;

		if (rval != DB_SUCCESS) {
			res->status = DB_BADTABLE;
			goto bail;
		}

		if (dptr->status == STATUS_NORMAL) {
			xff_add_log_entry(dptr, &entry, numattrs,
			    attrs, XLE_DELETE);
		}

		/*
		 * Update the relation indexes only after the log has
		 * has been updated: xff_add_log_entry() may need to
		 * consult the indexes.
		 */
		nis_db_unmerge_rel(dptr, &entry, numattrs, attrs, w[i+1]);
	}

bail:
	gettimeofday(&t_end, NULL);
	res->ticks +=
	    (t_end.tv_usec - t_start.tv_usec) +
	    ((t_end.tv_sec - t_start.tv_sec) * 1000000);

	return(res);
}
