/*
 * Copyright (c) 1996
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#ifndef RPC_HDR
%#ifndef lint
%__RCSID("@(#) $FreeBSD$");
%#endif /* not lint */
#endif

/*
 * This protocol definition file is only used to
 * generate some XDR functions. We don't actually
 * define any RPC services here.
 */
%#include <rpcsvc/nis.h>

const PAGESPERARRAY  = 4;
const RELSPERPAGE  = 256;

/* ----------------------------------------------------------------------- */

/*
 * This section describes the structure of a table descriptor. These
 * descriptors are stored in the catalog; by reading the catalog, we
 * can determine the nature of the table before we actually go to
 * open it.
 */

struct table_desc {
	table_obj	*table;
	string		table_name<NIS_MAXNAMELEN>;
	string		table_path<>;
	opaque		index_hashes<>;
	uint32_t	vers;
};

/* ----------------------------------------------------------------------- */

/*
 * This section describes the structures used by the relation cache.
 * We maintain all the information that describes the relations between
 * table entries in memory to speed up processing. This cache of relation
 * information is loaded whenever we open the table and dumped whenever
 * we close or checkpoint it (if it has changed).
 */

struct relation {
	uint32_t chaincount;
	uint32_t hashkey;
	uint32_t hashvals<>;
	struct relation *next;
};

typedef struct relation *relation_p;

struct relation_page {
	relation_p relptr[RELSPERPAGE];
	int dirty;
};

struct relation_array {
	relation_page page[PAGESPERARRAY];
	uint32_t pagecount;
	int dirty;
};

/* ----------------------------------------------------------------------- */

/*
 * This section describes the structures used in our table file storage
 * format. The first implementation of this library used Berkeley DB
 * as its underlying storage method, but this proved to be too slow
 * and inefficient (with respect to disk space usage). Berkeley DB is
 * fine by itself, but it gets messy when you try to use it for a
 * relational database.
 *
 * The new format actually (ab)uses the XDR library for storing data.
 * The XDR library is ideal for transforming data into an opaque,
 * machine-independent form and back again. All we have to do is
 * store our data using the XDR library as a filter. As a bonus,
 * the resulting file is automatically machine-independent, allowing
 * it to be read on any platform that supports RPC and/or the XDR
 * library.
 *
 * I strongly suspect that Sun used a similar technique for their
 * implementation, but I can't prove this since their XDR structure
 * definitions are not publically available.
 */

/*
 * Rpcgen and the XDR library don't understand 64-bit quantities
 * (long longs), so we have to teach them ourselves.
 */
struct longlong {
	unsigned char bytes[8];
};

enum xff_file_type {
	XFF_FILE = 0,
	XFF_LOG	 = 1,
	XFF_TMP  = 2
};

/*
 * XXX We're keeping our offsets at 32-bits for now. This should allow
 * us to manipulate files up to 4GB in size, which should be plenty.
 * The Sun NIS+ man pages indicate that NIS+ tables were designed to
 * hold around 10,000 entries; unless somebody creates tables with
 * massively large chunks of data, we should be okay.
 */
struct xff_header {
	uint32_t vers;			/* XFF version stamp */
	uint32_t log_present;		/* XFF file is shadowed by log */
	xff_file_type type;		/* file type: file, log or tmp */
	uint32_t table_size;		/* total file size */
	uint32_t start_of_entries; 	/* offset to first entry */
	uint32_t end_of_entries;	/* offset to end of last entry */
	uint32_t entry_just_added;	/* offset to entry just added */
	uint32_t start_of_relcache;	/* offset to start of relation data */
};

enum xle_status {
	XLE_PENDING = 0,
	XLE_COMMITTED = 1,
	XLE_OBSOLETE = 2
};

enum xle_action {
	XLE_ADD    = 0,
	XLE_UPDATE = 1,
	XLE_DELETE = 2
};

struct xff_log_entry {
	uint64_t	xle_time;	/* time log entry was created */
	xle_action	xle_action;	/* action being performed */
	xle_status	xle_status;	/* status of entry */
	nis_attr	xle_attrs<>;	/* supplied attributes */
	entry_obj 	*xle_entry;	/* supplied entry */
};
