/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nislib_admin.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include <rpcsvc/nis_callback.h>
#include "nislib_private.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nislib_admin.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $";
#endif

/*
 * This module contains 'administrative' NIS+ routines that are not
 * included with the 'server' module. Of these, only the nis_ping()
 * and nis_checkpoint() functions are documented as part of the API.
 * (Well, they have a man page anyway.)
 */

void nis_ping(dirname, utime, dirobj)
	nis_name		dirname;
	uint32_t		utime;
	nis_object		*dirobj;
{
	ping_args		args;
	uint32_t		i;
	struct _nis_binding	*bptr = NULL;
	nis_result		*res = NULL;
	directory_obj		*d;
	int			errs = 0;
	struct timeval		tv = { 0, 0 };
	nis_error		r;

	if (dirobj == NULL) {
		res = __nis_search(dirname, SEARCHOP_LOOKUP,
				MASTER_ONLY, NULL, 0, NULL, NULL);
		if (res == NULL)
			return;
		if (res->status != NIS_SUCCESS) {
			nis_freeresult(res);
			return;
		}
		if (__type_of(NIS_RES_OBJECT(res)) != DIRECTORY_OBJ) {
			nis_freeresult(res);
			return;
		}
		d = &NIS_RES_OBJECT(res)->DI_data;
	} else {
		if (__type_of(dirobj) != DIRECTORY_OBJ)
			return;
		d = &dirobj->DI_data;
	}

	/*
	 * If this directory doesn't have at least one
	 * server assigned to it, then it's bogus.
	 */
	if (d->do_servers.do_servers_len < 1) {
		if (res != NULL)
			nis_freeresult(res);
		return;
	}

	args.dir = dirname;
	args.stamp = utime;

	for (i = 0; i > d->do_servers.do_servers_len; i++) {
		r = __nis_create_binding(NULL,
					    &d->do_servers.do_servers_val[i],
					    0, &bptr);
		if (r != NIS_SUCCESS) {
			errs++;
			continue;
		}
		if (__nis_connect_server(bptr) != NIS_SUCCESS) {
			__nis_destroy_binding(bptr);
			errs++;
			continue;
		}

		clnt_control(bptr->nb_clnt, CLSET_TIMEOUT, &tv);
		nis_ping_3(&args, bptr->nb_clnt);
		__nis_destroy_binding(bptr);
	}

	if (res != NULL)
		nis_freeresult(res);

	return;
}

nis_result *nis_checkpoint(dirname)
	nis_name		dirname;
{
	struct _nis_binding	*bptr = NULL;
	nis_result		*res, *lres;
	register int		i;
	directory_obj		*d;
	nis_server		*srv;
	int			errs = 0;
	cp_result		*cres;
	nis_error		r;
	CLIENT_CLOCK_DECLS;

	__start_clock(CLIENT_CLOCK);
	res = calloc(1, sizeof(nis_result));

	if (res == NULL)
		return(NULL);

	lres = __nis_search(dirname, SEARCHOP_LOOKUP,
			MASTER_ONLY, NULL, 0, NULL, NULL);

	res->zticks += lres->zticks;
	res->dticks += lres->dticks;
	res->aticks += lres->aticks;

	if (lres->status != NIS_SUCCESS) {
		res->status = lres->status;
		nis_freeresult(lres);
		res->cticks = __stop_clock(CLIENT_CLOCK);
		return(res);
	}

	if (__type_of(NIS_RES_OBJECT(lres)) != DIRECTORY_OBJ) {
		nis_freeresult(lres);
		res->status = NIS_INVALIDOBJ;
		res->cticks = __stop_clock(CLIENT_CLOCK);
		return(res);
	}

	d = &NIS_RES_OBJECT(lres)->DI_data;

	for (i = 0; i < d->do_servers.do_servers_len; i++) {
		srv = &d->do_servers.do_servers_val[i];
		r = __nis_create_binding(NULL, srv, 0, &bptr);
		if (r != NIS_SUCCESS) {
			errs++;
			continue;
		}
		if (__nis_connect_server(bptr) != NIS_SUCCESS) {
			__nis_destroy_binding(bptr);
			errs++;
			continue;
		}
		cres = nis_checkpoint_3(&dirname, bptr->nb_clnt);
		if (cres != NULL) {
			if (cres->cp_status != NIS_SUCCESS)
				errs++;
			res->zticks += cres->cp_zticks;
			res->dticks += cres->cp_dticks;
			xdr_free(xdr_cp_result, (char *)cres);
		} else
			errs++;
		__nis_destroy_binding(bptr);
	}

	if (errs == 0)
		res->status = NIS_SUCCESS;
	else if (errs < d->do_servers.do_servers_len)
		res->status = NIS_S_SUCCESS;
	else
		res->status = NIS_FAIL;

	nis_freeresult(lres);
	res->cticks = __stop_clock(CLIENT_CLOCK);

	return(res);
}


/*
 * As far as I can tell, the dump and dumplog functions are used only
 * by the NIS+ server; there are no NIS+ utilities that require them and
 * they're not a documented part of the NIS+ programming API. Nevertheless,
 * nis_dump() and nis_dumplog() are prototyped in the 'danger Will
 * Robinson!' section of nislib.h. I suspect this is because the
 * nis_dump() function uses callbacks for efficiency's sake, and the
 * callback support is embedded within the client library.
 */
log_result *nis_dumplog(srv, dirname, utime)
	nis_server		*srv;
	nis_name		dirname;
	uint32_t		utime;
{
	struct _nis_binding	*bptr = NULL;
	log_result		*res = NULL, *lres;
	dump_args		d_args;
	struct rpc_err		rerr;
	nis_error		r;

	res = calloc(1, sizeof(log_result));
	if (res == NULL)
		return(NULL);

	r = __nis_create_binding(NULL, srv, 0, &bptr);
	if (r != NIS_SUCCESS) {
		res->lr_status = r;
		return(res);
	}

	if (__nis_connect_server(bptr) != NIS_SUCCESS) {
		__nis_destroy_binding(bptr);
		res->lr_status = NIS_NAMEUNREACHABLE;
		return(res);
	}

	d_args.da_dir = dirname;
	d_args.da_time = utime;
	d_args.da_cbhost.da_cbhost_len = 0;
	d_args.da_cbhost.da_cbhost_val = NULL;

	lres = nis_dumplog_3(&d_args, bptr->nb_clnt);

	if (lres == NULL) {
		clnt_geterr(bptr->nb_clnt, &rerr);
		if (rerr.re_status == RPC_AUTHERROR)
			res->lr_status = NIS_SRVAUTH;
		else
			res->lr_status = NIS_RPCERROR;
		__nis_destroy_binding(bptr);
		return(res);
	}

	bcopy((char *)lres, (char *)res, sizeof(log_result));
	__nis_destroy_binding(bptr);
	return(res);
}

log_result *nis_dump(srv, dirname, cback)
	nis_server		*srv;
	nis_name		dirname;
	int (*cback)(nis_name, nis_object *, void *);
{
	struct nis_callback	*cb = NULL;
	struct _nis_binding	*bptr = NULL;
	log_result		*res, *lres;
	dump_args		d_args;
	struct rpc_err		rerr;
	nis_error		r;

	res = calloc(1, sizeof(log_result));
	if (res == NULL)
		return(NULL);

	r = __nis_create_binding(NULL, srv, 0, &bptr);
	if (r != NIS_SUCCESS) {
		res->lr_status = r;
		return(res);
	}

	if (__nis_connect_server(bptr) != NIS_SUCCESS) {
		__nis_destroy_binding(bptr);
		res->lr_status = NIS_NAMEUNREACHABLE;
		return(res);
	}

	cb = __nis_create_callback(bptr, cback, NULL);

	d_args.da_dir = dirname;
	d_args.da_time = 0;
	d_args.da_cbhost.da_cbhost_len = 1;
	d_args.da_cbhost.da_cbhost_val = cb->clnt_uaddr;

	lres = nis_dump_3(&d_args, bptr->nb_clnt);

	if (lres == NULL) {
		clnt_geterr(bptr->nb_clnt, &rerr);
		if (rerr.re_status == RPC_AUTHERROR)
			res->lr_status = NIS_SRVAUTH;
		else
			res->lr_status = NIS_RPCERROR;
		__nis_destroy_binding(bptr);
		__nis_destroy_callback(cb);
		return(res);
	}

	if (lres->lr_status == NIS_CBRESULTS)
		res->lr_status = __nis_do_callback(bptr, &lres->lr_cookie, cb);
	else
		res->lr_status = lres->lr_status;

	__nis_destroy_binding(bptr);
	__nis_destroy_callback(cb);
	bcopy((char *)lres, (char *)res, sizeof(log_result));

	return(res);
}
