/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nislib_bind.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/cdefs.h>
#include <sys/fcntl.h>
#include <sys/queue.h>

#include <netinet/in.h>
#include <netinet/tcp.h>

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <syslog.h>

#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include <rpcsvc/nis_tags.h>
#include "nislib_private.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nislib_bind.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $";
#endif

#ifdef USE_SECURE_RPC
extern AUTH *authdes_pk_seccreate __P(( char *, netobj *, unsigned int,
					char *, des_block *, nis_server * ));
#endif

/*
 * This is the module that actually goes out and establishes RPC handles
 * to servers. Unlink NIS v2, we may use either UDP or TCP for any given
 * procedure. Also unlike v2, we do a lot of the work ourselves rather than
 * relying on ypbind(8) (or in this case, nis_cachemgr(8)).
 */

/*
 * The top five list.
 *
 * We try to keep a list of open CLIENT handles available for
 * servers that we know about. I chose five as the limit since
 * we don't want to eat up file descriptors within the client.
 *
 * We may have more than 5 binding though, since the user of a
 * binding may mark the structure as referenced; we can't delete
 * a referenced binding since that could lead to a NULL ptr
 * dereference. If all 5 bindings are in use and we still need
 * another, the list will be expanded.
 */
#ifndef NIS_SERV_CNT
#define NIS_SERV_CNT 5
#endif

/*
 * Circular queue definitions.
 */
#define CIRCLEQ_HEAD(name, type)					\
struct name {								\
	struct type *cqh_first;		/* first element */		\
	struct type *cqh_last;		/* last element */		\
}

#define CIRCLEQ_ENTRY(type)						\
struct {								\
	struct type *cqe_next;		/* next element */		\
	struct type *cqe_prev;		/* previous element */		\
}

/*
 * Circular queue functions.
 */
#define	CIRCLEQ_INIT(head) {						\
	(head)->cqh_first = (void *)(head);				\
	(head)->cqh_last = (void *)(head);				\
}

#define CIRCLEQ_INSERT_AFTER(head, listelm, elm, field) {		\
	(elm)->field.cqe_next = (listelm)->field.cqe_next;		\
	(elm)->field.cqe_prev = (listelm);				\
	if ((listelm)->field.cqe_next == (void *)(head))		\
		(head)->cqh_last = (elm);				\
	else								\
		(listelm)->field.cqe_next->field.cqe_prev = (elm);	\
	(listelm)->field.cqe_next = (elm);				\
}

#define CIRCLEQ_INSERT_BEFORE(head, listelm, elm, field) {		\
	(elm)->field.cqe_next = (listelm);				\
	(elm)->field.cqe_prev = (listelm)->field.cqe_prev;		\
	if ((listelm)->field.cqe_prev == (void *)(head))		\
		(head)->cqh_first = (elm);				\
	else								\
		(listelm)->field.cqe_prev->field.cqe_next = (elm);	\
	(listelm)->field.cqe_prev = (elm);				\
}

#define CIRCLEQ_INSERT_HEAD(head, elm, field) {				\
	(elm)->field.cqe_next = (head)->cqh_first;			\
	(elm)->field.cqe_prev = (void *)(head);				\
	if ((head)->cqh_last == (void *)(head))				\
		(head)->cqh_last = (elm);				\
	else								\
		(head)->cqh_first->field.cqe_prev = (elm);		\
	(head)->cqh_first = (elm);					\
}

#define CIRCLEQ_INSERT_TAIL(head, elm, field) {				\
	(elm)->field.cqe_next = (void *)(head);				\
	(elm)->field.cqe_prev = (head)->cqh_last;			\
	if ((head)->cqh_first == (void *)(head))			\
		(head)->cqh_first = (elm);				\
	else								\
		(head)->cqh_last->field.cqe_next = (elm);		\
	(head)->cqh_last = (elm);					\
}

#define	CIRCLEQ_REMOVE(head, elm, field) {				\
	if ((elm)->field.cqe_next == (void *)(head))			\
		(head)->cqh_last = (elm)->field.cqe_prev;		\
	else								\
		(elm)->field.cqe_next->field.cqe_prev =			\
		    (elm)->field.cqe_prev;				\
	if ((elm)->field.cqe_prev == (void *)(head))			\
		(head)->cqh_first = (elm)->field.cqe_next;		\
	else								\
		(elm)->field.cqe_prev->field.cqe_next =			\
		    (elm)->field.cqe_next;				\
}

static int __nis_srv_cnt = 0;

static CIRCLEQ_HEAD(bindhead, bindq_entry) bhead;

struct bindq_entry {
	nis_name		name;
	struct _nis_binding	*bptr;
	CIRCLEQ_ENTRY(bindq_entry) links;
};

/*
 * Initialize the circular binding queue. This is done only
 * once per process.
 */
static void __nis_init_bindings()
{
	static int		i = 0;

	if (!i) {
		CIRCLEQ_INIT(&bhead);
		i++;
	}
	return;
}

/*
 * Allocate an binding queue structure.
 */
static struct bindq_entry *__nis_alloc_binding()
{
	register struct bindq_entry	*q;

	q = calloc(1, sizeof(struct bindq_entry));

	if (q == NULL) {
		syslog(LOG_ERR, "NIS+: out of memory allocating binding");
		return(NULL);
	}

	return(q);
}

/*
 * Close down a binding and free its memory.
 */
void __nis_destroy_binding(bptr)
	struct _nis_binding	*bptr;
{
	if (bptr == NULL)
		return;

	if (bptr->nb_name != NULL)
		free(bptr->nb_name);

	if (bptr->nb_clnt != NULL) {
		if (bptr->nb_clnt->cl_auth != NULL) {
			auth_destroy(bptr->nb_clnt->cl_auth);
			bptr->nb_pkey = NULL;
		}
		clnt_destroy(bptr->nb_clnt);
		bptr->nb_clnt = NULL;
	}

	if (bptr->nb_type == NB_TYPE_CACHED) {
		__nis_destroy_cache_binding(bptr);
	}

	if (bptr->nb_type == NB_TYPE_MANUAL) {
		struct manual_binding	*mb;

		mb = bptr->nb_binding_manual;
		if (mb->mb_type == MB_SERVER) {
			if (mb->mb_srv != NULL) {
				NIS_FREESRV(mb->mb_srv);
				mb->mb_srv = NULL;
			}
		} else {
			if (mb->mb_dir != NULL) {
				NIS_FREEDIR(mb->mb_dir);
				mb->mb_dir = NULL;
			}
		}
	}

	free(bptr);

	return;
}

/*
 * Destroy the binding at the head of the queue.
 */
void __nis_flush_current_binding()
{
	register struct bindq_entry *q;

	q = bhead.cqh_first;
	if (q == NULL || q == (void *)&bhead)
		return;

	if (q->bptr == NULL) {
		syslog(LOG_ERR, "NIS+: binding queue entry with no binding!");
		return;
	}

	/*
	 * Can't flush binding if it's still
	 * being referenced.
	 */
	if (q->bptr->nb_refcnt)
		return;

	CIRCLEQ_REMOVE(&bhead, q, links);
	__nis_destroy_binding(q->bptr);
	free(q->name);
	q->bptr = NULL;
	q->name = NULL;
	free(q);
	__nis_srv_cnt--;

	return;
}

/*
 * Destroy a binding anywhere in the queue, starting from
 * the tail and working up. (If a binding has refcnt > 0, we can't
 * destroy it).
 */
static nis_error __nis_flush_binding(bptr)
	struct _nis_binding	*bptr;
{
	register struct bindq_entry *q;

	for (q = bhead.cqh_last; q != (void *)&bhead; q = q->links.cqe_prev) {
		if (q->bptr == NULL) {
			syslog(LOG_ERR, "NIS+: queue entry with no binding!");
			continue;
		}

		/*
		 * Can't flush binding if it's still
		 * being referenced.
		 */
		if (q->bptr->nb_refcnt)
			continue;
		if (bptr == NULL)
			break;
		else if (bptr == q->bptr)
			break;
	}

	/*
	 * All busy; try again later.
	 */
	if (q == (void *)&bhead)
		return(NIS_NOTFOUND);

	CIRCLEQ_REMOVE(&bhead, q, links);
	__nis_destroy_binding(q->bptr);
	free(q->name);
	q->bptr = NULL;
	q->name = NULL;
	free(q);
	__nis_srv_cnt--;

	return(NIS_SUCCESS);
}

/*
 * Add a new binding to the queue.
 */
static nis_error __nis_insert_binding(bptr, name)
	struct _nis_binding	*bptr;
	nis_name		name;
{
	register struct bindq_entry *q;

	if (__nis_srv_cnt >= NIS_SERV_CNT)
		__nis_flush_binding(NULL);

	q = __nis_alloc_binding();
	if (q == NULL)
		return(NIS_NOMEMORY);
	q->name = strdup(name);
	q->bptr = bptr;
	CIRCLEQ_INSERT_HEAD(&bhead, q, links);
	__nis_srv_cnt++;

	return(NIS_SUCCESS);
}

/*
 * Check the queue for a binding for 'name.'
 */
static struct _nis_binding *__nis_lookup_binding(name, srv, flags)
	nis_name		name;
	nis_server		*srv;
	uint32_t		flags;
{
	register struct bindq_entry	*q;
	register struct manual_binding	*m;

	for (q = bhead.cqh_first; q != (void *)&bhead; q = q->links.cqe_next) {
		m = q->bptr->nb_binding_manual;
		if ((q->bptr->nb_flags & ZMH_VC) && (flags & USE_DGRAM))
			continue;
		if (!(q->bptr->nb_flags & ZMH_VC) && !(flags & USE_DGRAM))
			continue;
		if ((q->bptr->nb_flags & ZMH_AUTH) && (flags & NO_AUTHINFO))
			continue;
		if (!(q->bptr->nb_flags & ZMH_AUTH) && !(flags & NO_AUTHINFO))
			continue;
		if (m->mb_srv_idx > 0 && (flags & MASTER_ONLY))
			continue;
		if ((srv != NULL && !strcasecmp(srv->name,
		 m->mb_dir->do_servers.do_servers_val[m->mb_srv_idx].name)) ||
			nis_dir_cmp(m->mb_dir->do_name, name) == SAME_NAME ||
			nis_dir_cmp(q->name, name) == SAME_NAME) {
			if (q != bhead.cqh_first) {
				CIRCLEQ_REMOVE(&bhead, q, links);
				CIRCLEQ_INSERT_HEAD(&bhead, q, links);
			}
			return(q->bptr);
		}
	}

	return(NULL);
}

/*
 * Hard won lesson from YP: be vewy, vewy careful with the sockets in
 * the the RPC CLIENT handles.
 */
static int __nis_check_binding(srv)
	struct _nis_binding	*srv;
{
	struct sockaddr_in	check;
	socklen_t		checklen = sizeof(struct sockaddr_in);

	if (getsockname(srv->nb_sock, (struct sockaddr *)&check,
		&checklen) == -1 || check.sin_family != AF_INET ||
		check.sin_port != srv->nb_port) {
		int save, sock;

		/*
		 * Somebody has stomped all over our sockets. Nuke this
		 * connection so we can make a new one.
		 */
		sock = srv->nb_sock;
		save = dup(srv->nb_sock);
		if (srv->nb_clnt != NULL) {
			if (srv->nb_clnt->cl_auth != NULL)
				auth_destroy(srv->nb_clnt->cl_auth);
			clnt_destroy(srv->nb_clnt);
		}
		srv->nb_clnt = NULL;
		sock = dup2(save, sock);
		close(save);
		return(1);
	}
	return(0);
}

/*
 * Check all bindings in the queue.
 */
static void __nis_check_bindings()
{
	register struct bindq_entry	*q;

	for (q = bhead.cqh_first; q != (void *)&bhead; q = q->links.cqe_next) {
		if (__nis_check_binding(q->bptr))
			__nis_connect_server(q->bptr);
	}
}

/*
 * Destroy all bindings in the queue.
 */
void __nis_destroy_bindings()
{
	register uint32_t	i;

	for (i = 0; i < __nis_srv_cnt; i++)
		__nis_flush_binding(NULL);

	return;
}

/*
 * Create a new nis_binding.
 */
nis_error __nis_create_binding(dir, srv, flags, res)
	directory_obj		*dir;
	nis_server		*srv;
	uint32_t		flags;
	struct _nis_binding	**res;
{
	struct _nis_binding	*bptr;
	struct manual_binding	*m;

	/* sanity check: only one or the other */
	if ((dir != NULL && srv != NULL) || (dir == NULL && srv == NULL))
		return(NIS_INVALIDOBJ);

	/* Allocate an nis_binding struct and set it up */
	bptr = calloc(1, sizeof(struct _nis_binding));
	if (bptr == NULL)
		return(NIS_NOMEMORY);

	bptr->nb_binding_manual = calloc(1, sizeof(struct manual_binding));
	if (bptr->nb_binding_manual == NULL) {
		free(bptr);
		return(NIS_NOMEMORY);
	}
	m = bptr->nb_binding_manual;
	bptr->nb_type = NB_TYPE_MANUAL;

	/* set up connection flags */
	if (flags & USE_DGRAM)
		bptr->nb_flags |= ZMH_DG;
	else
		bptr->nb_flags |= ZMH_VC;
	if (!(flags & NO_AUTHINFO))
		bptr->nb_flags |= ZMH_AUTH;

	 /* populate binding struct. */

	if (srv != NULL)
		m->mb_srv = __nis_copy_srv(srv);
	if (dir != NULL)
		m->mb_dir = __nis_copy_dir(dir);
	m->mb_srv_idx = -1;
	m->mb_ep_idx = -1;

	if (dir == NULL)
		m->mb_type = MB_SERVER;
	else {
		if (flags & MASTER_ONLY)
			m->mb_type = MB_MASTER;
		else
			m->mb_type = MB_REPLICA;
	}

	__nis_choose_server(bptr, flags);

	if (m->mb_srv_idx == -1 && m->mb_ep_idx == -1) {
		__nis_destroy_binding(bptr);
		return(NIS_NAMEUNREACHABLE);
	}

	if (m->mb_srv->key_type == NIS_PK_DH)
		bptr->nb_pkey = &m->mb_srv->pkey;

	*res = bptr;
	return(NIS_SUCCESS);
}

void __nis_disqualify_endpoint(srv, epname)
	nis_server		*srv;
	char			*epname;
/*
 * Remove an endpoint from an ep list in an nis_server struct.
 */
{
}

/*
 * Remove a server from a directory_obj server list.
 *
 * XXX We can't delete the entire server struct from the directory
 * object since that could screw up the ordering: if we delete the
 * master server (at index 0) then we might mistake the replica that
 * comes after it as the master. Instead, just free the server name
 * and make the name pointer NULL to indicate that the entry is no
 * longer valid.
 */
void __nis_disqualify_server(dir, srvname)
	directory_obj		*dir;
	nis_name		srvname;
{
	register int		i, srvcnt;
	nis_server		*srvlist;

	srvlist = dir->do_servers.do_servers_val;
	srvcnt = dir->do_servers.do_servers_len;

	for (i = 0; i < srvcnt; i++) {
		if (nis_dir_cmp(srvlist[i].name, srvname) == SAME_NAME) {
			free(srvlist[i].name);
			srvlist[i].name = NULL;
			return;
		}
	}

	return;
}
/*
 * Remove a server from a directory_obj server list.
 */
void __nis_disqualify_server2(dir, srvname)
	directory_obj		*dir;
	nis_name		srvname;
{
	register int		i, newcnt, srvcnt;
	nis_server		*srvlist, *newlist;
	char			sname[NIS_MAXNAMELEN];

	strncpy(sname, srvname, NIS_MAXNAMELEN);

	if (!dir->do_servers.do_servers_len)
		return;

	srvlist = dir->do_servers.do_servers_val;
	srvcnt = dir->do_servers.do_servers_len;
	newcnt = 0;
	newlist = NULL;

	for (i = 0; i < srvcnt; i++) {
		if (nis_dir_cmp(srvlist[i].name, sname) == SAME_NAME) {
			xdr_free((xdrproc_t)xdr_nis_server, (char *)&srvlist[i]);
			bzero((char *)&srvlist[i], sizeof(nis_server));
		} else {
			if (newlist == NULL)
				newlist = calloc(1, sizeof(nis_server));
			else
				newlist = realloc(newlist, (i+1) *
						sizeof(nis_server));
			bcopy((char *)&srvlist[i],
			      (char *)&newlist[newcnt],
			      sizeof(nis_server));
			newcnt++;
		}
	}

	free(dir->do_servers.do_servers_val);
	dir->do_servers.do_servers_len = newcnt;
	dir->do_servers.do_servers_val = newlist;

	return;
}

/*
 * This function closes down the current server connection in the
 * binding struct and tries to select a new server.
 */
static nis_error __nis_kill_current_srv(bptr, flags)
	struct _nis_binding	*bptr;
	uint32_t		flags;
{
	struct manual_binding	*m;

	if (bptr != NULL) {
		if (bptr->nb_clnt->cl_auth != NULL) {
			auth_destroy(bptr->nb_clnt->cl_auth);
			bptr->nb_pkey = NULL;
		}
		clnt_destroy(bptr->nb_clnt);
		bptr->nb_clnt = NULL;
	}

	m = bptr->nb_binding_manual;

	if (m->mb_type == MB_MASTER || m->mb_type == MB_REPLICA) {
		if (m->mb_dir->do_servers.do_servers_len)
			__nis_disqualify_server(m->mb_dir,
		m->mb_dir->do_servers.do_servers_val[m->mb_srv_idx].name);
	} else {
		NIS_FREESRV(m->mb_srv);
	}

	m->mb_srv = NULL;
	m->mb_srv_idx = -1;
	m->mb_ep_idx = -1;

	return(NIS_SUCCESS);
}

/*
 * This function connects to the server identified by the 'cur' index
 * in the binding handle.
 */
nis_error __nis_connect_server(bptr)
	struct _nis_binding	*bptr;
{
	struct sockaddr_in	check;
	socklen_t		checklen;
	struct timeval		tv = { NIS_CLIENT_TIMEOUT, 0 };
	nis_server		*srv;
	struct manual_binding	*m;

	if (bptr == NULL)
		return(NIS_FAIL);

	m = bptr->nb_binding_manual;

	if (m->mb_dir != NULL)
		srv = &m->mb_dir->do_servers.do_servers_val[m->mb_srv_idx];
	else
		srv = m->mb_srv;

	bptr->nb_srv_addr.sin_port = 0;
	bptr->nb_sock = RPC_ANYSOCK;
	/*
	 * NIS+, unlike NIS v2, uses TCP for most of its communications.
	 * With NIS v2, it's not likely that you will ever have a key
	 * or datum that will result in a transfer larger than 8K, so
	 * using UDP makes sense in terms of efficnency (there's less
	 * protocol processing overhead). But with NIS+, it's not unusual
	 * to have large request and reply structs being shipped back
	 * and forth, particularly when using nis_list() without a
	 * callback. In cases where you need to exchange more than 8K
	 * of data, it makes more sense to use the TCP transport and
	 * the xdrrec record marking stuff.
	 */
	if (bptr->nb_flags & ZMH_DG)
		bptr->nb_clnt = clntudp_bufcreate(&bptr->nb_srv_addr, NIS_PROG,
					NIS_VERSION, tv, &bptr->nb_sock,
					512, 8192);
	else
		bptr->nb_clnt = clnttcp_create(&bptr->nb_srv_addr, NIS_PROG,
					NIS_VERSION, &bptr->nb_sock,
					0, 0);
	if (bptr->nb_clnt == NULL)
		return(NIS_RPCERROR);

	clnt_control(bptr->nb_clnt, CLSET_TIMEOUT, &tv);
	if (bptr->nb_flags & ZMH_AUTH) {
#ifdef USE_SECURE_RPC
		if (bptr->nb_pkey != NULL && key_secretkey_is_set()) {
			bptr->nb_clnt->cl_auth =
				authdes_pk_seccreate(srv->name, bptr->nb_pkey,
						  NIS_WINDOW, NULL, NULL, srv);
		} else
#endif
			bptr->nb_clnt->cl_auth = authunix_create_default();
	}

	/* Make sockets close on exec. */
	fcntl(bptr->nb_sock, F_SETFD, 1);

	/*
	 * Bind UDP sockets to a local port so we can do sanity
	 * checks later.
	 */
	checklen = sizeof(struct sockaddr_in);
	bzero((char *)&check, checklen);
	if (bptr->nb_flags & ZMH_DG)
		bind(bptr->nb_sock, (struct sockaddr *)&check, checklen);
	else {
		/*
		 * Turn off the Nagle algorithm for TCP connections.
		 * Like X, communication between NIS+ clients and the
		 * server must be prompt in order to provide decent
		 * interactive response. With the Nagle algorithm
		 * turned on, you sometimes get significant delays.
		 */
		int			nodelay = 1;
		setsockopt(bptr->nb_sock, IPPROTO_TCP, TCP_NODELAY,
				(void *)&nodelay, sizeof(nodelay));
	}
	check.sin_family = AF_INET;
	if (!getsockname(bptr->nb_sock,
			(struct sockaddr *)&check, &checklen))
		bptr->nb_port = check.sin_port;

	return(NIS_SUCCESS);
}

/*
 * Find a directory_obj for the directory specified by 'name.' This
 * is the place where we begin a search for a server that we don't know
 * anything about starting with only our cold start object (i.e. nothing
 * useful turned up in the cache). What we do depends on our cold start
 * directory. Say we have a cold start object for 'foo.bar.baz.' and we
 * want to find 'gronk.bletch.foo.bar.baz.' The first thing we do is ask
 * the server for foo.bar.baz. if it knows anything about bletch.foo.bar.baz.
 * If so, it'll tell us who serves bletch.foo.bar.baz. Then we ask the
 * server for bletch.foo.bar.baz about gronk.bletch.foo.bar.baz. If it
 * knows about gronk.bletch.foo.bar.baz., we're done.
 *
 * Now let's say our cold start directory is gronk.bletch.foo.bar.baz. and
 * we want to know about foo.bar.baz.; in this case, we ask our server about
 * its parent (bletch.foo.bar.baz.) and move up the chain instead of down.
 *
 * If we are gronk.bleen.foo.bar.baz. and we want to know about
 * gronk.bletch.foo.bar.baz, we have to move up the chain until we
 * get to foo.bar.baz., make a sudden U-turn, and then start moving
 * down the other branch.
 *
 * If our cold start directory is gronk.bletch.foo.bar.baz. and we want
 * to know about microsoft.com., then we're screwed. :)
 *
 * Note that each time we learn about a new directory, we add it to the
 * cache so we don't have to generate network traffic all over again
 * later.
 *
 */

static nis_error nis_finddir(dir, name, res)
	directory_obj		*dir;
	nis_name		name;
	directory_obj		**res;
{
	fd_result		*r;
	directory_obj		*d;
	XDR			xdrs;
	nis_error		rstat;

	r = nis_finddirectory(dir, name);

	if (r == NULL) {
		NIS_FREEDIR(dir);
		*res = NULL;
		return(NIS_NAMEUNREACHABLE);
	}

	rstat = r->status;

	switch(rstat) {
	case NIS_SUCCESS:
		/* Decode result from nis_finddirectory(). */
		d = calloc(1, sizeof(directory_obj));
		xdrmem_create(&xdrs, r->dir_data.dir_data_val,
			    r->dir_data.dir_data_len, XDR_DECODE);
		xdr_directory_obj(&xdrs, d);
		xdr_destroy(&xdrs);
		xdr_free((xdrproc_t)xdr_fd_result, (char *)r);
		NIS_FREEDIR(dir);
		*res = d;
		return(NIS_SUCCESS);
		break;
	case NIS_INVALIDOBJ:
	case NIS_BADOBJECT:
		/*
		 * An invalid object error here means we tried
		 * to do an NIS_FINDDIRECTORY on something that
	 	 * isn't a directory object. This means we already
		 * found the right directory and went past it,
		 * trying to query a table or some other object
		 * by mistake. In this case, we already have the
		 * directory we need, so just return a pointer to
		 * it.
		 */
		xdr_free((xdrproc_t)xdr_fd_result, (char *)r);
		*res = dir;
		return(NIS_SUCCESS);
		break;
	case NIS_NOSUCHNAME:
	case NIS_NOTFOUND:
	case NIS_NOT_ME:
	default:
		NIS_FREEDIR(dir);
		*res = NULL;
		return(rstat);
		break;
	}

	return(NIS_NOTFOUND);
}

/*
 * Case: name == foo.bar.baz  cold->do_name == gronk.bletch.foo.bar.baz
 */
static nis_error nis_scan_up(cold, name, flags, res)
	directory_obj		*cold;
	nis_name		name;
	uint32_t		flags;
	directory_obj		**res;
{
	nis_name		ptr;
	directory_obj		*dir;
	char			__buf[NIS_MAXNAMELEN];
	nis_error		r = NIS_FAIL;

	if (name == NULL)
		return(NIS_BADNAME);

	dir = __nis_copy_dir(cold);
	strncpy(__buf, cold->do_name, NIS_MAXNAMELEN);
	ptr = __buf;

	while(strcmp(ptr, name)) {
		ptr = nis_domain_of(ptr);
		if ((r = nis_finddir(dir, ptr, &dir)) != NIS_SUCCESS)
			break;
	}

	*res = dir;

	return(r);
}

/*
 * Case: name == gronk.bletch.foo.bar.baz. cold->do_name == foo.bar.baz.
 * Recursion! Aaaahh!
 */

static nis_error nis_scan_down(cold, name, flags, res)
	directory_obj		*cold;
	nis_name		name;
	uint32_t		flags;
	directory_obj		**res;
{
	nis_name		ptr;
	directory_obj		*dir, *new = NULL;
	nis_error		r = NIS_FAIL;

	if (name == NULL)
		return(NIS_BADNAME);

	dir = __nis_copy_dir(cold);

	while(nis_dir_cmp(name, dir->do_name) == LOWER_NAME) {
		ptr = nis_domain_of(name);
		if (nis_dir_cmp(dir->do_name, ptr) == SAME_NAME) {
			if ((r = nis_finddir(dir, name, &new)) != NIS_SUCCESS)
				return(r);
			dir = new;
			break;
		} else {
			r = nis_scan_down(dir, ptr, flags, &new);
			if (r != NIS_SUCCESS)
				break;
			NIS_FREEDIR(dir);
			dir = new;
		}
		*res = dir;
		/*return(r);*/
	}

	*res = dir;
	return(r);
}

/*
 * Case: name = gronk.bleen.foo.bar.baz
 *       cold->do_name = gronk.bletch.foo.bar.baz.
 */
static nis_error nis_scan_branch(cold, name, flags, res)
	directory_obj		*cold;
	nis_name		name;
	uint32_t		flags;
	directory_obj		**res;
{

	char			*nptr, *cptr;
	directory_obj		*dir;
	nis_error		r;

	if (name == NULL)
		return(NIS_BADNAME);

	nptr = (name + strlen(name));
	cptr = (cold->do_name + strlen(cold->do_name));

	while (*nptr == *cptr) {
		nptr--; cptr--;
	}

	/* No common root */
	if (strlen(cptr) <= 1)
		return(NIS_FOREIGNNS);

	while(*cptr != '.')
		cptr++;
	cptr++;

	r = nis_scan_up(cold, cptr, flags, &dir);
	if (r != NIS_SUCCESS)
		return(r);
	r = nis_scan_down(dir, name, flags, &dir);

	*res = dir;
	return(r);
}

/*
 * Top level server locator function. This function is exported
 * as it is used by the nis_server(3) module.
 */
nis_error __nis_find_server(name, flags, d)
	nis_name		name;
	uint32_t		flags;
	directory_obj		**d;
{
	directory_obj		*coldstart = NULL;
	nis_error		cerr, res;

	if ((cerr = nis_get_cold_start(&coldstart)) != NIS_SUCCESS) {
		if (cerr == NIS_UNAVAIL)
			syslog(LOG_ERR, "NIS+ not installed");
		return(NIS_UNAVAIL);
	}

	switch (nis_dir_cmp(name, coldstart->do_name)) {
	case SAME_NAME:
		/* Easy case. */
		*d = coldstart;
		return(NIS_SUCCESS);
		break;
	case HIGHER_NAME:
		res = nis_scan_up(coldstart, name, flags, d);
		break;
	case LOWER_NAME:
		res = nis_scan_down(coldstart, name, flags, d);
		break;
	case NOT_SEQUENTIAL:
		res = nis_scan_branch(coldstart, name, flags, d);
		break;
	case BAD_NAME:
	default:
		NIS_FREEDIR(coldstart);
		return(NIS_BADNAME);
	}

	NIS_FREEDIR(coldstart);

	return(res);
}

/*
 * Return an RPC CLIENT handle attached to a server that specifies the
 * given name, chosen based on the rules in the specified flags. This
 * is the top-level interface to all the junk in this file.
 */
nis_error __nis_bind(name, flags, res)
	nis_name		name;
	uint32_t		flags;
	struct _nis_binding	**res;
{
	directory_obj 		*d = NULL;
	struct _nis_binding	*bptr = NULL;
	static pid_t		pid = -1;
	static uid_t		uid = -1;
	pid_t			gpid;
	uid_t			guid;
	nis_error		r = NIS_FAIL;

	*res = NULL;

	__nis_init_bindings();
	/*
	 * If we fork()ed or changed uid, we need to flush all
	 * saved binding info and start over.
	 */
	gpid = getpid();
	guid = geteuid();
	if ((pid != -1 && pid != gpid) || (uid != -1 && uid != guid)) {
		__nis_destroy_bindings();
		/*__nis_cache_close(); XXX */
	}
	pid = gpid;
	uid = guid;
	/* Make sure our sockets are still sane. */
	__nis_check_bindings();

	/*
	 * First, see if we have a handle already open to a server
	 * that serves the specified directory.
	 */
	if ((bptr = __nis_lookup_binding(name, NULL, flags)) != NULL) {
		bptr->nb_refcnt++;
		*res = bptr;
		return(NIS_SUCCESS);
	}

	/*
	 * If we don't have a handle already open, check the
	 * cache to see if we have a directory object that refers
	 * to the specified directory name. This can fail either
	 * because there's no matching entry in the cache or because
	 * there is no /var/nis/NIS_SHARED_DIRCACHE file (i.e. the
	 * cache manager isn't running.
	 */
#ifdef CACHEBIND
	/*
	 * Consult the cache manager cache before trying anything
	 * else. Only do this is the caller did _not_ set the
	 * NO_CACHE flag, othwewise we will recurse. This is here
	 * because the cache manager itself uses this code, and
	 * we don't want it to get caught in a loop.
	 */
	if (!(flags & NO_CACHE)) {
		r = __nis_cachebind(name, flags, res);
		if (r == NIS_SUCCESS)
			return(r);
	}
#endif

	/*
	 * If the cache lookup didn't pan out, we must grovel around
	 * for the directory object ourselves. Note that this adds
	 * all directory objects encountered during the search to the
	 * directory object cache.
	 */
	if (d == NULL) {
		r = __nis_find_server(name, flags, &d);
		if (r != NIS_SUCCESS)
			return(r);
	}


	/*
	 * By this time, we should have a directory_obj that tells us
	 * how to contact the servers that serve the directory we want.
	 * We now need to choose one from the list.
	 */
	r = __nis_create_binding(d, NULL, flags, &bptr);
	NIS_FREEDIR(d);

	if (r != NIS_SUCCESS)
		return(r);

	/* We chose a server; now try to establish a binding. */
	if (__nis_connect_server(bptr) == NIS_SUCCESS) {
		__nis_insert_binding(bptr, name);
		bptr->nb_refcnt++;
		*res = bptr;
		return(NIS_SUCCESS);
	} else {
		__nis_destroy_binding(bptr);
	}

	return(NIS_NAMEUNREACHABLE);
}

nis_error __nis_release_binding(bptr)
	struct _nis_binding	*bptr;
{
	if (bptr == NULL)
		return(NIS_FAIL);

	bptr->nb_refcnt--;

	return(NIS_SUCCESS);
}

/*
 * Sometimes errors are generated that indicate we have bound to an
 * unsuitable server (like a replica that isn't finished updating yet).
 * In this case, we have to nuke the existing binding, remove the
 * bad server from our list and try again.
 */
nis_error __nis_rebind(bptr, flags)
	struct _nis_binding	*bptr;
	uint32_t		flags;
{
	if (bptr == NULL)
		return(NIS_FAIL);

#ifdef CACHEBIND
	/*
	 * Consult the cache manager cache before trying anything
	 * else. Only do this is the caller did _not_ set the
	 * NO_CACHE flag, othwewise we will recurse. This is here
	 * because the cache manager itself uses this code, and
	 * we don't want it to get caught in a loop.
	 */
	if ((!flags & NO_CACHE)) {
		nis_error		r;
		r = __nis_cacherebind(bptr, flags);
		if (r == NIS_SUCCESS)
			return(r);
	}
#endif

	if (__nis_kill_current_srv(bptr, flags) != NIS_SUCCESS) {
		return(NIS_NAMEUNREACHABLE);
	}

	__nis_choose_server(bptr, flags);

	if (bptr->nb_binding_manual->mb_srv_idx == -1) {
		return(NIS_NAMEUNREACHABLE);
	}

	/*bptr->cur = bptr->srv_idx;*/

	if ((bptr->nb_flags & ZMH_AUTH) && bptr->nb_binding_manual->mb_srv->key_type == NIS_PK_DH)
		bptr->nb_pkey = &bptr->nb_binding_manual->mb_srv->pkey;

	return(__nis_connect_server(bptr));
}
