/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nislib_cache_bind.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $
 */

#include <sys/types.h>
#include <sys/cdefs.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>

#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include <rpcsvc/nis_tags.h>
#include "nislib_private.h"
#include "nis_cache.h"
#include "nis_cache_bind.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nislib_cache_bind.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $";
#endif

/*
 * This is version 2 of the client side caching module. Version 1 was
 * totally scrapped after Sun decided to change the way caching works in
 * Solaris 2.6. In 2.5.1 and earlier versions, caching was done in
 * the following way:
 *
 * - A client would go through all the usual binding gruntwork to find
 *   the directory object that describes a domain that it wants.
 * - To avoid having to repeat the gruntwork later, the client process
 *   hands a copy of this directory object to the local nis_cachemgr
 *   process, which saves in the NIS_SHARED_DIRCACHE file where other
 *   processes can find it.
 * - Each directory object is supposed to contain a signature that
 *   nis_cachemgr can verify. The signature is provided by the server
 *   that supplied the object. This mechanism is meant to prevent
 *   client processes from poisoning the cache with bogus objects.
 *
 * It turns out however that the signature mechanism never worked:
 * Sun NIS+ servers never actually signed their directory objects and
 * nis_cachemgr never checked the signatures. I don't know why.
 *
 * Sun promised to fix this in Solaris 2.6. But rather than fixing the
 * problem with signatures and making them work, Sun changed the whole
 * binding strategy so that the signatures aren't necessary anymore.
 * In effect, nis_cachemgr is now just a glorified ypbind. The new
 * approach is:
 *
 * - The client asks nis_cachemgr to lookup a directory object.
 * - The cache manager finds the object, saves it in NIS_SHARED_DIRCACHE.
 * - The client process reads the object from the cache and goes about
 *   its merry way.
 *
 * This is exactly how ypbind works, except nis_cachemgr is supposed
 * to make some more educated decisions about which server to bind to.
 *
 * The external interface to this module is in the form of three
 * functions: __nis_cachebind(), __nis_cachebind_srv and __nis_cacherebind().
 * __nis_cachebind() accepts an nis_name to bind and a set of flags
 * (passed down from a higher level routine). __nis_cachebind_srv()
 * accepts an nis_server struct instead of a name. The goal here is to
 * keep this module as simple as possible. The real gruntwork happens
 * in nis_cachemgr, so we shouldn't try to get too fancy here.
 */

#ifndef CACHEMGRSOCK
#define CACHEMGRSOCK "/var/run/cachemgrsock"
#endif

#ifdef foo
#define NB_TYPE_CACHED		1
#define NB_TYPE_MANUAL		2

#define MB_MASTER		CB_MASTER
#define MB_REPLICA		CB_REPLICA
#define MB_SERVER		CB_SERVER

struct manual_binding {
	cb_type			mb_type; /* Binding type */
	directory_obj		*mb_dir; /* Directory object. */
	nis_server		*mb_srv; /* Server object */
	int			mb_srv_idx;/* Server index (if dir is present)*/
	int			mb_ep_idx; /* Current Endpoint index. */
};

struct _nis_binding {
	/*
	 * Binding name. This is the name of the object
	 * for which this binding applies. This may be a
	 * directory or server name.
	 */
	nis_name		nb_name;
	/*
	 * Type of binding. CACHE bindings come from
	 * the cache manager. MANUAL bindings are
	 * generated on the fly.
	 */
	int			nb_type;

	/* A _copy_ of binding info if this is a CACHE binding */
	cachebind_hash_entry	*nb_binding_cached;

	/* A _copy_ of binding info if this is a MANUAL binding */
	struct manual_binding	*nb_binding_manual;

	/*
	 * Per-binding connection info. This includes
	 * the RPC client handle, public key of the
	 * guy on the other side, reference count.
	 */
	CLIENT			*nb_clnt;
	uint32_t		nb_flags; /* DG, VC, AUTH */
	int			nb_sock;
	netobj			*nb_pkey;
	int			nb_refcnt;
	unsigned short		nb_port;  /* local port */
};
#endif

/*
 * Saved cache information. Loaded from /var/nis/NIS_SHARED_DIRCACHE.
 * In theory, it should be safe for this to be here: if the process
 * fork()s, the child will share the cache with the parent until
 * it notices the on-disk version has been updated, at which point
 * parent and child will flush their copies and load new ones.
 */
static cachebind_table *__nis_cacheptr = NULL;

/*
 * FreeBSD uses a special AF_LOCAL RPC transport for communicating with
 * nis_cachemgr. This has two advantages: 1) it prevents any processes
 * on remote hosts on the network from communicating with nis_cachemgr
 * and 2) it avoids all the protocol processing overhead of TCP/IP.
 */
static CLIENT *nis_cachemgr_open()
{
	CLIENT			*cache_clnt;

#ifdef __FreeBSD__
	cache_clnt = clnt_create(CACHEMGRSOCK, CACHEPROG, CACHE_VER_2, "unix");
#else
	cache_clnt = clnt_create("localhost", CACHEPROG, CACHE_VER_2, "tcp");
#endif

	return(cache_clnt);
}

static bool_t nis_cachemgr_up()
{
	CLIENT			*clnt;

	clnt = nis_cachemgr_open();

	if (clnt == NULL)
		return(FALSE);
	else {
		clnt_destroy(clnt);
		return(TRUE);
	}
}

static nis_error nis_cachemgr_bind_master(dirname)
	nis_name		dirname;
{
	CLIENT			*clnt;
	nis_error		*r;

	clnt = nis_cachemgr_open();
	if (clnt == NULL)
		return(NIS_FAIL);

	r = nis_cache_bind_master_2(&dirname, clnt);
	clnt_destroy(clnt);
	if (r == NULL)
		return(NIS_FAIL);

	return (*r);
}

static nis_error nis_cachemgr_bind_replica(dirname)
	nis_name		dirname;
{
	CLIENT			*clnt;
	nis_error		*r;

	clnt = nis_cachemgr_open();
	if (clnt == NULL)
		return(NIS_FAIL);

	r = nis_cache_bind_replica_2(&dirname, clnt);
	clnt_destroy(clnt);
	if (r == NULL)
		return(NIS_FAIL);

	return (*r);
}

static nis_error nis_cachemgr_bind_server(srv)
	nis_server		*srv;
{
	CLIENT			*clnt;
	nis_error		*r;
	bind_server_arg		req;

	req.srv = srv;
	req.nsrv = 1;

	clnt = nis_cachemgr_open();
	if (clnt == NULL)
		return(NIS_FAIL);

	r = nis_cache_bind_server_2(&req, clnt);
	clnt_destroy(clnt);
	if (r == NULL)
		return(NIS_FAIL);

	return (*r);
}

static nis_error nis_cachebind_init(fd)
	int			*fd;
{
#ifdef DEBUG
	if ((*fd = open(CACHEFILE, O_RDONLY)) == -1)
#else
	if ((*fd = open("/tmp/NIS_SHARED_DIRCACHE", O_RDONLY)) == -1)
#endif
		return(NIS_FAIL);
	return(NIS_SUCCESS);
}

static int nis_cachebind_lock(fd)
	int			fd;
{
	struct flock            lock;
	pid_t			pid;

	pid = getpid();
	lock.l_start = 0;
	lock.l_len = 0;
	lock.l_pid = pid;
	lock.l_type = F_RDLCK;
	lock.l_whence = SEEK_SET;

	if (fcntl(fd, F_SETLKW, &lock) == -1)
		return(-1);
	return(0);
}

static int nis_cachebind_unlock(fd)
	int			fd;
{
	struct flock            lock;

	lock.l_start = 0;
	lock.l_len = 0;
	lock.l_pid = getpid();
	lock.l_type = F_UNLCK;
	lock.l_whence = SEEK_SET;

	if (fcntl(fd, F_SETLK, &lock) == -1)
		return(-1);
	return(0);
}

static caddr_t nis_cachebind_map(fd, size)
	int			fd;
	size_t			*size;
{
	caddr_t			maddr;
	struct stat		st;

	fstat(fd, &st);
	*size = st.st_size;
	maddr = mmap(0, *size, PROT_READ, MAP_SHARED, fd, 0);
	if (maddr == MAP_FAILED)
		return(NULL);

	return(maddr);
}	
	
static void nis_cachebind_unmap(fd, maddr, size)
	int			fd;
	caddr_t			maddr;
	size_t			size;
{
	munmap(maddr, size);
	nis_cachebind_unlock(fd);
	return;
}

static void nis_cachebind_free(c)
	cachebind_combined	*c;
{
	xdr_free(xdr_cachebind_combined, (char *)c);
	free(c);

	return;
}

static cachebind_hash_entry *nis_cachebind_copy(h)
	cachebind_hash_entry	*h;
{
	cachebind_hash_entry	*n;
	caddr_t			addr;
	int			size;
	XDR			xdrs;

	size = xdr_sizeof(xdr_cachebind_hash_entry, (char *)h);
	addr = calloc(1, size);
	if (addr == NULL)
		return(NULL);

	xdrmem_create(&xdrs, addr, size, XDR_ENCODE);
	if (!xdr_cachebind_hash_entry(&xdrs, h)) {
		xdr_destroy(&xdrs);
		free(addr);
		return(NULL);
	}
	xdr_destroy(&xdrs);
	n = calloc(1, sizeof(cachebind_hash_entry));
	if (n == NULL) {
		free(addr);
		return(NULL);
	}
	xdrmem_create(&xdrs, addr, size, XDR_DECODE);
	if (!xdr_cachebind_hash_entry(&xdrs, n)) {
		xdr_destroy(&xdrs);
		free(addr);
		return(NULL);
	}
	xdr_destroy(&xdrs);
	free(addr);

	return(n);
}

/*
 * - Check to see if the cache file has been mapped.
 * - If not, map it, build the hash table, then return.
 * - If so, check to see if it's been modified.
 * - If no, return.
 * - If yes, see if the cache has changed size.
 * - If yes, remap the file.
 * - Rebuild our copy of the hash table and return.
 */
static nis_error nis_cachebind_load(ptr)
	void			**ptr;
{
	int			fd;
	caddr_t			maddr;
	size_t			size;
	XDR			xdrs;
	cachebind_combined	*c = NULL;

	/* Cache hasn't been opened yet. */
	if (__nis_cacheptr == NULL) {
		cachebind_table		*cbt;
		cachebind_header	*h;

		if (nis_cachebind_init(&fd) == NIS_FAIL)
			return(NIS_FAIL);
		if ((maddr = nis_cachebind_map(fd, &size)) == NULL) {
			close(fd);
			return(NIS_FAIL);
		}
		cbt = (cachebind_table *)maddr;
		h = &cbt->cb_header;
		if (h->cb_vers != CB_VERSION &&
			h->cb_magic != CB_HEADER_MAGIC) {
			nis_cachebind_unmap(fd, maddr, size);
			close(fd);
			return(NIS_FAIL);
		}
		__nis_cacheptr = (cachebind_table *)maddr;
		*ptr = __nis_cacheptr->cb_table;
		return(NIS_SUCCESS);
	} else {

	}

#ifdef foo
	/*
	 * If the cache is already loaded, see if the disk version
	 * is newer than the loaded version and reload if it is,
	 * otherwise just return a pointer to the loaded cache.
	 */
	if (__nis_cache != NULL) {
		struct stat		st;
		cachebind_header	*header;

		header = &__nis_cache->cachebind_header;
		fstat(fd, &st);
		if (st.st_mtimespec.tv_sec > header->cb_updatetime) {
			nis_cachebind_free(__nis_cache);
			__nis_cache = NULL;
		} else {
			close(fd);
			*cache_combined = __nis_cache;
			return(NIS_SUCCESS);
		}
	}

	c = calloc(1, sizeof(cachebind_combined));
	if (c == NULL) {
		close(fd);
		return(NIS_FAIL);
	}

	if ((maddr = nis_cachebind_map(fd, &size)) == NULL) {
		close(fd);
		return(NIS_FAIL);
	}

	if (nis_cachebind_lock(fd) == NIS_FAIL)
		return(NIS_FAIL);
	xdrmem_create(&xdrs, maddr, size, XDR_DECODE);
	if (!xdr_cachebind_combined(&xdrs, c)) {
		xdr_destroy(&xdrs);
		nis_cachebind_unlock(fd);
		nis_cachebind_unmap(fd, maddr);
		close(fd);
		return(NIS_FAIL);
	}
	xdr_destroy(&xdrs);
	nis_cachebind_unlock(fd);
	nis_cachebind_unmap(fd, maddr);
	close(fd);

	*cache_combined = __nis_cache = c;
	__nis_cache->cachebind_header.cb_updatetime = time(NULL);
#endif

	return(NIS_SUCCESS);
}

static u_int32_t hashkey(key)
	char			*key;
{
	if (key == NULL)
		return(-1);
	return(__nis_hash((void *)key, strlen(key)));
}

static nis_error nis_cachebind_tsearch(dirname, srv, type, match)
	nis_name		dirname;
	nis_server		*srv;
	cb_type			type;
	cachebind_hash_entry	**match;
{
	uint32_t		*table;
	cachebind_header	*header;
	cachebind_hash_entry	*cur = NULL;
	uint32_t		i = -1, idx;
	cachebind_table		*cbt;
	void			*ptr;

	if (nis_cachebind_load(&ptr) != NIS_SUCCESS)
		return(NIS_FAIL);

	cbt = ptr;
	table = cbt->cb_table;
#ifdef should_be_in_cachebind_load
	header = cbt->cb_header;
	if (header->cb_vers != CB_VERSION &&
		header->cb_magic != CB_HEADER_MAGIC) {
		nis_cachebind_free(c);
		return(NIS_FAIL);
	}
#endif
	if (type == CB_SERVER)
		i = hashkey(srv->name);
	else
		i = hashkey(dirname);
	idx = (uint32_t)table[i & HASH_MASK];

	/* No hash table entry for this key; return. */
	if (idx == 0)
		return(NIS_FAIL);

#ifdef foo
	cur = (cachebind_hash_entry *)table[i & HASH_MASK];
	while(cur != NULL) {
		if (cur->hval == i && cur->cb_type == type) {
			*match = nis_cachebind_copy(cur);
			if (*match == NULL)
				return(NIS_NOMEMORY);
			return(NIS_SUCCESS);
		}
		cur = cur->next;
	}
#endif
	return(NIS_NOTFOUND);
}

static nis_error nis_cachebind_lookup(dirname, srv, type, match)
	nis_name		dirname;
	nis_server		*srv;
	cb_type			type;
	cachebind_hash_entry	**match;
{
	nis_error		r;

	/*
 	 * See if the information we want is already in the
	 * cache (either in core, or on disk if the disk copy
	 * is newer).
	 */
	r = nis_cachebind_tsearch(dirname, srv, type, match);
	if (r == NIS_SUCCESS)
		return(r);

	/*
	 * Cache didn't have what we wanted: prod the cache
	 * manager into doing a lookup for us.
	 */
	switch(type) {
	case CB_MASTER:
		r = nis_cachemgr_bind_master(dirname);
		break;
	case CB_REPLICA:
		r = nis_cachemgr_bind_replica(dirname);
		break;
	case CB_SERVER:
		r = nis_cachemgr_bind_server(srv);
		break;
	default:
		break;
	}

	if (r != NIS_SUCCESS)
		return(r);

	/*
	 * Now check the cache again. If the data we want
	 * _still_ isn't there, then return an error.
	 */
	return(nis_cachebind_tsearch(dirname, srv, type, match));
}

static nis_error nis_cachebind_replica(dirname, bptr)
	nis_name		dirname;
	struct _nis_binding	*bptr;
{
	cachebind_hash_entry	*match = NULL, *p;
	nis_error		r;
	nis_server		*s;
	int			i;
	directory_obj		*d;

	r = nis_cachebind_lookup(dirname, NULL, CB_REPLICA, &match);

	if (r != NIS_SUCCESS)
		return(r);

	bptr->nb_name = strdup(dirname);
	bptr->nb_type = NB_TYPE_CACHED;
	bptr->nb_binding_cached = match;
	p = bptr->nb_binding_cached;
	i = p->bdr_ptr->bep_val[0].hostnum;
	d = &p->bdr_ptr->dobj;
	s = &d->do_servers.do_servers_val[i];
	if (s->key_type == NIS_PK_DH)
		bptr->nb_pkey = &s->pkey;
	return(r);
}

static nis_error nis_cachebind_master(dirname, bptr)
	nis_name		dirname;
	struct _nis_binding	*bptr;
{
	cachebind_hash_entry	*match = NULL, *p;
	nis_error		r;
	nis_server		*s;
	int			i;
	directory_obj		*d;

	r = nis_cachebind_lookup(dirname, NULL, CB_MASTER, &match);

	if (r != NIS_SUCCESS)
		return(r);

	bptr->nb_name = strdup(dirname);
	bptr->nb_type = NB_TYPE_CACHED;
	bptr->nb_binding_cached = match;
	p = bptr->nb_binding_cached;
	i = p->bdr_ptr->bep_val[0].hostnum;
	d = &p->bdr_ptr->dobj;
	s = &d->do_servers.do_servers_val[i];
	if (s->key_type == NIS_PK_DH)
		bptr->nb_pkey = &s->pkey;

	return(r);
}

static nis_error nis_cachebind_server(srv, bptr)
	nis_server		*srv;
	struct _nis_binding	*bptr;
{
	cachebind_hash_entry	*match = NULL;
	nis_error		r;

	r = nis_cachebind_lookup(NULL, srv, CB_SERVER, &match);

	bptr->nb_name = strdup(srv->name);
	bptr->nb_type = NB_TYPE_CACHED;
	bptr->nb_pkey = &match->srv_ptr->pkey;
	bptr->nb_binding_cached = match;
	return(r);
}

nis_error __nis_cachebind_refresh(bptr)
	struct _nis_binding	*bptr;
{
	return(NIS_FAIL);
}

void __nis_destroy_cache_binding(bptr)
	struct _nis_binding	*bptr;
{
}

nis_error __nis_cachebind(name, flags, res)
	nis_name		name;
	uint32_t		flags;
	struct _nis_binding	**res;
{
	struct _nis_binding	*bptr;
	nis_error		r;

	if (nis_cachemgr_up() == FALSE)
		return(NIS_UNAVAIL);

	bptr = calloc(1, sizeof(struct _nis_binding));

	if (bptr == NULL)
		return(NIS_NOMEMORY);

	if (!(flags & NO_AUTHINFO))
		bptr->nb_flags |= ZMH_AUTH;
	if (flags & USE_DGRAM)
		bptr->nb_flags |= ZMH_DG;
	else
		bptr->nb_flags |= ZMH_VC;

	if (flags & MASTER_ONLY)
		r = nis_cachebind_master(name, bptr);
	else
		r = nis_cachebind_replica(name, bptr);

	if (r != NIS_SUCCESS) {
		free(bptr);
		bptr = NULL;
	}

	*res = bptr;

	return(r);
}

nis_error __nis_cachebind_srv(srv, flags, res)
	nis_server		*srv;
	uint32_t		flags;
	struct _nis_binding	**res;
{
	struct _nis_binding	*bptr;

	bptr = calloc(1, sizeof(struct _nis_binding));

	if (bptr == NULL)
		return(NIS_NOMEMORY);

	*res = bptr;

	if (!(flags & NO_AUTHINFO))
		bptr->nb_flags |= ZMH_AUTH;
	if (flags & USE_DGRAM)
		bptr->nb_flags |= ZMH_DG;
	else
		bptr->nb_flags |= ZMH_VC;

	return(nis_cachebind_server(srv, *res));
}

nis_error __nis_cacherebind(bptr, flags)
	struct _nis_binding	*bptr;
	uint32_t		flags;
{
	if (nis_cachemgr_up() == FALSE)
		return(NIS_UNAVAIL);

	return(NIS_FAIL);
}
