/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: nislib_callback.c,v 1.2 1998/12/13 17:49:32 wpaul Exp wpaul $
 */

#include <syslog.h>
#include <rpc/rpc.h>
#include <rpc/rpc_com.h>
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/cdefs.h>
#include <sys/filio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nislib_private.h"
#include "nis_callback.h"
 
#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nislib_callback.c,v 1.2 1998/12/13 17:49:32 wpaul Exp wpaul $";
#endif /* not lint */

extern void cb_prog_1 __P(( struct svc_req *, register SVCXPRT * ));

/*
 * This module contains the client side callback support.
 * There are three basic routines: create a callback instance, set a
 * callback in motion, and destroy a callback when done. All we do is
 * set ourselves up as an RPC server and allow the rpc.nisd on the other
 * side to connect to us and feed us data.
 *
 * Note that we actually tell the server what port we're bound to so
 * it's not necessary to register ourselves with the local portmapper.
 * This not only saves us some work, it saves rpc.nisd from calling the
 * portmapper to figure out where we are.
 */

static struct nis_callback *__curcb = NULL;

struct nis_callback *__nis_create_callback(bptr, callback, userdata)
	struct _nis_binding	*bptr;
	int (*callback)(nis_name, nis_object *, void *);
	void			*userdata;
{
	struct nis_callback	*cb;
	int			sock = RPC_ANYSOCK;
	struct sockaddr_in	sin;
	int			len = sizeof(struct sockaddr_in);
	char			addr[NIS_MAXNAMELEN];
	unsigned short		port;

	cb = calloc(1,sizeof(struct nis_callback));
	if (cb == NULL) {
		syslog(LOG_ERR,"NIS+: out of memory allocating callback");
		return(NULL);
	}

	cb->clnt_uaddr = calloc(1,sizeof(nis_server));
	if (cb->clnt_uaddr == NULL) {
		free(cb);
		syslog(LOG_ERR,"NIS+: out of memory allocating callback");
		return(NULL);
	}
	cb->clnt_uaddr->name = strdup(nis_local_principal());
	cb->clnt_uaddr->ep.ep_val = calloc(1,sizeof(endpoint));
	cb->clnt_uaddr->ep.ep_len = 1;
	cb->clnt_uaddr->ep.ep_val->family = strdup("inet");
	cb->callback = callback;
	cb->userdata = userdata;

#ifdef USE_SECURE_RPC
#ifdef __FreeBSD__
	if (svc_auth_reg(AUTH_DES, _svcauth_des) == -1)
		syslog(LOG_ERR, "NIS+: failed to register DES authenticator");
#endif
	if (bptr->nb_flags & ZMH_AUTH)
		cb->clnt_uaddr->key_type = NIS_PK_DH;
	else
		cb->clnt_uaddr->key_type = NIS_PK_NONE;
#endif

	if (bptr->nb_flags & ZMH_DG) {
		cb->clnt_uaddr->ep.ep_val->proto = strdup("udp");
		cb->xprt = svcudp_bufcreate(sock, 100, 8192);
	} else {
		cb->clnt_uaddr->ep.ep_val->proto = strdup("tcp");
		cb->xprt = svctcp_create(sock, 100, 8192);
	}
	cb->sock = cb->xprt->xp_fd;
	if (!svc_register(cb->xprt, CB_PROG, CB_VERS, cb_prog_1, 0)) {
		svc_destroy(cb->xprt);
		xdr_free(xdr_nis_server, (char *)cb->clnt_uaddr);
		free(cb->clnt_uaddr);
		free(cb);
		syslog(LOG_ERR, "NIS+: failed to register callback dispatcher");
		return(NULL);
	}

	if (getsockname(cb->sock, (struct sockaddr *)&sin, &len) == -1) {
		svc_destroy(cb->xprt);
		xdr_free(xdr_nis_server, (char *)cb->clnt_uaddr);
		free(cb->clnt_uaddr);
		free(cb);
		syslog(LOG_ERR, "NIS+: failed to read local socket info");
		return(NULL);
	}
	port = sin.sin_port;
	get_myaddress(&sin);
#if BYTE_ORDER == LITTLE_ENDIAN
	snprintf(addr, sizeof(addr), "%s.%d.%d", inet_ntoa(sin.sin_addr),
				port & 0x00FF, (port & 0xFF00) >> 8);
#else
	snprintf(addr, sizeof(addr), "%s.%d.%d", inet_ntoa(sin.sin_addr),
				(port & 0xFF00) >> 8, port & 0x00FF);
#endif
	cb->clnt_uaddr->ep.ep_val->uaddr = strdup(addr);

	return(cb);
}

#ifdef __FreeBSD__
/*
 * Cribbed from svc_run.c. We go to an excruciating amount of trouble
 * to select() on and have svc_getreqset process only those descriptors
 * that actually relate to this callback instance. This is handy in the
 * event that NIS+ callbacks are used within an RPC server itself; if
 * we didn't do this, we could end up calling out to the server's own
 * RPC service procedures from here. This may not necessarily cause
 * problems, but it can lead to confusion.
 */
nis_error __nis_do_callback(bptr, cookie, cb)
	struct _nis_binding	*bptr;
	netobj			*cookie;
	struct nis_callback	*cb;
{
	fd_set			*fds = NULL;
	struct timeval		tv;
/*
	extern int		__svc_fdsetsize;
	extern fd_set		*__svc_fdset;
*/
	int			*cb_is_running = 0;
	fd_set			cb_fdset;
	int			i, newguy = 0;
	int			setsize = 0;

	__curcb = cb;

	for (;;) {
		tv.tv_sec = 25;
		tv.tv_usec = 0;

		if (&svc_fdset != NULL) {
			/*int bytes = howmany(__svc_fdsetsize, NFDBITS) * */
			int bytes = _howmany(FD_SETSIZE, NFDBITS) *
						sizeof(fd_mask);
			fds = calloc(1, bytes);
			/*memcpy(fds, __svc_fdset, bytes);*/
			memcpy(fds, &svc_fdset, bytes);
		} else
			fds = NULL;

		/*setsize = __svc_fdsetsize;*/
		setsize = FD_SETSIZE;
		FD_ZERO(&cb_fdset);
		FD_SET(__curcb->sock, &cb_fdset);
		FD_CLR(__curcb->sock, &svc_fdset);
		if (__curcb->sock < FD_SETSIZE)
			FD_CLR(__curcb->sock, &svc_fdset);
		if (newguy) {
			FD_SET(newguy, &cb_fdset);
			FD_CLR(newguy, &svc_fdset);
			if (newguy < FD_SETSIZE)
				FD_CLR(newguy, &svc_fdset);
		}

		switch (select(svc_maxfd + 1, &cb_fdset, NULL, NULL, &tv)) {
		case -1:
			if (errno == EINTR) {
				if (fds)
					free(fds);
				continue;
			}
			if (fds != NULL)
				free(fds);
			/*svc_getreqset2(&cb_fdset, svc_maxfd + 1);*/
			svc_getreqset(&cb_fdset);
			return(NIS_CBERROR);
		case 0:
			if (fds != NULL)
				free(fds);
			/*
			 * See if callback 'thread' in the server is still
			 * alive. (Note: our server doesn't really use
			 * threads.)
			 */
			cb_is_running = nis_callback_3(cookie, bptr->nb_clnt);
			if (cb_is_running == NULL || *cb_is_running == FALSE) {
				syslog(LOG_ERR, "NIS+: callback timed out");
				/*svc_getreqset2(&cb_fdset, svc_maxfd + 1);*/
				svc_getreqset(&cb_fdset);
				return(NIS_CBERROR);
			}
			break;
		default:
			/*svc_getreqset2(&cb_fdset, svc_maxfd + 1);*/
			svc_getreqset(&cb_fdset);
			/*
			 * A new socket may miraculously appear if
			 * this is a TCP callback. Be sure to account
			 * for it.
			 */
			for (i = 0; i < svc_maxfd + 1; i++) {
				if (FD_ISSET(i, &svc_fdset) &&
					(i > setsize || /* !!! */
					!FD_ISSET(i, fds))) {
					newguy = i;
					break;
				}
			}
			free(fds);
			if (cb->nomore) {
				/*
				 * If this is a TCP connection, this
				 * will destroy the extra transport
				 * handle. If not, no harm done.
				 */
				/*svc_getreqset2(&cb_fdset, svc_maxfd + 1);*/
				svc_getreqset(&cb_fdset);
				return(cb->err);
			}
		}
	}
}
#else
nis_error __nis_do_callback(bptr, cookie, cb)
	struct _nis_binding	*bptr;
	netobj			*cookie;
	struct nis_callback	*cb;
{
#ifdef FD_SETSIZE
	fd_set readfds;
#else
	int readfds;
#endif /* def FD_SETSIZE */

	struct timeval		tv;
	int			*cb_is_running = 0;
	fd_set			cb_fdset;
	int			i, newguy = 0;

	__curcb = cb;

	for (;;) {
#ifdef FD_SETSIZE
		readfds = svc_fdset;
#else
		readfds = svc_fds;
#endif /* def FD_SETSIZE */
		FD_ZERO(&cb_fdset);
		FD_SET(__curcb->sock, &cb_fdset);
		/*FD_CLR(__curcb->sock, &svc_fdset);*/
		if (newguy) {
			FD_SET(newguy, &cb_fdset);
			/*FD_CLR(newguy, &svc_fdset);*/
		}
		tv.tv_sec = 25;
		tv.tv_usec = 0;
#ifdef __sgi
		switch (select(getdtablehi(), &cb_fdset, NULL, NULL, &tv)) {
#else
		switch (select(_rpc_dtablesize(), &cb_fdset, NULL, NULL, &tv)) {
#endif
		case -1:
			if (errno == EINTR) {
				continue;
			}
			svc_getreqset(&cb_fdset);
			return(NIS_CBERROR);
		case 0:
			/*
			 * See if callback 'thread' in the server is still
			 * alive. (Note: our server doesn't really use
			 * threads.)
			 */
			cb_is_running = nis_callback_3(cookie, bptr->nb_clnt);
			if (cb_is_running == NULL || *cb_is_running == FALSE) {
				syslog(LOG_ERR, "NIS+: callback timed out");
				svc_getreqset(&cb_fdset);
				return(NIS_CBERROR);
			}
			break;
		default:
			svc_getreqset(&cb_fdset);
			if (!newguy) {
#ifdef __sgi
				for (i = 0; i < getdtablehi(); i++) {
#else
				for (i = 0; i < FD_SETSIZE; i++) {
#endif
					if (FD_ISSET(i, &svc_fdset) &&
						!FD_ISSET(i, &readfds)) {
						newguy = i;
						break;
					}
				}
			}
			if (cb->nomore) {
				svc_getreqset(&cb_fdset);
				return(cb->err);
			}
		}
	}
}
#endif

nis_error __nis_destroy_callback(cb)
	struct nis_callback	*cb;
{
	svc_destroy(cb->xprt);
	close(cb->sock);
	xdr_free(xdr_nis_server, (char *)cb->clnt_uaddr);
	free(cb->clnt_uaddr);
	free(cb);

	return(NIS_SUCCESS);
}

bool_t *
cbproc_receive_1_svc(cback_data *argp, struct svc_req *rqstp)
{
	static bool_t		result;
	unsigned long		i;
	int			r = 0;
	char			tname[NIS_MAXNAMELEN + 1];

	result = FALSE;
	for (i = 0; i < argp->entries.entries_len; i++) {
		snprintf(tname, sizeof(tname), "%s.%s",
				argp->entries.entries_val[i]->zo_name,
				argp->entries.entries_val[i]->zo_domain);
		r = (*__curcb->callback)(tname, argp->entries.entries_val[i],
					__curcb->userdata);
		if (r) {
			result = TRUE;
			__curcb->nomore = 1;
			__curcb->err = NIS_SUCCESS;
			break;
		}
	}

	return (&result);
}

void *
cbproc_finish_1_svc(void *argp, struct svc_req *rqstp)
{
	static char		*result;

	__curcb->nomore = 1;
	__curcb->err = NIS_SUCCESS;

	return((void *)&result);
}

void *
cbproc_error_1_svc(nis_error *argp, struct svc_req *rqstp)
{
	static char		*result;

	__curcb->nomore = 1;
	__curcb->err = *argp;

	return((void *)&result);
}
