/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: nislib_core.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <syslog.h>
#include <sys/cdefs.h>
#include <sys/fcntl.h>
#include <sys/queue.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include <rpcsvc/nis_tags.h>
#include "nislib_private.h"

#ifndef lint
static const char rcsid[] = "@(#) $Id: nislib_core.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $";
#endif

/*
 * This module contains three routines that are central to the
 * nis_list() and nis_lookup() functions:
 *
 * nis_finddirectory()
 * __nis_search()
 * __nis_choose_server()
 *
 * Noth nis_lookup() and nis_list() eventually call __nis_search()
 * to do most of the real gruntwork. The other two functions are
 * needed by the binder code.
 *
 * This is done largely for the benefit of the NIS+ server, which
 * needs to be able to use the nis_list() and nis_lookup() interfaces
 * in order to function correctly. The problem is that the client
 * routines ultimately need to make RPCs to a server, and if the
 * server tries to make an RPC to itself, it will deadlock.
 *
 * To deal with this, rpc.nisd(8) has its own implementation of these
 * three functions that have been modified to short-circuit any case
 * there this deadlock might occur.
 *
 * Note that this hackery is accomplished by abusing a the fact that
 * if a program has a local symbol with the same name as a symbol in
 * a shared library, the local symbol will override the one in the
 * shared lib. Also note that this only works correctly if the function
 * resides in a seperate object module all by itself; if this code
 * were to appear as part of another object module, we'd end up with
 * 'multiple definition' errors from the linker.
 */

/*
 * Given a directory_obj, scan the list of servers that serve
 * dir->do_name and choose the one that replies the fastest.
 */
nis_server *__nis_choose_server_core(bptr, flags)
	struct _nis_binding	*bptr;
	uint32_t		flags;
{
	nis_error		pstat;
	directory_obj		*dir;
	struct manual_binding	*m;

	m = bptr->nb_binding_manual;
	dir = m->mb_dir;

	if (dir == NULL) {
		if (m->mb_srv == NULL)
			return(NULL);
		else {
			__nis_ping(bptr);
			return(m->mb_srv);
		}
	}

	if (!dir->do_servers.do_servers_len)
		return(NULL);

	if (flags & MASTER_ONLY || dir->do_servers.do_servers_len == 1) {
		uint32_t		cnt;
		cnt = dir->do_servers.do_servers_len;
		dir->do_servers.do_servers_len = 1;
		__nis_ping(bptr);
		dir->do_servers.do_servers_len = cnt;
		return(m->mb_srv);
	}

	pstat = __nis_ping(bptr);

	return(m->mb_srv);
}

fd_result *nis_finddirectory_core(dir, name)
	directory_obj		*dir;
	nis_name		name;
{
	struct _nis_binding	*bptr;
	fd_args			fd_args;
	fd_result		*fd_res = NULL;
	nis_error		r;

	r = __nis_create_binding(dir, NULL, NO_AUTHINFO|USE_DGRAM, &bptr);

	if (r != NIS_SUCCESS)
		return(NULL);

	if (__nis_connect_server(bptr) != NIS_SUCCESS) {
		__nis_destroy_binding(bptr);
		return(NULL);
	}

	fd_args.dir_name = name;
	fd_args.requester = nis_local_host();

	while(1) {
		fd_res = nis_finddirectory_3(&fd_args, bptr->nb_clnt);
		if (fd_res == NULL || fd_res->status != NIS_SUCCESS) {
			if (fd_res != NULL)
				xdr_free(xdr_fd_result, (char *)fd_res);
		} else
			break;
		if (__nis_rebind(bptr, NO_AUTHINFO|USE_DGRAM) != NIS_SUCCESS)
			break;
	}

	if (bptr != NULL)
		__nis_destroy_binding(bptr);

	return(fd_res);
}

nis_result *__nis_search_core(name, op, flags, attrs, numattrs,
					callback, userdata)
	nis_name		name;
	uint32_t		op;
	uint32_t		flags;
	nis_attr		*attrs;
	uint32_t		numattrs;
	int (*callback)(nis_name, nis_object *, void *);
	void			*userdata;
{
	struct _nis_binding	*bptr = NULL;
	nis_result		save, *res, *lres;
	ns_request		nsreq = { NULL, { 0, NULL } };
	ib_request		ibreq = { 	NULL,
						{ 0, NULL },
						0,
						{ 0, NULL },
						{ 0, NULL },
						0,
						{ 0, NULL },
					};
	int			linkcount = 0;
	struct nis_callback	*cb = NULL;
	char			*lname;
	int			done = 0;
	nis_error		r;
	CACHE_CLOCK_DECLS;

	ibreq.ibr_srch.ibr_srch_len = numattrs;
	ibreq.ibr_srch.ibr_srch_val = attrs;
	ibreq.ibr_flags = flags;
	res = calloc(1, sizeof(nis_result));
	if (res == NULL) {
		syslog(LOG_ERR, "NIS+: out of memory allocating result");
		return(NULL);
	}

	if (name == NULL) {
		res->status = NIS_BADNAME;
		return(res);
	}

	lname = name;
	bzero((char *)&save, sizeof(nis_result));

	__start_clock(CACHE_CLOCK);
	if (op == SEARCHOP_LOOKUP) {
		r = __nis_bind(nis_domain_of(lname), flags, &bptr);
		if (r != NIS_SUCCESS)
			r = __nis_bind(lname, flags, &bptr);
	} else {
		r = __nis_bind(lname, flags, &bptr);
		if (r != NIS_SUCCESS)
			r = __nis_bind(nis_domain_of(lname), flags, &bptr);
	}
	res->aticks += __stop_clock(CACHE_CLOCK);

	while(!done) {
		if (r != NIS_SUCCESS) {
			if (flags & HARD_LOOKUP) {
				syslog(LOG_WARNING, "NIS servers for %s not "
					"responding -- still trying", lname);
				continue;
			} else {
				if (r == NIS_NAMEUNREACHABLE)
					syslog(LOG_WARNING,
						"NIS+ servers unreachable");
				res->status = r;
				break;
			}
		}

		/* Perform the actual RPC. */
		if (op == SEARCHOP_LOOKUP) {
			nsreq.ns_name = lname;
			lres = nis_lookup_3(&nsreq, bptr->nb_clnt);
		} else {
			ibreq.ibr_name = lname;
			if (callback != NULL) {
				cb = __nis_create_callback(bptr,
							   callback,
						 	   userdata);
				ibreq.ibr_cbhost.ibr_cbhost_len = 1;
				ibreq.ibr_cbhost.ibr_cbhost_val =
							cb->clnt_uaddr;
			}
			lres = nis_iblist_3(&ibreq, bptr->nb_clnt);
		}

		/* Check if the RPC failed */
		if (lres == NULL) {
			struct rpc_err err;

			clnt_geterr(bptr->nb_clnt, &err);
			if (err.re_status == RPC_AUTHERROR)
				res->status = NIS_SRVAUTH;
			else
				res->status = NIS_RPCERROR;
			__start_clock(CACHE_CLOCK);
			if ((r = __nis_rebind(bptr, flags)) != NIS_SUCCESS)
				bptr = NULL;
			res->aticks += __stop_clock(CACHE_CLOCK);
			res->status = r;
			continue;
		}

		/* Keep all our counters current */
		res->cticks += lres->cticks;
		res->dticks += lres->dticks;
		res->zticks += lres->zticks;
		res->aticks += lres->aticks;

		xdr_free(xdr_nis_result, (char *)&save);
		bcopy((char *)lres, (char *)&save, sizeof(nis_result));

		switch(lres->status) {
		case NIS_CBRESULTS:
			lres->status = __nis_do_callback(bptr,
						 &lres->cookie, cb);
			__nis_destroy_callback(cb);
			cb = NULL;
		/*
		 * NIS_PARTIAL can happen if we are performing a LIST
		 * operation and the server is able to resolve the
		 * table name, but it can't find the entry in it.
		 * This most often happens when the table name is a
		 * link to something else, though it can also happen
		 * that the table name is just the wrong kind of
		 * object.
		 */
		case NIS_PARTIAL:
		case NIS_SUCCESS:
			/* Are we following links? */
			if (NIS_RES_NUMOBJ(lres) &&
			__type_of(NIS_RES_OBJECT(lres)) == LINK_OBJ &&
					flags & FOLLOW_LINKS) {
				/* if we hit the link limit, bail */
				if (linkcount > NIS_MAXLINKS) {
					res->status = NIS_LINKNAMEERROR;
					done++;
					break;
				}
				/*
				 * If we haven't hit the limit,
				 * attempt to resolve the link.
				 */
#			define LINK NIS_RES_OBJECT(lres)->LI_data
				linkcount++;
				lname = LINK.li_name;
				if (LINK.li_attrs.li_attrs_len) {
					ibreq.ibr_name = lname;
					ibreq.ibr_srch.ibr_srch_len =
						LINK.li_attrs.li_attrs_len;
					ibreq.ibr_srch.ibr_srch_val =
						LINK.li_attrs.li_attrs_val;
					op = SEARCHOP_LIST;
				} /*else {
					numattrs = 0; attrs = NULL;
					op = SEARCHOP_LOOKUP;
				} */
			} else {
				if (NIS_RES_NUMOBJ(lres) &&
					__type_of(NIS_RES_OBJECT(lres)) ==
					LINK_OBJ &&
					op == SEARCHOP_LIST &&
					(flags & FOLLOW_LINKS) == 0)
					res->status = NIS_NOTSEARCHABLE;
				else
					res->status = lres->status;
				NIS_RES_OBJECT(res) = NIS_RES_OBJECT(lres);
				NIS_RES_NUMOBJ(res) = NIS_RES_NUMOBJ(lres);
				NIS_RES_OBJECT(lres) = NULL;
				NIS_RES_NUMOBJ(lres) = 0;
				xdr_free(xdr_nis_result, (char *)lres);
				bzero((char *)&save, sizeof(nis_result));
				done++;
			}
			break;
		case NIS_NOT_ME:
		case NIS_TRYAGAIN:
		case NIS_RPCERROR:
		case NIS_SYSTEMERROR:
			xdr_free(xdr_nis_result, (char *)lres);
			/* Something fishy here. Try another server. */
			if (__nis_rebind(bptr, flags) == NIS_SUCCESS)
				continue;
			else {
				bptr = NULL;
				res->status = lres->status;
				done++;
			}
			break;
		default:
			/*
			 * Unspecified failure case.
			 */
			res->status = lres->status;
			done++;
			break;
		}
		if (cb != NULL)
			__nis_destroy_callback(cb);

	}

	xdr_free(xdr_nis_result, (char *)&save);

	if (bptr != NULL)
		__nis_release_binding(bptr);
	return(res);
}
