/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nislib_groups.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nislib_private.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nislib_groups.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $";
#endif

/*
 * This module implements the NIS+ group object manipulation functions.
 * It also maintains a cache, largely for the benefit of the server.
 *
 * Caching group lookups involves turning the group information inside-out:
 * while the group_obj maps from group to members, we want to map from
 * members to groups. Our main goal is to keep the amount of memory comsumed
 * by the cache to a minimum. It should be noted then that caching a
 * group_obj could consume a lot more memory than just caching member names
 * and the groups to which they belong: a group_obj may contain a _lot_
 * of members (hundreds maybe) and there's no pointi lugging all that
 * information around when only a small portion of it will be needed.
 *
 * The cache consists of a hash table with member names as keys. Each
 * table entry has a linked list of group names. So if the caller asks
 * to see if 'member' is belongs to 'group' and 'member' has been cached,
 * we can check 'group' against the cached group list instead of slogging
 * though the group's member list. If 'member' has not been cached, then
 * we have to do an nis_lookup() on the group, see if 'member' is in its
 * member list and cache the results.
 *
 * The tricky part comes when somebody changes group information. We
 * then have to update the cache to reflect the change. This can be done
 * immediately in the server context, but in the client context the
 * updates may not be noticed until the ttl on the group object runs
 * out.
 */

#define TABLESIZE 128
#define HASH_MASK 0x0000007F

struct glist {
	nis_name		gname;
	uint32_t		flags;
	time_t			ttl;
	struct glist		*next;
};

struct mem_entry {
	nis_name		member;
	struct glist		*list;
};

static struct mem_entry	*mtab[TABLESIZE];

static uint32_t __hits = 0;
static uint32_t __misses = 0;
static uint32_t __calls = 0;

void __nis_clear_grpstats()
{
	__hits = __misses = __calls = 0;

	return;
}

void __nis_get_grpstats(hits, misses, calls)
	uint32_t		*hits,
				*misses,
				*calls;
{
	*hits = __hits;
	*misses = __misses;
	*calls = __calls;

	return;
}

static void __nis_init_grpcache()
{
	static int		i = 0;

	if (!i) {
		bzero((char *)&mtab, sizeof(mtab));
		i++;
	}

	return;
}

static uint32_t hashkey(key)
	char			*key;
{
	if (key == NULL)
		 return (-1);
	return(__nis_hash((void *)key, strlen(key)) & HASH_MASK);
}

static struct mem_entry *__nis_mem_lookup(key)
	nis_name		key;
{
	struct mem_entry	*cur;

	cur = mtab[hashkey(key)];

	if (cur != NULL)
		return(cur);
	return(NULL);
}

static void __nis_mem_store(key, grp, ttl, flags)
	nis_name		key;
	nis_name		grp;
	uint32_t		ttl;
	uint32_t		flags;
{
	struct mem_entry	*new = NULL, *cur;
	uint32_t		i;
	struct glist		*gnew, *tmp;

	gnew = malloc(sizeof(struct glist));
	gnew->gname = strdup(grp);
	gnew->ttl = time(NULL) + ttl;
	gnew->flags = flags;
	gnew->next = NULL;

	i = hashkey(key);
	cur = mtab[i];

	if (cur == NULL) {
		new = (struct mem_entry *)malloc(sizeof(struct mem_entry));
		new->list = NULL;
		new->member = NULL;
	} else {
		if (!strcmp(cur->member, key)) {
			tmp = cur->list;
			while(tmp->next != NULL) {
				if (!strcmp(tmp->gname, gnew->gname)) {
					tmp->ttl = gnew->ttl;
					tmp->flags = gnew->flags;
					free(gnew->gname);
					free(gnew);
					return;
				}
				tmp = tmp->next;
			}
			gnew->next = cur->list;
			cur->list = gnew;
			return;
		} else {
			while(cur->list) {
				tmp = cur->list->next;
				free(cur->list->gname);
				free(cur->list);
				cur->list = tmp;
			}
			free(cur->member);
			cur->member = NULL;
			cur->list = NULL;
		}
	}

	new->member = strdup(key);
	new->list = gnew;
	mtab[i] = new;

	return;
}

static char *__convert_grp(grp)
	nis_name		grp;
{
	static char		newgrp[NIS_MAXNAMELEN + 1];
	char			*ptr1, *ptr2;

	ptr1 = nis_leaf_of(grp);
	ptr2 = nis_domain_of(grp);

	snprintf(newgrp, sizeof(newgrp), "%s.groups_dir.%s", ptr1, ptr2);

	return(newgrp);
}

static nis_error __nis_find_group(name, grp)
	nis_name		name;
	nis_object		**grp;
{
	nis_result		*res;
	nis_name		gname;
	nis_error		err;

	gname = __convert_grp(name);

	res = __nis_search(gname, SEARCHOP_LOOKUP,
			FOLLOW_LINKS, NULL, 0, NULL, NULL);

	if (res == NULL)
		return(NIS_NOMEMORY);

	err = res->status;

	if (res->status != NIS_SUCCESS) {
		nis_freeresult(res);
		return(err);
	}

	/* Sanity check. This can't happen. */
	if (NIS_RES_NUMOBJ(res) > 1 ||
		__type_of(NIS_RES_OBJECT(res)) != GROUP_OBJ) {
		nis_freeresult(res);
		return(NIS_INVALIDOBJ);
	}

	*grp = NIS_RES_OBJECT(res);
	NIS_RES_OBJECT(res) = NULL;
	NIS_RES_NUMOBJ(res) = 0;
	nis_freeresult(res);

	return(NIS_SUCCESS);
}

/*
 * This function is not documented in the Sun manual pages but it
 * is defined within the NSL library. (Do nm /usr/lib/libnsl.a on
 * a handy Solaris machine if you don't believe me.)
 */
void nis_flushgroups()
{
	register uint32_t	i;
	struct mem_entry	*cur;
	struct glist		*tmp, *gptr;

	for (i = 0; i < TABLESIZE; i++) {
		cur = mtab[i];
		if (cur == NULL)
			continue;
		gptr = cur->list;
		while(gptr) {
			tmp = gptr->next;
			free(gptr->gname);
			free(gptr);
			gptr = tmp;
		}
		free(cur->member);
		free(cur);
		mtab[i] = NULL;
	}
}

/*
 * This function is also undocumented but exists in libnsl.
 */
void __nis_flush_one_group(group)
	nis_name		group;
{
	register uint32_t	i;
	struct mem_entry	*cur = NULL;
	struct glist		*gptr, *prev, *next;

	for (i = 0; i < TABLESIZE; i++) {
		if (mtab[i] == NULL)
			continue;
		gptr = mtab[i]->list;
		prev = NULL;
		cur = mtab[i];
		while(gptr) {
			next = gptr->next;
			if (!strcmp(gptr->gname, group)) {
				free(gptr->gname);
				free(gptr);
				if (prev == NULL)
					mtab[i]->list = next;
				else
					prev->next = next;
				break;
			}
			prev = gptr;
			gptr = gptr->next;
		}

		/*
		 * If this was the only group, zap the whole
		 * member entry.
		 */
		if (cur->list == NULL) {
			free(cur->member);
			free(cur);
			mtab[i] = NULL;
		}
	}

	return;
}

enum gstat { MEM_YES = 0, MEM_NO = 1, MEM_NOTFOUND = 2, MEM_EXPIRED = 3 };

static enum gstat __nis_search_grcache(mem, grp)
	nis_name		mem;
	nis_name		grp;
{
	struct mem_entry	*mptr;
	struct glist		*gptr;
	time_t			now = time(NULL);

	__calls++;

	mptr = __nis_mem_lookup(mem);

	if (mptr == NULL) {
		__misses++;
		return(MEM_NOTFOUND);
	}

	__hits++;

	gptr = mptr->list;
	while(gptr != NULL) {
		if (!strcmp(gptr->gname, grp)) {
			if (now > gptr->ttl)
				return(MEM_EXPIRED);
			else
				if (!(gptr->flags & NEGMEM_GROUPS))
					return(MEM_YES);
		}
		gptr = gptr->next;
	}

	return(MEM_NO);
}

/*
 * We have to handle explicit, implicit, recursive and negative
 * groups here. The following rules apply:
 *
 * - A member name that is just a normal NIS+ principal name is
 *   an explicit member. Most members we encounter will probably
 *   be like this.
 *
 * - A member of the form *.foo.org. indicates that all principals
 *   in the foo.org. domain are members. This is an implicit member.
 *
 * - A member of the form @somegroup.foo.org. indicates that all
 *   principals that are members of the somegroup.foo.org. group
 *   are also members of this group. This is a recursive member.
 *
 * - Lastly, all of these may be preceded by a minus sign ('-')
 *   to indicate negative membership. That is, the member is explicltly
 *   excluded rather than included.
 *
 * Exclusion entries are created by doing a 'null insert,' where
 * we insert a record for the member into the hash table with an empty
 * group name. This allows the table search routine to positively locate
 * the member without showing it as being in in the group.
 *
 * Bletch.
 */

static bool_t __nis_update_grcache(mem, grp)
	nis_name		mem;
	nis_name		grp;
{
	nis_object		*obj;
	uint32_t		i;
	group_obj		*g;
	bool_t			r = FALSE;
	nis_name		gmem, mdom;
	int			test_exclude = 0;

	/*
	 * Sanity check on member principal.
	 */
	mdom = strchr(mem, '.');
	if (mdom == NULL)
		return(r);
	mdom++;

	if (__nis_find_group(grp, &obj) != NIS_SUCCESS)
		return(r);

	__nis_init_grpcache();
	g = &obj->GR_data;

	for (i = 0; i < g->gr_members.gr_members_len; i++) {
		gmem = g->gr_members.gr_members_val[i];
		if (gmem[0] == '-') {
			test_exclude++;
			gmem++;
		}
		switch(gmem[0]) {
		case '*':
			if (!strcmp(mdom, gmem + 2)) {
				if (test_exclude)
					__nis_mem_store(mem, grp, obj->zo_ttl,
						IMPMEM_GROUPS|NEGMEM_GROUPS);
				else
					__nis_mem_store(mem, grp, obj->zo_ttl,
							IMPMEM_GROUPS);
				r = TRUE;
			}
			break;
		case '@':
			/* Recursion! Aaaaahh! */
			if (__nis_update_grcache(mem, gmem + 1) == TRUE) {
				if (test_exclude)
					__nis_mem_store(mem, grp, obj->zo_ttl,
						RECURS_GROUPS|NEGMEM_GROUPS);
				else
					__nis_mem_store(mem, grp, obj->zo_ttl,
							RECURS_GROUPS);
				r = TRUE;
			}
			break;
		default:
			if (!strcmp(gmem, mem)) {
				if (test_exclude)
					__nis_mem_store(mem, grp, obj->zo_ttl,
							NEGMEM_GROUPS);
				else
					__nis_mem_store(mem, grp, obj->zo_ttl,
								0);
				r = TRUE;
			}
			break;
		}
	}

	nis_destroy_object(obj);

	return(r);
}

bool_t nis_ismember(principal, group)
	nis_name		principal;
	nis_name		group;
{
	enum gstat		r;

	if (group == NULL || !strlen(group))
		return(FALSE);

	if (principal == NULL || !strlen(principal))
		return(FALSE);

	switch(__nis_search_grcache(principal, group)) {
	case MEM_YES:
		return(TRUE);
		break;
	case MEM_NO:
		return(FALSE);
		break;
	case MEM_EXPIRED:
		__nis_flush_one_group(group);
	case MEM_NOTFOUND:
		__nis_update_grcache(principal, group);
		r = __nis_search_grcache(principal, group);
		if (r == MEM_YES)
			return(TRUE);
		else
			return(FALSE);
		break;
	default:
		syslog(LOG_ERR, "NIS+: check_grcache returned bogus code");
		return(FALSE);
		break;
	}

	return(FALSE);
}

nis_error nis_addmember(member, group)
	nis_name		member;
	nis_name		group;
{
	nis_result		*res;
	nis_object		*obj;
	nis_error		stat;
	group_obj		*g;
	nis_name		*members;
	int			i;

	stat =  __nis_find_group(group, &obj);

	if (stat != NIS_SUCCESS)
		return(stat);

	g = &obj->GR_data;

	/*
	 * Check that principal isn't already a member.
	 */
	for (i = 0; i < g->gr_members.gr_members_len; i++) {
		if (nis_dir_cmp(g->gr_members.gr_members_val[i],
				member) == SAME_NAME) {
			nis_destroy_object(obj);
			return(NIS_NAMEEXISTS);
		}
	}

	members = realloc(g->gr_members.gr_members_val,
			  (g->gr_members.gr_members_len + 1) *
				 sizeof(nis_name *));

	if (members == NULL) {
		nis_destroy_object(obj);
		return(NIS_NOMEMORY);
	}

	members[g->gr_members.gr_members_len] = strdup(member);
	g->gr_members.gr_members_val = members;
	g->gr_members.gr_members_len++;

	res = nis_modify(__convert_grp(group), obj);

	nis_destroy_object(obj);

	if (res == NULL)
		return(NIS_NOMEMORY);

	stat = res->status;
	nis_freeresult(res);

	__nis_flush_one_group(group);
	__nis_update_grcache(member, group);

	return(stat);
}

nis_error nis_removemember(member, group)
	nis_name		member;
	nis_name		group;
{
	nis_result		*res;
	nis_object		*obj;
	nis_error		stat;
	uint32_t		i;
	group_obj		*g;
	char			*save = NULL;

	stat =  __nis_find_group(group, &obj);

	if (stat != NIS_SUCCESS)
		return(stat);

	g = &obj->GR_data;

	for (i = 0; i < g->gr_members.gr_members_len; i++) {
		if (!strcmp(g->gr_members.gr_members_val[i], member)) {
			save = g->gr_members.gr_members_val[i];
			break;
		}
	}

	if (save == NULL) {
		nis_destroy_object(obj);
		return(NIS_BADREQUEST);
	}

	for (; i < g->gr_members.gr_members_len - 1; i++) {
		g->gr_members.gr_members_val[i] =
		g->gr_members.gr_members_val[i + 1];
	}

	g->gr_members.gr_members_val[i + 1] = NULL;
	g->gr_members.gr_members_len--;

	res = nis_modify(__convert_grp(group), obj);

	g->gr_members.gr_members_val[g->gr_members.gr_members_len] = save;
	g->gr_members.gr_members_len++;

	nis_destroy_object(obj);

	if (res == NULL)
		return(NIS_NOMEMORY);

	stat = res->status;
	nis_freeresult(res);
	__nis_flush_one_group(group);

	return(stat);
}

nis_error nis_creategroup(group, flags)
	nis_name		group;
	uint32_t		flags;
{
	group_obj		grp;
	nis_object		obj;
	nis_name		gname;
	nis_error		stat;
	nis_result		*res;

	gname = __convert_grp(group);

	bzero((char *)&grp, sizeof(group_obj));
	bzero((char *)&obj, sizeof(nis_object));

	grp.gr_flags = flags;
	__type_of(&obj) = GROUP_OBJ;
	obj.GR_data = grp;
	obj.zo_name = nis_leaf_of(gname);
	obj.zo_domain = nis_domain_of(gname);
	obj.zo_owner = nis_local_principal();
	obj.zo_group = nis_local_group();
	obj.zo_ttl = 60*60;	/* One hour ttl */
	obj.zo_access = DEFAULT_RIGHTS;

	res = nis_add(gname, &obj);

	if (res == NULL)
		return(NIS_NOMEMORY);

	stat = res->status;

	nis_freeresult(res);
	__nis_flush_one_group(group);

	return(stat);
}

nis_error nis_destroygroup(group)
	nis_name		group;
{
	nis_result		*res;
	nis_name		gname;
	nis_error		stat;

	gname = __convert_grp(group);

	res = nis_remove(gname, NULL);

	if (res == NULL)
		return(NIS_NOMEMORY);
	stat = res->status;

	nis_freeresult(res);
	__nis_flush_one_group(group);

	return(stat);
}

void nis_print_group_entry(group)
	nis_name		group;
{
	nis_object		*obj;
	group_obj		*grp;
	register int		i;
	nis_name		gmem;
	uint32_t		*mem_exp, mem_exp_cnt = 0;
	uint32_t		*mem_imp, mem_imp_cnt = 0;
	uint32_t		*mem_rec, mem_rec_cnt = 0;
	uint32_t		*nomem_exp, nomem_exp_cnt = 0;
	uint32_t		*nomem_imp, nomem_imp_cnt = 0;
	uint32_t		*nomem_rec, nomem_rec_cnt = 0;
	int			neg = 0;

	mem_exp = malloc(sizeof(uint32_t));
	mem_imp = malloc(sizeof(uint32_t));
	mem_rec = malloc(sizeof(uint32_t));
	nomem_exp = malloc(sizeof(uint32_t));
	nomem_imp = malloc(sizeof(uint32_t));
	nomem_rec = malloc(sizeof(uint32_t));

	if (__nis_find_group(group, &obj) != NIS_SUCCESS)
		return;
	grp = &obj->GR_data;

	for (i = 0; i < grp->gr_members.gr_members_len; i++) {
		gmem = grp->gr_members.gr_members_val[i];
		neg = 0;
		if (gmem[0] == '-') {
			gmem++;
			neg = 1;
		} else
			neg = 0;
		switch(gmem[0]) {
		case '*':
			if (neg) {
				nomem_imp = realloc(nomem_imp,
						 (mem_imp_cnt + 1) *
						sizeof(uint32_t));
				nomem_imp[nomem_imp_cnt] = i;
				nomem_imp_cnt++;
			} else {
				mem_imp = realloc(mem_imp, (mem_imp_cnt + 1) *
						sizeof(uint32_t));
				mem_imp[mem_imp_cnt] = i;
				mem_imp_cnt++;
			}
			break;
		case '@':
			if (neg) {
				nomem_rec = realloc(nomem_rec,
						 (nomem_rec_cnt + 1) *
						sizeof(uint32_t));
				nomem_rec[nomem_rec_cnt] = i;
				nomem_rec_cnt++;
			} else {
				mem_rec = realloc(mem_rec, (mem_rec_cnt + 1) *
						sizeof(uint32_t));
				mem_rec[mem_rec_cnt] = i;
				mem_rec_cnt++;
			}
			break;
		default:
			if (neg) {
				nomem_exp = realloc(nomem_exp,
						 (nomem_exp_cnt + 1) *
						sizeof(uint32_t));
				nomem_exp[nomem_exp_cnt] = i;
				nomem_exp_cnt++;
			} else {
				mem_exp = realloc(mem_exp, (mem_exp_cnt + 1) *
						sizeof(uint32_t));
				mem_exp[mem_exp_cnt] = i;
				mem_exp_cnt++;
			}
			break;
		}
	}

#define GVAL	grp->gr_members.gr_members_val
	printf("Group entry for \"%s\" group:\n", group);

	/* Explicit */
	if (mem_exp_cnt) {
		printf("    Explicit members:\n");
		for(i = 0; i < mem_exp_cnt; i++)
			printf("\t%s\n", GVAL[mem_exp[i]]);
	} else
		printf("    No explicit members\n");

	/* Implicit */
	if (mem_imp_cnt) {
		printf("    Implicit members:\n");
		for(i = 0; i < mem_imp_cnt; i++)
			printf("\t%s\n", GVAL[mem_imp[i]] + 2);
	} else
		printf("    No implicit members\n");

	/* Recursive */
	if (mem_rec_cnt) {
		printf("    Recursive members:\n");
		for(i = 0; i < mem_rec_cnt; i++)
			printf("\t%s\n", GVAL[mem_rec[i]] + 1);
	} else
		printf("    No recursive members\n");

	/* Explicit-not */
	if (nomem_exp_cnt) {
		printf("    Explicit nonmembers:\n");
		for(i = 0; i < nomem_exp_cnt; i++)
			printf("\t%s\n", GVAL[nomem_exp[i]] + 1);
	} else
		printf("    No explicit nonmembers\n");

	/* Implicit-not */
	if (nomem_imp_cnt) {
		printf("    Implicit nonmembers:\n");
		for(i = 0; i < nomem_imp_cnt; i++)
			printf("\t%s\n", GVAL[mem_imp[i]] + 3);
	} else
		printf("    No implicit nonmembers\n");

	/* Recursive-not */
	if (nomem_rec_cnt) {
		printf("    Explicit nonmembers:\n");
		for(i = 0; i < nomem_rec_cnt; i++)
			printf("\t%s\n", GVAL[nomem_rec[i]] + 2);
	} else
		printf("    No recursive nonmembers\n");

	free(mem_exp); free(mem_imp); free(mem_rec);
	free(nomem_exp); free(nomem_imp); free(nomem_rec);
	nis_destroy_object(obj);

	return;
}

nis_error nis_verifygroup(group)
	nis_name		group;
{
	nis_error		stat;
	nis_object		*obj = NULL;

	stat = __nis_find_group(group, &obj);

	if (obj != NULL)
		nis_destroy_object(obj);

	return(stat);
}
