/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: nislib_names.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $
 */

#include <syslog.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nislib_private.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nislib_names.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $";
#endif

static nis_result *__nis_nameops(name, op, obj)
	nis_name		name;
	uint32_t		op;
	nis_object		*obj;
	
{
	nis_result		*res, *lres;
	struct _nis_binding	*bptr = NULL;
	ns_request		req;
	struct rpc_err		rerr;
	uint32_t		aticks;
	nis_error		r;
	CACHE_CLOCK_DECLS;

	/*
	 * Prepare a locally allocated result struct.
	 */
	res = calloc(1, sizeof(nis_result));

	if (res == NULL) {
		syslog(LOG_ERR, "NIS+: out of memory doing namespace op");
		return(NULL);
	}

	if (name == NULL || strchr(name, '[') != NULL) {
		res->status = NIS_BADNAME;
		return(res);
	}

	/* Sanity check */
	if (obj != NULL && __type_of(obj) == ENTRY_OBJ) {
		res->status = NIS_INVALIDOBJ;
		return(res);
	}

	if (name == NULL) {
		res->status = NIS_BADREQUEST;
		return(res);
	}

	/*
	 * Bind to the server. The binding is not free()d. The binding
	 * module maintains a queue of a handful of recently used
	 * bindings; if we need this one again, we'll have it handy,
	 * and if we don't need it again, it'll get shoved out of the
	 * queue eventually.
	 */
	__start_clock(CACHE_CLOCK);
	r = __nis_bind(nis_domain_of(name), MASTER_ONLY, &bptr);
	if (r != NIS_SUCCESS)
		r = __nis_bind(name, MASTER_ONLY, &bptr);
	res->aticks = aticks = __stop_clock(CACHE_CLOCK);

	if (r != NIS_SUCCESS) {
		res->status = r;
		return(res);
	}

	/* Set up the request */

	req.ns_name = name;
	if (obj == NULL) {
		req.ns_object.ns_object_len = 0;
		req.ns_object.ns_object_val = NULL;
	} else {
		req.ns_object.ns_object_len = 1;
		req.ns_object.ns_object_val = obj;
	}

	switch(op) {
	case NIS_ADD:
		lres = nis_add_3(&req, bptr->nb_clnt);
		break;
	case NIS_REMOVE:
		lres = nis_remove_3(&req, bptr->nb_clnt);
		break;
	case NIS_MODIFY:
		lres = nis_modify_3(&req, bptr->nb_clnt);
		break;
	default:
		syslog(LOG_ERR, "NIS+: bad nameop: %d", op);
		res->status = NIS_BADREQUEST;
		__nis_release_binding(bptr);
		return(res);
		break;
	}

	__nis_release_binding(bptr);

	if (lres == NULL) {
		clnt_geterr(bptr->nb_clnt, &rerr);
		if (rerr.re_status == RPC_AUTHERROR)
			res->status = NIS_SRVAUTH;
		else
			res->status = NIS_RPCERROR;
		return(res);
	}

	/*
	 * lres is really a pointer to a static nis_result struct
	 * which itself contains pointers to other dynamically
	 * allocated data. We want the data and don't care about
	 * the static nis_result struct, so all we do us copy the
	 * pointers and stuff over to our local nis_result struct.
	 * No need to free the static lres, so we're done.
	 */
	bcopy((char *)lres, (char *)res, sizeof(nis_result));
	res->aticks = aticks;

	return(res);
}

void nis_freeresult(res)
	nis_result		*res;
{
	if (res == NULL)
		return;

	xdr_free(xdr_nis_result, (char *)res);
	free(res);

	return;
}


nis_result *nis_lookup(name, flags)
	const nis_name		name;
	const uint32_t		flags;
{
	nis_name		*nlist = NULL;
	nis_result		*res = NULL;
	int			i;
	nis_error		stat = NIS_NOTFOUND;
	CLIENT_CLOCK_DECLS;

	__start_clock(CLIENT_CLOCK);

	if (name == NULL || strchr(name, '[') != NULL) {
		res = calloc(1, sizeof(nis_result));
		if (res == NULL)
			return(NULL);
		res->status = NIS_BADNAME;
		res->cticks = __stop_clock(CLIENT_CLOCK);
		return(res);
	}

	/*
	 * If the name is fully qualified, just handle it
	 * quick and get out.
	 */
	if (!(flags & EXPAND_NAME) && name[strlen(name) - 1] == '.') {
		res = __nis_search(name, SEARCHOP_LOOKUP,
					flags, NULL, 0, NULL, NULL);
		if (res == NULL)
			return(NULL);
		res->cticks = __stop_clock(CLIENT_CLOCK);
		return(res);
	}

	nlist = nis_getnames(name);
	if (nlist == NULL) {
		res = calloc(1, sizeof(nis_result));
		if (res == NULL)
			return(NULL);
		res->status = NIS_BADNAME;
		res->cticks = __stop_clock(CLIENT_CLOCK);
		return(res);
	}

	/*
	 * Name was not fully qualified or caller specified
	 * EXPAND_NAME. We have a list of names to search;
	 * first one that yields success wins.
	 */
	for (i = 0; nlist[i] != NULL; i++) {
		res = __nis_search(nlist[i], SEARCHOP_LOOKUP,
					flags, NULL, 0, NULL, NULL);
		if (res == NULL)
			return(NULL);
		if (res->status != NIS_NAMEUNREACHABLE ||
			res->status != NIS_NOTFOUND)
			break;
		else {
			stat = res->status;
			nis_freeresult(res);
			res = NULL;
		}
	}

	if (res == NULL) {
		res = calloc(1, sizeof(nis_result));
		res->status = NIS_NOSUCHNAME;
	}

	res->cticks = __stop_clock(CLIENT_CLOCK);
	return(res);
}

nis_result *nis_add(name, obj)
	nis_name		name;
	nis_object		*obj;
{
	nis_result		*res;
	CLIENT_CLOCK_DECLS;

	__start_clock(CLIENT_CLOCK);

	res = __nis_nameops(name, NIS_ADD, obj);
	if (res == NULL)
		return(NULL);

	res->cticks = __stop_clock(CLIENT_CLOCK);

	return(res);
}

nis_result *nis_remove(name, obj)
	nis_name		name;
	nis_object		*obj;
{
	nis_result		*res;
	CLIENT_CLOCK_DECLS;

	__start_clock(CLIENT_CLOCK);

	res = __nis_nameops(name, NIS_REMOVE, obj);
	if (res == NULL)
		return(NULL);

	res->cticks = __stop_clock(CLIENT_CLOCK);

	return(res);
}

nis_result *nis_modify(name, obj)
	nis_name		name;
	nis_object		*obj;
{
	nis_result		*res;
	nis_name		oname = name;
	CLIENT_CLOCK_DECLS;

	__start_clock(CLIENT_CLOCK);

#ifdef foo
	/*
	 * XXX The man page says that when renaming an object,
	 * if obj != NULL and obj->zo_name != NULL, obj->zo_name
	 * is used as the new name instead of the supplied 'name'
	 * argument. I don't know yet if this is enforced in the
	 * client library or the server.
	 */
	if (obj != NULL && obj->zo_name != NULL)
		oname = obj->zo_name;
#endif

	res = __nis_nameops(oname, NIS_MODIFY, obj);
	if (res == NULL)
		return(NULL);

	res->cticks = __stop_clock(CLIENT_CLOCK);

	return(res);
}

