/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nislib_server.c,v 1.2 1999/02/09 01:56:04 wpaul Exp $
 */

#include <syslog.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nislib_private.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nislib_server.c,v 1.2 1999/02/09 01:56:04 wpaul Exp $";
#endif

static nis_error nis_dirop __P(( nis_name, nis_server *, unsigned long ));

nis_error nis_mkdir(dirname, machine)
	nis_name		dirname;
	nis_server		*machine;
{
	return(nis_dirop(dirname, machine, NIS_MKDIR));
}

nis_error nis_rmdir(dirname, machine)
	nis_name		dirname;
	nis_server		*machine;
{
	return(nis_dirop(dirname, machine, NIS_RMDIR));
}

static nis_error nis_dirop(dirname, machine, op)
	nis_name		dirname;
	nis_server		*machine;
	unsigned long		op;
{
	struct _nis_binding	*bptr;
	nis_error		*res;
	struct rpc_err		rerr;
	nis_error		r;

	if (machine == NULL || dirname == NULL)
		return(NIS_BADREQUEST);

	r = __nis_create_binding(NULL, machine, 0, &bptr);
	if (r != NIS_SUCCESS)
		return(r);
	if (__nis_connect_server(bptr) != NIS_SUCCESS) {
		__nis_destroy_binding(bptr);
		return(NIS_NAMEUNREACHABLE);
	}

	switch (op) {
	case NIS_MKDIR:
		res = nis_mkdir_3(&dirname, bptr->nb_clnt);
		break;
	case NIS_RMDIR:
		res = nis_rmdir_3(&dirname, bptr->nb_clnt);
		break;
	default:
		syslog(LOG_ERR, "NIS+: bad dirop value: %d", op);
		__nis_destroy_binding(bptr);
		return(NIS_BADREQUEST);
		break;
	}

	if (res == NULL) {
		clnt_geterr(bptr->nb_clnt, &rerr);
		__nis_destroy_binding(bptr);
		if (rerr.re_status == RPC_AUTHERROR)
			return(NIS_SRVAUTH);
		else
			return(NIS_RPCERROR);
	}

	__nis_destroy_binding(bptr);

	return(*res);
}

nis_error nis_servstate(machine, tags, numtags, result)
	nis_server		*machine;
	nis_tag			*tags;
	int			numtags;
	nis_tag			**result;
{
	nis_taglist		tlist, *rlist;
	struct _nis_binding	*bptr;
	nis_error		res;
	struct rpc_err		err;

	if (machine == NULL)
		return(NIS_BADREQUEST);

	res = __nis_create_binding(NULL, machine, 0, &bptr);
	if (res != NIS_SUCCESS)
		return(res);
	res = __nis_connect_server(bptr);

	if (res != NIS_SUCCESS)
		return(NIS_NAMEUNREACHABLE);

	tlist.tags.tags_len = numtags;
	tlist.tags.tags_val = tags;

	rlist = nis_servstate_3(&tlist, bptr->nb_clnt);

	if (rlist == NULL) {
		clnt_geterr(bptr->nb_clnt, &err);
		__nis_destroy_binding(bptr);
		if (err.re_status == RPC_AUTHERROR)
			res = NIS_SRVAUTH;
		else
			res = NIS_RPCERROR;
		return(res);
	} else {
		*result = rlist->tags.tags_val;
		res = NIS_SUCCESS;
	}
		
	__nis_destroy_binding(bptr);

	return(res);
}

nis_error nis_stats(machine, tags, numtags, result)
	nis_server		*machine;
	nis_tag			*tags;
	int			numtags;
	nis_tag			**result;
{
	nis_taglist		tlist, *rlist;
	struct _nis_binding	*bptr;
	nis_error		res;
	struct rpc_err		rerr;

	if (machine == NULL)
		return(NIS_BADREQUEST);

	res = __nis_create_binding(NULL, machine, 0, &bptr);
	if (res != NIS_SUCCESS)
		return(res);
	res = __nis_connect_server(bptr);

	if (res != NIS_SUCCESS)
		return(NIS_NAMEUNREACHABLE);

	tlist.tags.tags_len = numtags;
	tlist.tags.tags_val = tags;

	rlist = nis_status_3(&tlist, bptr->nb_clnt);

	if (rlist == NULL) {
		clnt_geterr(bptr->nb_clnt, &rerr);
		__nis_destroy_binding(bptr);
		if (rerr.re_status == RPC_AUTHERROR)
			res = NIS_SRVAUTH;
		else
			res = NIS_RPCERROR;
		return(res);
	} else {
		*result = rlist->tags.tags_val;
		res = NIS_SUCCESS;
	}
		
	__nis_destroy_binding(bptr);

	return(res);
}

nis_server **nis_getservlist(dirname)
	nis_name		dirname;
{
	nis_server		**srvlist;
	directory_obj		*dir;
	unsigned long		i;
	nis_error		r;

	/*
	 * This calls into the low-level server locator code
	 * in the binder. This code may in turn call the cache
	 * code.
	 */
	r = __nis_find_server(dirname, 0, &dir);

	if (r != NIS_SUCCESS)
		return(NULL);

	i = dir->do_servers.do_servers_len;

	srvlist = calloc(i + 1, sizeof(nis_server **));

	if (srvlist == NULL) {
		syslog(LOG_ERR, "NIS+: out of memory allocating srvlist");
		xdr_free(xdr_directory_obj, (char *)dir);
		free(dir);
		return(NULL);
	}

	/*
	 * copy pointers -- dir is dynamically allocated, no
	 * need to allocate more memory when we can just re-use
	 * what's already there.
	 */
	for (i = 0; i < dir->do_servers.do_servers_len; i++) 
		srvlist[i] = &dir->do_servers.do_servers_val[i];
	
	dir->do_servers.do_servers_len = 0;
	dir->do_servers.do_servers_val = NULL;

	xdr_free(xdr_directory_obj, (char *)dir);
	free(dir);

	return(srvlist);
}

void nis_freeservlist(machines)
	nis_server **machines;
{
	nis_server		**m = machines;

	while(*m != NULL) {
		xdr_free(xdr_nis_server, (char *)*m);
		free(*m);
		m++;
	}

	free(machines);

	return;
}

void nis_freetags(tags, numtags)
	nis_tag			*tags;
	int			numtags;
{
	nis_taglist		tlist;

	tlist.tags.tags_val = tags;
	tlist.tags.tags_len = numtags;

	xdr_free(xdr_nis_taglist, (char *)&tlist);

	return;
}
