/*
 * Copyright (c) 1996
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: nislib_subr.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $
 */

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <sys/cdefs.h>
#include <sys/param.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nislib_private.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nislib_subr.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $";
#endif

static int validname(str)
	char *str;
{
	if (str == NULL)	/* string pointer shouldn't be NULL */
		return(1);

	if (!strlen(str))
		return(1);

	switch(str[0]) {
	case ' ':
	case '\t':	/* tab -- whitespace */
	case '.':
#ifdef foo
	case '[':	/* XXX */
#endif
	case ']':
	case ',':
	case '=':
	case '@':		
	case '+':		/* 'The initial character may not be a */
/*	case '-': */		/* terminal o character or the characters */
		return(1);	/* '@', '+' or '-'. (No terminals?) */
		break;
	default:
		break;
	}

	if (strchr(str, '/'))
		return(1);	/* No slashes allowed anywhere in */
				/* an NIS+ name. */
	return(0);
}

static char terminators[] = { /*'[',*/ ']', ',', '=', ' ', '\t', '.' };
#define NUMTERMS 6

char *break_name(str, action)
	char *str;
	int action;
{
	static char buf[NIS_MAXNAMELEN];
	char *bptr, *optr;
	int quote = 0;
	int size = strlen(str);
	int terminals = 0;
	int has_attrs = 0;

	bptr = (char *)&buf;
	optr = str;

	bzero(bptr, NIS_MAXNAMELEN);

	while(1) {
		switch(*str) {
		case '[':
			if (quote) {
				*bptr = *str;
				bptr++; str++;
				break;
			}
			/* signal start of entry */
			if (!strchr(str, ']'))
				return(NULL);
			has_attrs++;
			str = strchr(str, ']');
			str++;
			break;
		case ']':
		case ',':
		case '.':
		case ' ':
		case '\t':	/* tab -- whitespace */
		case '=':
		case '@':		
		case '+':
	/*	case '-': */
			if (quote) {
				*bptr = *str;
				bptr++; str++;
				break;
			}
			if (*str == '.')
				terminals++;
			str++;
			break;
		case '"':
			if (quote)
				quote = 0;
			else
				quote = 1;
			str++;
			break;
		default:
			*bptr = *str;
			bptr++; str++;
			break;
		}
		if ((str - optr) == size)
			break;
		if (terminals)
			break;
	}
	if (action && !has_attrs)
		return(NULL);

	return((char *)&buf);
}

static char *next_term(str)
	char *str;
{
	int i;
	char *ptr = NULL;

	for (i = 0; i < NUMTERMS; i++) {
		if ((ptr = strchr(str, terminators[i])) != NULL) {
			return(ptr);
		}
	}

	return(NULL);
}

nis_name nis_leaf_of(name)
	const nis_name name;
{
	static nis_name res;

	if (validname(name))
		return(NULL);

	res = break_name(name, 0);

	return(res);
}

nis_name nis_name_of(name)
	const nis_name name;
{
	char *ptr;
	static char new[NIS_MAXNAMELEN];
	nis_name		localdom = nis_local_directory();
	nis_name res;

	if (validname(name))
		return(NULL);

	if (!strchr(name, '.'))
		return(NULL);

	if ((ptr = strstr(name, localdom)) == NULL)
		return(name);

	bzero((char *)&new, sizeof(new));
	if (ptr != name)
		snprintf(new, sizeof(new), "%.*s", (ptr - 1) - name, name);

	res = (char *)&new;

	return(strlen(res) ? res : NULL);
}

nis_name nis_domain_of(name)
	const nis_name name;
{
	char *ptr1, *ptr2;

	if (validname(name))
		return(NULL);

	ptr1 = next_term(name);

	if (ptr1 == NULL)
		return(".");

	while(ptr1 && *ptr1 != '.') {
		ptr1++;
		ptr2 = next_term(ptr1);
		ptr1 = ptr2;
	}

	ptr1++;

	return(ptr1);
}

static int __dots(name, suff)
	nis_name		name;
	nis_name		suff;
{
	char			*ptr;
	int			dots = 0;
	char			dbuf[NIS_MAXNAMELEN];

	if (suff != NULL) {
		strncpy(dbuf, name, NIS_MAXNAMELEN);
		strncat(dbuf, ".", NIS_MAXNAMELEN);
		strncat(dbuf, suff, NIS_MAXNAMELEN);
	}
	if (dbuf[strlen(dbuf) - 1] != '.')
		strncat(dbuf, ".", NIS_MAXNAMELEN);

	ptr = dbuf;

	while(ptr != dbuf + strlen(dbuf)) {
		if (*ptr == '.')
			dots++;
		ptr++;
	}

	return(dots);
}

#define NEWNAME \
	names = realloc(names, (namecount + 2) * sizeof(nis_name *)); \
	strncpy(nbuf, name, NIS_MAXNAMELEN); \
	strncat(nbuf, ".", NIS_MAXNAMELEN); \
	strncat(nbuf, ptr, NIS_MAXNAMELEN); \
	if (ptr[strlen(ptr) - 1] != '.') \
		strncat(nbuf, ".", NIS_MAXNAMELEN); \
	names[namecount] = strdup(nbuf); \
	namecount++; \
	names[namecount] = NULL;

#define NEWLOCNAME \
	names = realloc(names, (namecount + 2) * sizeof(nis_name *)); \
	strncpy(nbuf, name, NIS_MAXNAMELEN); \
	strncat(nbuf, ".", NIS_MAXNAMELEN); \
	strncat(nbuf, ptr, NIS_MAXNAMELEN); \
	strncat(nbuf, locdir, NIS_MAXNAMELEN); \
	names[namecount] = strdup(nbuf); \
	namecount++; \
	names[namecount] = NULL;

nis_name *nis_getnames(name)
	const nis_name		name;
{
        char                    *path, *save;
        nis_name                *names;
	nis_name		locdir = nis_local_directory();
	char			*ptr;
	int			namecount = 0;
	char			nbuf[NIS_MAXNAMELEN];

        if (name == NULL)
                return(NULL);

	names = calloc(2, sizeof(nis_name *));

	if (name[strlen(name) - 1] == '.') {
		names[0] = strdup(name);
		names[1] = NULL;
		return(names);
	}

	names[1] = NULL;

        path = getenv("NIS_PATH");
        if (path == NULL)
                save = path = strdup("$");
        else
                save = path = strdup(path);

	ptr = strsep(&path, ":");

	while(ptr != NULL) {
		if (strchr(ptr, '$')) {
			if (strlen(ptr) > 1) {
				*strchr(ptr, '$') = '\0';
				NEWLOCNAME;
			} else {
				ptr = locdir;
				while(__dots(name, ptr) > 2 && ptr != NULL) {
					NEWNAME;
					ptr = nis_domain_of(ptr);
				}
			}
		} else {
			NEWNAME;
		}
		ptr = strsep(&path, ":");
	}

	free(save);
	return(names);
}

void nis_freenames(namelist)
	nis_name		*namelist;
{
	nis_name		*list = namelist;

	if (namelist == NULL)
		return;

	while(*list != NULL) {
		free(*list);
		list++;
	}

	free(namelist);

	return;
}

name_pos nis_dir_cmp(n1, n2)
	const nis_name		n1;
	const nis_name		n2;
{
	int			len1,
				len2;

	/* Check for bad names */
	if (validname(n1) || validname(n2))
		return(BAD_NAME);

	len1 = strlen(n1);
	len2 = strlen(n2);

	if (len1 > len2) {
		/*
		 * If n1 is longer than n2, then n1 may be:
		 * - a child of n2, or
		 * - a child of a sibling of n2.
		 * Either way, all we know for sure is that it
		 * can't be a parent of n2, so HIGHER_NAME is
		 * eliminated.
		 */
		if (!strncasecmp(n1 + (len1 - len2), n2, len2))
			return(LOWER_NAME);
	} else if (len1 < len2) {
		/*
		 * If n1 is shorter than n2, then n1 may be:
		 * - a parent of n2, or
		 * - a parent of a sibling of n2
		 * Either way, LOWER_NAME is eliminated.
		 */
		if (!strncasecmp(n1, n2 + (len2 - len1), len1))
			return(HIGHER_NAME);
	}

	/*
	 * If n1 and n2 are both the same size, they may
	 * be the same name. Either way, both HIGHER_NAME
	 * and LOWER_NAME are eliminated.
	 */
	if (!strcasecmp(n1, n2))
		return(SAME_NAME);

	return(NOT_SEQUENTIAL);
}

static unsigned char *nislib_encode_nis_obj(obj, s)
	nis_object *obj;
	uint32_t *s;
{
	uint32_t size = sizeof(nis_object);
	char *addr = NULL;
	XDR xdrs;

	size = xdr_sizeof(xdr_nis_object, (char *)obj);

	if ((addr = calloc(1, size)) == NULL)
		return(NULL);

	xdrmem_create(&xdrs, addr, size, XDR_ENCODE);
	if (!xdr_nis_object(&xdrs, obj)) {
		xdr_destroy(&xdrs);
		free(addr);
		return(NULL);
	}

	*s = size;
	xdr_destroy(&xdrs);

	/* Note: must be free()'ed later! */
	return(addr);
}

nis_object *nis_clone_object(src, dest)
	nis_object *src;
	nis_object *dest;
{
	unsigned char *addr;
	unsigned int s;
	XDR xdrs;
	nis_object *obj;

	if (src == NULL)
		return (NULL);

	addr = nislib_encode_nis_obj(src, &s);

	if (addr == NULL)
		return(NULL);

	if (dest == NULL) {
		if ((obj = calloc(1, sizeof(nis_object))) == NULL) {
			free(addr);
			return(NULL);
		}
	} else
		obj = dest;
	
	xdrmem_create(&xdrs, addr, s, XDR_DECODE);
	if (!xdr_nis_object(&xdrs, obj)) {
		xdr_destroy(&xdrs);
		free(addr);
		return(NULL);
	}
	xdr_destroy(&xdrs);
	free(addr);

	return(obj);
}

void nis_destroy_object(obj)
	nis_object *obj;
{
	if (obj == NULL)
		return;
	xdr_free(xdr_nis_object, (char *)obj);
	free(obj);
}

void nis_print_rights(acc)
	uint32_t		acc;
{
	int			i, j;
	char			str[20];

	j = 0;
	for (i = 3; i >= 0; i--) {
		if ((acc >> (i * 8) & 0x000000FF) & NIS_READ_ACC)
			str[j] = 'r';
		else
			str[j] = '-';
		j++;
		if ((acc >> (i * 8) & 0x000000FF) & NIS_MODIFY_ACC)
			str[j] = 'm';
		else
			str[j] = '-';
		j++;
		if ((acc >> (i * 8) & 0x000000FF) & NIS_CREATE_ACC)
			str[j] = 'c';
		else
			str[j] = '-';
		j++;
		if ((acc >> (i * 8) & 0x000000FF) & NIS_DESTROY_ACC)
			str[j] = 'd';
		else
			str[j] = '-';
		j++;
	}

	str[j] = '\0';

	printf("%s", str);
	return;
}

static char *ttltostr(ttl)
	uint32_t		ttl;
{
	int			hours = 0,
				mins = 0,
				secs = 0;

	static char		str[20];

	hours = ttl / 3600;
	mins = (ttl - (hours * 3600)) / 60;
	secs = ttl - (hours * 3600) - (mins * 60);

	snprintf(str, sizeof(str), "%d:%d:%d", hours, mins, secs);

	return(str);
}

static void print_serv(serv)
	nis_server		*serv;
{
	char *keytype[] = { "None", "Diffie-Hellman", "RSA",
			    "Kerberos", "<undefined>", NULL };
	int			i;

	printf("        Name       : %s\n", serv->name != NULL ?
					serv->name : "(nil)");
	printf("        Public Key : %s (%d bits)\n",
					keytype[serv->key_type],
					serv->pkey.n_len * 4);
	printf("        Universal addresses (%d)\n", serv->ep.ep_len);
	for (i = 0; i < serv->ep.ep_len; i++) {
		printf("        [%d] - ", i);
		printf("%s, ", serv->ep.ep_val[i].proto != NULL ?
				serv->ep.ep_val[i].proto : "-");
		printf("%s, ", serv->ep.ep_val[i].family != NULL ?
				serv->ep.ep_val[i].family : "-");
		printf("%s\n", serv->ep.ep_val[i].uaddr != NULL ?
				serv->ep.ep_val[i].uaddr : "-");
	}

	return;
}

void nis_print_directory(dir)
	directory_obj		*dir;
{
	char *dirtype[] = { "UNKNOWN", "NIS", "SUNYP", "IVY", "DNS", "X500",
			    "DNANS", "XCHS", "CDS", "<undefined>", NULL };
	int			i;

	if (dir == NULL)
		return;

	printf("Name : '%s'\n", dir->do_name != NULL ? dir->do_name : "(nil)");
	printf("Type : %s\n", dirtype[dir->do_type]);

	for (i = 0; i < dir->do_servers.do_servers_len; i++) {
		if (i == 0)
			printf("Master Server :\n");
		else
			printf("Replicate :\n");
		print_serv(&dir->do_servers.do_servers_val[i]);
	}

	printf("Time to live : %s\n", ttltostr(dir->do_ttl));
	printf("Default Access Rights :\n");
	for (i = 0; i < dir->do_armask.do_armask_len; i++) {
		struct oar_mask *oa;

		oa = &dir->do_armask.do_armask_val[i];
		printf("        Type         : %s\n", dirtype[oa->oa_otype]);
		printf("        Access rights: ");
		nis_print_rights(oa->oa_rights);
		printf("\n");
	}
	return;
}

static char *tattr(flags)
	uint32_t		flags;
{
	static char		str[NIS_MAXNAMELEN];

	bzero((char *)&str, sizeof(str));

	strcat(str, "(");

	if (flags & TA_SEARCHABLE)
		strcat(str, "SEARCHABLE, ");
	if (flags & TA_BINARY) {
		strcat(str, "BINARY DATA");
		if (flags & TA_XDR)
			strcat(str, ", XDR ENCODED");
		if (flags & TA_ASN1)
			strcat(str, ", ASN.1 ENCODED");
		if (flags & TA_CRYPT)
			strcat(str, ", ENCRYPTED");
	} else {
		strcat(str, "TEXTUAL DATA");
		if (flags & TA_SEARCHABLE) {
			if (flags & TA_CASE)
				strcat(str, ", CASE INSENSITIVE");
			else
				strcat(str, ", CASE SENSITIVE");
		}
	}

	strcat(str, ")");

	return(str);
}

void nis_print_table(tab)
	table_obj		*tab;
{
	int			i;
	table_col		*tc;

	printf("Table type          : %s\n", tab->ta_type);
	printf("Number of columns   : %d\n", tab->ta_maxcol);
	printf("Character Separator : %c\n", tab->ta_sep);
	printf("Search path         : %s\n", tab->ta_path);
	printf("Columns             :\n");

	for (i = 0; i < tab->ta_cols.ta_cols_len; i++) {
		tc = &tab->ta_cols.ta_cols_val[i];
		printf("        [%d]     Name          : %s\n", i, tc->tc_name);
		printf("                Attributes    : %s\n",
							  tattr(tc->tc_flags));
		printf("                Access Rights : ");
		nis_print_rights(tc->tc_rights);
		printf("\n");
	}

	return;
}

void nis_print_entry(ent)
	entry_obj		*ent;
{
	int			i;
	entry_col		*col;

	printf("        Entry of type %s\n", ent->en_type);

	for (i = 0; i < ent->en_cols.en_cols_len; i++) {
		col = &ent->en_cols.en_cols_val[i];
		printf("        [%d] - [%d bytes] ", i,
					col->ec_value.ec_value_len);
		switch(col->ec_flags) {
		case EN_CRYPT:
			printf("Encrypted data\n");
			break;
		case EN_XDR:
			printf("XDR'd Data\n");
			break;
		case EN_BINARY:
			printf("Binary data\n");
			break;
		case EN_ASN1:
			printf("ASN.1 Encoded data\n");
			break;
		default:
			if (col->ec_value.ec_value_val == NULL)
				printf("'(nil)'\n");
			else
				printf("'%.*s'\n", col->ec_value.ec_value_len,
						   col->ec_value.ec_value_val);
			break;
		}
	}

	return;
}

void nis_print_group(grp)
	group_obj		*grp;
{
	int			i;

	printf("Group Flags :");
	if (grp->gr_flags & IMPMEM_GROUPS)
		printf(" Implicit Memberships Allowed");
	if (grp->gr_flags & RECURS_GROUPS)
		printf(" Recursive Memberships Allowed");
	if (grp->gr_flags & NEGMEM_GROUPS)
		printf(" Negative Memberships Allowed");
	printf("\n");

	printf("Group Members :\n");
	for (i = 0; i < grp->gr_members.gr_members_len; i++)
		printf("%s\n", grp->gr_members.gr_members_val[i]);

	return;
}

void nis_print_link(link)
	link_obj		*link;
{
	char *objtype[] = { "BOGUS", "NO", "DIRECTORY",
			    "GROUP", "TABLE", "ENTRY",
			    "LINK", "PRIVATE", "<undefined>", NULL };

	printf("Linked Object Type : %s\n", objtype[link->li_rtype]);
	printf("Link to : %s\n", link->li_name);

	return;
}

void nis_print_object(obj)
	nis_object		*obj;
{
	char *objtype[] = { "BOGUS", "NO", "DIRECTORY",
			    "GROUP", "TABLE", "ENTRY",
			    "LINK", "PRIVATE", "<undefined>", NULL };
	time_t			tmp;

	printf("Object Name   : %s\n", obj->zo_name);
	printf("Directory     : %s\n", obj->zo_domain);
	printf("Owner         : %s\n", obj->zo_owner);
	printf("Group         : %s\n", obj->zo_group);
	printf("Access Rights : ");
	nis_print_rights(obj->zo_access);
	printf("\n");
	printf("Time to Live  : %s\n", ttltostr(obj->zo_ttl));
	tmp = obj->zo_oid.ctime;
	printf("Creation Time : %s", ctime(&tmp));
	tmp = obj->zo_oid.mtime;
	printf("Mod. Time     : %s", ctime(&tmp));
	printf("Object Type   : %s\n", objtype[__type_of(obj)]);

	switch(__type_of(obj)) {
	case DIRECTORY_OBJ:
		nis_print_directory(&obj->DI_data);
		break;
	case LINK_OBJ:
		nis_print_link(&obj->LI_data);
		break;
	case TABLE_OBJ:
		nis_print_table(&obj->TA_data);
		break;
	case ENTRY_OBJ:
		nis_print_entry(&obj->EN_data);
		break;
	case GROUP_OBJ:
		nis_print_group(&obj->GR_data);
		break;
	default:
		printf("unknown object type\n");
		break;
	}

	return;
}
