/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: nislib_tables.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nislib_private.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nislib_tables.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $";
#endif

static nis_result *__nis_ibops(name, obj, op, flags)
	nis_name		name;
	nis_object		*obj;
	uint32_t		op;
	uint32_t		flags;
{
	struct _nis_binding	*bptr = NULL;
	nis_attr		*attrs = NULL;
	uint32_t		numattrs = 0;
	ib_request		ibreq;
	nis_result		*lres = NULL, *res = NULL;
	nis_name		tname;
	struct rpc_err		rerr;
	uint32_t		aticks, lflags = 0;
	char			buf[NIS_MAXNAMELEN];
	nis_error		r;
	CACHE_CLOCK_DECLS;

	res = (nis_result *)calloc(1, sizeof(nis_result));

	if (res == NULL)
		return(NULL);

	if (name == NULL && obj == NULL) {
		res->status = NIS_BADREQUEST;
		return(res);
	}

	if (name == NULL) {
		snprintf(buf, sizeof(buf), "%s.%s",
					obj->zo_name, obj->zo_domain);
		tname = buf;
	} else {
		if (__nis_parse_request(name, &attrs, &numattrs,
					&tname) != NIS_SUCCESS) {
			res->status = NIS_BADNAME;
			return(res);
		}
	}

	if (op != NIS_IBFIRST)
		lflags = flags | MASTER_ONLY;

	__start_clock(CACHE_CLOCK);
	r = __nis_bind(tname, lflags, &bptr);
	if (r != NIS_SUCCESS)
		r = __nis_bind(nis_domain_of(tname), lflags, &bptr);
	res->aticks = aticks = __stop_clock(CACHE_CLOCK);

	if (r != NIS_SUCCESS) {
		if (attrs != NULL) {
			xdr_free((xdrproc_t)xdr_nis_attr, (char *)attrs);
			free(attrs);
			free(tname);
		}
		res->status = r;
		return(res);
	}

	bzero((char *)&ibreq, sizeof(ib_request));

	ibreq.ibr_name = tname;
	ibreq.ibr_srch.ibr_srch_len = numattrs;
	ibreq.ibr_srch.ibr_srch_val = attrs;
	ibreq.ibr_flags = flags;
	if (obj != NULL) {
		ibreq.ibr_obj.ibr_obj_len = 1;
		ibreq.ibr_obj.ibr_obj_val = obj;
	}

	switch(op) {
	case NIS_IBFIRST:
		lres = nis_ibfirst_3(&ibreq, bptr->nb_clnt);
		break;
	case NIS_IBADD:
		lres = nis_ibadd_3(&ibreq, bptr->nb_clnt);
		break;
	case NIS_IBMODIFY:
		lres = nis_ibmodify_3(&ibreq, bptr->nb_clnt);
		break;
	case NIS_IBREMOVE:
		lres = nis_ibremove_3(&ibreq, bptr->nb_clnt);
		break;
	default:
		syslog(LOG_ERR, "NIS+: bad ibop value: %d", op);
		if (attrs != NULL) {
			xdr_free((xdrproc_t)xdr_nis_attr, (char *)attrs);
			free(attrs);
			free(tname);
		}
		res->status = NIS_BADREQUEST;
		__nis_release_binding(bptr);
		return(res);
		break;
	}

	__nis_release_binding(bptr);

	if (attrs != NULL) {
		xdr_free((xdrproc_t)xdr_nis_attr, (char *)attrs);
		free(attrs);
		free(tname);
	}

	if (lres == NULL) {
		clnt_geterr(bptr->nb_clnt, &rerr);
		if (rerr.re_status == RPC_AUTHERROR)
			res->status = NIS_SRVAUTH;
		else
			res->status = NIS_RPCERROR;
		return(res);
	}

	bcopy((char *)lres, (char *)res, sizeof(nis_result));
	res->aticks = aticks;

	return(res);
}

/*
 * Alright, I'd like to state right now, for the record, that I'm pissed.
 * Between the convoluted binding strategy, the inherent complexity of
 * the list/lookup procedures, and all the hoops we have to jump through
 * inside the server, it's virtually impossible to get any of this crap
 * to work with any kind of speed. It was bad enough in NIS v2; at least
 * there the procedures were nice and simple. Here, we have to bend over
 * backwards to establish a binding to the right server, then twist ourselves
 * into a pretzel to decode the stupid 'indexed name' passed to us by the
 * caller, then handle the zillion semantics flags (follow the table
 * path? all results? expand names? mommy?), spin our heads 180 degrees
 * around to deal with callbacks, all the while performing umpteen sanity
 * checks to make sure we end up with a sensible result.
 *
 * Sun could have done this a lot easier by just tweaking NIS v2 a little
 * instead of inventing this monstrosity. But nooooooooooooo....
 */

static void __copy_objs(res, lres)
	nis_result		*res;
	nis_result		*lres;
{
	register int		i, j;
	register char		*from, *to;

	j = NIS_RES_NUMOBJ(res);

	for (i = 0; i < NIS_RES_NUMOBJ(lres); i++) {
		if (!NIS_RES_NUMOBJ(res))
			NIS_RES_OBJECT(res) = malloc(sizeof(nis_object));
		else
			NIS_RES_OBJECT(res) = 
				realloc(NIS_RES_OBJECT(res),
					(NIS_RES_NUMOBJ(res)+1) *
					sizeof(nis_object));

		from = (char *)&lres->objects.objects_val[i];
		to = (char *)&res->objects.objects_val[j];
		bcopy(from, to, sizeof(nis_object));
		bzero(from, sizeof(nis_object));
		j++;
	}

	NIS_RES_NUMOBJ(res) = j;

	NIS_RES_NUMOBJ(lres) = 0;
	free(lres->objects.objects_val);
	lres->objects.objects_val = NULL;

	return;
}

static nis_result *__follow_path(tname, attrs, numattrs, cback, cdata, flags)
	nis_name		tname;
	nis_attr		*attrs;
	uint32_t		numattrs;
	int (*cback)(nis_name, nis_object *, void *);
	void			*cdata;
	uint32_t		flags;
{
	nis_result		*res, *lres = NULL;
	char			tpath[NIS_MAXPATH + 1], *p;
	int			failures = 0, done = 0;
	nis_name		tptr;

	/*
	 * First, we need to get the table object to learn the
	 * table path.
	 * We could call nis_lookup() directly here, but I want
	 * to keep the separation of modules cleaner. It is assumed
	 * that this function will be called only with fully qualified
	 * NIS+ names, so it's okay for us to do this here. (The
	 * internal search routine does not expand names for us.)
	 */
	res = __nis_search(tname, SEARCHOP_LOOKUP, flags,
			    NULL, 0, NULL, NULL);

	if (res == NULL)
		return(NULL);

	if (res->status != NIS_SUCCESS)
		return(res);

	/*
	 * If the returned object isn't a table, we can't
	 * continue. It may be the returned object is a link,
	 * but if the caller had wanted us to follow links, they
	 * would have set the proper flag and the internal search
	 * routine would have followed the link for us.
	 */
	if (__type_of(NIS_RES_OBJECT(res)) != TABLE_OBJ) {
		nis_freeresult(res);
		res = calloc(1, sizeof(nis_result));
		res->status = NIS_INVALIDOBJ;
		return(res);
	}

	/*
	 * Save the path, discard everything else.
	 */
	snprintf(tpath, NIS_MAXPATH, "%s:%s", tname,
				NIS_RES_OBJECT(res)->TA_data.ta_path);
	p = tpath;
	nis_freeresult(res);
	res = NULL;

	while((tptr = strsep(&p, ":")) != NULL) {
		lres = __nis_search(tptr, SEARCHOP_LIST, flags,
		    	attrs, numattrs, cback, cdata);
		/*
		 * Out of memory; we should stop processing.
		 * The ALL_RESULTS flag is supposed to mean
		 * 'read all tables even if you get an error reading
		 * from one,' but if internal search ran out of
		 * memory, it is unlikely that we will be able
		 * to continue.
	 	 */
		if (lres == NULL)
			break;
		switch(lres->status) {
		case NIS_SUCCESS:
			if (flags & ALL_RESULTS) {
				if (res == NULL)
					res = calloc(1, sizeof(nis_result));
				__copy_objs(res, lres);
			} else {
				res = lres;
				done++;
			}
			break;
		case NIS_CBRESULTS:
			if (!(flags & ALL_RESULTS)) {
				res = lres;
				done++;
			}
			break;
		default:
			res = lres;
			if (flags & ALL_RESULTS || flags & FOLLOW_PATH)
				failures++;
			else
				done++;
			break;
		}
		if (res != lres)
			nis_freeresult(lres);
		lres = NULL;
		if (done)
			break;
	}

	if (res == NULL)
		return(NULL);

	if (res->status == NIS_SUCCESS && failures)
		res->status = NIS_S_SUCCESS;
	if (res->status == NIS_NOTFOUND && failures)
		res->status = NIS_S_NOTFOUND;

	return(res);
}

static void __free_attrs(attrs, numattrs)
	nis_attr		*attrs;
	uint32_t           numattrs;
{
	register unsigned int	i;

	for (i = 0; i < numattrs; i++)
		xdr_free((xdrproc_t)xdr_nis_attr, (char *)(attrs + i));

	free(attrs);

	return;
}

nis_result *nis_list(name, flags, callback, userdata)
	nis_name		name;
	uint32_t		flags;
	int (*callback)(nis_name, nis_object *, void *);
	void			*userdata;
{
	nis_attr		*attrs = NULL;
	uint32_t		numattrs = 0;
	nis_result		*res = NULL;
	nis_name		tname;
	nis_name		*nlist = NULL, *nptr;

#ifdef foo
	if (res == NULL) {
		syslog(LOG_ERR, "NIS+: out of memory allocating result");
		return(NULL);
	}
#endif

	if (__nis_parse_request(name, &attrs,
				&numattrs, &tname) != NIS_SUCCESS) {
		res = (nis_result *)calloc(1, sizeof(nis_result));
		res->status = NIS_BADNAME;
		return(res);
	}

	/*
	 * Simple case.
	 */
	if (tname[strlen(tname) - 1] == '.' && !(flags & EXPAND_NAME)) {
		res = __nis_search(tname, SEARCHOP_LIST, flags,
				attrs, numattrs, callback, userdata);
		__free_attrs(attrs, numattrs);
		free(tname);
		return(res);
	}

	if (flags & FOLLOW_PATH) {
		if (flags & EXPAND_NAME) {
			nptr = nlist = nis_getnames(tname);
			while(*nptr != NULL) {
				res = __follow_path(*nptr, attrs, numattrs,
						callback, userdata, flags);
				if (res->status != NIS_NAMEUNREACHABLE ||
					res->status != NIS_NOTFOUND)
					break;
				nis_freeresult(res);
				nptr++;
			}
			nis_freenames(nlist);
		} else
			res = __follow_path(tname, attrs, numattrs,
						callback, userdata, flags);
	} else {
		if (flags & EXPAND_NAME) {
			nptr = nlist = nis_getnames(tname);
			while(*nptr != NULL) {
				res = __nis_search(*nptr,
						    SEARCHOP_LIST,
						    flags, attrs,
						    numattrs, callback,
						    userdata);
				if (res->status != NIS_NAMEUNREACHABLE ||
					res->status != NIS_NOTFOUND)
					break;
				nis_freeresult(res);
				nptr++;
			}
			nis_freenames(nlist);
		} else
			res = __nis_search(tname, SEARCHOP_LIST,
					    flags, attrs, numattrs,
					    callback, userdata);
	}

	if (attrs != NULL)
		__free_attrs(attrs, numattrs);
	if (tname != NULL)
		free(tname);

	return(res);
}

nis_result *nis_add_entry(name, obj, flags)
	nis_name		name;
	nis_object		*obj;
	uint32_t		flags;
{
	nis_result		*res;
	CLIENT_CLOCK_DECLS;

	__start_clock(CLIENT_CLOCK);

	res = __nis_ibops(name, obj, NIS_IBADD, flags);

	if (res == NULL)
		return(NULL);

	res->cticks = __stop_clock(CLIENT_CLOCK);

	return(res);
}

nis_result *nis_modify_entry(name, obj, flags)
	nis_name		name;
	nis_object		*obj;
	uint32_t		flags;
{
	nis_result		*res;
	CLIENT_CLOCK_DECLS;

	__start_clock(CLIENT_CLOCK);

	res = __nis_ibops(name, obj, NIS_IBMODIFY, flags);

	if (res == NULL)
		return(NULL);

	res->cticks = __stop_clock(CLIENT_CLOCK);

	return(res);
}

nis_result *nis_remove_entry(name, obj, flags)
	nis_name		name;
	nis_object		*obj;
	uint32_t		flags;
{
	nis_result		*res;
	CLIENT_CLOCK_DECLS;

	__start_clock(CLIENT_CLOCK);

	res = __nis_ibops(name, obj, NIS_IBREMOVE, flags);

	if (res == NULL)
		return(NULL);

	res->cticks = __stop_clock(CLIENT_CLOCK);

	return(res);
}

nis_result *nis_first_entry(name)
	nis_name		name;
{
	nis_result		*res;
	CLIENT_CLOCK_DECLS;

	__start_clock(CLIENT_CLOCK);

	res = __nis_ibops(name, NULL, NIS_IBFIRST, 0);

	if (res == NULL)
		return(NULL);

	res->cticks = __stop_clock(CLIENT_CLOCK);

	return(res);
}

nis_result *nis_next_entry(name, next)
	nis_name		name;
	netobj			*next;
{
	struct _nis_binding	*bptr = NULL;
	ib_request		ibreq;
	nis_result		*res,*lres;
	struct rpc_err		rerr;
	uint32_t		aticks;
	nis_error		r;
	CLIENT_CLOCK_DECLS;
	CACHE_CLOCK_DECLS;

	__start_clock(CLIENT_CLOCK);

	res = calloc(1, sizeof(nis_result));

	if (res == NULL)
		return(NULL);

	__start_clock(CACHE_CLOCK);
	r = __nis_bind(name, 0, &bptr);
	if (r != NIS_SUCCESS)
		r = __nis_bind(nis_domain_of(name), 0, &bptr);
	res->aticks = aticks = __stop_clock(CACHE_CLOCK);

	if (r != NIS_SUCCESS) {
		res->status = r;
		return(res);
	}

	bzero((char *)&ibreq, sizeof(ib_request));

	ibreq.ibr_name = name;
	ibreq.ibr_cookie.n_len = next->n_len;
	ibreq.ibr_cookie.n_bytes = next->n_bytes;

	lres = nis_ibnext_3(&ibreq, bptr->nb_clnt);

	__nis_release_binding(bptr);

	if (lres == NULL) {
		clnt_geterr(bptr->nb_clnt, &rerr);
		if (rerr.re_status == RPC_AUTHERROR)
			res->status = NIS_SRVAUTH;
		else
			res->status = NIS_RPCERROR;
		res->cticks = __stop_clock(CLIENT_CLOCK);
		return(res);
	}

	bcopy((char *)lres, (char *)res, sizeof(nis_result));
	res->cticks = __stop_clock(CLIENT_CLOCK);
	res->aticks = aticks;

	return(res);
}
