/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nislib_util.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $
 */

#include <sys/types.h>
#include <sys/cdefs.h>
#include <sys/socket.h>

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <fcntl.h>
#include <rpc/rpc.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nislib_private.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nislib_util.c,v 1.1 1998/07/08 19:16:49 wpaul Exp wpaul $";
#endif

/*
 * This hash function is stolen directly from the
 * Berkeley DB package. It already exists inside libc, but
 * it's declared static which prevents us from calling it
 * from here.
 */
/*
 * OZ's original sdbm hash
 */
uint32_t
__nis_hash(keyarg, len)
	const void *keyarg;
	size_t len;
{
	register const u_char *key;
	register size_t loop;
	register u_int32_t h;

#define HASHC   h = *key++ + 65599 * h

	h = 0;
	key = keyarg;
	if (len > 0) {
		loop = (len + 8 - 1) >> 3;

		switch (len & (8 - 1)) {
		case 0:
			do {
				HASHC;
				/* FALLTHROUGH */
		case 7:
				HASHC;
				/* FALLTHROUGH */
		case 6:
				HASHC;
				/* FALLTHROUGH */
		case 5:
				HASHC;
				/* FALLTHROUGH */
		case 4:
				HASHC;
				/* FALLTHROUGH */
		case 3:
				HASHC;
				/* FALLTHROUGH */
		case 2:
				HASHC;
				/* FALLTHROUGH */
		case 1:
				HASHC;
			} while (--loop);
		}
	}
	return (h);
}


/*
 * Turn a 'universal address' into a struct sockaddr_in.
 * Bletch.
 */
nis_error __nis_uaddr_to_sockaddr(endpt, sin)
	endpoint *endpt;
	struct sockaddr_in *sin;
{
	unsigned char p_bytes[2];
	int i;
	uint32_t a[6];

#if BYTE_ORDER == LITTLE_ENDIAN
	i = sscanf(endpt->uaddr, "%u.%u.%u.%u.%u.%u", &a[0], &a[1], &a[2],
						&a[3], &a[4], &a[5]);
#else
	i = sscanf(endpt->uaddr, "%u.%u.%u.%u.%y.%y", &a[3], &a[2], &a[1],
						&a[0], &a[4], &a[5]);
#endif
	if (i < 6)
		return(NIS_FAIL);

	bzero((char *)sin, sizeof(struct sockaddr_in));

	for (i = 0; i < 4; i++)
		sin->sin_addr.s_addr |= (a[i] & 0x000000FF) << (8 * i);

	p_bytes[0] = (unsigned char)a[4] & 0x000000FF;
	p_bytes[1] = (unsigned char)a[5] & 0x000000FF;

	sin->sin_family = AF_INET; /* always */
	bcopy((char *)&p_bytes, (char *)&sin->sin_port, 2);

	return (NIS_SUCCESS);
}

nis_error __nis_parse_request(name, attr, numattr, tname)
	nis_name		name;
	nis_attr		**attr;
	uint32_t		*numattr;
	nis_name		*tname;
{
	nis_name		str = name;
	nis_attr		*attrs = NULL, *ptr = NULL;
	uint32_t		numattrs = 0;
	char			buf[NIS_MAXNAMELEN];
	char			*bptr, *optr;
	int			quote = 0;
	int			size = strlen(str);
	int			done = 0;

	if (name == NULL || !strlen(name))
		return(NIS_BADNAME);

	bptr = buf;
	optr = str;

	bzero(buf, NIS_MAXNAMELEN);

	while(1) {
		switch(*str) {
		case '[':
			if (quote) {
				*bptr = *str;
				bptr++; str++;
				break;
			}
			ptr = attrs = malloc(sizeof(nis_attr));
			str++;
			break;
		case ',':
		case ']':
			if (quote) {
				*bptr = *str;
				bptr++; str++;
				break;
			}
			if (strlen(buf)) {
				ptr->zattr_val.zattr_val_val = strdup(buf);
				ptr->zattr_val.zattr_val_len = strlen(buf) + 1;
				bzero(buf, NIS_MAXNAMELEN);
				bptr = buf;
				numattrs++;
			}
			if (*str == ']')
				done++;
			else {
				attrs = realloc(attrs, sizeof(nis_attr) *
						(numattrs + 1));
				ptr = attrs + numattrs;
/*				numattrs++; */
			}
			str++;
			break;
/*		case '.': */
		case ' ':
		case '\t':
/*		case '@': */
/*		case '-': */
		case '+':
			if (quote) {
				*bptr = *str;
				bptr++; str++;
				break;
			}
			str++;
			break;
		case '=':
			if (quote) {
				*bptr = *str;
				bptr++; str++;
				break;
			}
			ptr->zattr_ndx = strdup(buf);
			bzero(buf, NIS_MAXNAMELEN);
			bptr = buf;
			str++;
			break;
		case '"':
			if (quote)
				quote = 0;
			else
				quote = 1;
			str++;
			break;
		default:
			*bptr = *str;
			bptr++; str++;
			break;
		}
		if ((str - optr) == size || done)
			break;
	}

	if (numattrs) {
		*attr = attrs;
		*numattr = numattrs;
	} else {
		*attr = 0;
		/**numattr = NULL;*/
		if (attrs != NULL)
			free(attrs);
	}

	if (done)
		*tname = strdup(str + 1);
	else
		*tname = strdup(name);

	return(NIS_SUCCESS);
}

int __start_clock(c)
	struct timeval		*c;
{
	if (c == NULL)
		return(-1);

	gettimeofday(c, NULL);
	return(0);
}

uint32_t __stop_clock(c)
	struct timeval		*c;
{
	struct timeval		t_end;

	if (c == NULL)
		return(0);

	gettimeofday(&t_end, NULL);

	return((t_end.tv_usec - c->tv_usec) +
	       ((t_end.tv_sec - c->tv_sec) * 1000000));
}

nis_error nis_put_cold_start(cold_start_obj)
	directory_obj *cold_start_obj;
{
	FILE			*fp;
	XDR			xdrs;
	int			fd;
	struct flock		lock;

	if ((fp = fopen(_PATH_NIS_COLD_START, "w+")) == NULL)
		return(NIS_UNAVAIL);

	fd = fileno(fp);
	lock.l_start = 0;
	lock.l_len = 0;
	lock.l_pid = getpid();
	lock.l_type = F_WRLCK;
	lock.l_whence = SEEK_SET;

	/* XXX I don't like this. */
	if (fcntl(fd, F_SETLKW, &lock) == -1) {
		syslog(LOG_ERR, "NIS+: couldn't get \
exclusive lock on cold start file: %m");
		fclose(fp);
		return(NIS_TRYAGAIN);
	}

	lock.l_type = F_UNLCK;

	xdrstdio_create(&xdrs, fp, XDR_ENCODE);

	if (!xdr_directory_obj(&xdrs, cold_start_obj)) {
		xdr_destroy(&xdrs);
		fcntl(fd, F_SETLK, &lock);
		fclose(fp);
		return(NIS_SYSTEMERROR);
	}

	xdr_destroy(&xdrs);
	fcntl(fd, F_SETLK, &lock);
	fclose(fp);

	return(NIS_SUCCESS);
}

nis_error nis_get_cold_start(cold_start_obj)
	directory_obj **cold_start_obj;
{
	FILE			*fp;
	XDR			xdrs;
	int			fd;
	struct flock		lock;

	*cold_start_obj = calloc(1, sizeof(directory_obj));
	if (cold_start_obj == NULL)
		return(NIS_NOMEMORY);

	if ((fp = fopen(_PATH_NIS_COLD_START, "r")) == NULL)
		return(NIS_UNAVAIL);

	fd = fileno(fp);
	lock.l_start = 0;
	lock.l_len = 0;
	lock.l_pid = getpid();
	lock.l_type = F_RDLCK;
	lock.l_whence = SEEK_SET;

	if (fcntl(fd, F_SETLKW, &lock) == -1) {
		syslog(LOG_ERR, "NIS+: couldn't get \
shared lock on cache file: %m");
		fclose(fp);
		free(*cold_start_obj);
		return(NIS_TRYAGAIN);
	}

	lock.l_type = F_UNLCK;

	xdrstdio_create(&xdrs, fp, XDR_DECODE);

	if (!xdr_directory_obj(&xdrs, *cold_start_obj)) {
		xdr_destroy(&xdrs);
		free(*cold_start_obj);
		fcntl(fd, F_SETLK, &lock);
		return(NIS_SYSTEMERROR);
	}

	xdr_destroy(&xdrs);
	fcntl(fd, F_SETLK, &lock);
	fclose(fp);

	/* Must be free()d later. */

	return(NIS_SUCCESS);
}

char *__nis_getpkey(pname, netname)
	char			*pname;
	char			*netname;
{
	char			buf[NIS_MAXNAMELEN];
	char			pkey[HEXKEYBYTES + 1];
	char			*p = NULL;
	nis_result		*res;
	int			len = 0;
	struct nis_attr		attr_p[2] = { { "cname" ,    { 0, NULL } },
					      { "auth_type", { 4, "DES" } } };
	struct nis_attr		attr_n[2] = { { "auth_name" , { 0, NULL } },
					      { "auth_type",  { 4, "DES" } } };
	struct nis_attr		*attr = NULL;

	if (netname == NULL && pname == NULL)
		return(NULL);
	if (netname != NULL && pname != NULL)
		return(NULL);

	if (netname != NULL) {
		p = strchr(netname, '@');
		if (p == NULL)
			return(NULL);
		p++;
	
		snprintf(buf, sizeof(buf), "cred.org_dir.%s.", p);

		attr_n[0].zattr_val.zattr_val_len = strlen(netname) + 1;
		attr_n[0].zattr_val.zattr_val_val = netname;
		attr = attr_n;
	} else {
		p = strchr(pname, '.');
		if (p == NULL)
			return(NULL);

		p++;

		snprintf(buf, sizeof(buf), "cred.org_dir.%s", p);

		attr_p[0].zattr_val.zattr_val_len = strlen(pname) + 1;
		attr_p[0].zattr_val.zattr_val_val = pname;
		attr = attr_p;
	}

	res = __nis_search(buf, SEARCHOP_LIST, 0, attr, 2, NULL, NULL);

	if (res == NULL)
		return(NULL);

	if (res->status != NIS_SUCCESS) {
		nis_freeresult(res);
		return(NULL);
	}

	len = ENTRY_LEN(NIS_RES_OBJECT(res), 3);
	strncpy(pkey, ENTRY_VAL(NIS_RES_OBJECT(res), 3), len);
	pkey[len] = NUL;
	p = strchr(pkey, ':');
	if (p != NULL)
		*p = NUL;

	nis_freeresult(res);

	return(strdup(pkey));
}

directory_obj *__nis_copy_dir(dir)
	directory_obj		*dir;
{
	char			*addr;
	directory_obj		*newdir = NULL;
	XDR			xdrs;
	int			size;

	size = xdr_sizeof((xdrproc_t)xdr_directory_obj, (char *)dir);
	addr = malloc(size);
	xdrmem_create(&xdrs, addr, size, XDR_ENCODE);
	xdr_directory_obj(&xdrs, dir);
	xdr_destroy(&xdrs);
	xdrmem_create(&xdrs, addr, size, XDR_DECODE);
	newdir = calloc(1, sizeof(directory_obj));
	xdr_directory_obj(&xdrs, newdir);
	xdr_destroy(&xdrs);
	free(addr);

	return(newdir);
}

nis_server *__nis_copy_srv(srv)
	nis_server		*srv;
{
	char			*addr;
	nis_server		*newsrv = NULL;
	XDR			xdrs;
	int			size;

	size = xdr_sizeof((xdrproc_t)xdr_nis_server, (char *)srv);
	addr = malloc(size);
	xdrmem_create(&xdrs, addr, size, XDR_ENCODE);
	xdr_nis_server(&xdrs, srv);
	xdr_destroy(&xdrs);
	xdrmem_create(&xdrs, addr, size, XDR_DECODE);
	newsrv = calloc(1, sizeof(nis_server));
	xdr_nis_server(&xdrs, newsrv);
	xdr_destroy(&xdrs);
	free(addr);

	return(newsrv);
}
