/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nis_common.c,v 1.7 1998/05/27 19:00:23 wpaul Exp $
 */

#include <sys/cdefs.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <pwd.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nis_common.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nis_common.c,v 1.7 1998/05/27 19:00:23 wpaul Exp $";
#endif


struct passwd *getpwuid_nisplus_master(uid, domain)
	uid_t			uid;
	nis_name		domain;
{
	nis_object		*o;
	nis_result		*res;
	nis_error		r;
	char			tname[NIS_MAXNAMELEN];
	unsigned long		flags = MASTER_ONLY|EXPAND_NAME|FOLLOW_PATH;
	static struct passwd	pw;

        snprintf(tname, sizeof(tname), "[uid=%lu],passwd.org_dir.%s",
						(u_long)uid, domain);
        res = nis_list(tname, flags, NULL, NULL);
        if (res == NULL)
		return(NULL);
        r = res->status;
        if (r != NIS_SUCCESS) {
                nis_freeresult(res);
		return(NULL);
        }
        if (NIS_RES_NUMOBJ(res) > 1) {
                nis_freeresult(res);
		return(NULL);
        }
        o = NIS_RES_OBJECT(res);
        if (__type_of(o) != ENTRY_OBJ) {
                nis_freeresult(res);
		return(NULL);
        }

	bzero((char *)&pw, sizeof(struct passwd));

	if (ENTRY_LEN(o, 0)) {
		pw.pw_name = strdup(ENTRY_VAL(o, 0));
		pw.pw_fields |= _PWF_NAME;
	}
	if (ENTRY_LEN(o, 1)) {
		pw.pw_passwd = strdup(ENTRY_VAL(o, 1));
		pw.pw_fields |= _PWF_PASSWD;
	}
	if (ENTRY_LEN(o, 2)) {
		pw.pw_uid = strtol(ENTRY_VAL(o, 2), NULL, 10);
		pw.pw_fields |= _PWF_UID;
	}
	if (ENTRY_LEN(o, 3)) {
		pw.pw_uid = strtol(ENTRY_VAL(o, 3), NULL, 10);
		pw.pw_fields |= _PWF_GID;
	}
	if (ENTRY_LEN(o, 4)) {
		pw.pw_gecos = strdup(ENTRY_VAL(o, 4));
		pw.pw_fields |= _PWF_GECOS;
	}
	if (ENTRY_LEN(o, 5)) {
		pw.pw_dir = strdup(ENTRY_VAL(o, 5));
		pw.pw_fields |= _PWF_DIR;
	}
	if (ENTRY_LEN(o, 6)) {
		pw.pw_shell = strdup(ENTRY_VAL(o, 6));
		pw.pw_fields |= _PWF_SHELL;
	}

	nis_freeresult(res);

	return(&pw);
}

/*
 * The Solaris nisgrpadm man page says we're supposed to allow the
 * -M flag to force MASTER_ONLY when doing group lookups, but none
 * of the documented NIS+ groups routines in the API library accept
 * a 'flags' argument that allows us to specify this. *sigh* So much
 * for a well thought-out design.
 */
nis_error nis_verifygroup_local(group, flags, grp, gname)
	nis_name		group;
	unsigned long		flags;
	group_obj		**grp;
	char			**gname;
{
	nis_result		*res;
	char			buf[NIS_MAXNAMELEN];
	nis_error		r;
	nis_error		*g = NULL;
	char			*n;

	if (group[strlen(group) - 1] != '.') {
		snprintf(buf, sizeof(buf), "%s.groups_dir", group);
		res = nis_lookup(buf, flags);
	} else {
		snprintf(buf, sizeof(buf), "%s.groups_dir.%s",
			nis_leaf_of(group), nis_domain_of(group));
		res = nis_lookup(buf, flags);
	}

	if (res == NULL)
		return(NIS_NOMEMORY);

	r = res->status;

	if (res->status != NIS_SUCCESS) {
		nis_freeresult(res);
		return(r);
	}

	if (__type_of(NIS_RES_OBJECT(res)) != GROUP_OBJ) {
		nis_freeresult(res);
		return(NIS_BADOBJECT);
	}

	if (grp != NULL) {
		g = malloc(sizeof(group_obj));
		bcopy((char *)&NIS_RES_OBJECT(res)->GR_data,
			(char *)g, sizeof(group_obj));
		bzero((char *)&NIS_RES_OBJECT(res)->GR_data, sizeof(group_obj));
		*(caddr_t *)grp = (caddr_t)g;
	}

	if (gname != NULL) {
		nis_object		*o = NIS_RES_OBJECT(res);
		n = malloc(strlen(o->zo_name) +
			strlen(nis_domain_of(o->zo_domain)) + 1);
		strcpy(n, o->zo_name);
		strcat(n, ".");
		strcat(n, nis_domain_of(o->zo_domain));
		*(caddr_t *)gname = (caddr_t)n;
	}

	nis_freeresult(res);
	return(r);
}

nis_error nis_verifydom(dom, flags)
	nis_name		dom;
	unsigned long		flags;
{
	nis_result		*res;
	nis_error		r;
	char			buf[NIS_MAXNAMELEN];

	snprintf(buf, sizeof(buf), "cred.org_dir.%s", dom);

	res = nis_lookup(buf, flags);

	if (res == NULL)
		return(NIS_NOMEMORY);

	r = res->status;
	nis_freeresult(res);
	return(r);
}

nis_error nis_verifyprinc(princ, flags, pname)
	nis_name		princ;
	unsigned long		flags;
	char			**pname;
{
	nis_result		*res;
	char			buf[NIS_MAXNAMELEN];
	nis_error		r;

	if (princ[0] == '-')
		princ++;
	if (princ[0] == '@')
		return(nis_verifygroup_local(princ + 1, flags, NULL, pname));
	else if (princ[0] == '*')
		return(nis_verifydom(princ + 2, flags));
	else {
		snprintf(buf, sizeof(buf), "[cname=%s],cred.org_dir", princ);

		res = nis_list(buf, flags, NULL, NULL);

		if (res == NULL)
			return(NIS_NOMEMORY);

		r = res->status;

		if (res->status != NIS_SUCCESS) {
			nis_freeresult(res);
			return(r);
		}

		if (__type_of(NIS_RES_OBJECT(res)) != ENTRY_OBJ) {
			nis_freeresult(res);
			return(NIS_BADOBJECT);
		}

		nis_freeresult(res);
		return(r);
	}

	return(NIS_FAIL);
}

nis_error nis_expandprinc(princ, flags, pname)
	nis_name		princ;
	unsigned long		flags;
	char			**pname;
{
	nis_name		*nlist;
	int			i;
	nis_error		r = NIS_NOTFOUND;

	nlist = nis_getnames(princ);

	if (nlist == NULL)
		return(NIS_NOMEMORY);

	for (i = 0; nlist[i] != NULL; i++) {
		r = nis_verifyprinc(nlist[i], flags, NULL);
		if (r == NIS_SUCCESS)
			break;
	}

	if (r == NIS_SUCCESS)
		*(caddr_t *)pname = (caddr_t)strdup(nlist[i]);

	nis_freenames(nlist);
	return(r);
}

int nis_parse_modes(str, omode, a_mask, s_mask)
	char			*str;
	unsigned long		*omode;
	unsigned long		*a_mask, *s_mask;
{
	int			n = 0, o = 0, g = 0, w = 0, a = 0;
	int			r = 0, m = 0, c = 0, d = 0;
	unsigned long		mode = 0, imode = 0, mask = 0;
	int			done = 0;
	char			*ptr = str;
	unsigned long		add_mask = 0, sub_mask = 0;
	int			add = 0, sub = 0, equals = 0;

	if (ptr == NULL || !strlen(ptr) || omode == NULL)
		return(-1);

	while(!done) {
		switch(*ptr) {
		case 'n':		/* nobody */
			if (r || m || c || d)
				return(-1);
			n++;
			break;
		case 'o':		/* owner */
			if (r || m || c || d)
				return(-1);
			o++;
			break;
		case 'g':		/* group */
			if (r || m || c || d)
				return(-1);
			g++;
			break;
		case 'w':		/* world */
			if (r || m || c || d)
				return(-1);
			w++;
			break;
		case 'a':		/* all (owg) */
			o++; w++; g++;
			break;
		case '+':		/* grant right */
			if (sub || r || m || c || d)
				return(-1);
			add++;
			break;
		case '-':		/* revoke right */
			if (add || r || m || c || d)
				return(-1);
			sub++;
			break;
		case '=':		/* explicitly set rights */
			if (add || sub || r || m || c || d)
				return(-1);
			equals++;
			break;
		case 'r':		/* read */
			r++;
			break;
		case 'm':		/* write */
			m++;
			break;
		case 'c':		/* create */
			c++;
			break;
		case 'd':		/* destroy */
			d++;
			break;
		case ',':
		case ' ':
		case '\t':
		case '\n':
		case '\0':
			if (r)
				imode |= NIS_READ_ACC;
			if (m)
				imode |= NIS_MODIFY_ACC;
			if (c)
				imode |= NIS_CREATE_ACC;
			if (d)
				imode |= NIS_DESTROY_ACC;
			if (!n && !o && !g && !w) {
				o++; w++; g++;
			}
			if (n) {
				mode |= (imode << 24);
				mask |= 0xFF000000;
			}
			if (o) {
				mode |= (imode << 16);
				mask |= 0x00FF0000;
			}
			if (g) {
				mode |= (imode << 8);
				mask |= 0x0000FF00;
			}
			if (w) {
				mode |= imode;
				mask |= 0x000000FF;
			}

			if (add) {
				add_mask |= mode;
				sub_mask &= ~mode;
			}
			if (sub) {
				sub_mask |= mode;
				add_mask &= ~mode;
			}
			if (equals) {
				add_mask |= mode;
				sub_mask |= (~mode & mask);
			}

			n = o = g = w = a = 0;
			r = m = c = d = 0;
			mode = 0; imode = 0; mask = 0;
			add = 0; sub = 0; equals = 0;
			if (*ptr == '\n' || *ptr == '\0')
				done++;
			break;
		default:
			return(-1);
			break;
		}
		ptr++;
	}

	if (a_mask != NULL)
		*a_mask = add_mask;
	if (s_mask != NULL)
		*s_mask = sub_mask;
	if (add_mask)
		*omode |= add_mask;
	if (sub_mask)
		*omode &= ~sub_mask;

	return(0);
}

int nis_parse_ttl(str, ttl)
	char			*str;
	unsigned long		*ttl;
{
	int			d = 0, h = 0, m = 0, s = 0;
	int			done = 0;
	char			*ptr = str;
	char			digits[1024];
	unsigned long		t = 0;

	if (ptr == NULL)
		return(-1);
	bzero(digits, sizeof(digits));

	while(!done) {
		switch(*ptr) {
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			strncat(digits, ptr, 1);
			break;
		case 'd':
			if (d)
				return(-1);
			d++;
			t += atol(digits) * 86400;
			bzero(digits, sizeof(digits));
			break;
		case 'h':
			if (h)
				return(-1);
			h++;
			t += atol(digits) * 3600;
			bzero(digits, sizeof(digits));
			break;
		case 'm':
			if (m)
				return(-1);
			m++;
			t += atol(digits) * 60;
			bzero(digits, sizeof(digits));
			break;
		case 's':
			if (s)
				return(-1);
			s++;
			t += atol(digits);
			bzero(digits, sizeof(digits));
			break;
		case ' ':
		case '\t':
		case '\n':
		case '\0':
			if (!d && !h && !m && !s)
				t += atol(digits);
			bzero(digits, sizeof(digits));
			if (*ptr == '\n' || *ptr == '\0')
				done++;
			break;
		default:
			return(-1);
			break;
		}
		ptr++;
	}

	*ttl = t;
	return(0);
}
