/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nis_defaults.c,v 1.14 1998/05/31 15:24:48 wpaul Exp $
 */

#include <stdio.h>
#include <err.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nis_defaults.h"
#include "nis_common.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nis_defaults.c,v 1.14 1998/05/31 15:24:48 wpaul Exp $";
#endif

/*
 * This module implements a set of NIS+ default value selection routines
 * needed by some of the utilities. Contrary to what you might think, these
 * are not part of the NIS+ library proper, and consequently they appear
 * here in a seperate module outside the lib directory.
 */

#ifndef DEF_TTL
#define DEF_TTL	43200
#endif

#ifndef DEF_RIGHTS
#define DEF_RIGHTS	(DEFAULT_RIGHTS)
#endif

#define	PARSED_TTL	1
#define PARSED_RIGHTS	2
#define PARSED_GROUP	4
#define PARSED_OWNER	8

static char			nbuf[NIS_MAXNAMELEN];
static char			gbuf[NIS_MAXNAMELEN];
static nis_name			parsed_owner;
static nis_name			parsed_group;
static unsigned long		parsed_ttl;
static unsigned long		parsed_rights;
static unsigned long		parser_bits = 0;
static int			parser_called = 0;


static int p_rights(str)
	char			*str;
{
	if (parser_bits & PARSED_RIGHTS)
		return(0);

	parsed_rights = 0;
	if (nis_parse_modes(str, &parsed_rights, NULL, NULL) == -1)
		return(1);

	parser_bits |= PARSED_RIGHTS;

	return(0);
}


static int p_ttl(str)
	char			*str;
{
	if (parser_bits & PARSED_TTL)
		return(0);
	if (nis_parse_ttl(str, &parsed_ttl) == -1)
		return(1);
	parser_bits |= PARSED_TTL;
	return(0);
}

static int p_owner(str)
	char			*str;
{
	if (parser_bits & PARSED_OWNER)
		return(0);
	if (str == NULL || !strlen(str) || str[strlen(str) - 1] != '.')
		return(1);
	snprintf(nbuf, sizeof(nbuf), "%s", str);
	parsed_owner = nbuf;
	parser_bits |= PARSED_OWNER;

	return(0);
}

static int p_group(str)
	char			*str;
{
	if (parser_bits & PARSED_GROUP)
		return(0);
	if (str == NULL || !strlen(str) || str[strlen(str) - 1] != '.')
		return(1);
	snprintf(gbuf, sizeof(gbuf), "%s", str);
	parsed_group = gbuf;
	parser_bits |= PARSED_GROUP;

	return(0);
}

struct parser_mod {
	char			*token;
	int			(*pfunc)(char *);
};

static int nis_parse_defs(str)
	char			*str;
{
	struct parser_mod	p[] = { { "ttl=",	p_ttl },
					{ "access=",	p_rights },
					{ "owner=",	p_owner },
					{ "group=",	p_group },
					{ NULL,		NULL} };
	struct parser_mod	*ptmp = p;
	char			*defs, *ptr = NULL;
	int			r;

	if (parser_called)
		return(parser_bits);

	parser_called++;
	if (str != NULL)
		defs = str;
	else {
		defs = getenv("NIS_DEFAULTS");
		if (defs == NULL)
			return(0);
	}

	while((ptr = strsep(&defs, ":")) != NULL) {
		ptmp = p;
		while(ptmp->token != NULL) {
			if (strstr(ptr, ptmp->token) != NULL) {
				r = (*ptmp->pfunc)(ptr + strlen(ptmp->token));
				if (r)
					return(-1);
			}
			ptmp++;
		}
	}

	if ((defs != NULL && strlen(defs)) && !parser_bits)
		return(-1);

	return(0);
}

int nis_init_defs(str)
	char			*str;
{
	int			r;
	parser_called = parser_bits = 0;
	r = nis_parse_defs(str);
	if (r) {
		if (str == NULL)
			warnx("can't parse NIS_DEFAULTS environment variable.");
		else
			warnx("can't parse nis_defaults argument.");
		return(1);
	}

	return(0);
}

unsigned long nis_default_ttl()
{
	if (nis_parse_defs(NULL) & PARSED_TTL)
		return(parsed_ttl);
	return(DEF_TTL);
}

nis_name nis_default_owner()
{
	if (nis_parse_defs(NULL) & PARSED_OWNER)
		return(parsed_owner);
	return(nis_local_principal());
}

nis_name nis_default_group()
{
	if (nis_parse_defs(NULL) & PARSED_GROUP)
		return(parsed_group);
	return(nis_local_group());
}

unsigned long nis_default_rights()
{
	if (nis_parse_defs(NULL) & PARSED_RIGHTS)
		return(parsed_rights);
	return(DEF_RIGHTS);
}

