/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nisaddcred.c,v 1.8 1998/05/26 21:26:35 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <err.h>
#include <pwd.h>
#include <sys/param.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include <nis_common.h>
#include <rpc/key_prot.h>
#include "nisaddcred_extern.h"

#ifndef lint
static const char copyright[] = "@(#) Copyright (c) 1996, 1997\
	Bill Paul. All rights reserved.";
static const char rcsid[] =
	"@(#) $Id: nisaddcred.c,v 1.8 1998/05/26 21:26:35 wpaul Exp wpaul $";
#endif

char	program_name[] = "nisaddcred";

static void usage(pname)
	char			*pname;
{
	fprintf(stderr, "usage: %s [ -p principal ] [ -P nis principal ]\n",
								 pname);
	fprintf(stderr, "                 [ -l login password ] flavor "
						"[ domain name ]\n");
	fprintf(stderr, "       %s -r [ -P nis principal ] [ domain name ]\n",
								pname);
	exit(1);
}

/*
 * The following two functions are a kludge. In order to create
 * AUTH_UNIX credentials for a principal, we need three things:
 * - the NIS+ principal name
 * - the AUTH_UNIX principal, which in this case is the user's
 *   UNIX real UID
 * - the GIDs of all the groups of which the user is a member
 *
 * There are certain restrictions to consider: the UID, GID and credential
 * information must all be resident in NIS+, since were are in effect
 * tying together informatiom from the passwd.org_dir, group.org_dir and
 * cred.org_dir tables. This is required since the NIS+ server only considers
 * NIS+ data sources when authenticating a principal. However, this
 * makes it hard for us to track down the required information. If
 * a user is creating credentials for himself, then all the relevant
 * UID and GID information can be obtained from the process using the
 * getuid() and getgid() syscalls. But if someone else is creating creds
 * for a different user, the information has to be obtained in a different
 * manner.
 *
 * It may seem that we can simply call getpwuid() or getgrouplist() to
 * quickly obtain this information, but that's not the case. We have
 * no way on knowing that getpwuid() and getgrouplist() will actually
 * obtain the required data from NIS+, and _only_ NIS+: they may also
 * search the local passwd and group databases, or roll over to NIS v2
 * or LDAP, or Hesiod, or whatever). Consequently, we have to grope
 * through the passwd.org_dir and group.org_dir tables ourselves to find
 * the information we want.
 *
 * Searching for the list of GIDs is necessarily messy: the group.org_dir
 * table is searchable only by group name and GID. We can't search by
 * members. The only thing we can do is scan the entire group.org_dir
 * table and keep a running list of membership information as we go.
 * The most efficient way to do this is with a callback, hence that's
 * what we use here.
 */
 
struct u_data {
	char			*username;
	char			*groups;
};

static int get_groups_cb(name, obj, udata)
	nis_name		name;
	nis_object		*obj;
	void			*udata;
{
	struct u_data		*p;

	p = udata;
	if (strcmp(p->username, ENTRY_VAL(obj, 3)))
		return(0);

	if (p->groups == NULL) {
		p->groups = malloc(ENTRY_LEN(obj, 1));
		if (p->groups == NULL)
			errx(1, "out of memory scanning groups");
	} else {
		p->groups = realloc(p->groups, strlen(p->groups) +
				ENTRY_LEN(obj, 1) + 1);
		if (p->groups == NULL)
			errx(1, "out of memory scanning groups");
		strcat(p->groups, ",");
	}

	strcat(p->groups, ENTRY_VAL(obj, 1));

	return(0);
}

static nis_error get_groups(user, g,  domain)
	nis_name		user;
	char			**g;
	nis_name		domain;
{
	char			tname[NIS_MAXNAMELEN];
	nis_result		*res;
	nis_error		r;
	struct u_data		u;
	struct passwd		*pw;
	uid_t			uid;

	uid = strtol(user, NULL, 10);

	if ((pw = getpwuid_nisplus_master(uid, domain)) == NULL)
		errx(1, "couldn't find passwd.org_dir entry for UID %s", user);

	if (!(pw->pw_fields & _PWF_NAME) || !(pw->pw_fields & _PWF_GID))
		errx(1, "malformed passwd entry for UID %s!", user);

	u.username = pw->pw_name;
	snprintf(tname, sizeof(tname), "%lu", pw->pw_gid);
	u.groups = strdup(tname);

	/*
	 * Now find all the additional GIDs from the group table.
	 */
	snprintf(tname, sizeof(tname), "group.org_dir.%s", domain);
	res = nis_list(tname, 0, get_groups_cb, (void *)&u);
	if (res == NULL)
		errx(1, "out of memory doing nis_list");
	r = res->status;
	nis_freeresult(res);

	if (r != NIS_SUCCESS)
		errx(1, "nis_list failed: %s: %s", tname, nis_sperrno(r));
	*g = u.groups;
	return(r);
}

static void remove_creds(cname, domain)
	nis_name		cname;
	nis_name		domain;
{
	nis_result		*res;
	char			buf[NIS_MAXNAMELEN];

	if (cname == NULL)
		cname = strdup(nis_local_principal());

	snprintf(buf, sizeof(buf), "[cname=%s],cred.org_dir.%s",
			cname, domain);

	if (buf[strlen(buf) - 1] != '.')
		strcat(buf, ".");

	res = nis_remove_entry(buf, NULL, REM_MULTIPLE);

	if (res == NULL)
		errx(1, "out of memory doing nis_remove_entry()");

	if (res->status != NIS_SUCCESS) {
		warnx("%s: %s", cname, nis_sperrno(res->status));
		nis_freeresult(res);
		exit(1);
	}

	nis_freeresult(res);
	exit(0);
}

int add_local_cred(princ, nis_princ, domain)
	char			*princ;
	char			*nis_princ;
	char			*domain;
{
	char			*g;
	char			pbuf[NIS_MAXNAMELEN];

	if (princ == NULL) {
		snprintf(pbuf, sizeof(pbuf), "%lu", geteuid());
		princ = strdup(pbuf);
	}

	if (nis_princ == NULL) {
		struct passwd		*pw;

		pw = getpwuid_nisplus_master(geteuid(), domain);
		if (pw == NULL || !(pw->pw_fields & _PWF_NAME))
			errx(1, "failed to read NIS+ passwd info for %lu",
								geteuid());
		snprintf(pbuf, sizeof(pbuf), "%s.%s", pw->pw_name, domain);
		nis_princ = strdup(pbuf);
	}

	get_groups(princ, &g, domain);
	nisplus_update_local(princ, g, nis_princ);

	exit(0);
}

int add_des_cred(princ, nis_princ, passwd, domain)
	char			*princ;
	char			*nis_princ;
	char			*passwd;
	char			*domain;
{
	nis_error		status;
	char			prompt[NIS_MAXNAMELEN];
	char			public[HEXKEYBYTES + 1];
	char			secret[HEXKEYBYTES + 1];
	uid_t			uid;
	gid_t			gid;
	int			glen;
	gid_t			glist[NGROUPS];
	char			pbuf[NIS_MAXNAMELEN];

	if (princ == NULL) {
		char			hname[NIS_MAXNAMELEN];
		gethostname(hname, NIS_MAXNAMELEN);

		if (geteuid()) {
			snprintf(pbuf, sizeof(pbuf), "unix.%lu@%s",
					geteuid(), domain);
		} else {
			snprintf(pbuf, sizeof(pbuf), "unix.s@%s",
					hname, domain);
		}
		if (pbuf[strlen(pbuf) - 1] == '.')
			pbuf[strlen(pbuf) - 1] = NUL;

		princ = strdup(pbuf);
	}

	if (nis_princ == NULL)
		nis_princ = strdup(nis_local_principal());

	if (netname2user(princ, &uid, &gid, &glen, glist)) {
		status = cred_exists(nis_princ, "local", domain, NULL);
		if (status != NIS_SUCCESS)
			errx(1,"you must create a \"local\" credential first.");
	}

	status = cred_exists(nis_princ, "des", domain, NULL);
	if (status == NIS_SUCCESS)
		printf("Modifying ");
	else
		printf("Adding ");

	printf("key pair for %s (%s).\n", princ, nis_princ);

	if (passwd == NULL) {
		snprintf(prompt, sizeof(prompt),
			"Enter %s's login password: ", princ);
		passwd = getpass(prompt);
	}
	genkeys(public, secret, passwd);
	nisplus_update_des(princ, public, secret, nis_princ);
	if (!keylogin(princ, secret))
		exit(1);

	exit(0);
}

int main (argc, argv)
	int			argc;
	char			*argv[];
{
	int			ch;
	char			*princ = NULL;
	char			*nis_princ = NULL;
	char			*passwd = NULL;
	char			*domain = nis_local_directory();
	int			removecred = 0;
	char			*pname = argv[0];
	int			flavor = 0;

	/*
	 * Grr.... only Sun could invent such a wacky syntax.
	 */
	while((ch = getopt(argc, argv, "rp:P:l:")) != -1) {
		switch(ch) {
		case 'p':
			princ = optarg;
			break;
		case 'P':
			nis_princ = optarg;
			break;
		case 'l':
			passwd = optarg;
			break;
		case 'r':
			removecred++;
			break;
		default:
			usage(argv[0]);
		}
	}

	argc -= optind;
	argv += optind;

	if (removecred) {
		if (argc)
			domain = argv[0];
		remove_creds(nis_princ, domain);
	}

	/*
	 * Check for flavor type.
	 */
	if (argc) {
		if (!strcasecmp(argv[0], "local"))
			flavor = AUTH_UNIX;
		else if (!strcasecmp(argv[0], "des"))
			flavor = AUTH_DES;
		else {
			warnx("Authentication Flavor name required.");
			usage(pname);
		}
		argc -= 1;
		argv += 1;
	}

	/*
	 * See if user specified optional domain.
	 */
	if (argc)
		domain = argv[0];

	/*
	 * Passwords required for AUTH_DES only.
	 */
	if (flavor == AUTH_SYS)
		add_local_cred(princ, nis_princ, domain);

	if (flavor == AUTH_DES)
		add_des_cred(princ, nis_princ, passwd, domain);

	usage(pname);
	exit(0);
}
