/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: niscat.c,v 1.8 1998/05/29 03:01:32 wpaul Exp $
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>

#ifndef lint
static const char copyright[] = "@(#) Copyright (c) 1996, 1997\
	Bill Paul. All rights reserved.";
static const char rcsid[] = "@(#) $Id: niscat.c,v 1.8 1998/05/29 03:01:32 wpaul Exp $";
#endif

static int rawbin = 0;

static int display_obj(name, obj, udata)
	nis_name		name;
	nis_object		*obj;
	void			*udata;
{
	register unsigned int	i;
	entry_col		*e;
	char			*sep = (char *)udata;

	for (i = 0; i < obj->EN_data.en_cols.en_cols_len; i++) {
		e = &obj->EN_data.en_cols.en_cols_val[i];
		if (e->ec_flags & EN_BINARY && !rawbin)
			printf("*BINARY*");
		else {
			/* If field has 0 length, don't print it */
			if (ENTRY_LEN(obj, i))
				printf("%.*s", (int)ENTRY_LEN(obj, i),
						 ENTRY_VAL(obj, i));
		}
		if (i < obj->EN_data.en_cols.en_cols_len - 1)
			printf("%c", *sep);
	}
	printf("\n");
	return(0);
}

static int do_list(name, flags, print_obj, header)
	nis_name		name;
	unsigned long		flags;
	int			print_obj;
	int			header;
{
	nis_result		*res;
	char			sep;

	if (strchr(name, '[') == NULL)
		res = nis_lookup(name, flags);
	else
		res = nis_list(name, flags, NULL, NULL);

	if (res == NULL) {
		warnx("out of memory doing nis_lookup");
		return(-1);
	}

	if (res->status != NIS_SUCCESS && res->status != NIS_S_SUCCESS) {
		warnx("%s: %s", name, nis_sperrno(res->status));
		nis_freeresult(res);
		return(1);
	}

	if (print_obj) {
		unsigned long		i;
		for (i = 0; i < NIS_RES_NUMOBJ(res); i++)
			nis_print_object(&res->objects.objects_val[i]);
		nis_freeresult(res);
		return(0);
	} else if (__type_of(NIS_RES_OBJECT(res)) != TABLE_OBJ) {
		nis_freeresult(res);
		warnx("%s is not a table!", name);
		return(1);
	}

	sep = NIS_RES_OBJECT(res)->TA_data.ta_sep;

	if (header) {
		int			i;
		table_obj		*t = &NIS_RES_OBJECT(res)->TA_data;
		printf("# ");
		for (i = 0; i < t->ta_cols.ta_cols_len; i++) {
			printf("%s", t->ta_cols.ta_cols_val[i].tc_name);
			if (i < t->ta_cols.ta_cols_len - 1)
				printf("%c", t->ta_sep);
		}
		printf("\n");
	}

	nis_freeresult(res);

	res = nis_list(name, flags, display_obj, (void *)&sep);

	if (res == NULL) {	
		warnx("out of memory doing nis_list");
		return(-1);
	}

	if (res->status != NIS_SUCCESS && res->status != NIS_CBRESULTS
		&& res->status != NIS_S_SUCCESS) {
		warnx("%s: %s", name, nis_sperrno(res->status));
		nis_freeresult(res);
		return(1);
	}

	nis_freeresult(res);

	return(0);
}

static void usage(pname)
	char			*pname;
{
	fprintf(stderr, "usage: %s [-LAMhv] tablename ...\n", pname);
	fprintf(stderr, "       %s [-LPAM] -o name ...\n", pname);
	exit(1);
}

int main(argc, argv)
	int			argc;
	char			*argv[];
{
	int			ch, cnt;
	unsigned long		flags = EXPAND_NAME;
	int			header = 0, print_obj = 0;
	char			*pname = argv[0];

	while((ch = getopt(argc, argv, "LPAMhvo")) != -1) {
		switch(ch) {
		case 'L':
			flags |= FOLLOW_LINKS;
			break;
		case 'A':
			flags |= ALL_RESULTS;
			break;
		case 'M':
			flags |= MASTER_ONLY;
			break;
		case 'P':
			flags |= FOLLOW_PATH;
			break;
		case 'h':
			header++;
			break;
		case 'v':
			rawbin++;
			break;
		case 'o':
			print_obj++;
			break;
		default:
			usage(pname);
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if (!argc)
		usage(pname);

	for (cnt = 0; cnt < argc; cnt++)
		do_list(argv[cnt], flags, print_obj, header);

	exit(0);
}
