/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nischattrs.c,v 1.23 1998/01/17 20:26:36 wpaul Exp $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nis_common.h"

#ifndef lint
static const char copyright[] = "@(#) Copyright (c) 1996, 1997, 1998\
	Bill Paul. All rights reserved.";
static const char rcsid[] =
	"@(#) $Id: nischattrs.c,v 1.23 1998/01/17 20:26:36 wpaul Exp $";
#endif

static unsigned long		add_mask = 0, sub_mask = 0;

unsigned long			flags = EXPAND_NAME;
int				ch, force = 0;

nis_error nis_do_update(argc, argv, flags, force, mode, ttl, owner, group)
	int			argc;
	char			*argv[];
	unsigned long		flags;
	int			force;
	int			mode;
	unsigned long		ttl;
	nis_name		owner;
	nis_name		group;
{
	int			i, indexed = 0;
	nis_result		*res, *mres;
	nis_object		*obj;
	char			*name, buf[NIS_MAXNAMELEN];


	for (i = 0; i < argc; i++) {
		name = argv[i];
		if (strchr(name, '[') == NULL) {
			res = nis_lookup(name, flags);
			indexed = 0;
		} else {
			indexed = 1;
			res = nis_list(name, flags, NULL, NULL);
		}
		if (res == NULL) {
			warnx("out of memory looking up %s", name);
			continue;
		}
		if (res->status != NIS_SUCCESS) {
			warnx("%s: %s", name, nis_sperrno(res->status));
			nis_freeresult(res);
			continue;
		}
		obj = NIS_RES_OBJECT(res);
		if (mode) {
			if (add_mask)
				obj->zo_access |= add_mask;
			if (sub_mask)
				obj->zo_access &= ~sub_mask;
		}
		if (ttl)
			obj->zo_ttl = ttl;
		if (owner) {
			free(obj->zo_owner);
			obj->zo_owner = strdup(owner);
		}
		if (group) {
			free(obj->zo_group);
			obj->zo_group = strdup(group);
		}
		snprintf(buf, sizeof(buf), "%s.%s", obj->zo_name, 
						    obj->zo_domain);
		if (indexed)
			mres = nis_modify_entry(NULL, obj, 0);
		else
			mres = nis_modify(buf, obj);
		nis_freeresult(res);
		if (mres == NULL) {
			warnx("out of memory modifying %s", name);
			continue;
		}
		if (mres->status != NIS_SUCCESS) {
			warnx("failed to modify %s: %s", name,
						nis_sperrno(mres->status));
			nis_freeresult(mres);
			continue;
		}
	}

	return(NIS_SUCCESS);
}

static void usage(pname)
	char			*pname;
{
	char			*n = NULL;

	if (strstr(pname, "nischmod") != NULL)
		n = "mode";
	else if (strstr(pname, "nischttl") != NULL)
		n = "ttl";
	else if (strstr(pname, "nischown") != NULL)
		n = "owner";
	else if (strstr(pname, "nischgrp") != NULL)
		n = "group";

	fprintf(stderr, "usage: %s [-ALPf] %s name ...\n", pname, n);
	exit(1);
}

static int nischmod_main(argc, argv, envp)
	int			argc;
	char			*argv[];
	char			*envp[];
{
	unsigned long		dummy = 0;

	/* Compute rights masks */
	if (nis_parse_modes(*argv, &dummy, &add_mask, &sub_mask) == -1) {
		warnx("%s: malformed mode", *argv);
		exit(1);
	}

	argc -= 1;
	argv += 1;

	if (argc < 2)
		
	nis_do_update(argc, argv, flags, force, 1, 0, NULL, NULL);

	exit(0);
}

static int nischttl_main(argc, argv, envp)
	int			argc;
	char			*argv[];
	char			*envp[];
{
	unsigned long		ttl;
	/* Compute ttl value */

	if (nis_parse_ttl(*argv, &ttl) == -1) {
		warnx("%s: malformed ttl", *argv);
		exit(1);
	}

	argc -= 1;
	argv += 1;

	nis_do_update(argc, argv, flags, force, 0, ttl, NULL, NULL);

	exit(0);
}

static int nischown_main(argc, argv, envp)
	int			argc;
	char			*argv[];
	char			*envp[];
{
	char			*owner;

	/*
	 * Verify the principal exists (except for 'nobody':
	 * 'nobody' always exists).
	 */
	if (strcmp(argv[0], "nobody"))
		if (nis_expandprinc(argv[0], flags, &owner) != NIS_SUCCESS)
			errx(1, "%s: no such principal", argv[0]);

	argc -= 1;
	argv += 1;

	nis_do_update(argc, argv, flags, force, 0, 0, owner, NULL);

	exit(0);
}

static int nischgrp_main(argc, argv, envp)
	int			argc;
	char			*argv[];
	char			*envp[];
{
	char			*group;

	/*
	 * Verify that group exists.
	 */
	if (nis_verifygroup_local(argv[0], flags, NULL, &group) != NIS_SUCCESS)
		errx(1, "%s: no such group", argv[0]);

	argc -= 1;
	argv += 1;

	nis_do_update(argc, argv, flags, force, 0, 0, NULL, group);

	exit(0);
}

int main(argc, argv, envp)
	int			argc;
	char			*argv[];
	char			*envp[];
{
	char			*pname;
	char			*gchars;


	if (strstr(argv[0], "nischmod") == NULL) {
		gchars = "AfLP";
	} else {
		gchars = "AfLPrmcd";
	}

	while((ch = getopt(argc, argv, gchars)) != -1) {
		switch(ch) {
		case 'A':
			flags |= FOLLOW_PATH|ALL_RESULTS;
			break;
		case 'f':
			force++;
			break;
		case 'L':
			flags |= FOLLOW_LINKS;
			break;
		case 'P':
			flags |= FOLLOW_PATH;
			break;
		case 'r': case 'm':
			if (argv[optind - 1][0] == '-' &&
			    argv[optind - 1][1] == ch &&
			    argv[optind - 1][2] == '\0')
				optind--;
			goto done;
			break;
		default:
			usage(argv[0]);
			break;
		}
	}

done:
	pname = argv[0];
	argc -= optind;
	argv += optind;

	if (argc < 2)
		usage(pname);

	if (strstr(pname, "nischmod") != NULL)
		exit(nischmod_main(argc, argv, envp));
	else if (strstr(pname, "nischttl") != NULL)
		exit(nischttl_main(argc, argv, envp));
	else if (strstr(pname, "nischown") != NULL)
		exit(nischown_main(argc, argv, envp));
	else if (strstr(pname, "nischgrp") != NULL)
		exit(nischgrp_main(argc, argv, envp));
	else
		err(1, "%s: invalid command", argv[0]);

	exit(1);
}
