/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nisgrpadm.c,v 1.7 1998/02/27 05:18:01 wpaul Exp $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nis_common.h"

#ifndef lint
static const char copyright[] = "@(#) Copyright (c) 1996, 1997\
	Bill Paul. All rights reserved.";
static const char rcsid[] =
	"@(#) $Id: nisgrpadm.c,v 1.7 1998/02/27 05:18:01 wpaul Exp $";
#endif

static void usage(pname)
	char			*pname;
{
	fprintf(stderr, "usage: %s -a | -r | -t [-s] group principal ...\n",
								pname);
	fprintf(stderr, "       %s -c | -d | -l [-M] [-s] group\n", pname);
	exit(1);
}

#define ACTION_NEWGRP	1
#define ACTION_NEWMEM	2
#define ACTION_REMGRP	3
#define ACTION_REMMEM	4
#define ACTION_LSGRP	5
#define ACTION_TESTMEM	6

static void newgrp(gname, flags, silent)
	nis_name		gname;
	unsigned long		flags;
	int			silent;
{
	char			buf[NIS_MAXNAMELEN];
	nis_error		r;

	if (gname[strlen(gname) - 1] != '.')
		snprintf(buf, sizeof(buf), "%s.%s", gname,
					nis_local_directory());
	else
		snprintf(buf, sizeof(buf), "%s", gname);

	r = nis_creategroup(buf, 0);

	if (r != NIS_SUCCESS) {
		warnx("Unable to create group \"%s\": %s", buf,
						nis_sperrno(r));
		exit(r);
	}

	if (!silent)
		printf("Group \"%s\" created.\n", buf);

	exit(NIS_SUCCESS);
}

static void rmgrp(gname, flags, silent)
	nis_name		gname;
	unsigned long		flags;
	int			silent;
{
	nis_error		r;
	nis_name		g;

	r = nis_verifygroup_local(gname, flags, NULL, &g);

	if (r != NIS_SUCCESS)
		err(NIS_NOTFOUND, "Group \"%s\" not found.", gname);

	r = nis_destroygroup(g);

	if (r != NIS_SUCCESS) {
		warnx("Unable to destroy group \"%s\": %s", g,
						  nis_sperrno(r));
		free(g);
		exit(r);
	}

	if (!silent)
		printf("Group \"%s\" destroyed.\n", g);

	free(g);
	exit(NIS_SUCCESS);
}

static void newmem(gname, members, cnt, flags, silent)
	nis_name		gname;
	char			*members[];
	int			cnt;
	unsigned long		flags;
	int			silent;
{
	nis_error		r;
	int			i;
	nis_name		g;
	char			*p;

	r = nis_verifygroup_local(gname, flags, NULL, &g);

	if (r != NIS_SUCCESS)
		errx(NIS_NOTFOUND, "Group \"%s\" not found.", gname);

	if (!cnt) {
		warnx("Missing principal name(s).");
		free(g);
		return;
	}

	for (i = 0; i < cnt; i++) {
		r = nis_expandprinc(members[i], flags, &p);
		if (r != NIS_SUCCESS) {
			warnx("Principal \"%s\" not found.", members[i]);
			continue;
		}
		r = nis_addmember(p, g);
		if (r != NIS_SUCCESS) {
			warnx("Unable to add \"%s\" to group \"%s\": %s",
					p, g, nis_sperrno(r));
		} else {
			if (!silent)
				printf("Added \"%s\" to group \"%s\".\n",
						p, g);
		}
		free(p);
	}

	free(g);
	exit(r);
}

static void remmem(gname, members, cnt, flags, silent)
	nis_name		gname;
	char			*members[];
	int			cnt;
	unsigned long		flags;
	int			silent;
{
	nis_error		r;
	int			i;
	nis_name		g;
	char			*p;

	r = nis_verifygroup_local(gname, flags, NULL, &g);

	if (r != NIS_SUCCESS)
		errx(NIS_NOTFOUND, "Group \"%s\" not found.", gname);

	if (!cnt) {
		warnx("Missing principal name(s).");
		free(g);
		return;
	}

	for (i = 0; i < cnt; i++) {
		r = nis_expandprinc(members[i], flags, &p);
		if (r != NIS_SUCCESS) {
			warnx("Principal \"%s\" not found.", members[i]);
			continue;
		}
		r = nis_removemember(p, g);
		if (r != NIS_SUCCESS) {
			warnx("Unable to remove \"%s\" from group \"%s\": %s",
						p, g, nis_sperrno(r));
		} else {
			if (!silent)
				printf("Removed \"%s\" from group \"%s\".\n",
								p, g);
		}
		free(p);
	}

	free(g);
	exit(r);
}

static void testmem(gname, members, cnt, flags, silent)
	nis_name		gname;
	char			*members[];
	int			cnt;
	unsigned long		flags;
	int			silent;
{
	bool_t			r;
	int			i;
	nis_name		g;
	char			*p;

	r = nis_verifygroup_local(gname, flags, NULL, &g);

	if (r != NIS_SUCCESS)
		errx(NIS_NOTFOUND, "Group \"%s\" not found.", gname);

	if (!cnt) {
		warnx("Missing principal name(s).");
		free(g);
		return;
	}

	for (i = 0; i < cnt; i++) {
		r = nis_expandprinc(members[i], flags, &p);
		if (r != NIS_SUCCESS) {
			warnx("Principal \"%s\" not found.", members[i]);
			continue;
		}
		r = nis_ismember(p, g);
		if (r == TRUE) {
			printf("\"%s\" is a member of group \"%s\".\n", p, g);
		} else {
			printf("\"%s\" is not a member of group \"%s\".\n",
									p, g);
		}
		free(p);
	}

	free(g);
	exit(r);
}

static void listgrp(gname, flags)
	nis_name		gname;
	unsigned long		flags;
{
	nis_error		r;
	char			*g = NULL;

	r = nis_verifygroup_local(gname, flags, NULL, &g);

	if (r != NIS_SUCCESS) {
		warnx("Group \"%s\" not found (%s).", gname, nis_sperrno(r));
		exit(NIS_NOTFOUND);
	}

	nis_print_group_entry(g);

	free(g);
	exit(0);
}

int main(argc, argv)
	int			argc;
	char			*argv[];
{
	int			ch;
	int			silent = 0;
	int			flags = EXPAND_NAME|FOLLOW_LINKS;
	char			*pname = argv[0];
	int			action = 0;

	while((ch = getopt(argc, argv, "artcdlM")) != -1) {
		switch(ch) {
		case 'a':
			if (action) {
				warnx("c, d, a, r, t, and l options "
					"are mutually exclusive.");
				usage(argv[0]);
			}
			action = ACTION_NEWMEM;
			break;
		case 'c':
			if (action) {
				warnx("c, d, a, r, t, and l options "
					"are mutually exclusive.");
				usage(argv[0]);
			}
			action = ACTION_NEWGRP;
			break;
		case 'd':
			if (action) {
				warnx("c, d, a, r, t, and l options "
					"are mutually exclusive.");
				usage(argv[0]);
			}
			action = ACTION_REMGRP;
			break;
		case 'l':
			if (action) {
				warnx("c, d, a, r, t, and l options "
					"are mutually exclusive.");
				usage(argv[0]);
			}
			action = ACTION_LSGRP;
			break;
		case 'M':
			flags |= MASTER_ONLY;
			break;
		case 'r':
			if (action) {
				warnx("c, d, a, r, t, and l options "
					"are mutually exclusive.");
				usage(argv[0]);
			}
			action = ACTION_REMMEM;
			break;
		case 's':
			silent++;
			break;
		case 't':
			if (action) {
				warnx("c, d, a, r, t, and l options "
					"are mutually exclusive.");
				usage(argv[0]);
			}
			action = ACTION_TESTMEM;
			break;
		default:
			usage(argv[0]);
		}
	}

	argc -= optind;
	argv += optind;

	if (!argc)
		usage(pname);

	switch(action) {
	case ACTION_NEWGRP:
		if (argc > 1)
			usage(pname);
		newgrp(argv[0], flags, silent);
		break;
	case ACTION_REMGRP:
		if (argc > 1)
			usage(pname);
		rmgrp(argv[0], flags, silent);
		break;
	case ACTION_NEWMEM:
		if (argc < 2)
			usage(pname);
		newmem(argv[0], argv + 1, argc - 1, flags, silent);
		break;
	case ACTION_REMMEM:
		if (argc < 2)
			usage(pname);
		remmem(argv[0], argv + 1, argc - 1, flags, silent);
		break;
	case ACTION_LSGRP:
		if (argc > 1)
			usage(pname);
		listgrp(argv[0], flags);
		break;
	case ACTION_TESTMEM:
		if (argc < 2)
			usage(pname);
		testmem(argv[0], argv + 1, argc - 1, flags, silent);
		break;
	default:
		break;
	}

	usage(pname);

	exit(0);
}

