/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nisls.c,v 1.8 1998/06/07 17:55:31 wpaul Exp $
 */

#include <sys/types.h>
#include <sys/cdefs.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nisls.c,v 1.8 1998/06/07 17:55:31 wpaul Exp $";
#endif

static void print_otype(otype)
	zotypes			otype;
{
	switch(otype) {
	case NO_OBJ:
		printf("N ");
		break;
	case DIRECTORY_OBJ:
		printf("D ");
		break;
	case GROUP_OBJ:
		printf("G ");
		break;
	case TABLE_OBJ:
		printf("T ");
		break;
	case ENTRY_OBJ:
		printf("E ");
		break;
	case LINK_OBJ:
		printf("L ");
		break;
	case PRIVATE_OBJ:
		printf("P ");
		break;
	case BOGUS_OBJ:
	default:
		printf("B ");
		break;
	}

	return;
}

static void print_obj(obj, longlist, group, mod, dobj)
	nis_object		*obj;
	int			longlist;
	int			group;
	int			mod;
	int			dobj;
{
	char			buf[NIS_MAXNAMELEN];
	time_t			tmp;

	if (__type_of(obj) != ENTRY_OBJ) {
		warnx("object should have been an entry but wasn't.");
		return;
	}

	bzero (buf, sizeof(buf));

	if (longlist) {
		if (dobj)
			print_otype(obj->zo_data.zo_type);
		else
			print_otype(ntohl(*(zotypes *)ENTRY_VAL(obj, 0)));
		nis_print_rights(obj->zo_access);
		printf(" ");
		printf("%s ", group ? obj->zo_group : obj->zo_owner);
		tmp = mod ? obj->zo_oid.mtime : obj->zo_oid.ctime;
		strncpy(buf, ctime(&tmp), sizeof(buf) - 1);
		*strchr(buf, '\n') = '\0';
		printf("%s ", buf);
	}

	if (dobj) {
		snprintf(buf, sizeof(buf), "%s.%s", obj->zo_name,
						    obj->zo_domain);
		printf("%s\n", buf);
	} else
		printf("%s\n", obj->zo_name);

	return;
}

static int show_dir(name, flags, longlist, group, mod)
	nis_name		name;
	unsigned long		flags;
	int			longlist;
	int			group;
	int			mod;
{
	nis_result		*res;

	res = nis_lookup(name, flags);

	if (res == NULL) {
		warnx("out of memory doing nis_lookup");
		return(-1);
	}

	if (res->status != NIS_SUCCESS) {
		warnx("%s: %s", name, nis_sperrno(res->status));
		nis_freeresult(res);
		return(-1);
	}

	print_obj(NIS_RES_OBJECT(res), longlist, group, mod, 1);

	return(0);
}

static int list_dir(name, flags, recursive, longlist, group, mod)
	nis_name		name;
	unsigned long		flags;
	int			recursive;
	int			longlist;
	int			group;
	int			mod;

{
	nis_result		*res1, *res2;
	long			i;
	char			buf[NIS_MAXNAMELEN];
	nis_object		*obj;

	res1 = nis_lookup(name, flags);

	if (res1 == NULL) {
		warnx("out of memory doing nis_lookup");
		return(-1);
	}

	if (res1->status != NIS_SUCCESS) {
		warnx("%s: %s", name, nis_sperrno(res1->status));
		nis_freeresult(res1);
		return(-1);
	}

	snprintf(buf, sizeof(buf), "%s.%s", NIS_RES_OBJECT(res1)->zo_name,
					    NIS_RES_OBJECT(res1)->zo_domain);


	if (__type_of(NIS_RES_OBJECT(res1)) == DIRECTORY_OBJ)
		res2 = nis_list(buf, flags, NULL, NULL);
	else {
		nis_freeresult(res1);
		return(show_dir(buf, flags, longlist, group, mod));
	}

	printf("%s:\n", buf);
	if (res2 == NULL) {
		warnx("out of memory doing nis_list");
		nis_freeresult(res1);
		return(-1);
	}

	if (res2->status == NIS_PARTIAL) {
		nis_freeresult(res1);
		return(0);
	}

	if (res2->status != NIS_SUCCESS) {
		warnx("%s: %s", NIS_RES_OBJECT(res1)->zo_name,
				 nis_sperrno(res2->status));
		nis_freeresult(res1);
		nis_freeresult(res2);
		return(-1);
	}

	for (i = 0; i < NIS_RES_NUMOBJ(res2); i++) {
		obj = &res2->objects.objects_val[i];
		print_obj(obj, longlist, group, mod, 0);
	}

	if (recursive) {
		printf("\n");
		for (i--; i > -1; i--) {
			obj = &res2->objects.objects_val[i];
			if (ntohl(*(zotypes *)ENTRY_VAL(obj, 0))
							 == DIRECTORY_OBJ) {

				snprintf(buf, sizeof(buf), "%s.%s",
							obj->zo_name,
							obj->zo_domain);
				list_dir(buf, flags, recursive,
					 longlist, group, mod);
			}
		}
	}

	nis_freeresult(res1);
	nis_freeresult(res2);
	return(0);
}

static void usage(pname)
	char			*pname;
{
	fprintf(stderr, "%s: [-LMlmgdR] [name ...]\n", pname);
	exit(1);
}

int main(argc, argv)
	int			argc;
	char			*argv[];
{
	int			ch, cnt;
	unsigned long		flags = EXPAND_NAME;
	char			*pname = argv[0];
	int			recursive = 0, longlist = 0;
	int			group = 0, mod = 0, dobj = 0;

	while((ch = getopt(argc, argv, "LMlmgdR")) != -1) {
		switch(ch) {
		case 'r':
			recursive++;
			break;
		case 'L':
			flags |= FOLLOW_LINKS;
			break;
		case 'l':
			longlist++;
			break;
		case 'M':
			flags |= MASTER_ONLY;
			break;
		case 'd':
			dobj++;
			break;
		case 'g':
			group++;
			break;
		case 'm':
			mod++;
			break;
		case 'R':
			recursive++;
			break;
		default:
			usage(pname);
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if (!argc) {
		if (dobj)
			show_dir(nis_local_directory(), flags,
				 longlist, group, mod);
		else
			list_dir(nis_local_directory(), flags,
				 recursive, longlist, group, mod);
	} else {
		for (cnt = 0; cnt < argc; cnt++) {
			if (dobj)
				show_dir(argv[cnt], flags, 
					 longlist, group, mod);
			else
				list_dir(argv[cnt], flags, recursive,
					 longlist, group, mod);
		}
	}

	exit(0);
}
