/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nismatch.c,v 1.3 1998/01/18 19:05:21 wpaul Exp $
 */

#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>

#ifndef lint
static const char copyright[] = "@(#) Copyright (c) 1996, 1997, 1998\
	Bill Paul. All rights reserved.";
static const char rcsid[] =
	"@(#) $Id: nismatch.c,v 1.3 1998/01/18 19:05:21 wpaul Exp $";
#endif

static int			rawbin = 0;

static int			print_obj = 0, count_only = 0;
static int			match_cnt = 0;

static int display_obj(name, obj, udata)
	nis_name		name;
	nis_object		*obj;
	void			*udata;
{
	register unsigned int	i;
	entry_col		*e;
	char			*sep = (char *)udata;

	if (count_only) {
		match_cnt++;
		return(0);
	}

	if (print_obj) {
		nis_print_object(obj);
		return(0);
	}

	for (i = 0; i < obj->EN_data.en_cols.en_cols_len; i++) {
		e = &obj->EN_data.en_cols.en_cols_val[i];
		if (e->ec_flags & EN_BINARY && !rawbin)
			printf("*BINARY*");
		else {
			/* If field has 0 length, don't print it */
			if (ENTRY_LEN(obj, i))
				printf("%.*s", (int)ENTRY_LEN(obj, i),
						 ENTRY_VAL(obj, i));
		}
		if (i < obj->EN_data.en_cols.en_cols_len - 1)
			printf("%c", *sep);
	}
	printf("\n");
	return(0);
}

static int do_list(name, t, flags, header)
	nis_name		name;
	table_obj		*t;
	unsigned long		flags;
	int			header;
{
	nis_result		*res;
	char			sep;

	sep = t->ta_sep;

	if (header) {
		int			i;
		printf("# ");
		for (i = 0; i < t->ta_cols.ta_cols_len; i++) {
			printf("%s", t->ta_cols.ta_cols_val[i].tc_name);
			if (i < t->ta_cols.ta_cols_len - 1)
				printf("%c", t->ta_sep);
		}
		printf("\n");
	}

	res = nis_list(name, flags, display_obj, (void *)&sep);

	if (res == NULL) {	
		warnx("out of memory doing nis_list");
		return(-1);
	}

	if (res->status != NIS_SUCCESS && res->status != NIS_CBRESULTS) {
		warnx("%s: %s", name, nis_sperrno(res->status));
		nis_freeresult(res);
		return(1);
	}

	nis_freeresult(res);

	return(0);
}

static void usage(pname)
	char			*pname;
{
	fprintf(stderr, "usage: %s [-AMchvo] key tablename\n", pname);
	fprintf(stderr, "       %s [-AMchvo] colname=keypat ... tablename\n",
								 pname);
	fprintf(stderr, "       %s [-AMchvo] indexname\n", pname);
	exit(1);
}

int main(argc, argv)
	int			argc;
	char			*argv[];
{
	int			ch, cnt, i, j;
	unsigned long		flags = EXPAND_NAME;
	int			header = 0;
	char			*pname = argv[0], *cname, *tname, *equals;
	nis_result		*res;
	table_obj		*t;
	char			tmpname[NIS_MAXNAMELEN],
				fullname[NIS_MAXNAMELEN];

	while((ch = getopt(argc, argv, "AMLPchvo")) != -1) {
		switch(ch) {
		case 'L':
			flags |= FOLLOW_LINKS;
			break;
		case 'A':
			flags |= ALL_RESULTS;
			break;
		case 'M':
			flags |= MASTER_ONLY;
			break;
		case 'P':
			flags |= FOLLOW_PATH;
			break;
		case 'h':
			header++;
			break;
		case 'v':
			rawbin++;
			break;
		case 'o':
			print_obj++;
			break;
		case 'c':
			count_only++;
			break;
		default:
			usage(pname);
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if (argc < 1)
		usage(pname);

	if (argc == 1) {
		tname = strchr(argv[0], ']');
		if (tname == NULL)
			errx(1, "badly formatted index name");
		tname += 2;
	} else
		tname = argv[argc - 1];

	/*
	 * Try to find the specified table object.
	 */
	res = nis_lookup(tname, flags);

	if (res == NULL)
		errx(1, "out of memory doing nis_lookup");

	if (res->status != NIS_SUCCESS) {
		warnx("%s: %s", tname, nis_sperrno(res->status));
		nis_freeresult(res);
		exit(1);
	}

	if (__type_of(NIS_RES_OBJECT(res)) != TABLE_OBJ) {
		warnx("%s is not a table!", tname);
		nis_freeresult(res);
		exit(1);
	}

	t = &NIS_RES_OBJECT(res)->TA_data;

	if (argc == 1 && strchr(argv[0], '=') != NULL) {
		do_list(argv[0], t, flags, header);
		if (count_only)
			printf("%d\n", match_cnt);
		exit(0);
	}

	/*
	 * Sanity check.
	 */
	for (i = 0; i < t->ta_cols.ta_cols_len; i++) {
		if (t->ta_cols.ta_cols_val[i].tc_flags & TA_SEARCHABLE)
			break;
	}

	if (i == t->ta_cols.ta_cols_len)
		errx(1, "Invalid table schema: At least one column "
			"must be searchable");

	/*
	 * Quick case.
	 */
	if (argc == 2 && strchr(argv[0], '=') == NULL) {
		if (!(t->ta_cols.ta_cols_val[0].tc_flags & TA_SEARCHABLE)) {
			warnx("table column is not searchable");
			nis_freeresult(res);
			exit(1);
		}
		snprintf(fullname, sizeof(fullname), "[%s=%s],%s.%s",
				t->ta_cols.ta_cols_val[0].tc_name,
				argv[0],
				NIS_RES_OBJECT(res)->zo_name,
				NIS_RES_OBJECT(res)->zo_domain);
		do_list(fullname, t, flags, header);

		if (count_only)
			printf("%d\n", match_cnt);
		exit(0);
	}

	/*
	 * Not so quick case.
	 */
	cnt = 0;
	strcpy(tmpname, "[");
	for(j = 0; j < argc - 1; j++) {
		equals = strchr(argv[j], '=');
		if (equals == NULL) {
			warnx("%s badly formatted search pattern", argv[j]);
			usage(pname);
		}
		for (i = 0; i < t->ta_cols.ta_cols_len; i++) {
			cname = t->ta_cols.ta_cols_val[i].tc_name;
			if (!strncasecmp(cname, argv[j], equals - argv[j])) {
				strncat(tmpname, argv[j], sizeof(tmpname));
				break;
			}
		}
		if (i == t->ta_cols.ta_cols_len)
			errx(1, "table '%s' has no such column '%s'",
						tname, argv[j]);
		else
			strncat(tmpname, ",", sizeof(fullname));
	}

	snprintf(fullname, sizeof(fullname), "%s],%s.%s", tmpname,
				NIS_RES_OBJECT(res)->zo_name,
				NIS_RES_OBJECT(res)->zo_domain);
	do_list(fullname, t, flags, header);

	if (count_only)
		printf("%d\n", match_cnt);
	exit(0);
}
