/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nismkdir.c,v 1.5 1998/06/07 04:28:17 wpaul Exp $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include <rpc/key_prot.h>
#include "nis_defaults.h"

#ifndef lint
static const char copyright[] = "@(#) Copyright (c) 1996, 1997\
	Bill Paul. All rights reserved.";
static const char rcsid[] =
	"@(#) $Id: nismkdir.c,v 1.5 1998/06/07 04:28:17 wpaul Exp $";
#endif

void usage(pname)
	char			*pname;
{
	fprintf(stderr, "usage: %s [-D defaults] [-m hostname] "
				"[-s hostname] dirname\n", pname);
	exit(1);
}

static void newdir(name)
	char			*name;
{
	nis_result		*res, *ares;
	nis_object		obj;
	directory_obj		*d, *pd;
	nis_error		mkstat;

	bzero((char *)&obj, sizeof(obj));

	res = nis_lookup(nis_domain_of(name), MASTER_ONLY);

	if (res == NULL)
		errx(1, "out of memory doing nis_lookup");

	if (res->status != NIS_SUCCESS) {
		warnx("%s: %s", nis_domain_of(name), nis_sperrno(res->status));
		nis_freeresult(res);
		exit(1);
	}

	if (__type_of(NIS_RES_OBJECT(res)) != DIRECTORY_OBJ) {
		warnx("%s: is not a directory!", nis_domain_of(name));
		nis_freeresult(res);
		exit(1);
	}

	pd = &NIS_RES_OBJECT(res)->DI_data;

	obj.zo_name = strdup(nis_leaf_of(name));
	obj.zo_domain = strdup(nis_domain_of(name));
	obj.zo_owner = strdup(nis_default_owner());
	obj.zo_group = strdup(nis_default_group());
	obj.zo_ttl = nis_default_ttl();
	obj.zo_access = nis_default_rights();
	__type_of(&obj) = DIRECTORY_OBJ;
	d = &obj.DI_data;

	d->do_name = strdup(name);
	d->do_type = NIS;
	d->do_ttl = nis_default_ttl();
	d->do_servers.do_servers_len = pd->do_servers.do_servers_len;
	d->do_servers.do_servers_val = pd->do_servers.do_servers_val;

	ares = nis_add(name, &obj);

	d->do_servers.do_servers_len = 0;
	d->do_servers.do_servers_val = NULL;

	if (ares == NULL) {
		nis_freeresult(res);
		errx(1, "out of memory doing nis_add");
	}

	if (ares->status != NIS_SUCCESS) {
		warnx("can't add directory: %s", nis_sperrno(ares->status));
		nis_freeresult(ares);
		nis_freeresult(res);
		exit(1);
	}

	nis_freeresult(ares);

	mkstat = nis_mkdir(name, &pd->do_servers.do_servers_val[0]);

	nis_freeresult(res);

	if (mkstat != NIS_SUCCESS) {
		warnx("%s: %s", name, nis_sperrno(res->status));
		nis_remove(name, NULL);
		exit(1);
	}

	exit(0);
}

static nis_server *makesrv(name)
	char			*name;
{
	struct hostent		*h;
	int			i;
	endpoint		*ep;
	nis_server		*server;
	char			netname[MAXNETNAMELEN];
	char			pkey[HEXKEYBYTES + 1];
	char			haddr[NIS_MAXNAMELEN];
	char			hname[NIS_MAXNAMELEN];

	h = gethostbyname(name);

	if (h == NULL)
		err(1, "%s: gethostbyname failed", name);

	ep = malloc(sizeof(endpoint));
	i = 0;

	while(*h->h_addr_list != NULL) {
		struct in_addr		*in;
		endpoint		*e;
		in = *(struct in_addr **)h->h_addr_list;
		e = ep = realloc(ep, sizeof(endpoint) * (i + 3));
		e += i;
		snprintf(haddr, sizeof(haddr), "%s.0.111", inet_ntoa(*in));
		e->uaddr = strdup(haddr);
		e->family = strdup("inet");
		e->proto = strdup("tcp");	/* For TCP. */
		e++;
		e->uaddr = strdup(haddr);
		e->family = strdup("inet");
		e->proto = strdup("udp");	/* For UDP. */
		e++;
		e->uaddr = strdup(haddr);
		e->family = strdup("inet");
		e->proto = strdup("-");		/* For broadcast (?) */
		i += 3;
		h->h_addr_list++;
	}

	server = calloc(1, sizeof(nis_server));


	/*
	 * This is something I'm not to happy with. In NIS+,
	 * a host's principal name contains the NIS+ domain,
	 * not the DNS domain. For example, if we have a host
	 * registered as rootmaster.sales.widget.org in the DNS,
	 * but it's a member of the internal.widget-nisplus.org.
	 * NIS+ domain, then the hostname that appears when you
	 * run nisdefaults(1) is sales.internal.widget-nisplus.org.
	 * This is somewhat counter-intuitive.
	 */
	snprintf(hname, sizeof(hname), "%s", h->h_name);
	if (strchr(hname, '.') == NULL) {
		strncat(hname, ".", sizeof(hname));
		strncat(hname, nis_local_directory(), sizeof(hname));
	} else {
		*strchr(hname, '.') = NUL;
		strncat(hname, ".", sizeof(hname));
		strncat(hname, nis_local_directory(), sizeof(hname));
	}
	server->name = strdup(hname);
	server->ep.ep_len = i;
	server->ep.ep_val = ep;

	if (host2netname(netname, hname, NULL) == TRUE &&
		getpublickey(netname, pkey) == TRUE) {
		server->key_type = NIS_PK_DH;
		server->pkey.n_bytes = strdup(pkey);
		server->pkey.n_len = strlen(pkey) + 1;
	} else {
		server->key_type = NIS_PK_NONE;
		server->pkey.n_bytes = NULL;
		server->pkey.n_len = 0;
	}

	return(server);
}

static void newmaster(name, mastername)
	char			*name;
	char			*mastername;
{
	nis_result		*res, *ares;
	directory_obj		*d;
	nis_server		*master, *old, *new;
	int			i;

	master = makesrv(mastername);

	res = nis_lookup(name, MASTER_ONLY);

	if (res == NULL)
		errx(1, "out of memory doing nis_lookup");

	if (res->status != NIS_SUCCESS) {
		warnx("%s: %s", name, nis_sperrno(res->status));
		nis_freeresult(res);
		exit(1);
	}

	if (__type_of(NIS_RES_OBJECT(res)) != DIRECTORY_OBJ) {
		warnx("%s: is not a directory!", name);
		nis_freeresult(res);
		exit(1);
	}

	d = &NIS_RES_OBJECT(res)->DI_data;

	old = d->do_servers.do_servers_val;
	if (nis_dir_cmp(old[0].name, master->name) == SAME_NAME) {
		warnx("%s is already the master for %s!",
			master->name, d->do_name);
		nis_freeresult(res);
		xdr_free(xdr_nis_server, (char *)master);
		exit(1);
	}

	for (i = 0; i < d->do_servers.do_servers_len; i++) {
		if (nis_dir_cmp(old[i].name, master->name) == SAME_NAME)
			break;
	}

	if (i < d->do_servers.do_servers_len) {
		nis_server		tmp;
		bcopy((char *)&old[i], (char *)&tmp, sizeof(nis_server));
		bcopy((char *)&old[0], (char *)&old[i], sizeof(nis_server));
		bcopy((char *)&tmp, (char *)&old[0], sizeof(nis_server));
	} else {
		d->do_servers.do_servers_len++;
		new = calloc(1, sizeof(nis_server) *
			d->do_servers.do_servers_len);
		for (i = 0; i < d->do_servers.do_servers_len; i++) {
			if (i) {
				bcopy((char *)&old[i-1],
					(char *)&new[i], sizeof(nis_server));
			} else {
				bcopy((char *)master, (char *)&new[i],
					sizeof(nis_server));
			}
		}
		d->do_servers.do_servers_val = new;
		free(old);
	}

	free(master);

	ares = nis_modify(name, NIS_RES_OBJECT(res));

	nis_freeresult(res);

	if (res == NULL)
		errx(1, "out of memory doing nis_modify");

	if (ares->status != NIS_SUCCESS) {
		warnx("couldn't change master: %s", nis_sperrno(ares->status));
		nis_freeresult(ares);
		exit(1);
	}

	nis_freeresult(ares);

	exit(0);
}

static void newslave(name, slavename)
	char			*name;
	char			*slavename;
{
	nis_result		*res, *ares;
	directory_obj		*d;
	int			i;
	char			*sname;
	nis_server		*replica;

	replica = makesrv(slavename);

	res = nis_lookup(name, MASTER_ONLY);

	if (res == NULL)
		errx(1, "out of memory doing nis_lookup");

	if (res->status != NIS_SUCCESS) {
		warnx("%s: %s", name, nis_sperrno(res->status));
		nis_freeresult(res);
		exit(1);
	}

	if (__type_of(NIS_RES_OBJECT(res)) != DIRECTORY_OBJ) {
		warnx("%s: is not a directory!", name);
		nis_freeresult(res);
		exit(1);
	}

	d = &NIS_RES_OBJECT(res)->DI_data;
	for(i = 0; i < d->do_servers.do_servers_len; i++) {
		sname = d->do_servers.do_servers_val[i].name;
		if (nis_dir_cmp(replica->name, sname) == SAME_NAME) {
			warnx("%s already serves %s!",
				sname, name);
			nis_freeresult(res);
			exit(1);
		}
	}

	d->do_servers.do_servers_len++;
	d->do_servers.do_servers_val =
		realloc(d->do_servers.do_servers_val,
			d->do_servers.do_servers_len *
			sizeof(nis_server));
	bcopy(replica, (char *)&d->do_servers.do_servers_val[i],
			 sizeof(nis_server));

	ares = nis_modify(name, NIS_RES_OBJECT(res));

	nis_freeresult(res);
	free(replica);

	if (res == NULL)
		errx(1, "out of memory doing nis_modify");

	if (ares->status != NIS_SUCCESS) {
		warnx("couldn't add replica: %s", nis_sperrno(ares->status));
		nis_freeresult(ares);
		exit(1);
	}

	nis_freeresult(ares);
	exit(0);
}

int main(argc, argv)
	int			argc;
	char			*argv[];
{
	int			ch;
	char			*pname = argv[0];
	char			*mastername = NULL, *slavename = NULL;
	char			*defs = NULL;

	while((ch = getopt(argc, argv, "D:m:s:")) != -1) {
		switch(ch) {
		case 'D':
			defs = optarg;
			break;
		case 'm':
			if (slavename != NULL) {
				warnx("can only use one of -m and -s");
				usage(argv[0]);
			}
			mastername = optarg;
			break;
		case 's':
			if (mastername != NULL) {
				warnx("can only use one of -m and -s");
				usage(argv[0]);
			}
			slavename = optarg;
			break;
		default:
			usage(argv[0]);
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if (!argc)
		usage(pname);

	if (argv[0][strlen(argv[0]) - 1] != '.')
		errx(1, "dirname must be fully qualified.");

	if (defs != NULL)
		nis_init_defs(defs);

	if (mastername != NULL)
		newmaster(argv[0], mastername);

	if (slavename != NULL)
		newslave(argv[0], slavename);

	newdir(argv[0]);

	exit(0);
}
