/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nisrmdir.c,v 1.3 1998/06/08 04:51:56 wpaul Exp $
 */

#include <sys/types.h>
#include <sys/cdefs.h>
#include <sys/socket.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include <rpc/key_prot.h>

#ifndef lint
static const char copyright[] = "@(#) Copyright (c) 1996, 1997\
	Bill Paul. All rights reserved.";
static const char rcsid[] =
	"@(#) $Id: nisrmdir.c,v 1.3 1998/06/08 04:51:56 wpaul Exp $";
#endif

void usage(pname)
	char			*pname;
{
	fprintf(stderr, "usage: %s [-if] [-s hostname] dirname\n", pname);
	exit(1);
}

static void removedir(name, inter, force)
	char			*name;
	int			inter;
	int			force;
{
	nis_result		*res, *rres;
	nis_object		*obj;
	directory_obj		*d;
	nis_error		rmstat = NIS_SUCCESS;
	char			fullname[NIS_MAXNAMELEN];
	unsigned long		i;

	bzero((char *)&obj, sizeof(obj));

	res = nis_lookup(name, MASTER_ONLY|EXPAND_NAME);

	if (res == NULL)
		errx(1, "out of memory doing nis_lookup");

	if (res->status != NIS_SUCCESS) {
		warnx("%s: %s", name, nis_sperrno(res->status));
		nis_freeresult(res);
		exit(1);
	}

	obj = NIS_RES_OBJECT(res);

	if (__type_of(obj) != DIRECTORY_OBJ) {
		warnx("%s: is not a directory!", name);
		nis_freeresult(res);
		exit(1);
	}

	snprintf(fullname, sizeof(fullname), "%s.%s", obj->zo_name,
							obj->zo_domain);
	if (inter) {
		int                     ch, first;
		fprintf(stderr, "remove %s? ", fullname);
		fflush(stderr);
		first = ch = getchar();
		while (ch != '\n' && ch != -1)
			ch = getchar();
		if (first != 'y') {
			nis_freeresult(res);
			exit(0);
		}
	}

	d = &obj->DI_data;

	rres = nis_remove(fullname, obj);

	if (rres == NULL) {
		nis_freeresult(res);
		errx(1, "out of memory doing nis_remove");
	}

	if (rres->status != NIS_SUCCESS) {
		warnx("cannot remove master: %s", nis_sperrno(rres->status));
		nis_freeresult(res);
		exit(1);
	}

	nis_freeresult(rres);

	for (i = 0; i < d->do_servers.do_servers_len; i++) {
		rmstat = nis_rmdir(fullname,
			&d->do_servers.do_servers_val[i]);
		if (rmstat != NIS_SUCCESS)
			break;
	}

	/*
 	 * This is kinda dumb: if nis_rmdir() fails somewhere
	 * along the line, we have to put the directory object
	 * back, otherwise we will have a bunch of orphaned
	 * directory databases left lying around. Relying on the
	 * client to handle this is just asking for trouble.
	 */
	if (rmstat != NIS_SUCCESS) {
		warnx("%s: %s", name, nis_sperrno(res->status));
		nis_add(name, NIS_RES_OBJECT(res));
		nis_freeresult(res);
		exit(1);
	}

	nis_freeresult(res);
	exit(0);
}

static nis_server *makesrv(name)
	char			*name;
{
	struct hostent		*h;
	int			i;
	endpoint		*ep;
	nis_server		*server;
	char			netname[MAXNETNAMELEN];
	char			pkey[HEXKEYBYTES + 1];
	char			haddr[NIS_MAXNAMELEN];
	char			hname[NIS_MAXNAMELEN];

	h = gethostbyname(name);

	if (h == NULL)
		err(1, "%s: gethostbyname failed", name);

	ep = malloc(sizeof(endpoint));
	i = 0;

	while(*h->h_addr_list != NULL) {
		struct in_addr		*in;
		endpoint		*e;
		in = *(struct in_addr **)h->h_addr_list;
		e = ep = realloc(ep, sizeof(endpoint) * (i + 3));
		e += i;
		snprintf(haddr, sizeof(haddr), "%s.0.111", inet_ntoa(*in));
		e->uaddr = strdup(haddr);
		e->family = strdup("inet");
		e->proto = strdup("tcp");	/* For TCP. */
		e++;
		e->uaddr = strdup(haddr);
		e->family = strdup("inet");
		e->proto = strdup("udp");	/* For UDP. */
		e++;
		e->uaddr = strdup(haddr);
		e->family = strdup("inet");
		e->proto = strdup("-");		/* For broadcast (?) */
		i += 3;
		h->h_addr_list++;
	}

	server = calloc(1, sizeof(nis_server));

	snprintf(hname, sizeof(hname), "%s.", h->h_name);
	if (strchr(hname, '.') == NULL) {
		strncat(hname, ".", sizeof(hname) - 1);
		strncat(hname, nis_local_directory(), sizeof(hname) - 1);
	} else {
		*strchr(hname, '.') = NUL;
		strncat(hname, ".", sizeof(hname) - 1);
		strncat(hname, nis_local_directory(), sizeof(hname) - 1);
	}
	server->name = strdup(hname);
	server->ep.ep_len = i;
	server->ep.ep_val = ep;

	if (host2netname(netname, h->h_name, NULL) == TRUE &&
		getpublickey(netname, pkey) == TRUE) {
		server->key_type = NIS_PK_DH;
		server->pkey.n_bytes = strdup(pkey);
		server->pkey.n_len = strlen(pkey) + 1;
	} else {
		server->key_type = NIS_PK_NONE;
		server->pkey.n_bytes = NULL;
		server->pkey.n_len = 0;
	}

	return(server);
}

static void removeslave(name, slavename, inter, force)
	char			*name;
	char			*slavename;
	int			inter;
	int			force;
{
	nis_result		*res, *ares;
	directory_obj		*d;
	int			i;
	char			*sname;
	nis_server		*replica;
	int			srvs;
	nis_server		*new, *old;
	char			fullname[NIS_MAXNAMELEN];
	nis_object		*obj, *save;
	nis_error		rmstat = NIS_SUCCESS;

	replica = makesrv(slavename);

	res = nis_lookup(name, MASTER_ONLY|EXPAND_NAME);

	if (res == NULL)
		errx(1, "out of memory doing nis_lookup");

	if (res->status != NIS_SUCCESS) {
		warnx("%s: %s", name, nis_sperrno(res->status));
		nis_freeresult(res);
		exit(1);
	}

	if (__type_of(NIS_RES_OBJECT(res)) != DIRECTORY_OBJ) {
		warnx("%s: is not a directory!", name);
		nis_freeresult(res);
		exit(1);
	}

	obj = NIS_RES_OBJECT(res);
	save = nis_clone_object(obj, NULL);

	snprintf(fullname, sizeof(fullname), "%s.%s", obj->zo_name,
							obj->zo_domain);

	if (inter) {
		int                     ch, first;
		fprintf(stderr, "remove %s? ", fullname);
		fflush(stderr);
		first = ch = getchar();
		while (ch != '\n' && ch != -1)
			ch = getchar();
		if (first != 'y') {
			nis_freeresult(res);
			exit(0);
		}
	}

	d = &obj->DI_data;

	srvs = 0;
	new = calloc(1, sizeof(nis_server) *
		(d->do_servers.do_servers_len - 1));
	old = d->do_servers.do_servers_val;

	for (i = 0; i < d->do_servers.do_servers_len; i++) {
		sname = old[i].name;
		if (nis_dir_cmp(replica->name, sname) != SAME_NAME) {
			bcopy((char *)&old[i],
				(char *)&new[srvs], sizeof(nis_server));
			srvs++;
		} else
			xdr_free(xdr_nis_server, (char *)&old[i]);
	}


	if (srvs == d->do_servers.do_servers_len && !force) {
		warnx("Host %s is not listed as a replica for %s.",
			slavename, fullname);
		warnx("Use the -fs option to attempt rmdir anyhow.");
		nis_freeresult(res);
		free(new);
		exit(1);
	}

	d->do_servers.do_servers_len--;
	d->do_servers.do_servers_val = new;
	free(old);

	ares = nis_modify(fullname, obj);

	nis_freeresult(res);

	if (res == NULL)
		errx(1, "out of memory doing nis_modify");

	if (ares->status != NIS_SUCCESS) {
		warnx("cannot remove replica: %s", nis_sperrno(ares->status));
		nis_freeresult(ares);
		res = nis_modify(fullname, save);
		nis_freeresult(res);
		nis_destroy_object(save);
		exit(1);
	}

	nis_freeresult(ares);
	rmstat = nis_rmdir(fullname, replica);
	xdr_free(xdr_nis_server, (char *)replica);
	free(replica);

	if (rmstat != NIS_SUCCESS) {
		warnx("Failed to remove replica %s: %s", name,
				nis_sperrno(rmstat));
		res = nis_modify(fullname, save);
		nis_freeresult(res);
		nis_destroy_object(save);
		exit(1);
	}

	nis_destroy_object(save);

	exit(0);
}

int main(argc, argv)
	int			argc;
	char			*argv[];
{
	int			ch, interactive = 0, force = 0;
	char			*pname = argv[0];
	char			*slavename = NULL;

	while((ch = getopt(argc, argv, "ifs:")) != -1) {
		switch(ch) {
		case 'i':
			interactive++;
			break;
		case 'f':
			interactive = 0;
			force++;
			break;
		case 's':
			slavename = optarg;
			break;
		default:
			usage(argv[0]);
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if (!argc)
		usage(pname);

	if (argv[0][strlen(argv[0]) - 1] != '.')
		interactive++;

	if (slavename != NULL)
		removeslave(argv[0], slavename, interactive, force);

	removedir(argv[0], interactive, force);

	exit(0);
}
