/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nistbladm.c,v 1.9 1998/05/31 15:25:04 wpaul Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <string.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nis_common.h"
#include "nis_defaults.h"

#ifndef lint
static const char copyright[] = "@(#) Copyright (c) 1996, 1997, 1998\
	Bill Paul. All rights reserved.";
static const char rcsid[] =
	"@(#) $Id: nistbladm.c,v 1.9 1998/05/31 15:25:04 wpaul Exp $";
#endif

#define ACTION_NEWTAB	1
#define ACTION_REMTAB	2
#define ACTION_ADDENT	3
#define ACTION_MODENT	4
#define ACTION_REMENT	5
#define ACTION_MODTAB	6

static char			*pname;

void usage()
{
	fprintf(stderr, "usage:	%s [-D defaults] -c [-p path] "
			"[-s sep] type\n", pname);
	fprintf(stderr, "		colname=[flags][,access] "
					"... tablename\n");
	fprintf(stderr, "	%s -u [-p path] [-s sep] [-t type] "
				"[colname=access ...] tablename\n", pname);
	fprintf(stderr, "	%s -d tablename\n", pname);
	fprintf(stderr, "	%s [-D defaults] -a|A colname=val "
				"... tablename\n", pname);
	fprintf(stderr, "	%s [-D defaults] -a|A indexedname\n", pname);
	fprintf(stderr, "	%s -e|E colname=val ... indexedname\n", pname);
	fprintf(stderr, "	%s -m colname=val ... indexedname\n", pname);
	fprintf(stderr, "	%s -r|R [colname=val ...] tablename\n", pname);
	fprintf(stderr, "	%s -r|R indexedname\n", pname);

	exit(1);
}

extern nis_error __nis_parse_request __P(( nis_name, nis_attr **,
				unsigned long *, nis_name * ));

static void check_table_sanity(t)
	table_obj		*t;
{
	int			i;
	table_col		*tc;
	int			has_search = 0;

	/*
	 * Check that table schema is sane.
	 */

	if (t->ta_maxcol > NIS_MAXCOLUMNS ||
		t->ta_cols.ta_cols_len > NIS_MAXCOLUMNS)
		errx(1, "invalid table schema: maximum "
			"number of columns (%d) exceeded", NIS_MAXCOLUMNS);

	if (t->ta_maxcol <= 0 || t->ta_cols.ta_cols_len <= 0)
		errx(1, "invalid table schema: invalid number of columns");

	/*
	 * Should have at least one column that
	 * is searchable.
	 */

	for (i = 0; i < t->ta_cols.ta_cols_len; i++) {
		tc = &t->ta_cols.ta_cols_val[i];
		if (tc->tc_flags & TA_SEARCHABLE)
			has_search++;
	}

	if (!has_search)
		errx(1, "invalid table schema: at least "
			"one column should be searchable");

	return;
}

/*
 * Create an entry based on supplied key/value paris and table
 * name. Make sure table exists, make sure key/value pair make
 * sense with respect to table object definition.
 */
static nis_error make_entry(argc, argv, e, table)
	int			argc;
	char			*argv[];
	nis_object		*e;
	nis_object		**table;
{
	nis_name		tname;
	nis_object		*tobj;
	table_obj		*t;
	table_col		*tc;
	entry_col		*ec = NULL, *c = NULL;
	nis_result		*res;
	nis_error		r = NIS_FAIL;
	int			i, j, a;
	nis_attr		*attrs = NULL;
	unsigned long		numattrs;
	char			*p;

	if (!argc) {
		warnx("not enough arguments");
		usage();
	}

	if (argc == 1) {
		r = __nis_parse_request(argv[0], &attrs, &numattrs, &tname);
		if (r != NIS_SUCCESS)
			return(r);
	} else {
		if (argc < 2) {
			warnx("not enough arguments");
			usage();
		}
		tname = argv[argc - 1];
		argc--;
		for (i = 0; i < argc; i++) {
			p = strchr(argv[i], '=');
			if (p == NULL || !strlen(p + 1)) {
				warnx("%s: not an attribute/value pair",
								argv[i]);
				usage();
			}
			*p = '\0';
			if (attrs == NULL)
				attrs = malloc(sizeof(nis_attr));
			else
				attrs = realloc(attrs,
						sizeof(nis_attr) * (i + 1));
			if (attrs == NULL)
				errx(1, "out of memory allocating nis_attr");
			attrs[i].zattr_ndx = strdup(argv[i]);
			attrs[i].zattr_val.zattr_val_len = strlen(p + 1) + 1;
			attrs[i].zattr_val.zattr_val_val = strdup(p + 1);
		}
		numattrs = i;
	}

	res = nis_lookup(tname, MASTER_ONLY|FOLLOW_LINKS|EXPAND_NAME);

	if (res != NULL)
		r = res->status;

	if (r != NIS_SUCCESS) {
		nis_freeresult(res);
		warnx("%s: %s", tname, nis_sperrno(r));
		return(r);
	}

	tobj = NIS_RES_OBJECT(res);

	if (__type_of(tobj) != TABLE_OBJ) {
		nis_freeresult(res);
		errx(1, "%s is not a table!", tname);
	}

	t = &tobj->TA_data;

	/*
	 * Sanity check.
	 */
	for (i = 0; i < numattrs; i++) {
		for (j = 0; j < t->ta_cols.ta_cols_len; j++) {
			tc = &t->ta_cols.ta_cols_val[j];
			if (!strcmp(attrs[i].zattr_ndx, tc->tc_name))
				break;
		}
		if (j == t->ta_cols.ta_cols_len)
			errx(1, "table %s.%s has no column named %s",
				tobj->zo_name, tobj->zo_domain,
					attrs[i].zattr_ndx);
	}

	j = 0;
	for (i = 0; i < t->ta_cols.ta_cols_len; i++) {
		tc = &t->ta_cols.ta_cols_val[i];
		if (ec == NULL)
			ec = calloc(1, sizeof(entry_col));
		else
			ec = realloc(ec, sizeof(entry_col) * (j + 2));
		c = &ec[j];
		for(a = 0; a < numattrs; a++) {
			if (!strcmp(attrs[a].zattr_ndx, tc->tc_name))
				break;
		}
		c->ec_flags = 0;
		if (a == numattrs) {
			c->ec_value.ec_value_len = 0;
			c->ec_value.ec_value_val = NULL;

		} else {
			c->ec_value.ec_value_len =
					attrs[a].zattr_val.zattr_val_len;
			c->ec_value.ec_value_val =
					malloc(c->ec_value.ec_value_len);
			bcopy(attrs[a].zattr_val.zattr_val_val,
				c->ec_value.ec_value_val,
				c->ec_value.ec_value_len);
		}
		j++;
	}

	__type_of(e) = ENTRY_OBJ;
	e->EN_data.en_type = strdup(t->ta_type);
	e->EN_data.en_cols.en_cols_len = j;
	e->EN_data.en_cols.en_cols_val = ec;
	e->zo_name = strdup(tobj->zo_name);
	e->zo_domain = strdup(tobj->zo_domain);
	e->zo_owner = nis_default_owner();
	e->zo_group = nis_default_group();
	e->zo_access = nis_default_rights();
	e->zo_ttl = nis_default_ttl();

	*table = tobj;
	NIS_RES_OBJECT(res) = NULL;
	NIS_RES_NUMOBJ(res) = 0;

	nis_freeresult(res);

	return(NIS_SUCCESS);
}

static void destroy_table(name)
	nis_name		name;
{
	nis_result		*r;
	nis_error		s;
	nis_object		*t;
	char			nbuf[NIS_MAXNAMELEN];

	r = nis_lookup(name, MASTER_ONLY|FOLLOW_LINKS|EXPAND_NAME);

	if (r == NULL)
		errx(1, "out of memory doing nis_lookup");

	s = r->status;

	if (s != NIS_SUCCESS) {
		nis_freeresult(r);
		errx(1, "lookup of %s failed: %s", name, nis_sperrno(s));
	}

	if (NIS_RES_NUMOBJ(r) > 1) {
		nis_freeresult(r);
		errx(1, "lookup of %s yielded more than one object!", name);
	}

	t = NIS_RES_OBJECT(r);
	NIS_RES_OBJECT(r) = NULL;
	NIS_RES_NUMOBJ(r) = 0;
	nis_freeresult(r);

	if (__type_of(t) != TABLE_OBJ) {
		nis_destroy_object(t);
		errx(1, "%s is not a table!", name);
	}

	snprintf(nbuf, sizeof(nbuf), "%s.%s", t->zo_name, t->zo_domain);
	r = nis_remove(nbuf, t);
	nis_destroy_object(t);

	if (r == NULL)
		errx(1, "out of memory doing nis_remove");

	s = r->status;
	nis_freeresult(r);

	if (s != NIS_SUCCESS)
		errx(1, "failed to destroy %s: %s", nbuf, nis_sperrno(s));

	exit(0);
}

static unsigned long parse_flags(str)
	char			*str;
{
	char			*p = str;
	unsigned long		flags = 0;

	if (p == NULL)
		return(-1);

	while(*p != NUL && *p != '\n') {
		switch(*p) {
		case 'S':
		case 's':
			flags |= TA_SEARCHABLE;
			break;
		case 'C':
		case 'c':
			flags |= TA_CRYPT;
			break;
		case 'I':
		case 'i':
			flags |= TA_CASE;
			break;
		case 'B':
		case 'b':
			flags |= TA_BINARY;
			break;
		case 'X':
		case 'x':
			flags |= TA_CRYPT;
			break;
		default:
			return(-1);
			break;
		}
		p++;
	}

	return(flags);
}

static void parse_coldata(str, cols, colcnt, rights_only)
	char			*str;
	table_col		**cols;
	unsigned long		*colcnt;
	int			rights_only;
{
	char			*p1 = NULL, *p2 = NULL;
	table_col		*t;
	unsigned long		flags = -1, rights = 0;

	if (*cols == NULL)
		*cols = (table_col *)calloc(1, sizeof(table_col));
	else
		*cols = (table_col *)realloc(*cols, (*colcnt + 1) *
						sizeof(table_col));
	if (*cols == NULL)
		errx(1, "out of memory allocating column data");

	t = *cols;

	/*
	 * First, look for equals sign; this
	 * denotes a column name/data pair.
	 */
	p1 = strchr(str, '=');
	*p1 = NUL;
	p2 = p1 + 1;

	t[*colcnt].tc_name = strdup(str);

	if (!strlen(p2)) {
		*p1 = '=';
		(*colcnt)++;
		return;
	}

	*p1 = '=';
	p1 = strchr(p2, ',');
	if (p1 != NULL)
		*p1 = NUL;
	flags = parse_flags(p2);
	if (p1 != NULL)
		*p1 = ',';

	if (flags == -1 && !rights_only)
		errx(1, "column flags not specified: %s", str);

	t[*colcnt].tc_flags = flags;

	if (p1 == NULL || !strlen(p1))
		rights = nis_default_rights();
	else {
		p1++;
		if (nis_parse_modes(p1, &rights, NULL, NULL) == -1)
			errx(1, "Couldn't parse access rights (%s) "
				"for column %s", p1, str);
	}

	t[*colcnt].tc_rights = rights;

	(*colcnt)++;

	return;
}

static void create_table(argc, argv, path, sep)
	int			argc;
	char			*argv[];
	char			*path;
	char			sep;
{
	nis_result		*r;
	nis_object		o;
	table_obj		*t;
	nis_error		s;
	char			*tname = NULL;
	table_col		*tcols = NULL;
	unsigned long		colcnt = 0;
	char			*type = NULL;
	int			i;

	bzero((char *)&o, sizeof(o));
	bzero((char *)&t, sizeof(t));

	type = argv[0];

	argc--; argv++;

	for (i = 0; i < argc; i++) {
		if (strchr(argv[i], '=') != NULL)
			parse_coldata(argv[i], &tcols, &colcnt);
		else
			break;
	}


	if (!colcnt)
		errx(1, "no column information specified");

	argc -= i;
	argv += i;

	if (!argc)
		errx(1, "no table name specified");

	tname = argv[0];

	if (tname[strlen(tname) - 1] != '.')
		errx(1, "table name must be fully qualified");

	/*
	 * See if object with same name exists.
	 */
	r = nis_lookup(tname, MASTER_ONLY|FOLLOW_LINKS|EXPAND_NAME);

	if (r == NULL)
		errx(1, "out of memory doing nis_lookup");

	s = r->status;

	if (s == NIS_SUCCESS) {
		if (__type_of(NIS_RES_OBJECT(r)) == TABLE_OBJ) {
			nis_freeresult(r);
			errx(1, "table already exists!");
		} else {
			nis_freeresult(r);
			errx(1, "object with same name already exists");
		}
	}

	nis_freeresult(r);
/*
	if (s != NIS_NOTFOUND) {
		errx(1, "%s: %s", tname, nis_sperrno(s));
	}
*/
	o.zo_name = strdup(nis_leaf_of(tname));
	o.zo_domain = strdup(nis_domain_of(tname));
	o.zo_owner = strdup(nis_default_owner());
	o.zo_group = strdup(nis_default_group());
	o.zo_access = nis_default_rights();
	o.zo_ttl = nis_default_ttl();
	__type_of(&o) = TABLE_OBJ;
	t = &o.TA_data;

	if (sep == NUL)
		t->ta_sep = ' ';
	else
		t->ta_sep = sep;

	if (path == NULL)
		t->ta_path = strdup("");
	else
		t->ta_path = strdup(path);

	t->ta_type = strdup(type);
	t->ta_cols.ta_cols_len = colcnt;
	t->ta_cols.ta_cols_val = tcols;
	t->ta_maxcol = colcnt;

	check_table_sanity(t);

	r = nis_add(tname, &o);
	xdr_free((xdrproc_t)xdr_nis_object, (char *)&o);

	if (r == NULL)
		errx(1, "couldn't create table: out of memory doing nis_add");

	s = r->status;
	nis_freeresult(r);

	if (s != NIS_SUCCESS)
		errx(1, "couldn't create table: %s", nis_sperrno(s));

	exit(0);
}

static void update_entry(argc, argv, upper)
	int			argc;
	char			*argv[];
	int			upper;
{
	unsigned long		flags = MASTER_ONLY|EXPAND_NAME;
	nis_object		*e_obj, *t_obj;
	nis_result		*res;
	nis_error		s;
	char			nbuf[NIS_MAXNAMELEN];
	table_obj		*t;
	table_col		*tc;
	entry_obj		*e;
	entry_col		*ec;
	int			i, j;
	nis_name		tname;

	if (argc < 2) {
		warnx("not enough arguments");
		usage();
	}

	tname = argv[argc - 1];
	argc--;

	res = nis_list(tname, flags, NULL, NULL);

	if (res == NULL)
		errx(1, "out of memory doing nis_list()");

	s = res->status;
	if (s != NIS_SUCCESS) {
		nis_freeresult(res);
		errx(1, "failed to look up %s: %s", argv[argc - 1],
							nis_sperrno(s));
	}

	if (NIS_RES_NUMOBJ(res) > 1) {
		nis_freeresult(res);
		errx(1, "nis_list() for %s returned more than one object",
					argv[argc - 1]);
	}

	if (__type_of(NIS_RES_OBJECT(res)) != ENTRY_OBJ) {
		nis_freeresult(res);
		errx(1, "%s is not an entry!", argv[argc - 1]);
	}

	e_obj = NIS_RES_OBJECT(res);
	e = &e_obj->EN_data;
	NIS_RES_OBJECT(res) = NULL;
	NIS_RES_NUMOBJ(res) = 0;
	nis_freeresult(res);
	snprintf(nbuf, sizeof(nbuf), "%s.%s", e_obj->zo_name,
						e_obj->zo_domain);

	/* Now get the table object; we need that too. */
	res = nis_lookup(nbuf, MASTER_ONLY|FOLLOW_LINKS|EXPAND_NAME);

	if (res == NULL) {
		nis_destroy_object(e_obj);
		errx(1, "out of memory doing nis_lookup()");
	}

	s = res->status;

	if (s != NIS_SUCCESS) {
		nis_destroy_object(e_obj);
		nis_freeresult(res);
		errx(1, "lookup of %s failed: %s", nbuf, nis_sperrno(s));
	}

	t_obj = NIS_RES_OBJECT(res);

	if (__type_of(t_obj) != TABLE_OBJ) {
		nis_destroy_object(e_obj);
		nis_freeresult(res);
		errx(1, "%s is not a table!", nbuf);
	}

	NIS_RES_OBJECT(res) = NULL;
	NIS_RES_NUMOBJ(res) = 0;
	nis_freeresult(res);
	
	e = &e_obj->EN_data;
	t = &t_obj->TA_data;

	for (i = 0; i < argc; i++) {
		char			*attr;
		char			*val;

		attr = argv[i];
		val = strchr(attr, '=');
		if (val == NULL || !strlen(val + 1)) {
			nis_destroy_object(e_obj);
			nis_destroy_object(t_obj);
			errx(1, "malformed attr/val pair: %s", argv[i]);
		}
		*val = NUL;
		val++;
		for (j = 0; j < e->en_cols.en_cols_len; j++) {
			ec = &e->en_cols.en_cols_val[j];
			tc = &t->ta_cols.ta_cols_val[j];
			if (!strcmp(tc->tc_name, attr)) {
				free(ec->ec_value.ec_value_val);
				ec->ec_value.ec_value_val = strdup(val);
				ec->ec_value.ec_value_len = strlen(val) + 1;
				ec->ec_flags |= EN_MODIFIED;
				break;
			}
		}
		val--;
		*val = '=';
		if (j == e->en_cols.en_cols_len) {
			nis_destroy_object(e_obj);
			nis_destroy_object(t_obj);
			errx(1, "table %s has no column named %s", nbuf, attr);
		}
	}

	if (tname[strlen(tname) - 1] != '.') {
		char			*p;

		p = strchr(tname, ']');
		if (p == NULL) {
			nis_destroy_object(e_obj);
			nis_destroy_object(t_obj);
			errx(1, "bad name: %s\n", tname);
		}
		p++;
		*p = NUL;
		snprintf(nbuf, sizeof(nbuf), "%s,%s.%s", tname,
					t_obj->zo_name, t_obj->zo_domain);
		*p = ',';
		p--;
	} else {
		snprintf(nbuf, sizeof(nbuf), "%s", tname);
	}

	if (upper)
		flags |= ADD_OVERWRITE;

	res = nis_modify_entry(nbuf, e_obj, flags);

	nis_destroy_object(e_obj);
	nis_destroy_object(t_obj);

	if (res == NULL)
		errx(1, "out of memory doing nis_modify_entry()");

	s = res->status;

	nis_freeresult(res);

	if (s != NIS_SUCCESS)
		errx(1, "modify failed: %s", nis_sperrno(s));

	exit(0);
}

/*
 * Update a table object. There are only a few things that can be
 * changed:
 * - The separator character
 * - The concatenation path
 * - The type
 * - The access rights for a given column
 * Nothing else can be modified. To add or remove columns or change
 * column attributes, you need to destroy the table and create a new
 * one to take its place.
 */
static void update_table(argc, argv, path, sep, type)
	int			argc;
	char			*argv[];
	char			*path;
	char			sep;
	char			*type;
{
	nis_result		*res;
	nis_object		*o;
	table_obj		*t;
	table_col		*tc;
	nis_error		s;
	char			*tname = NULL;
	table_col		*tcols = NULL;
	unsigned long		colcnt = 0;
	int			i, j;
	char			nbuf[NIS_MAXNAMELEN];

	if (!argc) {
		warnx("not enough arguments");
		usage();
	}

	for (i = 0; i < argc; i++) {
		if (strchr(argv[i], '=') != NULL) {
			parse_coldata(argv[i], &tcols, &colcnt);
		} else
			break;
	}

	argc -= i;
	argv += i;

	if (!argc)
		errx(1, "no table name specified");

	tname = argv[0];

	res = nis_lookup(tname, MASTER_ONLY|FOLLOW_LINKS|EXPAND_NAME);

	if (res == NULL)
		errx(1, "out of memory doing nis_lookup");

	s = res->status;

	if (s != NIS_SUCCESS) {
		nis_freeresult(res);
		errx(1, "couldn't modify table: %s", nis_sperrno(s));
	}

	o = NIS_RES_OBJECT(res);
	snprintf(nbuf, sizeof(nbuf), "%s.%s", o->zo_name, o->zo_domain);

	if (__type_of(o) != TABLE_OBJ) {
		nis_freeresult(res);
		errx(1, "%s is not a table!", nbuf);
	}

	t = &o->TA_data;
	check_table_sanity(t);

	for (i = 0; i < t->ta_cols.ta_cols_len; i++) {
		tc = &t->ta_cols.ta_cols_val[i];
		for(j = 0; j < colcnt; j++) {
			if (!strcmp(tc->tc_name, tcols[j].tc_name))
				tc->tc_rights = tcols[j].tc_rights;
		}
	}

	if (path != NULL) {
		free(t->ta_path);
		t->ta_path = strdup(path);
	}
	if (sep)
		t->ta_sep = sep;
	if (type != NULL) {
		free(t->ta_type);
		t->ta_type = strdup(type);
	}
	NIS_RES_OBJECT(res) = NULL;
	NIS_RES_NUMOBJ(res) = 0;
	nis_freeresult(res);

	res = nis_modify(nbuf, o);

	if (res == NULL)
		errx(1, "out of memory doing nis_modify()");

	s = res->status;

	nis_freeresult(res);

	if (s != NIS_SUCCESS)
		errx(1, "couldn't update table: %s", nis_sperrno(s));

	exit(0);
}

static void add_entry(argc, argv, upper)
	int			argc;
	char			*argv[];
	int			upper;
{
	unsigned long		flags = MASTER_ONLY|EXPAND_NAME;
	nis_object		entry, *table = NULL;
	nis_result		*res;
	nis_error		r;
	char			nbuf[NIS_MAXNAMELEN];

	r = make_entry(argc, argv, &entry, &table);

	if (r == NIS_BADREQUEST)
		usage();

	if (r != NIS_SUCCESS)
		exit(1);

	if (upper)
		flags |= ADD_OVERWRITE;

	snprintf(nbuf, sizeof(nbuf), "%s.%s", table->zo_name,
						table->zo_domain);
	res = nis_add_entry(nbuf, &entry, flags);

	if (res == NULL)
		errx(1, "out of memory doing nis_add_entry");

	r = res->status;
	nis_freeresult(res);

	if (r != NIS_SUCCESS)
		errx(1, "%s: %s", nbuf, nis_sperrno(r));

	return;
}

static void remove_entry(argc, argv, upper)
	int			argc;
	char			*argv[];
	int			upper;
{
	unsigned long		flags = MASTER_ONLY|EXPAND_NAME;
	nis_result		*res;
	nis_error		r;
	char			nbuf[NIS_MAXNAMELEN];
	nis_attr		*attrs = NULL;
	unsigned long		nattrs = 0;
	nis_name		tname;
	int			i;
	nis_object		*t;

	if (!argc) {
		warnx("not enough arguments");
		usage();
	}

	if (argc == 1 && argv[0][0] == '[') {
		r = __nis_parse_request(argv[0], &attrs, &nattrs, &tname);
		if (r != NIS_SUCCESS)
			errx(1, "remove failed: %s", nis_sperrno(r));
		strcat(nbuf, "[");
		for (i = 0; i < nattrs; i++) {
			if (strlen(nbuf) > 1)
				strncat(nbuf, ",", sizeof(nbuf));
			strncat(nbuf, attrs[i].zattr_ndx, sizeof(nbuf));
			strncat(nbuf, "=", sizeof(nbuf));
			strncat(nbuf, attrs[i].zattr_val.zattr_val_val,
				attrs[i].zattr_val.zattr_val_len);
		}
	} else {
		char			*p;
		tname = argv[argc - 1];
		argc--;
		strcat(nbuf, "[");
		for (i = 0; i < argc; i++) {
			p = strchr(argv[i], '=');
			if (p == NULL || !strlen(p + 1)) {
				warnx("%s: not an attribute/value pair",
								argv[i]);
				usage();
			}
			if (strlen(nbuf) > 1)
				strncat(nbuf, ",", sizeof(nbuf));
			strncat(nbuf, argv[i], sizeof(nbuf));
		}
	}

	res = nis_lookup(tname, MASTER_ONLY|FOLLOW_LINKS|EXPAND_NAME);
	if (res == NULL)
		errx(1, "out of memory doing nis_lookup()");
	r = res->status;
	if (r != NIS_SUCCESS) {
		nis_freeresult(res);
		errx(1, "remove failed: %s", nis_sperrno(r));
	}

	if (__type_of(NIS_RES_OBJECT(res)) != TABLE_OBJ) {
		nis_freeresult(res);
		errx(1, "%s is not a table!", tname);
	}

	t = NIS_RES_OBJECT(res);

	strncat(nbuf, "],", sizeof(nbuf));
	strncat(nbuf, t->zo_name, sizeof(nbuf));
	strncat(nbuf, ".", sizeof(nbuf));
	strncat(nbuf, t->zo_domain, sizeof(nbuf));

	nis_freeresult(res);

	if (upper)
		flags |= REM_MULTIPLE;

	res = nis_remove_entry(nbuf, NULL, flags);

	if (res == NULL)
		errx(1, "out of memory doing nis_remove_entry()");

	r = res->status;
	nis_freeresult(res);

	if (r != NIS_SUCCESS)
		errx(1, "can't remove entry: %s", nis_sperrno(r));

	exit(0);
}

int main(argc, argv)
	int			argc;
	char			*argv[];
{
	int			ch;
	char			*defs = NULL;
	char			*path = NULL;
	int			action = 0;
	unsigned char		sep = 0;
	char			*type = NULL;
	int			upper = 0;

	pname = argv[0];

	while((ch = getopt(argc, argv, "aAcdD:eEmrRs:t:u")) != -1) {
		switch(ch) {
		case 'A':
			upper++;
		case 'a':
			action = ACTION_ADDENT;
			break;
		case 'c':
			action = ACTION_NEWTAB;
			break;
		case 'd':
			action = ACTION_REMTAB;
			break;
		case 'D':
			defs = optarg;
			break;
		case 'E':
		case 'm': /* backwards compat */
			upper++;
		case 'e':
			action = ACTION_MODENT;
			break;
		case 'p':
			path = optarg;
			break;
		case 'R':
			upper++;
		case 'r':
			action = ACTION_REMENT;
			break;
		case 's':
			if (strlen(optarg) > 1) {
				warnx("separator must be a single character");
				usage();
			}
			sep = *optarg;
			break;
		case 't':
			type = optarg;
			break;
		case 'u':
			action = ACTION_MODTAB;
			break;
		default:
			usage();
			break;
		}
	}


	argc -= optind;
	argv += optind;

	if (!argc || !action)
		usage();

	if (nis_init_defs(defs))
		exit(1);

	switch(action) {
	case ACTION_REMTAB:
		if (argc != 1)
			usage();
		destroy_table(argv[0]);
		break;
	case ACTION_MODTAB:
		update_table(argc, argv, path, sep, type);
		break;
	case ACTION_NEWTAB:
		create_table(argc, argv, path, sep);
		break;
	case ACTION_ADDENT:
		add_entry(argc, argv, upper);
		break;
	case ACTION_REMENT:
		remove_entry(argc, argv, upper);
		break;
	case ACTION_MODENT:
		update_entry(argc, argv, upper);
		break;
	default:
		errx(1, "unknown action");
		break;
	}


	exit(0);
}
