/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nistest.c,v 1.4 1998/01/24 04:14:56 wpaul Exp $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nis_common.h"

#ifndef lint
static const char copyright[] = "@(#) Copyright (c) 1996, 1997\
	Bill Paul. All rights reserved.";
static const char rcsid[] =
	"@(#) $Id: nistest.c,v 1.4 1998/01/24 04:14:56 wpaul Exp $";
#endif

static void usage(pname)
	char			*pname;
{
	fprintf(stderr, "%s: [-LPAM] [-t G|D|L|T|E|P] [-a mode] name\n", pname);
	exit(2);
}

static zotypes str2type(str)
	char			*str;
{
	switch(str[0]) {
	case 'T':
		return (TABLE_OBJ);
	case 'D':
		return (DIRECTORY_OBJ);
	case 'E':
		return (ENTRY_OBJ);
	case 'G':
		return (GROUP_OBJ);
	case 'L':
		return (LINK_OBJ);
	case 'P':
		return (PRIVATE_OBJ);
	default:
		return (BOGUS_OBJ);
	}
}

static void test_indexed(name, flags, access)
	nis_name		name;
	unsigned long		flags;
	unsigned long		access;
{
	nis_result		*res;
	nis_object		*o;

	res = nis_list(name, flags, NULL, NULL);

	if (res == NULL)
		errx(1, "out of memory doing nis_lookup");

	if (res->status != NIS_SUCCESS) {
		nis_freeresult(res);
		exit(1);
	}

	o = NIS_RES_OBJECT(res);

	if (access) {
		if ((o->zo_access & access) == access) {
			nis_freeresult(res);
			exit(1);
		}
	}

	nis_freeresult(res);
	exit(0);
}

static void test_object(name, flags, access, otype)
	nis_name		name;
	unsigned long		flags;
	unsigned long		access;
	zotypes			otype;
{
	nis_result		*res;
	nis_object		*o;

	res = nis_lookup(name, flags);

	if (res == NULL)
		errx(1, "out of memory doing nis_lookup");

	if (res->status != NIS_SUCCESS) {
		nis_freeresult(res);
		exit(1);
	}

	o = NIS_RES_OBJECT(res);

	if (otype != NO_OBJ) {
		if (__type_of(o) != otype) {
			nis_freeresult(res);
			exit(1);
		}
	}

	if (access) {
		if (o->zo_access & ~access) {
			nis_freeresult(res);
			exit(1);
		}
	}

	nis_freeresult(res);
	exit(0);
}

int main(argc, argv)
	int			argc;
	char			*argv[];
{
	int			ch;
	unsigned long		flags = 0;
	zotypes			otype = NO_OBJ;
	char			*pname = argv[0];
	unsigned long		omode = 0;

	while((ch = getopt(argc, argv, "LAPMt:a:")) != -1) {
		switch(ch) {
		case 'L':
			flags |= FOLLOW_LINKS;
			break;
		case 'P':
			flags |= FOLLOW_PATH;
			break;
		case 'A':
			flags |= ALL_RESULTS;
			break;
		case 'M':
			flags |= MASTER_ONLY;
			break;
		case 't':
			if (strlen(optarg) != 1)
				usage(argv[0]);
			otype = str2type(optarg);
			if (otype == BOGUS_OBJ)
				usage(argv[0]);
			break;
		case 'a':
			if (nis_parse_modes(optarg, &omode, NULL, NULL) == -1)
				usage(argv[0]);
			break;
		default:
			usage(argv[0]);
		break;
		}
	}

	argc -= optind;
	argv += optind;

	if (!argc)
		usage(pname);

	if (argv[0][0] == '[') {
		if (otype != NO_OBJ)
			usage(pname);
		test_indexed(argv[0], flags, omode);
	} else
		test_object(argv[0], flags, omode, otype);

	exit(1);
}

