#include <sys/cdefs.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <err.h>
#include <errno.h>
#include <signal.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nis_cache_bind.h"
#include "nis_cache_extern.h"
#include "nis_extern.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nis_cache_bind.c,v 1.1 1998/05/30 22:05:40 wpaul Exp wpaul $";
#endif

/*
int _rpcpmstart = 0;
char *progname = "foo";
*/
/*
 * In order to make cache file updates as painless as possible, we
 * (ab)use the XDR library to dump the cache data directly to an
 * mmap()ed memory region using xdrmem_create(). This involves devising
 * a set of structures that are all linked together to form one giant
 * structure that the XDR routines can serialize all in one shot.
 */

static cachebind_header		*header;
static cachebind_hash_entry_p	*table;
static cachebind_combined	*combined;
static int			cachebind_fd;
static caddr_t			laddr;
static unsigned long		csize;

static u_int32_t hashkey(key)
	char			*key;
{
	if (key == NULL)
		return(-1);
	return(__nis_hash((void *)key, strlen(key)));
}

static void nis_cachebind_alarm(sig)
	int		sig;
{
	if (sig != SIGALRM)
		nis_err("caught sgnal type %d by mistake", sig);

	return;
}

/*
 * Aquire exclusive lock on cache file.
 */
static void nis_cachebind_lock()
{
	struct flock		lock;

	lock.l_start = 0;
	lock.l_len = 0;
	lock.l_pid = getpid();
	lock.l_type = F_WRLCK;
	lock.l_whence = SEEK_SET;

	alarm(5);
	if (fcntl(cachebind_fd, F_SETLK, &lock) == -1)
		nis_err("help! flock failed: %s", strerror(errno));
	alarm(0);

	return;
}

/*
 * Give up exclusive lock on cache file.
 */
static void nis_cachebind_unlock()
{
	struct flock		lock;

	lock.l_start = 0;
	lock.l_len = 0;
	lock.l_pid = getpid();
	lock.l_type = F_UNLCK;
	lock.l_whence = SEEK_SET;

	if (fcntl(cachebind_fd, F_SETLK, &lock) == -1)
		nis_err("unlocking failed: %s", strerror(errno));

	return;
}

static nis_error nis_cachebind_dump()
{
	unsigned long		tsize, fsize;
	XDR			xdrs;

	tsize = xdr_sizeof(xdr_cachebind_combined, (char *)combined);
	if (tsize > CB_CHUNKSIZE)
		fsize = tsize + (CB_CHUNKSIZE - (tsize % CB_CHUNKSIZE));
	else
		fsize = CB_CHUNKSIZE;

	if (laddr != NULL)
		munmap(laddr, csize);

	nis_cachebind_lock();
	ftruncate(cachebind_fd, fsize);
	csize = fsize;

	laddr = mmap(0, csize, PROT_READ|PROT_WRITE,
			MAP_SHARED, cachebind_fd, 0);

	if (laddr == MAP_FAILED) {
		munmap(laddr, csize);
		laddr = NULL;
		nis_cachebind_unlock();
		return(NIS_SYSTEMERROR);
	}

	xdrmem_create(&xdrs, laddr, tsize, XDR_ENCODE);
	if (!xdr_cachebind_combined(&xdrs, combined)) {
		xdr_destroy(&xdrs);
		munmap(laddr, csize);
		laddr = NULL;
		nis_cachebind_unlock();
		return(NIS_SYSTEMERROR);
	}
	xdr_destroy(&xdrs);
	msync(laddr, csize, 0);
	munmap(laddr, csize);
	laddr = NULL;
	nis_cachebind_unlock();

	return(NIS_SUCCESS);
}

static nis_error nis_cachebind_init(fname)
	nis_name		fname;
{
	combined = (cachebind_combined *)calloc(1,sizeof(cachebind_combined));
	table = combined->cachebind_hash_table;
	header = &combined->cachebind_header;
	header->cb_vers  = CB_VERSION;
	header->cb_magic = CB_HEADER_MAGIC;

	cachebind_fd = open(fname, O_RDWR|O_CREAT, 0644);
	if (cachebind_fd == -1)
		return(NIS_SYSTEMERROR);

	return(nis_cachebind_dump(fname));
}

static nis_error nis_cachebind_load(fname)
	char			*fname;
{
	XDR			xdrs;
	struct stat		st;

	if (combined != NULL)
		return(NIS_SUCCESS);

	if (stat(fname, &st) == -1) {
		if (errno != ENOENT)
			return(NIS_SYSTEMERROR);
		return(nis_cachebind_init(fname));
	}

	csize = st.st_size;
	cachebind_fd = open(fname, O_RDWR, 0644);
	if (cachebind_fd == -1)
		return(NIS_SYSTEMERROR);

	laddr = mmap(0, csize, PROT_READ|PROT_WRITE,
			MAP_SHARED, cachebind_fd, 0);

	if (laddr == MAP_FAILED) {
		close(cachebind_fd);
		return(NIS_SYSTEMERROR);
	}

	combined = (cachebind_combined *)calloc(1,sizeof(cachebind_combined));
	xdrmem_create(&xdrs, laddr, csize, XDR_DECODE);
	if (!xdr_cachebind_combined(&xdrs, combined)) {
		xdr_destroy(&xdrs);
		munmap(laddr, csize);
		laddr = NULL;
		close(cachebind_fd);
		return(NIS_SYSTEMERROR);
	}
	xdr_destroy(&xdrs);
	msync(laddr, csize, 0);
	munmap(laddr, csize);
	laddr = NULL;

	table = combined->cachebind_hash_table;
	header = &combined->cachebind_header;

	return(NIS_SUCCESS);
}

nis_error nis_cachebind_open(fname)
	char			*fname;
{
	signal(SIGALRM, nis_cachebind_alarm);

	return(nis_cachebind_load(fname));
}

nis_error nis_cachebind_close()
{
	if (nis_cachebind_dump() != NIS_SUCCESS)
		return(NIS_SYSTEMERROR);

	close(cachebind_fd);

	return(NIS_SUCCESS);
}

nis_error nis_cachebind_insert(name, type, b, d, a)
	nis_name		name;
	cb_type			type;
	nis_bound_endpoint	*b;
	nis_bound_directory	*d;
	nis_active_endpoint	*a;
{
	cachebind_hash_entry	*new;
	u_int32_t		i;

	i = hashkey(name);

	new = (cachebind_hash_entry *)calloc(1, sizeof(cachebind_hash_entry));
	new->hval = i;
	new->cb_type = type;
	new->name = strdup(name);
	new->bep_ptr = b;
	new->bdr_ptr = d;
	new->aep_ptr = a;
	new->cr_time = time(NULL);
	new->next = NULL;

	new->next = (cachebind_hash_entry *)table[i & HASH_MASK];
	table[i & HASH_MASK] = (cachebind_hash_entry_p)new;

	nis_cachebind_dump();

	return(NIS_SUCCESS);
}

nis_error nis_cachebind_delete(name, type)
	nis_name		name;
	cb_type			type;
{
	cachebind_hash_entry	*cur, *prev;
	u_int32_t		i;

	i = hashkey(name);
	cur = (cachebind_hash_entry *)table[i & HASH_MASK];
	prev = NULL;

	while(cur != NULL){
		if (cur->hval == i && cur->cb_type == type)
			break;
		prev = cur;
		cur = cur->next;
	}

	if (cur == NULL)
		return(NIS_NOTFOUND);

	if (cur == table[i & HASH_MASK]) {
		table[i & HASH_MASK] = cur->next;
		free(cur->name);
		free(cur);
		return(NIS_SUCCESS);
	}

	if (prev == table[i &  HASH_MASK])
		table[i & HASH_MASK] = cur->next;
	else
		prev = cur->next;

	free(cur->name);
	free(cur);

	return(NIS_SUCCESS);
}

cachebind_hash_entry *nis_cachebind_lookup(name, type)
	nis_name		name;
	cb_type			type;
{
	cachebind_hash_entry	*cur;
	u_int32_t		i;

	i = hashkey(name);
	cur = (cachebind_hash_entry *)table[i & HASH_MASK];

	while(cur) {
		if (cur->hval == i &&  cur->cb_type == type)
			return(cur);
		cur = cur->next;
	}

	return(NULL);
}

nis_error nis_cachebind_insert_directory(dir)
	directory_obj		*dir;
{
	nis_bound_directory	d;

	bzero((char *)&d, sizeof(d));
	bcopy((char *)dir, (char *)&d.dobj, sizeof(directory_obj));
	nis_cachebind_delete(dir->do_name, CB_SERVER);
	nis_cachebind_insert(dir->do_name, CB_SERVER, NULL, &d, NULL);

	return(NIS_SUCCESS);
}

