/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: nis_cache_pkeys.c,v 1.2 1997/06/16 23:48:40 wpaul Exp wpaul $
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nis_cache.h>
#include <rpcsvc/nislib.h>

#ifndef lint
static const char rcsid[] = "@(#) $Id: nis_cache_pkeys.c,v 1.2 1997/06/16 23:48:40 wpaul Exp wpaul $";
#endif

#define TABLESIZE 256
#define HASH_MASK 0x000000FF

struct h_entry {
	unsigned long hashval;
	nis_name name;
	char *pkey;
	struct h_entry *next;
};

static struct h_entry *pkeys[TABLESIZE];

static u_int32_t hashkey(key)
	char			*key;
{
	if (key == NULL)
		return(-1);
	return(__nis_hash((void *)key, strlen(key)));
}

static void h_insert(name, pkey)
	nis_name		name;
	netobj			*pkey;
{
	struct h_entry		*new;
	u_int32_t		i;

	i = hashkey(name) & HASH_MASK;

	new = (struct h_entry *)malloc(sizeof(struct h_entry));
	new->hashval = i;
	new->name = strdup(name);
	new->pkey = calloc(1, pkey->n_len + 1);
	bcopy(pkey->n_bytes, new->pkey, pkey->n_len);
	new->next = NULL;

	new->next = pkeys[i];
	pkeys[i] = new;

	return;
}

static char *h_lookup(key)
	nis_name		key;
{
	struct h_entry		*cur;
	u_int32_t		i;

	i = hashkey(key) & HASH_MASK;
	cur = pkeys[i];

	while (cur) {
		if (cur->hashval == i)
			return(cur->pkey);
		cur = cur->next;
	}

	return(NULL);
}

static void h_delete(key)
	nis_name		key;
{
	struct h_entry		*cur, *prev;
	u_int32_t		i;

	i = hashkey(key) & HASH_MASK;
	cur = pkeys[i];
	prev = NULL;

	while (cur) {
		if (cur->hashval == i)
			break;
		prev = cur;
		cur = cur->next;
	}

	if (cur == NULL)
		return;

	if (cur == pkeys[i]) {
		pkeys[i] = cur->next;
		free(cur->name);
		free(cur->pkey);
		free(cur);
		return;
	}

	if (prev == pkeys[i])
		pkeys[i] = cur->next;
	else
		prev = cur->next;

	free(cur->name);
	free(cur->pkey);
	free(cur);

	return;
}

void nis_remove_pkeys(dir)
	directory_obj *dir;
{
	unsigned long		i;
	nis_server		*nptr;

	for (i = 0; i < dir->do_servers.do_servers_len; i++) {
		nptr = &dir->do_servers.do_servers_val[i];
		if (nptr->key_type != NIS_PK_DH)
			continue;
		h_delete(nptr->name);
	}

	return;
}

char *nis_lookup_pkey(name)
	nis_name		name;
{
	return(h_lookup(name));
}

void nis_cache_pkeys(dir)
	directory_obj *dir;
{
	unsigned long		i;
	nis_server		*nptr = NULL;

	for (i = 0; i < dir->do_servers.do_servers_len; i++) {
		nptr = &dir->do_servers.do_servers_val[i];
		if (nptr->key_type != NIS_PK_DH)
			continue;
		h_delete(nptr->name);
		h_insert(nptr->name, &nptr->pkey);
	}

	return;
}
