/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: nis_cache_subr.c,v 1.10 1997/09/03 05:09:05 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/cdefs.h>
#include <sys/queue.h>
#include <sys/file.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nis_cache.h>
#include "nis_cache_extern.h"
#include "nis_extern.h"
#include "../../lib/nislib/nislib_cache.h"

#ifndef lint
static const char rcsid[] = "@(#) $Id: nis_cache_subr.c,v 1.10 1997/09/03 05:09:05 wpaul Exp wpaul $";
#endif

static CIRCLEQ_HEAD(cachehead, cacheq_entry) qhead;

struct cacheq_entry {
	time_t			ctime;
	unsigned long		hashval;
	directory_obj		*dptr;		/* un-xdr'ed object */
	CIRCLEQ_ENTRY(cacheq_entry) links;
};


static int			cache_fd;	/* Cache file descriptor */
static struct cache_header	*hptr;	/* Pointer to cache file header */
static struct cache_spacer	*sptr;	/* Pointer to current spacer */

static caddr_t			laddr;	/* Address returned by mmap() */
static off_t			csize;	/* Size of cache file */

int nis_verify_sig(fd_res)
	fd_result *fd_res;
{
	struct signature *sig;
	char *pkey;

	if (insecure)
		return(0);

	if ((pkey = nis_lookup_pkey(fd_res->source)) == NULL)
		return(1);

	sig = __nis_calculate_encrypted_cksum(fd_res->dir_data.dir_data_val,
					     fd_res->dir_data.dir_data_len,
					     pkey, DES_ENCRYPT | DES_HW);

	if (sig->signature_len != fd_res->signature.signature_len ||
		bcmp(sig->signature_val, fd_res->signature.signature_val,
						sig->signature_len))
		return(1);

	return(0);
}

static void nis_cache_mark(state)
	enum cache_state	state;
{
	hptr->state = state;
	hptr->upd_time = time(NULL);

	msync(laddr, csize, 0);

	return;
}

/*
 * Extend the log file by another chunk.
 */
nis_error nis_extend_cache()
{
	munmap(laddr, csize);
	csize += CHUNKSIZE;
	lseek(cache_fd, csize, SEEK_SET);
	write(cache_fd, ".", 1);
	laddr = mmap(0, csize, PROT_READ|PROT_WRITE, MAP_SHARED, cache_fd, 0);
	/*
	 * Now that we've extended the file, we recalculate
	 * our address pointers just in case the file has been
	 * mmap()ed at a different location.
	 */
	hptr = (struct cache_header *)laddr;
	sptr = (struct cache_spacer *)(laddr + hptr->offset);
	while(sptr->next_offset) {
		caddr_t	ptr;
		if (sptr->magic != SPACER_MAGIC) {
			nis_err("cache is corrupt!");
			return(NIS_FAIL);
		}
		ptr = (caddr_t)sptr;
		ptr += sptr->next_offset + sizeof(struct cache_spacer);
		sptr = (struct cache_spacer *)ptr;
	}
	return(NIS_SUCCESS);
}

void nis_cache_init()
{

	CIRCLEQ_INIT(&qhead);
	return;
}

int nis_cache_open()
{
	int			coldstart = 0;
	struct stat             st;

	if (stat(CACHEFILE, &st) == -1) {
		if ((cache_fd = open(CACHEFILE,O_RDWR|O_CREAT,0644)) == -1) {
			nis_err("couldn't open cache: %s", strerror(errno));
			return(1);
		}
		lseek(cache_fd, CHUNKSIZE, SEEK_SET);
		write(cache_fd, ".", 1);
		csize = CHUNKSIZE;
		coldstart++;
	} else {
		if ((cache_fd = open(CACHEFILE, O_RDWR,0600)) == -1) {
			nis_err("couldn't open cache: %s", strerror(errno));
			return(1);
		}
		csize = st.st_size;
	}

	laddr = mmap(0, csize, PROT_READ|PROT_WRITE, MAP_SHARED, cache_fd, 0);
	if ((int)laddr == -1) {
		nis_err("mmap failed: %s", strerror(errno));
		close(cache_fd);
		return(NIS_FAIL);
	}

	hptr = (struct cache_header *)laddr;
	if (coldstart) {
		hptr->magic = HEADER_MAGIC;
		hptr->state = CACHE_STABLE;
		hptr->offset = sizeof(struct cache_header);
		hptr->size = sizeof(struct cache_header) +
				sizeof(struct cache_spacer);
		hptr->upd_time = 0;
		sptr = (struct cache_spacer *)(laddr + hptr->offset);
		sptr->magic = SPACER_MAGIC;
		nis_cache_mark(CACHE_STABLE);
	} else {
		sptr = (struct cache_spacer *)(laddr + hptr->offset);
		if (hptr->magic != HEADER_MAGIC ||
			sptr->magic != SPACER_MAGIC) {
			nis_err("cache is corrupt! aborting!");
			munmap(laddr, csize);
			close(cache_fd);
			return(1);
		}
	}

	return(0);
}

/*
 * Aquire advisory exclusive lock on cache file.
 */
void nis_lock_cache()
{
	struct flock		lock;

	lock.l_start = 0;
	lock.l_len = 0;
	lock.l_pid = getpid();
	lock.l_type = F_WRLCK;
	lock.l_whence = SEEK_SET;

	if (fcntl(cache_fd, F_SETLKW, &lock) == -1)
		nis_err("help! flock failed: %s", strerror(errno));

	return;
}

/*
 * Give up exclusive lock on cache file.
 */
void nis_unlock_cache()
{
	struct flock		lock;

	lock.l_start = 0;
	lock.l_len = 0;
	lock.l_pid = getpid();
	lock.l_type = F_UNLCK;
	lock.l_whence = SEEK_SET;

	if (fcntl(cache_fd, F_SETLK, &lock) == -1)
		nis_err("unlocking failed: %s", strerror(errno));

	return;
}

/*
 * Dump our linked list of directory objects into NIS_SHARED_DIRCACHE.
 */
int nis_cache_dump()
{
	register struct cacheq_entry	*q;
	register caddr_t		ptr;
	struct cache_spacer		*s;
	unsigned long			size = 0;
	XDR				xdrs;

	nis_cache_mark(CACHE_UPDATING);

	s = (struct cache_spacer *)(laddr + hptr->offset);
	for (q = qhead.cqh_first; q != (void *)&qhead; q = q->links.cqe_next) {
		ptr = (caddr_t)(s) + sizeof(struct cache_spacer);
		size = xdr_sizeof(xdr_directory_obj, q->dptr);
		s->next_offset = size;
		s->hashval = q->hashval;
		s->ttl = q->dptr->do_ttl;
		s->ctime = time(NULL);
		if (!s->ttl)	/* XXX hack to work around Sun bug */
			s->ttl = 60*60;
		s->ctime = q->ctime;
		xdrmem_create(&xdrs, ptr, size, XDR_ENCODE);
		if (!xdr_directory_obj(&xdrs, q->dptr)) {
			xdr_destroy(&xdrs);
			nis_err("updating cache failed!");
			return(1);
		}
		xdr_destroy(&xdrs);
		s = (struct cache_spacer *)(ptr + size);
		s->prev_offset = size;
		s->magic = SPACER_MAGIC;
		hptr->size += sizeof(struct cache_spacer) + size;
	}

	nis_cache_mark(CACHE_STABLE);
	return(0);
}

/*
 * Load all the directory objects from the NIS_SHARED_DIRCACHE file
 * into a linked list.
 */
int nis_cache_load()
{
	register struct cacheq_entry	*q = NULL;
	register caddr_t		ptr;
	struct cache_spacer		*s;
	unsigned long			size = 0;
	XDR				xdrs;

	s = (struct cache_spacer *)(laddr + hptr->offset);
	while(s->next_offset) {
		if (s->magic != SPACER_MAGIC) {
			nis_err("cache is corrupt!");
			return(1);
		}
		ptr = (caddr_t)(s) + sizeof(struct cache_spacer);
		size = s->next_offset;
		q = (struct cacheq_entry *)calloc(1,
				sizeof(struct cacheq_entry));
		q->dptr = (directory_obj *)calloc(1, sizeof(directory_obj));
		xdrmem_create(&xdrs, ptr, size, XDR_DECODE);
		if (!xdr_directory_obj(&xdrs, q->dptr)) {
			xdr_destroy(&xdrs);
			free(q->dptr);
			free(q);
			nis_err("reading cache failed!");
			return(1);
		}
		nis_cache_pkeys(q->dptr);
		xdr_destroy(&xdrs);
		q->hashval = s->hashval;
		q->ctime = s->ctime;
		CIRCLEQ_INSERT_TAIL(&qhead, q, links);
		q = NULL;
		ptr = (caddr_t)s;
		ptr += s->next_offset + sizeof(struct cache_spacer);
		s = (struct cache_spacer *)ptr;
	}

	return(0);
}

int nis_cache_refresh(dir)
	directory_obj *dir;
{
	register struct cacheq_entry	*q;
	unsigned long			h;

	h = __nis_hash(dir->do_name, strlen(dir->do_name));
	for (q = qhead.cqh_first; q != (void *)&qhead; q = q->links.cqe_next) {
		if (q->hashval == h) {
			xdr_free(xdr_directory_obj, (char *)q->dptr);
			bzero((char *)q->dptr, sizeof(directory_obj));
			bcopy((char *)dir, (char *)q->dptr,
				sizeof(directory_obj));
			q->ctime = time(NULL);
			nis_cache_pkeys(dir);
			return(0);
		}
	}

	return(1);
}

int nis_cache_add(dir)
	directory_obj *dir;
{
	register struct cacheq_entry	*q;
	char				*ptr;
	unsigned long			size = 0;
	XDR				xdrs;

	q = (struct cacheq_entry *)calloc(1, sizeof(struct cacheq_entry));
	q->dptr = (directory_obj *)calloc(1, sizeof(directory_obj));
	q->hashval = __nis_hash(dir->do_name, strlen(dir->do_name));
	q->ctime = time(NULL);

	size = xdr_sizeof(xdr_directory_obj, (char *)dir);
	ptr = calloc(1, size);
	xdrmem_create(&xdrs, ptr, size, XDR_ENCODE);
	xdr_directory_obj(&xdrs, dir);
	xdr_destroy(&xdrs);
	xdrmem_create(&xdrs, ptr, size, XDR_DECODE);
	xdr_directory_obj(&xdrs, q->dptr);
	xdr_destroy(&xdrs);
	free(ptr);

	nis_cache_pkeys(dir);
	CIRCLEQ_INSERT_TAIL(&qhead, q, links);

	return(0);
}


int nis_cache_remove(name)
	nis_name			name;
{
	register struct cacheq_entry	*q;
	unsigned long			h;

	h = __nis_hash(name, strlen(name));

	for (q = qhead.cqh_first; q != (void *)&qhead; q = q->links.cqe_next) {
		if (q->hashval == h) {
			nis_remove_pkeys(q->dptr);
			CIRCLEQ_REMOVE(&qhead, q, links);
			xdr_free(xdr_directory_obj, (char *)q->dptr);
			free(q);
			return(0);
		}
	}

	return(0);
}
