/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nisinit.c,v 1.14 1997/07/26 21:15:56 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include <rpc/pmap_clnt.h>

#ifndef lint
static const char rcsid[] = "@(#) $Id: nisinit.c,v 1.14 1997/07/26 21:15:56 wpaul Exp wpaul $ (Project Flatbush)";
static const char copyright[] = "@(#) Copyright (c) 1996, 1997\
	Bill Paul. All rights reserved.";
#endif

#ifndef _PATH_NIS
#define _PATH_NIS "/var/nis"
#endif

#ifndef _PATH_NIS_ROOT_OBJ
#define _PATH_NIS_ROOT_OBJ "/var/nis/data/root.object"
#endif

#ifndef _PATH_NIS_SRVLIST
#define _PATH_NIS_SRVLIST "/var/nis/data/serving_list"
#endif

static void usage(pname)
	char			*pname;
{
	fprintf(stderr, "usage: %s -c -H host|-B|-C coldstart\n", pname);
	fprintf(stderr, "       %s -r\n", pname);
	fprintf(stderr, "       %s -p Y|D|N parent_domain host ...\n", pname);
	exit(0);
}

static int init_serv()
{
	nis_object		root_obj;
	directory_obj		dir_obj;
	endpoint		*ep;
	nis_server		server;
	XDR			xdrs;
	FILE			*fp;
	char			hname[NIS_MAXNAMELEN], haddr[NIS_MAXNAMELEN];
	struct hostent		*h;
	int			i = 0;
	char			*dname, *dotptr;

	bzero((char *)&dir_obj, sizeof(directory_obj));
	bzero((char *)&root_obj, sizeof(root_obj));

	gethostname(hname, NIS_MAXNAMELEN);
	dname = nis_local_directory();
	if (dname == NULL || !strlen(dname)) {
		warnx("NIS domain name not set");
		return(-1);
	}

	h = gethostbyname(hname);
	if (h == NULL)
		errx(1, "gethostbyname() failed");

	strcpy(hname, nis_local_host());

	ep = malloc(sizeof(endpoint));
	i = 0;

	/*
	 * We're using RPC 4.0, which only understands two
	 * transports: TCP and UDP. TI-RPC also has loopback,
	 * but that's of no use to us.
	 */
	
	while(*h->h_addr_list != NULL) {
		struct in_addr		*in;
		endpoint		*e;
		in = *(struct in_addr **)h->h_addr_list;
		e = ep = realloc(ep, sizeof(endpoint) * (i + 3));
		e += i;
		snprintf(haddr, sizeof(haddr), "%s.0.111", inet_ntoa(*in));
		e->uaddr = strdup(haddr);
		e->family = "inet";
		e->proto = "tcp";		/* For TCP. */
		e++;
		e->uaddr = strdup(haddr);
		e->family = "inet";
		e->proto = "udp";		/* For UDP. */
		e++;
		e->uaddr = strdup(haddr);
		e->family = "inet";
		e->proto = "-";		/* For broadcast (?) */
		i += 3;
		h->h_addr_list++;
	}

	server.name = strdup(hname);
	server.ep.ep_len = i;
	server.ep.ep_val = ep;

	/*
	 * We can't initialize the credentials for the root server
	 * from nisinit(8) because this may be the first time we're
	 * initializing NIS anywhere, so there may not be any credentials
	 * to give it. If we want to assign it specific credentials,
	 * we have to do it with the usual NIS utilities once the server
	 * is started.
	 */
	server.key_type = NIS_PK_NONE;
	server.pkey.n_bytes = NULL;
	server.pkey.n_len = 0;

	dir_obj.do_name = strdup(dname);
	dir_obj.do_type = NIS;
	dir_obj.do_servers.do_servers_len = 1;
	dir_obj.do_servers.do_servers_val = &server;
	dir_obj.do_ttl = 60 * 60 * 12;

	nis_put_cold_start(&dir_obj);

	root_obj.zo_owner = strdup(hname);

	dotptr = strchr(dname, '.');
	if (dotptr != NULL)
		*dotptr = '\0';
	dotptr++;
	
	root_obj.zo_name = strdup(dname);
	root_obj.zo_domain = strdup(dotptr);
	root_obj.zo_group = "";
	root_obj.zo_access = DEFAULT_RIGHTS | (NIS_READ_ACC << 24);
	root_obj.zo_ttl = 60 * 60 * 24;
	root_obj.zo_data.zo_type = DIRECTORY_OBJ;
	root_obj.zo_oid.ctime = root_obj.zo_oid.ctime = time(NULL);
	bcopy((char *)&dir_obj, (char *)&root_obj.DI_data,
			 sizeof(directory_obj));

	mkdir(_PATH_NIS_DATA, 0750);
	if ((fp = fopen(_PATH_NIS_ROOT_OBJ, "w")) == NULL)
		err(1, "couldn't open %s for writing", _PATH_NIS_ROOT_OBJ);

	xdrstdio_create(&xdrs, fp, XDR_ENCODE);
	if (!xdr_nis_object(&xdrs, &root_obj))
		err(1, "creating root.object failed");

	xdr_destroy(&xdrs);
	fclose(fp);

	if ((fp = fopen(_PATH_NIS_SRVLIST, "w+")) == NULL) {
		warnx("%s: %s", _PATH_NIS_SRVLIST, strerror(errno));
		return(-1);
	}

	fprintf(fp, "%s.%s\n", dname, dotptr);

	fclose(fp);

	return(0);
}

static fd_result		fdr_bcast;

static bool_t eachresult(out, addr)
	fd_result		*out;
	struct sockaddr_in	*addr;
{
	bzero((char *)&fdr_bcast, sizeof(fdr_bcast));

	/*
	 * Keep trying until we find a server that knows about the
	 * directory we're asking for.
	 */
	if (out->status != NIS_SUCCESS)
		return(FALSE);

	fdr_bcast.source = strdup(out->source);
	fdr_bcast.status = out->status;
	fdr_bcast.dir_data.dir_data_len = out->dir_data.dir_data_len;
	fdr_bcast.dir_data.dir_data_val = malloc(out->dir_data.dir_data_len);
	bcopy(out->dir_data.dir_data_val,
	      fdr_bcast.dir_data.dir_data_val,
	      fdr_bcast.dir_data.dir_data_len);

	if (out->signature.signature_len) {
		fdr_bcast.signature.signature_len =
					out->signature.signature_len;
		fdr_bcast.signature.signature_val =
					malloc(out->signature.signature_len);
		bcopy(out->signature.signature_val,
		      fdr_bcast.signature.signature_val,
		      fdr_bcast.signature.signature_len);
	}

	return(TRUE);
}

static nis_error get_cold_from_bcast(cold)
	directory_obj		**cold;
{
	fd_args			fda;
	fd_result		fdr;
	enum clnt_stat		stat;
	static directory_obj	dir;
	XDR			xdrs;

	bzero((char *)&fdr, sizeof(fdr));
	bzero((char *)&dir, sizeof(dir));
	fda.requester = "broadcast";
	fda.dir_name = nis_local_directory();
	if (fda.dir_name == NULL || !strlen(fda.dir_name)) {
		warnx("NIS domain name not set");
		return(-1);
	}

	stat = clnt_broadcast(NIS_PROG, NIS_VERSION, NIS_FINDDIRECTORY,
			(xdrproc_t)xdr_fd_args, (char *)&fda,
			(xdrproc_t)xdr_fd_result, (char *)&fdr,
			(resultproc_t)eachresult);

	if (stat != RPC_SUCCESS)
		return(NIS_NAMEUNREACHABLE);

	if (fdr_bcast.status != NIS_SUCCESS) {
		warnx("%s: %s", nis_local_directory(),
			 nis_sperrno(fdr_bcast.status));
		return(fdr_bcast.status);
	}

	xdrmem_create(&xdrs, fdr_bcast.dir_data.dir_data_val,
			     fdr_bcast.dir_data.dir_data_len, XDR_DECODE);
	xdr_directory_obj(&xdrs, &dir);
	xdr_destroy(&xdrs);

	*(caddr_t *)cold = (caddr_t)&dir;

	return(NIS_SUCCESS);
}

static nis_error get_cold_from_host(cold, host)
	directory_obj		**cold;
	char			*host;
{
	fd_args			fda;
	fd_result		*fdr;
	static directory_obj	dir;
	XDR			xdrs;
	CLIENT			*clnt;
	struct rpc_err		rpcerr;

	bzero((char *)&fdr, sizeof(fdr));
	bzero((char *)&dir, sizeof(dir));
	fda.requester = "broadcast";
	fda.dir_name = nis_local_directory();
	if (fda.dir_name == NULL || !strlen(fda.dir_name)) {
		warnx("NIS domain name not set");
		return(-1);
	}


	clnt = clnt_create(host, NIS_PROG, NIS_VERSION, "tcp");

	if (clnt == NULL) {
		warnx("%s", clnt_spcreateerror(host));
		return(NIS_NAMEUNREACHABLE);
	}

	clnt->cl_auth = authunix_create_default();

	fdr = nis_finddirectory_3(&fda, clnt);

	auth_destroy(clnt->cl_auth);

	if (fdr == NULL) {
		clnt_geterr(clnt, &rpcerr);
		warnx("%s: %s", nis_local_directory(),
				clnt_sperrno(rpcerr.re_status));
		clnt_destroy(clnt);
		return(NIS_RPCERROR);
	}

	clnt_destroy(clnt);

	if (fdr->status != NIS_SUCCESS) {
		warnx("%s: %s", nis_local_directory(),
			 nis_sperrno(fdr->status));
		return(fdr->status);
	}

	xdrmem_create(&xdrs, fdr->dir_data.dir_data_val,
			     fdr->dir_data.dir_data_len, XDR_DECODE);
	xdr_directory_obj(&xdrs, &dir);
	xdr_destroy(&xdrs);

	*(caddr_t *)cold = (caddr_t)&dir;

	return(NIS_SUCCESS);
}

static nis_error get_cold_from_file(cold, file)
	directory_obj		**cold;
	char			*file;
{
	FILE			*fp;
	static directory_obj	dir;
	XDR			xdrs;

	bzero((char *)&dir, sizeof(dir));

	if ((fp = fopen(file, "r")) == NULL) {
		warnx("%s: %s", file, strerror(errno));
		return(-1);
	}

	xdrstdio_create(&xdrs, fp, XDR_DECODE);
	xdr_directory_obj(&xdrs, &dir);
	xdr_destroy(&xdrs);

	fclose(fp);

	*(caddr_t *)cold = (caddr_t)&dir;

	return(NIS_SUCCESS);
}

static int init_client(cstart_file, cstart_host, cstart_bcast)
	char			*cstart_file;
	char			*cstart_host;
	int			cstart_bcast;
{
	directory_obj		*cold;

	if (cstart_bcast) {
		if (get_cold_from_bcast(&cold) != NIS_SUCCESS) {
			warnx("Error: No servers responding, use -H option");
			return(-1);
		}
	}

	if (cstart_host) {
		if (get_cold_from_host(&cold, cstart_host) != NIS_SUCCESS) {
			warnx("Error: Host not responding, use -C option");
			return(-1);
		}
	}

	if (cstart_file) {
		if (get_cold_from_file(&cold, cstart_file) != NIS_SUCCESS) {
			warnx("Error: unable to read from file");
			return(-1);
		}
	}

	nis_put_cold_start(cold);

	return(0);
}

int main(argc, argv)
	int			argc;
	char			*argv[];
{
	int			ch;
	int			root_serv = 0;
	int			client = 0, cstart_bcast = 0;
	char			*cstart_host = NULL, *cstart_file = NULL;
	int			stat = -1;
#ifdef foo
	if (geteuid())
		errx(1, "This program must be run as superuser.");
#endif
	while((ch = getopt(argc, argv, "C:H:Bcr")) != -1) {
		switch(ch) {
		case 'r':
			root_serv++;
			break;
		case 'c':
			client++;
			break;
		case 'C':
			cstart_file = optarg;
			break;
		case 'H':
			cstart_host = optarg;
			break;
		case 'B':
			cstart_bcast++;
			break;
		default:
			usage(argv[0]);
			break;
		}
	}

	printf("This machine is in the %s NIS+ domain\n",
				 nis_local_directory());

	if (root_serv) {
		mkdir(_PATH_NIS, 0755);
		stat = init_serv();
	}
	if (client) {
		printf("Setting up NIS+ client ...\n");
		mkdir(_PATH_NIS, 0755);
		stat = init_client(cstart_file, cstart_host, cstart_bcast);
	}

	if (stat == 0)
		printf("All done.\n");
	exit(0);
}
