/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: nislog.c,v 1.17 1998/02/09 05:50:10 wpaul Exp $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <err.h>
#include <errno.h>
#include <rpcsvc/nis.h>
#include "nis_log.h"

#ifndef lint
static const char copyright[] = "@(#) Copyright (c) 1996, 1997, 1998\
	Bill Paul. All rights reserved.";
static const char rcsid[] =
	"@(#) $Id: nislog.c,v 1.17 1998/02/09 05:50:10 wpaul Exp $ (Project Flatbush)";
#endif

#ifndef _PATH_NIS_LOG
#define _PATH_NIS_LOG "/var/nis/trans.log"
#endif

static int			log_fd;
static struct log_header	*hptr;
static struct log_spacer	*sptr;
static caddr_t			laddr;
static off_t			lsize;

static void nis_position_log(count)
	int			count;
{
	int			cnt = count;
	struct log_spacer	*s;
	caddr_t			ptr;

	ptr = laddr + (hptr->size - sizeof(struct log_spacer));
	s = (struct log_spacer *)ptr;

	while(cnt) {
		if (s->magic != SPACER_MAGIC) {
			warnx("log is corrupt!");
			return;
		}
		if (!s->prev_offset)
			break;
		ptr = (caddr_t)s;
		ptr -= s->prev_offset + sizeof(struct log_spacer);
		s = (struct log_spacer *)ptr;
		cnt--;
	}

	sptr = s;

	return;
}

static void nis_dump_log(count)
	int			count;
{
	log_entry		ent;
	XDR			xdrs;
	caddr_t			addr;
	uint32_t		update = 0;
	int			i;
	time_t			tmp;
	char			*letype[] = { "NOP", "ADD Name", "REMOVE Name",
					      "MODIFY (Original Value)",
					      "MODIFY (New Value)",
					      "ADD Entry", "REMOVE Entry",
					      "MODIFY Entry",
					      "UPDATE time stamp" };
	while (sptr->next_offset) {
		caddr_t			ptr;
		if (sptr->magic != SPACER_MAGIC) {
			warnx("log is corrupted!");
			return;
		}
		bzero((char *)&ent, sizeof(ent));
		addr = (caddr_t)sptr + sizeof(struct log_spacer);
		xdrmem_create(&xdrs, addr, sptr->next_offset, XDR_DECODE);
		xdr_log_entry(&xdrs, &ent);
		xdr_destroy(&xdrs);

		printf("@@@@@@@@@@@@@@@@ Transaction @@@@@@@@@@@@@@@@@@\n");
		printf("#%05u, XID : %u\n", update, sptr->xid);
		tmp = ent.le_time;
		printf("Time        : %s", ctime(&tmp));
		printf("Directory   : %s\n", nis_domain_of(ent.le_name));
		printf("Entry type  : %s\n", letype[ent.le_type]);
		printf("Enrty timestamp : %s", ctime(&tmp));
		printf("Principal       : %s\n", ent.le_princp);
		printf("Object name     : ");
		if (ent.le_attrs.le_attrs_len) {
			printf("[ ");
			for (i = 0; i < ent.le_attrs.le_attrs_len; i++) {
				nis_attr		*p;
				p = &ent.le_attrs.le_attrs_val[i];
				printf("%s = %.*s", p->zattr_ndx,
					p->zattr_val.zattr_val_len,
					p->zattr_val.zattr_val_val);
				if (i != (ent.le_attrs.le_attrs_len - 1))
					printf(", ");
			}
			printf(" ],");
		}
		printf("%s\n", ent.le_name);
		printf(".................. Object .....................\n");
		nis_print_object(&ent.le_object);
		printf("...............................................\n");
		xdr_free((xdrproc_t)xdr_log_entry, (char *)&ent);
		if ((update + 1) == count)
			break;
		ptr = (caddr_t)sptr;
		ptr += sptr->next_offset + sizeof(struct log_spacer);
		sptr = (struct log_spacer *)ptr;
		update++;
	}

	return;
}

static void usage(pname)
	char			*pname;
{
	fprintf(stderr, "usage: %s\n", pname);
	fprintf(stderr, "       %s -h [number]\n", pname);
	fprintf(stderr, "       %s -t [number]\n", pname);

	exit(1);
}
		
int main(argc, argv)
	int			argc;
	char			*argv[];
{
	struct stat		st;
	int			ch;
	int			head = 0, tail = 0;
	int			count = 0;
	char			*pname;

	while((ch = getopt(argc, argv, "ht")) != -1) {
		switch(ch) {
		case 'h':
			if (tail) {
				warnx("specify either -h or -t");
				usage(argv[0]);
			}
			head++;
			if (argv[optind] != NULL &&
			    argv[optind][0] != '-' &&
			    isdigit(argv[optind][0])) {
				count = atoi(argv[optind]);
				optind++;
			}
			if (!count)
				count = 1;
			break;
		case 't':
			if (head) {
				warnx("specify either -h or -t");
				usage(argv[0]);
			}
			tail++;
			if (argv[optind] != NULL &&
			    argv[optind][0] != '-' &&
			    isdigit(argv[optind][0])) {
				count = atoi(argv[optind]);
				optind++;
			}
			if (!count)
				count = 1;
			break;
		default:
			usage(argv[0]);
			break;
		}
	}

	pname = argv[0];
	argc -= optind;
	argv += optind;

	if (argc)
		usage(pname);

	if (stat(_PATH_NIS_LOG, &st) == -1) {
			err(1, "couldn't open log");
	} else {
		if ((log_fd = open(_PATH_NIS_LOG, O_RDONLY,0600)) == -1) {
			err(1, "couldn't open log");
		}
		lsize = st.st_size;
	}

	laddr = mmap(0, lsize, PROT_READ, MAP_PRIVATE, log_fd, 0);
	if ((int)laddr == -1) {
		warn("mmap failed");
		close(log_fd);
		exit(1);
	}

	hptr = (struct log_header *)laddr;
	sptr = (struct log_spacer *)(laddr + hptr->offset);
	if (hptr->magic != HEADER_MAGIC ||
		sptr->magic != SPACER_MAGIC) {
		warnx("transaction log corrupt! aborting!");
		munmap(laddr, lsize);
		close(log_fd);
		exit(1);
	}

	printf("NIS Log printing facility\n");
	printf("NIS Log dump :\n");
	printf("	Log state : ");
	switch(hptr->state) {
	case LOG_STABLE:
		printf("STABLE\n");
		break;
	case LOG_UPDATING:
		printf("UPDATING\n");
		break;
	case LOG_SYNCING:
		printf("SYNCING\n");
		break;
	default:
		printf("unknown\n");
		break;
	}
	printf("Number of updates    : %u\n", hptr->count);
	printf("Current XID          : %u\n", hptr->xid);
	printf("Size of Log in bytes : %ld\n", hptr->size);

	printf("*** UPDATES ***\n");

	if (tail)
		nis_position_log(count);

	nis_dump_log(count);

	munmap(laddr, lsize);
	close(log_fd);
	
	exit(0);
}
