/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nisping.c,v 1.13 1998/02/01 03:56:02 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>

#ifndef lint
static const char copyright[] = "@(#) Copyright (c) 1996, 1997\
	Bill Paul. All rights reserved.";
static const char rcsid[] =
	"@(#) $Id: nisping.c,v 1.13 1998/02/01 03:56:02 wpaul Exp wpaul $ (Project Flatbush)";
#endif

void usage(pname)
	char			*pname;
{
	fprintf(stderr, "%s: [-uf] [-H hostname] [-r|directory]\n", pname);
	fprintf(stderr, "%s: -C [-a] [-H hostname] [directory]\n", pname);
	exit(-1);
}

/*
 * Get the update time from the master.
 */
static unsigned long *find_master_utime(d, master)
	directory_obj		*d;
	nis_server		*master;
{
	struct _nis_binding	*bptr = NULL;
	struct rpc_err		rerr;
	unsigned long		*utime;
	nis_error		r;

	printf("Master server is %s\n", master->name);

	r = __nis_create_binding(NULL, master, NO_AUTHINFO, &bptr);

	if (r != NIS_SUCCESS)
		return(NULL);

	__nis_connect_server(bptr);

	utime = nis_cptime_3(&d->do_name, bptr->nb_clnt);

	if (utime == NULL) {
		clnt_geterr(bptr->nb_clnt, &rerr);
		warnx("%s: %s", d->do_name, clnt_sperrno(rerr.re_status));
		__nis_destroy_binding(bptr);
		return(NULL);
	}

	__nis_destroy_binding(bptr);

	if (*utime == 0)
		printf("\tNo last update time available for directory %s.\n",
								d->do_name);
	else
		printf("\tLast update occurred at: %s\n", ctime(utime));

	return(utime);
}

static int send_cp(srv, dir, all)
	nis_server		*srv;
	nis_name		dir;
	int			all;
{
	struct _nis_binding	*bptr = NULL;
	struct rpc_err		rerr;
	char			*allstr = "";
	nis_error		r;

	r = __nis_create_binding(NULL, srv, NO_AUTHINFO, &bptr);

	if (r != NIS_SUCCESS) {
		printf("Unavailable.\n");
		return(1);
	}

	if (__nis_connect_server(bptr) != NIS_SUCCESS) {
		printf("Unavailable.\n");
		__nis_destroy_binding(bptr);
		return(1);
	}

	if (nis_checkpoint_3(all ? &allstr : &dir, bptr->nb_clnt) == NULL) {
		clnt_geterr(bptr->nb_clnt, &rerr);
		warnx("RPC error on server %s, %s", srv->name,
					clnt_sperrno(rerr.re_status));
		__nis_destroy_binding(bptr);
		return(1);
	}

	printf("checkpoint scheduled on %s\n", srv->name);
	__nis_destroy_binding(bptr);

	return(0);
}

int do_checkpoint(obj, host, all)
	nis_object		*obj;
	nis_name		host;
	int			all;
{
	char			hname[NIS_MAXNAMELEN];
	directory_obj		*d = &obj->DI_data;
	nis_server		*master = &d->do_servers.do_servers_val[0];
	nis_server		*srv = NULL;
	unsigned long		i;

	if (host != NULL) {
		if (host[strlen(host) - 1] == '.')
			snprintf(hname, sizeof(hname), "%s", host);
		else
			snprintf(hname, sizeof(hname), "%s.%s", host,
								d->do_name);

		for (i = 0; i < d->do_servers.do_servers_len; i++) {
			if (nis_dir_cmp(d->do_servers.do_servers_val[i].name,
						 hname) == SAME_NAME) {
				srv = &d->do_servers.do_servers_val[i];
				break;
			}
		}
		if (srv == NULL) {
			warnx("Host %s does not serve directory %s.",
						hname, d->do_name);
			return(-1);
		}
	}

	printf("Checkpointing replicas serving directory %s :\n", d->do_name);

	find_master_utime(d, master);

	for (i = 0; i < d->do_servers.do_servers_len; i++) {
		srv = &d->do_servers.do_servers_val[i];
		printf("%s server is %s\n", i ? "Replica" : "Master",
							 srv->name);
		send_cp(srv, d->do_name, all);
	}
	return(0);
}

static int send_ping(srv, dir, p_args, force, show_only)
	nis_server		*srv;
	nis_name		dir;
	ping_args		*p_args;
	int			force;
	int			show_only;
{
	struct _nis_binding	*bptr = NULL;
	struct timeval		tv = { 0, 0 };
	unsigned long		*rep_time;
	struct rpc_err		rerr;
	nis_error		r;

	printf("Replica server is %s\n", srv->name);
	r = __nis_create_binding(NULL, srv, NO_AUTHINFO, &bptr);

	if (r != NIS_SUCCESS) {
		printf("\tUnavailable.\n");
		return(1);
	}
	if (__nis_connect_server(bptr) != NIS_SUCCESS) {
		printf("\tUnavailable.\n");
		__nis_destroy_binding(bptr);
		return(1);
	}
	rep_time = nis_cptime_3(&dir, bptr->nb_clnt);
	if (rep_time == NULL) {
		clnt_geterr(bptr->nb_clnt, &rerr);
		printf("\tnis_cptime failed: %s\n", clnt_sperrno(rerr.re_status));
		__nis_destroy_binding(bptr);
		return(1);
	}
	if (rep_time)
		printf("\tLast update seen was: %s\n", ctime(rep_time));
	else
		printf("\tNo last update available for directory %s.\n", dir);
	if ((*rep_time < p_args->stamp || force) && !show_only) {
		printf("\tPinging ... %s\n", srv->name);
		clnt_control(bptr->nb_clnt, CLSET_TIMEOUT, &tv);
		nis_ping_3(p_args, bptr->nb_clnt);
	}
	__nis_destroy_binding(bptr);

	return(0);
}

int do_ping(obj, host, force, show_only)
	nis_object		*obj;
	nis_name		host;
	int			force;
	int			show_only;
{
	unsigned long		*utime = NULL;
	unsigned long		i;
	char			hname[NIS_MAXNAMELEN];
	directory_obj		*d = &obj->DI_data;
	nis_server		*master = &d->do_servers.do_servers_val[0];
	nis_server		*srv = NULL;
	ping_args		p_args;

	if (d->do_servers.do_servers_len == 1 && !show_only) {
		warnx("%s: no replicas.", d->do_name);
		return(-1);
	}

	if (host != NULL) {
		if (host[strlen(host) - 1] == '.')
			snprintf(hname, sizeof(hname), "%s", host);
		else
			snprintf(hname, sizeof(hname), "%s.%s", host,
								d->do_name);

		for (i = 0; i < d->do_servers.do_servers_len; i++) {
			if (nis_dir_cmp(d->do_servers.do_servers_val[i].name,
						 hname) == SAME_NAME) {
				srv = &d->do_servers.do_servers_val[i];
				break;
			}
		}
		if (srv == NULL) {
			warnx("Host %s does not serve directory %s.",
						hname, d->do_name);
			return(-1);
		}
		if (i == 0) {
			warnx("Host %s is the master server for %s.",
						hname, d->do_name);
			return(-1);
		}
	}

	if (show_only)
		printf("Last updates for directory %s:\n", d->do_name);
	else
		printf("Pinging replicas serving firectory %s :\n", d->do_name);
	utime = find_master_utime(d, master);

	p_args.dir = d->do_name;
	p_args.stamp = *utime;

	/*
	 * If there's just the one user-specified host, handle it and
	 * get out quick.
	 */
	if (host != NULL)
		return(send_ping(srv, d->do_name, &p_args, force, show_only));

	for (i = 1; i < d->do_servers.do_servers_len; i++) {
		srv = &d->do_servers.do_servers_val[i];
		send_ping(srv, d->do_name, &p_args, force, show_only);
	}

	return(0);
}

int main(argc, argv)
	int			argc;
	char			*argv[];
{
	int			ch;
	nis_result		*res;
	nis_name		host = NULL;
	nis_name		dir;
	int			force = 0, show_only = 0, checkpoint = 0;
	int			ping_root = 0, all = 0;
	int			exitval;

	while ((ch = getopt(argc, argv, "aufrCH:")) != -1) {
		switch(ch) {
		case 'a':
			all++;
			break;
		case 'f':
			force++;
			break;
		case 'u':
			show_only++;
			break;
		case 'C':
			checkpoint++;
			break;
		case 'H':
			host = optarg;
			break;
		case 'r':
			ping_root++;
			break;
		default:
			usage(argv[0]);
			break;
		}
	}	

	argc -= optind;
	argv += optind;

	if (argc)
		dir = argv[0];
	else
		dir = nis_local_directory();

	res = nis_lookup(dir, MASTER_ONLY|EXPAND_NAME);

	if (res == NULL)
		err(-1, "nis_lookup failed");

	if (res->status != NIS_SUCCESS) {
		warnx("%s: %s", dir, nis_sperrno(res->status));
		nis_freeresult(res);
		exit(-1);
	}

	if (__type_of(NIS_RES_OBJECT(res)) != DIRECTORY_OBJ) {
		warnx("%s is not a directory.", dir);
		nis_freeresult(res);
		exit(-1);
	}

	if (checkpoint)
		exitval = do_checkpoint(NIS_RES_OBJECT(res), host, all);
	else
		exitval = do_ping(NIS_RES_OBJECT(res), host, force, show_only);

	nis_freeresult(res);

	exit(exitval);
}
