/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nisshowcache.c,v 1.1 1998/06/01 03:35:21 wpaul Exp wpaul $
 */

#include <sys/cdefs.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <err.h>
#include <errno.h>
#include <signal.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nis_cache_bind.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nisshowcache.c,v 1.1 1998/06/01 03:35:21 wpaul Exp wpaul $";
#endif

#ifndef _PATH_NIS_SHARED_DIRCACHE
#define _PATH_NIS_SHARED_DIRCACHE "/var/nis/NIS_SHARED_DIRCACHE"
#endif

static cachebind_header		*header;
static cachebind_hash_entry_p	*table;
static cachebind_combined	*combined;
static int			cachebind_fd;
static caddr_t			laddr;
static unsigned long		csize;

static void nis_cachebind_alarm(sig)
	int		sig;
{
	if (sig != SIGALRM)
		warnx("caught signal type %d by mistake", sig);

	return;
}

/*
 * Aquire exclusive lock on cache file.
 */
static void nis_cachebind_lock()
{
	struct flock		lock;

	lock.l_start = 0;
	lock.l_len = 0;
	lock.l_pid = getpid();
	lock.l_type = F_RDLCK;
	lock.l_whence = SEEK_SET;

	alarm(5);
	if (fcntl(cachebind_fd, F_SETLK, &lock) == -1)
		warnx("help! flock failed: %s", strerror(errno));
	alarm(0);

	return;
}

/*
 * Give up exclusive lock on cache file.
 */
static void nis_cachebind_unlock()
{
	struct flock		lock;

	lock.l_start = 0;
	lock.l_len = 0;
	lock.l_pid = getpid();
	lock.l_type = F_UNLCK;
	lock.l_whence = SEEK_SET;

	if (fcntl(cachebind_fd, F_SETLK, &lock) == -1)
		warnx("unlocking failed: %s", strerror(errno));

	return;
}

static nis_error nis_cachebind_load(fname)
	char			*fname;
{
	XDR			xdrs;
	struct stat		st;

	if (stat(fname, &st) == -1) {
		if (errno != ENOENT)
			return(NIS_SYSTEMERROR);
	}

	csize = st.st_size;
	cachebind_fd = open(fname, O_RDONLY, 0644);
	if (cachebind_fd == -1)
		return(NIS_SYSTEMERROR);

	nis_cachebind_lock();
	laddr = mmap(0, csize, PROT_READ,
			MAP_SHARED, cachebind_fd, 0);
	nis_cachebind_unlock();

	if (laddr == MAP_FAILED) {
		close(cachebind_fd);
		return(NIS_SYSTEMERROR);
	}

	combined = (cachebind_combined *)calloc(1,sizeof(cachebind_combined));
	xdrmem_create(&xdrs, laddr, csize, XDR_DECODE);
	if (!xdr_cachebind_combined(&xdrs, combined)) {
		xdr_destroy(&xdrs);
		munmap(laddr, csize);
		laddr = NULL;
		close(cachebind_fd);
		return(NIS_SYSTEMERROR);
	}
	xdr_destroy(&xdrs);
	munmap(laddr, csize);
	laddr = NULL;

	table = combined->cachebind_hash_table;
	header = &combined->cachebind_header;

	return(NIS_SUCCESS);
}

nis_error nis_cachebind_open(fname)
	char			*fname;
{
	signal(SIGALRM, nis_cachebind_alarm);

	return(nis_cachebind_load(fname));
}

nis_error nis_cachebind_close()
{

	close(cachebind_fd);
	xdr_free(xdr_cachebind_combined, (char *)combined);
	free(combined);

	return(NIS_SUCCESS);
}

static void nis_cachebind_print(c)
	cachebind_hash_entry	*c;
{
	directory_obj		*d;

	if (c->cb_type == CB_SERVER)
		return;

	d = &c->bdr_ptr->dobj;
	nis_print_directory(d);

	return;
}

static void nis_cachebind_enum()
{
	u_int32_t		i;
	cachebind_hash_entry	*cur;

	for (i = 0; i < TABLESIZE; i++) {
		cur = (cachebind_hash_entry *)table[i];
		while(cur != NULL) {
			nis_cachebind_print(cur);
			cur = cur->next;
		}
	}

	return;
}

int main(argc, argv)
	int			argc;
	char			*argv[];
{
	if (nis_cachebind_open(_PATH_NIS_SHARED_DIRCACHE) != NIS_SUCCESS)
		exit(1);
	nis_cachebind_enum();
	nis_cachebind_close();

	exit(0);
}
