/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nisupdkeys.c,v 1.4 1997/09/02 23:39:59 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <err.h>
#include <sys/cdefs.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include <rpc/key_prot.h>

#ifndef lint
static const char copyright[] = "@(#) Copyright (c) 1996, 1997\
	Bill Paul. All rights reserved.";
static const char rcsid[] =
	"@(#) $Id: nisupdkeys.c,v 1.4 1997/09/02 23:39:59 wpaul Exp wpaul $";
#endif

static void usage(pname)
	char			*pname;
{
	fprintf(stderr, "usage: %s [ -a | -C ] [ -H host ] [ directory ]\n",
								pname);
	fprintf(stderr, "       %s -s [ -a | -C ] -H host\n", pname);
	exit(1);
}


static char *getpkey_nisplus(sname)
	char			*sname;
{
	char			buf[NIS_MAXNAMELEN];
	char			pkey[HEXKEYBYTES + 1];
	char			*p = NULL;
	nis_result		*res;
	int			len = 0;

	p = strchr(sname, '.');
	if (p == NULL)
		return(NULL);

	p++;

	snprintf(buf, sizeof(buf), "[cname=%s,auth_type=DES],cred.org_dir.%s",
		sname, p);

	res = nis_list(buf,
		MASTER_ONLY|USE_DGRAM|NO_AUTHINFO|FOLLOW_LINKS|FOLLOW_PATH,
		NULL, NULL);

	if (res == NULL)
		errx(1, "out of memory doing nis_list");

	if (res->status != NIS_SUCCESS) {
		warnx("%s: %s", sname, nis_sperrno(res->status));
		nis_freeresult(res);
		return(NULL);
	}

	len = ENTRY_LEN(NIS_RES_OBJECT(res), 3);
	strncpy(pkey, ENTRY_VAL(NIS_RES_OBJECT(res), 3), len);
	pkey[len] = NUL;
	p = strchr(pkey, ':');
	if (p != NULL)
		*p = NUL;

	nis_freeresult(res);

	return(strdup(pkey));
}

static nis_server *makesrv(name)
	char			*name;
{
	struct hostent		*h;
	int			i;
	endpoint		*ep;
	nis_server		*server;
	char			netname[MAXNETNAMELEN];
	char			*pkey;
	char			haddr[NIS_MAXNAMELEN];
	char			hname[NIS_MAXNAMELEN];

	h = gethostbyname(name);

	if (h == NULL)
		err(1, "%s: gethostbyname failed", name);

	ep = malloc(sizeof(endpoint));
	i = 0;

	while(*h->h_addr_list != NULL) {
		struct in_addr		*in;
		endpoint		*e;
		in = *(struct in_addr **)h->h_addr_list;
		e = ep = realloc(ep, sizeof(endpoint) * (i + 3));
		e += 3;
		snprintf(haddr, sizeof(haddr), "%s.0.111", inet_ntoa(*in));
		e->uaddr = strdup(haddr);
		e->family = strdup("inet");
		e->proto = strdup("tcp");	/* For TCP. */
		e++;
		e->uaddr = strdup(haddr);
		e->family = strdup("inet");
		e->proto = strdup("udp");	/* For UDP. */
		e++;
		e->uaddr = strdup(haddr);
		e->family = strdup("inet");
		e->proto = strdup("-");		/* For broadcast (?) */
		i += 3;
		h->h_addr_list++;
	}

	server = calloc(1, sizeof(nis_server));

	snprintf(hname, sizeof(hname), "%s.", h->h_name);
	if (strchr(hname, '.') == NULL) {
		strncat(hname, nis_local_directory(), sizeof(hname));
	} else {
		*strchr(hname, '.') = NUL;
		strncat(hname, nis_local_directory(), sizeof(hname));
	}
	server->name = strdup(hname);
	server->ep.ep_len = i;
	server->ep.ep_val = ep;

	if (host2netname(netname, hname, NULL) == TRUE &&
		(pkey = getpkey_nisplus(server->name)) != NULL) {
		server->key_type = NIS_PK_DH;
		server->pkey.n_bytes = pkey;
		server->pkey.n_len = strlen(pkey) + 1;
	} else {
		server->key_type = NIS_PK_NONE;
		server->pkey.n_bytes = NULL;
		server->pkey.n_len = 0;
	}

	return(server);
}

#define ACTION_CLEAR	1
#define ACTION_NEWKEY	2
#define ACTION_NEWIP	3

static int update_dir(name, sname, action)
	nis_name		name;
	nis_name		sname;
	int			action;
{
	nis_result		*res;
	nis_object		*new;
	directory_obj		*d;
	int			i, changes = 0;
	nis_server		*s, *n;
	char			*p;

	res = nis_lookup(name, MASTER_ONLY|NO_AUTHINFO);

	if (res == NULL)
		errx(1, "out of memory doing nis_lookup");

	if (res->status != NIS_SUCCESS) {
		warnx("%s: %s", name, nis_sperrno(res->status));
		nis_freeresult(res);
	}

	if (__type_of(NIS_RES_OBJECT(res)) != DIRECTORY_OBJ) {
		warnx("%s is not a directory!", name);
		nis_freeresult(res);
	}

	new = nis_clone_object(NIS_RES_OBJECT(res), NULL);
	nis_freeresult(res);

	d = &new->DI_data;

	for (i = 0; i < d->do_servers.do_servers_len; i++) {
		s = &d->do_servers.do_servers_val[i];
		if (sname != NULL && nis_dir_cmp(s->name, sname) != SAME_NAME)
			continue;
		changes++;
		switch(action) {
		case ACTION_CLEAR:
			s->key_type = NIS_PK_NONE;
			free(s->pkey.n_bytes);
			s->pkey.n_bytes = NULL;
			s->pkey.n_len = 0;
			break;
		case ACTION_NEWKEY:
			if (sname != NULL)
				p = getpkey_nisplus(sname);
			else
				p = getpkey_nisplus(s->name);
			if (p == NULL) {
				warnx("failed to get public key for %s!",
								sname);
				break;
			}
			s->key_type = NIS_PK_DH;
			if (s->pkey.n_len)
				free(s->pkey.n_bytes);
			s->pkey.n_bytes = p;
			s->pkey.n_len = strlen(p) + 1;
			break;
		case ACTION_NEWIP:
			if (sname == NULL)
				n = makesrv(s->name);
			else
				n = makesrv(sname);
			if (n == NULL) {
				warnx("failed to get IP address for %s!",
								sname);
				break;
			}
			xdr_free(xdr_nis_server, (char *)s);
			bcopy((char *)n, (char *)s, sizeof(nis_server));
			free(n);
			break;
		default:
			break;
		}
	}

	if (!changes && sname != NULL) {
		warnx("%s does not serve %s!", sname, name);
		nis_destroy_object(new);
		exit(1);
	}

	res = nis_modify(name, new);

	if (res == NULL)
		errx(1, "out of memory doing nis_modify");

	if (res->status != NIS_SUCCESS) {
		warnx("failed to modify %s: %s", name,
				 nis_sperrno(res->status));
		nis_freeresult(res);
		nis_destroy_object(new);
		exit(1);
	}

	nis_freeresult(res);
	nis_destroy_object(new);

	return(0);
}

static int update_alldirs(sname, action)
	nis_name		sname;
	int			action;
{
	nis_tag			t = { TAG_DIRLIST, "directory list" };
	nis_tag			*r;
	nis_error		terr;
	char			*p;
	nis_server		*srv;

	srv = makesrv(sname);

	if (srv == NULL)
		errx(1, "couldn't convert %s to nis_server", sname);

	terr = nis_stats(srv, &t, 1, &r);

	if (terr != NIS_SUCCESS) {
		warnx("couldn't get dirlist from %s: %s", sname,
							nis_sperrno(terr));
		nis_freetags(r, 1);
		exit(1);
	}

	if (r->tag_type != TAG_DIRLIST || strchr(r->tag_val, '.') == NULL) {
		warnx("bad dirlist from %s: %s", sname, r->tag_val);
		nis_freetags(r, 1);
		exit(1);
	}

	while((p = strsep(&r->tag_val, " \n")) != NULL)
		update_dir(p, srv->name, action);

	nis_freetags(r, 1);
	return(0);
}

int main(argc, argv)
	int			argc;
	char			*argv[];
{
	int			ch, alldirs = 0;
	char			*hname = NULL;
	int			action = ACTION_NEWKEY;
	char			*pname = argv[0];

	while((ch = getopt(argc, argv, "aCH:s")) != -1) {
		switch(ch) {
		case 'a':
			action = ACTION_NEWIP;
			break;
		case 'C':
			action = ACTION_CLEAR;
			break;
		case 'H':
			hname = optarg;
			break;
		case 's':
			alldirs++;
			break;
		default:
			usage(argv[0]);
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if (alldirs) {
		if (hname == NULL)
			usage(pname);
		update_alldirs(hname, action);
	} else {
		nis_server		*s;

		if (hname == NULL)
			usage(pname);
		s = makesrv(hname);
		if (s == NULL)
			errx(1, "failed to convert %s to nis_server", hname);
		if (argc)
			update_dir(argv[0], s->name, action);
		else
			update_dir(nis_local_directory(), s->name, action);
		xdr_free(xdr_nis_server, (char *)s);
		free(s);
	}

	usage(pname);
	exit(0);
}
