/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: nis_access.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <sys/cdefs.h>
#include <sys/types.h>
#include <rpc/rpc.h>
#include <rpc/auth.h>
#include <rpc/auth_unix.h>
#ifdef USE_SECURE_RPC
#include <rpc/auth_des.h>
#include <rpc/key_prot.h>
#endif
#include <rpcsvc/nis.h>
#include <rpcsvc/nis_db.h>
#include <rpcsvc/nislib.h>
#include "nis_extern.h"

#ifndef lint
static const char rcsid[] = "@(#) $Id: nis_access.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $";
#endif

/*
 * Access control stuff. This module provides support for determining
 * NIS+ principal names based on RPC netnames and for checking whether
 * said principals have permission to perform given functions on NIS+
 * 'objects.'
 */

#define OPSYS "unix"

/*
 * We provide some caching support since principal lookups are liable to
 * occur quite frequently. Rather than make a big production, we keep only
 * a small hash table (256 entries) and don't bother accounting for
 * collisions. This means the number of entries in the hash table will
 * never be greater than TABLESIZE, and our memory usage won't go through
 * the roof.
 */
#define TABLESIZE 256
#define HASH_MASK 0x000000FF

/* Entry in the cred cache */
struct cred_entry {
	char			*key;
	char			*data;
	uint32_t		ttl;
	time_t			ctime;
};

static struct cred_entry *table[TABLESIZE];

void nis_init_cred_cache()
{
	bzero((char *)&table, sizeof(table));
	return;
}

static u_int32_t hashkey(key)
	char		*key;
{
	if (key == NULL)
		return (-1);
	return(__nis_hash((void *)key, strlen(key)) & HASH_MASK);
}

static char *lookup(table, key)
	struct cred_entry	*table[];
	char			*key;
{
	struct cred_entry	*cur;

	cur = table[hashkey(key)];

	if (cur == NULL)
		return(NULL);

	/* See if ttl expired */
	if (time(NULL) > (cur->ttl + cur->ctime))
		return(NULL);

	return(table[hashkey(key)]->data);
}

static void store(table, key, data, ttl)
	struct cred_entry	*table[];
	char			*key,
				*data;
	uint32_t		ttl;
{
	struct cred_entry	*new;
	u_int32_t		i;

	i = hashkey(key);

	if (table[i] == NULL)
		new = malloc(sizeof(struct cred_entry));
	else {
		new = table[i];
		free(new->key);
		free(new->data);
	}

	new->key = strdup(key);
	new->data = strdup(data);
	new->ttl = ttl;
	new->ctime = time(NULL);
	table[i] = new;

	return;
}

static const char* PKTABLE = "cred.org_dir";

/*
 * Probe the cred table for the principal name that relates
 * to a given RPC netname.
 */

static char *nis_netid_to_princ(netid, domain, flavor)
	char			*netid;
	char			*domain;
	enum_t			flavor;
{
	static char 		princ[NIS_MAXNAMELEN + 1];
	char			*p;
	char			*type;
	char			buf[NIS_MAXNAMELEN];
	nis_result		*res;
	int			len;

	if ((p = lookup(&table, netid)) != NULL)
		return(p);

	if (flavor == AUTH_DES)
		type = "DES";
	else if (flavor == AUTH_UNIX)
		type = "LOCAL";
	else
		return(NULL);
	bzero(buf, sizeof(buf));
	snprintf(buf, sizeof(buf), "[auth_name=\"%s\",auth_type=%s],%s.%s",
			netid, type, PKTABLE, domain);
	if (buf[strlen(buf) - 1] != '.')
		strcat(buf, ".");

	res = nis_list(buf, USE_DGRAM|NO_AUTHINFO|FOLLOW_LINKS|FOLLOW_PATH,
								NULL, NULL);
	if (res == NULL)
		return(NULL);

	if (res->status != NIS_SUCCESS && res->status != NIS_S_SUCCESS) {
		nis_freeresult(res);
		return(NULL);
	}

	if (res->objects.objects_len > 1) {
		nis_err("%s entry for %s in directory % not unique",
				type, netid, domain);
		nis_freeresult(res);
		return(NULL);
	}

	len = ENTRY_LEN(res->objects.objects_val, 0);
	strncpy(princ, ENTRY_VAL(res->objects.objects_val, 0), len);
	princ[len] = '\0';

	store(&table, netid, princ, NIS_RES_OBJECT(res)->zo_ttl);

	nis_freeresult(res);
	return(princ);
}

/*
 * Construct a principal name given RPC auth credentials.
 */

nis_name nis_principal(rqstp, _pbuf)
	struct svc_req		*rqstp;
	char			*_pbuf;
{
#ifdef USE_SECURE_RPC
	struct authdes_cred	*dcreds;
#endif
	struct authunix_parms	*ucreds;
	nis_name		principal = NULL;
	static char		netname[NIS_MAXNAMELEN + 1];
	static char		domain[NIS_MAXNAMELEN + 1];
	char			*ptr;

	switch(rqstp->rq_cred.oa_flavor) {
#ifdef USE_SECURE_RPC
	case AUTH_DES:
		dcreds = (struct authdes_cred *) rqstp->rq_clntcred;
		ptr = strchr(dcreds->adc_fullname.name, '@');
		if (ptr == NULL) {
			nis_err("bogus netname: %s",
					dcreds->adc_fullname.name);
			principal = "nobody";
		}
		ptr++;
		strncpy(domain, ptr, NIS_MAXNAMELEN);
		if (domain[strlen(domain) - 1] != '.')
			strcat(domain, ".");
		principal = nis_netid_to_princ(dcreds->adc_fullname.name,
					domain, rqstp->rq_cred.oa_flavor);
		if (principal == NULL) {
			if (verbose)
				nis_err("warning: DES cred lookup failed");
			principal = "nobody";
		}
		break;
#endif
	case AUTH_UNIX:
		if (securelevel > 1) {
			principal = "nobody";
			break;
		}
		ucreds = (struct authunix_parms *)rqstp->rq_clntcred;
		ptr = nis_domain_of(ucreds->aup_machname);
		if (ptr == NULL) {
			principal = "nobody";
			nis_err("couldn't convert LOCAL creds");
			break;
		}
		if (!strcmp(ptr, "."))
			ptr = nis_local_directory();
		strncpy(domain, ptr, NIS_MAXNAMELEN);
		if (domain[strlen(domain) - 1] != '.')
			strcat(domain, ".");
		if (ucreds->aup_uid == 0) {
			strcpy(netname, ucreds->aup_machname);
			if (netname[strlen(netname) -1] != '.')
				strcat(netname, ".");
			principal = netname;
			break;
		}
		snprintf(netname, sizeof(netname), "%u", ucreds->aup_uid);
		principal = nis_netid_to_princ(netname, domain,
						rqstp->rq_cred.oa_flavor);
		if (principal == NULL) {
			if (verbose)
				nis_err("warning: LOCAL cred lookup failed");
			principal = "nobody";
		}
		break;
	case AUTH_NONE:
	default:
		principal = "nobody";
		break;
	}

	strcpy(_pbuf, principal);
	return(_pbuf);
}

/*
 * Core access check routine. Sun provides macros in nis.x/nis.h
 * for this stuff: may as well take advantage of them.
 */

bool_t nis_check_access(action, perms, princ, owner, group)
	uint32_t		action;
	uint32_t		perms;
	nis_name		princ;
	nis_name		owner;
	nis_name		group;
{
	/*
	 * If securelevel == 0, all bets are off.
	 */
	if (securelevel == 0)
		return(TRUE);

	/*
	 * If 'nobody' allows the action, then return
	 * success.
	 */
	if (NOBODY(perms, action))
		return(TRUE);

	/*
	 * If 'world' allows the access and the principal has
	 * been authenticated (i.e. the principal name is not
	 * 'nobody'), then return success.
	 */
	if (WORLD(perms, action) && strcmp(princ, "nobody"))
		return(TRUE);

	/*
	 * If 'owner' allows access and the principal name is the
	 * same as the owner name, return success.
	 */
	if (OWNER(perms, action) && !strcmp(princ, owner))
		return(TRUE);

	/*
	 * Lastly, if 'group' allows access and the principal
	 * is a member of the group, return success.
	 */
	if (GROUP(perms, action) && nis_ismember(princ, group) == TRUE)
		return(TRUE);

	/*
	 * Nothing worked: you're out of luck.
	 */
	return(FALSE);
}

nis_error nis_can_read(principal, obj)
	nis_name		principal;
	nis_object		*obj;
{
	if (nis_check_access(NIS_READ_ACC,
			     obj->zo_access,
			     principal,
			     obj->zo_owner,
			     obj->zo_group) == TRUE)
		return(NIS_SUCCESS);
	return(NIS_PERMISSION);
}

nis_error nis_can_create(principal, obj)
	nis_name		principal;
	nis_object		*obj;
{
	if (nis_check_access(NIS_CREATE_ACC,
			     obj->zo_access,
			     principal,
			     obj->zo_owner,
			     obj->zo_group) == TRUE)
		return(NIS_SUCCESS);
	return(NIS_PERMISSION);
}

nis_error nis_can_modify(principal, obj)
	nis_name		principal;
	nis_object		*obj;
{
	if (nis_check_access(NIS_MODIFY_ACC,
			     obj->zo_access,
			     principal,
			     obj->zo_owner,
			     obj->zo_group) == TRUE)
		return(NIS_SUCCESS);
	return(NIS_PERMISSION);
}

nis_error nis_can_delete(principal, obj)
	nis_name		principal;
	nis_object		*obj;
{
	if (nis_check_access(NIS_DESTROY_ACC,
			     obj->zo_access,
			     principal,
			     obj->zo_owner,
			     obj->zo_group) == TRUE)
		return(NIS_SUCCESS);
	return(NIS_PERMISSION);
}
