/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nis_cache.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nis_extern.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nis_cache.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $";
#endif

/*
 * This module implements a generic object cache. This is
 * _NOT_ the same thing as the cache manager. We keep this cache
 * in order to make the server's life easier and that's it.
 *
 * The most important routines exported by this module are
 * __nis_dcache_lookup() and __nis_ocache_lookup(). The idea is
 * to use these within the server instead of nis_lookup(). This
 * allows us to cache the result returned by nis_lookup() so that
 * we can avoid having to do a full nis_lookup() again the next
 * time we need the same object.
 *
 * Notes:
 *
 * - We don't cache entry objects. There's no point. We may cache
 *   table objects however.
 *
 * - We have both dcache and ocache lookup routines: the former returns
 *   only a directory object without any other cruft surrounding it
 *   and uses somewhat different sanity checks. The latter returns
 *   the entire nis_object structure.
 *
 * - All objects returned by this module are _READ ONLY_. Callers
 *   should never free objects returned from here. Instead, the
 *   cache module will free an object if:
 *
 *	o its TTL expires
 *      o one of the cache flush routines is called
 *	o the lookup routines are called with LOOKUP_ONLY; this
 *        will force the lookup to be done without with nis_lookup()
 *        and cause the cache entry for an object to be refreshed.
 */

#define TABLESIZE 256
#define HASH_MASK 0x000000FF

#define	I_SERVE			0x01
#define I_AM_MASTER		0x02

struct ocache_ent {
	time_t			c_time;		/* Creation time */
	nis_object		*o;		/* Ptr to object */
	u_int32_t		hval;		/* Hash key */
	int			ostat;		/* Status bits */
	struct ocache_ent	*next;		/* Collision pointer */
};

static struct ocache_ent *ocache[TABLESIZE];
static uint32_t		d_calls = 0, d_hits = 0, d_miss = 0;
static uint32_t		o_calls = 0, o_hits = 0, o_miss = 0;

static u_int32_t hashkey(key)
	char			*key;
{
	if (key == NULL)
		return(-1);
	return(__nis_hash((void *)key, strlen(key)));
}

static struct ocache_ent *ocache_lookup(name)
	nis_name		name;
{
	struct ocache_ent	*cur;
	u_int32_t		i;

	i = hashkey(name);
	cur = ocache[i & HASH_MASK];

	while(cur != NULL) {
		if (cur->hval == i)
			return(cur);
		cur = cur->next;
	}

	return(NULL);
}

static void ocache_delete(name)
	nis_name		name;
{
	struct ocache_ent	*cur, *prev;
	u_int32_t		i;

	i = hashkey(name);
	cur = ocache[i & HASH_MASK];
	prev = NULL;

	while(cur != NULL) {
		if (cur->hval == i)
			break;
		prev = cur;
		cur = cur->next;
	}

	if (cur == NULL)
		return;

	if (cur == ocache[i & HASH_MASK]) {
		ocache[i & HASH_MASK] = cur->next;
		xdr_free(xdr_nis_object, (char *)cur->o);
		free(cur->o);
		free(cur);
		return;
	}

	if (prev == ocache[i & HASH_MASK])
		ocache[i & HASH_MASK] = cur->next;
	else
		prev = cur->next;

	xdr_free(xdr_nis_object, (char *)cur->o);
	free(cur->o);
	free(cur);

	return;
}

void __nis_ocache_insert(o)
	nis_object		*o;
{
	struct ocache_ent	*new;
	u_int32_t		i;
	char			buf[NIS_MAXNAMELEN];
	nis_error		oerr;
	directory_obj		*d;

	snprintf(buf, sizeof(buf), "%s.%s", o->zo_name, o->zo_domain);

	if (verbose)
		nis_err("inserting %s into cache", buf);

	if (ocache_lookup(buf) != NULL)
		ocache_delete(buf);

	i = hashkey(buf);

	new = calloc(1, sizeof(struct ocache_ent));
	new->hval = i;
	new->o = nis_clone_object(o, NULL);
	new->c_time = time(NULL);
	new->next = NULL;

	if (__type_of(new->o) == DIRECTORY_OBJ) {
		if (__i_am_master(&new->o->DI_data) == TRUE)
			new->ostat |= I_AM_MASTER;
		if (__is_local_dir(&new->o->DI_data) == TRUE)
			new->ostat |= I_SERVE;
	} else {
		oerr = __nis_dcache_lookup(o->zo_domain, 0, 0, &d);

		if (oerr != NIS_SUCCESS)
			nis_err("warning: couldn't determine object dir");
		else {
			if (__i_am_master(d) == TRUE)
				new->ostat |= I_AM_MASTER;
			if (__is_local_dir(d) == TRUE)
				new->ostat |= I_SERVE;
		}
	}

	new->next = ocache[i & HASH_MASK];
	ocache[i & HASH_MASK] = new;

	return;
}

void __nis_get_dcache_stats(hits, misses, calls)
	uint32_t		*hits,
				*misses,
				*calls;
{
	*hits = d_hits;
	*misses = d_miss;
	*calls = d_calls;

	return;
}

void __nis_get_ocache_stats(hits, misses, calls)
	uint32_t		*hits,
				*misses,
				*calls;
{
	*hits = o_hits;
	*misses = o_miss;
	*calls = o_calls;

	return;
}

/*
 * Meaning of special cache lookup flags:
 *
 * CHECK_MASTER		Make sure I am the master server, return
 *			NIS_NOTMASTER otherwise.
 * CHECK_SERVER		Make sure I am a server (master or replica),
 *			return NIS_NOT_ME otherwise.
 * CACHE_ONLY		Check the cache only; don't rollover to
 *			nis_lookup().
 * LOOKUP_ONLY		Opposite of CACHE_ONLY: do an nis_lookup()
 *			and refresh the cache on success (but don't
 *			check the cache for lookup matches).
 */ 
nis_error __nis_dcache_lookup(name, flags, lflags, dir)
	nis_name		name;
	uint32_t		flags;
	uint32_t		lflags;
	directory_obj		**dir;
{
	nis_result		*res;
	struct ocache_ent	*d_ent;
	nis_error		derr;

	/*
	 * See if the directory is in the cache.
	 */
	if (!(flags & LOOKUP_ONLY)) {
		d_ent = ocache_lookup(name);
		d_calls++;
		if (d_ent == NULL) {
			d_miss++;
		} else {
			d_hits++;
			if (__type_of(d_ent->o) != DIRECTORY_OBJ)
				return(NIS_BADNAME);
			if (time(NULL) > (d_ent->c_time + d_ent->o->zo_ttl)) {
				if (flags & CACHE_ONLY) {
					*(caddr_t *)dir =
						(caddr_t)&d_ent->o->DI_data;
					return(NIS_CACHEEXPIRED);
				} else
					ocache_delete(name);
			} else {
				if (flags & CHECK_MASTER) {
					if (!(d_ent->ostat & I_AM_MASTER))
						return(NIS_NOTMASTER);
				}
				if (flags & CHECK_SERVER) {
					if (!(d_ent->ostat & I_SERVE))
						return(NIS_NOT_ME);
				}
				*(caddr_t *)dir = (caddr_t)&d_ent->o->DI_data;
				return(NIS_SUCCESS);
			}
		}
	}

	if (flags & CACHE_ONLY)
		return(NIS_NOTFOUND);

	ocache_delete(name);
	res = nis_lookup(name, lflags);

	if (res == NULL)
		return(NIS_NOMEMORY);

	if (res->status != NIS_SUCCESS) {
		derr = res->status;
		nis_freeresult(res);
		return(derr);
	}

	if (NIS_RES_NUMOBJ(res) > 1 ||
	    __type_of(NIS_RES_OBJECT(res)) != DIRECTORY_OBJ) {
		nis_freeresult(res);
		return(NIS_BADNAME);
	}

	/* __nis_search() will do an insert for us */

	nis_freeresult(res);

	d_ent = ocache_lookup(name);

	if (flags & CHECK_MASTER && !(d_ent->ostat & I_AM_MASTER))
		return(NIS_NOTMASTER);

	if (flags & CHECK_SERVER && !(d_ent->ostat & I_SERVE))
		return(NIS_NOT_ME);

	*(caddr_t *)dir = (caddr_t)&d_ent->o->DI_data;

	return(NIS_SUCCESS);
}

nis_error __nis_ocache_lookup(name, flags, lflags, obj)
	nis_name		name;
	uint32_t		flags;
	uint32_t		lflags;
	nis_object		**obj;
{
	nis_result		*res;
	struct ocache_ent	*o_ent;
	nis_error		oerr;

	/*
	 * See if the directory is in the cache.
	 */
	if (!(flags & LOOKUP_ONLY)) {
		o_ent = ocache_lookup(name);
		o_calls++;
		if (o_ent == NULL) {
			o_miss++;
		} else {
			o_hits++;
			if (time(NULL) > (o_ent->c_time + o_ent->o->zo_ttl)) {
				if (flags & CACHE_ONLY) {
					*(caddr_t *)obj =
						 (caddr_t)o_ent->o;
					return(NIS_CACHEEXPIRED);
				} else
					ocache_delete(name);
			} else {
				if (flags & CHECK_MASTER) {
					if (!(o_ent->ostat & I_AM_MASTER))
						return(NIS_NOTMASTER);
				}
				if (flags & CHECK_SERVER) {
					if (!(o_ent->ostat & I_SERVE))
						return(NIS_NOT_ME);
				}
				*(caddr_t *)obj = (caddr_t)o_ent->o;
				return(NIS_SUCCESS);
			}
		}
	}

	if (flags & CACHE_ONLY)
		return(NIS_NOTFOUND);

	res = nis_lookup(name, lflags);

	if (res == NULL)
		return(NIS_NOMEMORY);

	if (res->status != NIS_SUCCESS) {
		oerr = res->status;
		nis_freeresult(res);
		return(oerr);
	}

	/* __nis_search() will do an insert for us */

	nis_freeresult(res);

	o_ent = ocache_lookup(name);

	if (flags & CHECK_MASTER && !(o_ent->ostat & I_AM_MASTER))
		return(NIS_NOTMASTER);

	if (flags & CHECK_SERVER && !(o_ent->ostat & I_SERVE))
		return(NIS_NOT_ME);

	*(caddr_t *)obj = (caddr_t)o_ent->o;

	return(NIS_SUCCESS);
}

void __nis_flush_one_dir(name)
	nis_name		name;
{
	ocache_delete(name);
	if (verbose)
		nis_err("flushing %s from cache", name);

	/*
	 * If the domain being flushed is our local domain, then
	 * we need to update the cold start object. If the cache
	 * manager is running then it will try to do it too, but
	 * we do it anyway in case that fails.
	 */
	if (nis_dir_cmp(name, nis_local_directory()) == SAME_NAME) {
		nis_update_cold();
		ocache_delete(name);
	}
#ifdef notyet
	__nis_cache_refresh(name);
#endif

	return;
}

nis_error __nis_flush_dcache()
{
	register unsigned int	i;
	struct ocache_ent	*cur, *tmp;

	for (i = 0; i < TABLESIZE; i++) {
		cur = ocache[i];
		while(cur != NULL) {
			if (__type_of(cur->o) == DIRECTORY_OBJ) {
#ifdef notyet
				__nis_cache_remove(&cur->o->DI_data);
#endif
				xdr_free(xdr_nis_object, (char *)cur->o);
				free(cur->o);
				tmp = cur->next;
				free(cur);
				cur = tmp;
			} else
				cur = cur->next;
		}
	}
			
	return(NIS_SUCCESS);
}

nis_error __nis_flush_ocache()
{
	register unsigned int	i;
	struct ocache_ent	*cur, *tmp;

	for (i = 0; i < TABLESIZE; i++) {
		cur = ocache[i];
		while(cur != NULL) {
#ifdef notyet
			if (__type_of(cur->o) == DIRECTORY_OBJ)
				__nis_cache_remove(&cur->o->DI_data);
#endif
			xdr_free(xdr_nis_object, (char *)cur->o);
			free(cur->o);
			tmp = cur->next;
			free(cur);
			cur = tmp;
		}
	}

	return(NIS_SUCCESS);
}

void __nis_init_ocache()
{
	bzero((char *)&ocache, sizeof(ocache));

	return;
}
