/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id$
 */

#include <sys/types.h>
#include <sys/cdefs.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <rpcsvc/nis.h>
#include <rpcsvc/nis_db.h>
#include <rpcsvc/nislib.h>
#include "nis_extern.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nis_db.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $";
#endif

#define DB_TTL 5

struct db_usag {
	char			*table;
	unsigned int		ttl;
};

static bool_t nis_db_cb(args, id, reason)
	void			*args;
	uint32_t		id;
	int			reason;
{
	struct db_usag		*ptr;

	ptr = args;

	if (ptr->ttl) {
		ptr->ttl--;
		return(TRUE);
	}

	db_unload_table(ptr->table);
	free(ptr->table);
	free(ptr);
	return(FALSE);
}

static nis_error nis_db_sched_flush(table)
	char			*table;
{
	uint32_t		id;
	struct db_usag		*ptr;

	id = nis_event_find(table, (void **)&ptr);

	if (id) {
		free(ptr->table);
		free(ptr);
		nis_event_del(id);
	}

	return(NIS_SUCCESS);
}

static nis_error nis_db_sched_update(table)
	char			*table;
{
	uint32_t		id;
	struct db_usag		*ptr;

	id = nis_event_find(table, (void **)&ptr);

	if (id) {
		ptr->ttl = DB_TTL;
		return(NIS_SUCCESS);
	}

	ptr = calloc(1, sizeof(struct db_usag));
	ptr->ttl = DB_TTL;
	ptr->table = strdup(table);

	nis_event_add_conditional(nis_db_cb, ptr, NULL, 30, table);

	return(NIS_SUCCESS);
}

/*
 * The add routine handles both creation and modification
 * of tables.
 */
db_status db_add(name, obj, res)
	nis_name		name;
	nis_object		*obj;
	struct db_objs		*res;
{
	char			*db_name = NULL,
				*newname;
	nis_attr		attr;
	db_result		*db_res;
	db_status		db_stat;
	entry_obj		dir_ent;
	entry_col		dir_ent_col[2];
	char			*addr;
	uint32_t		s;
	char			__buf[NIS_MAXNAMELEN];

	if (res != NULL)
		bzero((char *)res, sizeof(struct db_objs));
	db_name = __nis_table(nis_domain_of(name), __buf);
	newname = nis_leaf_of(name);

	/*
	 * XDR the supplied object struct.
	 */
	addr = nis_encode_nis_obj(obj, &s);

	/*
	 * Create an entry to be stored in the parent
	 * directory 'table.'
	 */
	dir_ent_col[0].ec_flags = 0;
	dir_ent_col[0].ec_value.ec_value_val = newname;
	dir_ent_col[0].ec_value.ec_value_len = strlen(newname);
	dir_ent_col[1].ec_flags = EN_BINARY | EN_XDR;
	dir_ent_col[1].ec_value.ec_value_val = addr;
	dir_ent_col[1].ec_value.ec_value_len = s;

	dir_ent.en_type = "IN_DIRECTORY";
	dir_ent.en_cols.en_cols_len = 2;
	dir_ent.en_cols.en_cols_val = (entry_col *)&dir_ent_col;

	attr.zattr_ndx = "IN_DIRECTORY";
	attr.zattr_val.zattr_val_len = strlen(newname);
	attr.zattr_val.zattr_val_val = newname;

	/*
	 * Update the 'object' in the underlying database. This
	 * will actually turn into a remove and an add within the
	 * database library.
	 */


	db_res = db_add_entry(db_name, 1, &attr, &dir_ent);
	nis_db_sched_update(db_name);

	free(addr);

	if (db_res == NULL)
		return(DB_MEMORY_LIMIT);

	db_stat = db_res->status;
	if (res != NULL)
		res->dticks = db_res->ticks;

	db_free_result(db_res);

	return(db_stat);
}

db_status db_remove(name, res)
	nis_name		name;
	struct db_objs		*res;
{
	nis_attr                attr;
	char			*db_name,
				*oldname;
	char                    __buf[NIS_MAXNAMELEN];
	db_result		*db_res;
	db_status		db_stat;

	if (res != NULL)
		bzero((char *)res, sizeof(struct db_objs));
	oldname = nis_leaf_of(name);
	db_name = __nis_table(nis_domain_of(name), __buf);

	attr.zattr_ndx = "IN_DIRECTORY";
	attr.zattr_val.zattr_val_len = strlen(oldname);
	attr.zattr_val.zattr_val_val = oldname;

	db_res = db_remove_entry(db_name, 1, &attr);
	nis_db_sched_update(db_name);

	if (db_res == NULL)
		return(DB_MEMORY_LIMIT);

	db_stat = db_res->status;
	if (res != NULL)
		res->dticks += db_res->ticks;

	db_free_result(db_res);

	return(db_stat);
}

db_status db_newdir(name)
	nis_name		name;
{
	db_status		db_stat;
	table_obj		dir_table;
	table_col		dir_col[2];
	nis_attr		attr;
	char			*newdir;
	char			__buf[NIS_MAXNAMELEN];

	newdir = __nis_table(name, __buf);

	db_stat = db_table_exists(newdir);

	if (db_stat == DB_SUCCESS)
		return(DB_NOTUNIQUE);

	dir_table.ta_type = "directory_tbl";
	dir_table.ta_maxcol = 2;
	dir_table.ta_sep = ' ';
	dir_table.ta_path = "";

	dir_col[0].tc_name = "IN_DIRECTORY";
	dir_col[0].tc_flags = TA_SEARCHABLE;
	dir_col[0].tc_rights = DEFAULT_RIGHTS;

	dir_col[1].tc_name = "xdr_nis_object";
	dir_col[1].tc_flags = TA_BINARY | TA_XDR;
	dir_col[1].tc_rights = DEFAULT_RIGHTS;

	dir_table.ta_cols.ta_cols_len = 2;
	dir_table.ta_cols.ta_cols_val = (table_col *)&dir_col;

	attr.zattr_ndx = "IN_DIRECTORY";
	attr.zattr_val.zattr_val_len = 0;
	attr.zattr_val.zattr_val_val = "";

	db_stat = db_create_table(newdir, &dir_table);

	nis_db_sched_update(newdir);
	__nis_flush_one_dir(name);

	return(db_stat);
}

db_status db_newtab(name, obj)
	nis_name		name;
	nis_object		*obj;
{
	db_status		db_stat;
	char			*newdir;
	char			__buf[NIS_MAXNAMELEN];

	newdir = __nis_table(name, __buf);

	db_stat = db_table_exists(newdir);

	if (db_stat == DB_SUCCESS)
		return(DB_NOTUNIQUE);

	db_stat = db_create_table(newdir, &obj->TA_data);

	return(db_stat);
}

db_status db_rm(name)
	nis_name		name;
{
	db_status		db_stat;
	char			*newdir;
	char			__buf[NIS_MAXNAMELEN];

	newdir = __nis_table(name, __buf);

	db_stat = db_table_exists(newdir);

	if (db_stat != DB_SUCCESS)
		return(db_stat);

	db_stat = db_destroy_table(newdir);
	nis_db_sched_flush(newdir);

	return(db_stat);
}

db_status db_exists(name)
	nis_name		name;
{
	db_status		db_stat;
	char			*newdir;
	char			__buf[NIS_MAXNAMELEN];

	newdir = __nis_table(name, __buf);

	db_stat = db_table_exists(newdir);

	return(db_stat);
}

db_status db_lookup(name, res)
	nis_name		name;
	struct db_objs		*res;
{
	char			*db_name;
	char			*lname;
	char			__buf[NIS_MAXNAMELEN];
	nis_attr		attr;
	db_result		*db_res;
	db_status		db_stat;
	DATABASE_CLOCK_DECLS;

	bzero((char *)res, sizeof(struct db_objs));
	/*
	 * Sometimes we are asked to look up the root object.
	 * The root object is stored in a file by itself rather
	 * than as part of a directory, so we have to handle it
	 * specially.
	 */

	if (nis_dir_cmp(name, rootname) == SAME_NAME) {
		__start_clock(DATABASE_CLOCK);
		if (root_obj == NULL) {
			db_stat = DB_NOTFOUND;
		} else {
			NIS_RES_OBJECT(res) = nis_clone_object(root_obj, NULL);
			NIS_RES_NUMOBJ(res) = 1;
			db_stat = DB_SUCCESS;
		}
		res->dticks = __stop_clock(DATABASE_CLOCK);
		return(db_stat);
	}

	db_name = __nis_table(nis_domain_of(name), __buf);
	lname = nis_leaf_of(name);

	if (db_name == NULL)
		return(DB_NOTFOUND);

	db_stat = db_table_exists(db_name);

	if (db_stat != DB_SUCCESS)
		return(db_stat);

	attr.zattr_ndx = "IN_DIRECTORY";
	attr.zattr_val.zattr_val_val = lname;
	attr.zattr_val.zattr_val_len = strlen(lname);

	db_res = db_list_entries(db_name, 1, &attr);
	nis_db_sched_update(db_name);

	if (db_res == NULL)
		return(DB_MEMORY_LIMIT);

	db_stat = db_res->status;

	if (db_stat != DB_SUCCESS) {
		db_free_result(db_res);
		return(db_stat);
	}

	res->dticks += db_res->ticks;
	NIS_RES_OBJECT(res) = nis_decode_nis_obj(OBJ_VAL(db_res, 0, 1),
						 OBJ_LEN(db_res, 0, 1));
	NIS_RES_NUMOBJ(res) = 1;

	db_free_result(db_res);

	return(db_stat);
}

db_status db_iblookup(name, attrs, numattrs, res)
	nis_name		name;
	nis_attr		*attrs;
	uint32_t		numattrs;
	struct db_objs		*res;
{
	db_result		*db_res;
	db_status		db_stat;
	char			__buf[NIS_MAXNAMELEN];
	char			*db_name;
	int			i;
	nis_object		*obj, *list;

	bzero((char *)res, sizeof(struct db_objs));
	db_name = __nis_table(name, __buf);
	db_res = db_list_entries(db_name, numattrs, attrs);
	nis_db_sched_update(db_name);

	if (db_res == NULL)
		return(DB_MEMORY_LIMIT);

	db_stat = db_res->status;

	if (db_stat != DB_SUCCESS) {
		db_free_result(db_res);
		return(db_stat);
	}

	list = calloc(db_res->objects.objects_len, sizeof(nis_object));

	if (list == NULL) {
		db_free_result(db_res);
		return(DB_MEMORY_LIMIT);
	}

	for (i = 0; i < db_res->objects.objects_len; i++) {
		if (strcmp(ENTRY_TYPE(db_res), "IN_DIRECTORY") == NULL)
			obj = nis_encap_obj(db_res, i);
		else if (strcmp(ENTRY_TYPE(db_res), "IN_TABLE") == NULL)
			obj = nis_decode_nis_obj(OBJ_VAL(db_res, i, 1),
						 OBJ_LEN(db_res, i, 1));
		else {
			nis_err("warning: bogus internal entry type: %s",
				     		ENTRY_TYPE(db_res));
			continue;
		}
		bcopy(obj, list + i, sizeof(nis_object));
		free(obj);
	}

	NIS_RES_NUMOBJ(res) = i;
	NIS_RES_OBJECT(res) = list;
	res->dticks = db_res->ticks;

	db_free_result(db_res);

	return(db_stat);
}

db_status db_ibfirst(name, attrs, numattrs, res)
	nis_name		name;
	nis_attr		*attrs;
	uint32_t		numattrs;
	struct db_objs		*res;
{
	db_result		*db_res;
	db_status		db_stat;
	char			__buf[NIS_MAXNAMELEN];
	char			*db_name;
	int			i;
	nis_object		*obj, *list;

	bzero((char *)res, sizeof(struct db_objs));
	db_name = __nis_table(name, __buf);
	db_res = db_first_entry(db_name, numattrs, attrs);
	nis_db_sched_update(db_name);

	if (db_res == NULL)
		return(DB_MEMORY_LIMIT);

	db_stat = db_res->status;

	if (db_stat != DB_SUCCESS) {
		db_free_result(db_res);
		return(db_stat);
	}

	list = calloc(db_res->objects.objects_len, sizeof(nis_object));

	if (list == NULL) {
		db_free_result(db_res);
		return(DB_MEMORY_LIMIT);
	}

	for (i = 0; i < db_res->objects.objects_len; i++) {
		if (strcmp(ENTRY_TYPE(db_res), "IN_DIRECTORY") == NULL)
			obj = nis_encap_obj(db_res, i);
		else if (strcmp(ENTRY_TYPE(db_res), "IN_TABLE") == NULL)
			obj = nis_decode_nis_obj(OBJ_VAL(db_res, i, 1),
						 OBJ_LEN(db_res, i, 1));
		else {
			nis_err("warning: bogus internal entry type: %s",
				     		ENTRY_TYPE(db_res));
			continue;
		}
		bcopy(obj, list + i, sizeof(nis_object));
		free(obj);
	}

	res->cookie.n_len = db_res->nextinfo.db_next_desc_len;
	res->cookie.n_bytes = db_res->nextinfo.db_next_desc_val;
	db_res->nextinfo.db_next_desc_len = 0;
	db_res->nextinfo.db_next_desc_val = NULL;
	NIS_RES_NUMOBJ(res) = i;
	NIS_RES_OBJECT(res) = list;

	db_free_result(db_res);

	return(db_stat);
}

db_status db_ibnext(name, nextptr, res)
	nis_name		name;
	netobj			*nextptr;
	struct db_objs		*res;
{
	db_result		*db_res;
	db_status		db_stat;
	db_next_desc		db_next;
	char			__buf[NIS_MAXNAMELEN];
	char			*db_name;
	nis_object		*obj;

	bzero((char *)res, sizeof(struct db_objs));
	db_name = __nis_table(name, __buf);
	db_next.db_next_desc_len = nextptr->n_len;
	db_next.db_next_desc_val = nextptr->n_bytes;

	db_res = db_next_entry(db_name, &db_next);
	nis_db_sched_update(db_name);

	if (db_res == NULL) {
		free(nextptr->n_bytes);
		nextptr->n_bytes = NULL;
		nextptr->n_len = 0;
		return(DB_MEMORY_LIMIT);
	}

	db_stat = db_res->status;
	nextptr->n_len = 0;
	nextptr->n_bytes = NULL;

	if (db_stat != DB_SUCCESS) {
		db_free_result(db_res);
		return(db_stat);
	}

	res->dticks += db_res->ticks;
	res->cookie.n_len = db_res->nextinfo.db_next_desc_len;
	res->cookie.n_bytes = db_res->nextinfo.db_next_desc_val;
	db_res->nextinfo.db_next_desc_len = 0;
	db_res->nextinfo.db_next_desc_val = NULL;

	if (strcmp(ENTRY_TYPE(db_res), "IN_DIRECTORY") == NULL)
		obj = nis_encap_obj(db_res, 0);
	else if (strcmp(ENTRY_TYPE(db_res), "IN_TABLE") == NULL)
		obj = nis_decode_nis_obj(OBJ_VAL(db_res, 0, 1),
					 OBJ_LEN(db_res, 0, 1));
	else {
		nis_err("bogus internal entry type: %s",
					ENTRY_TYPE(db_res));
		db_free_result(db_res);
		return(DB_BADTABLE);
	}

	NIS_RES_NUMOBJ(res) = 1;
	NIS_RES_OBJECT(res) = obj;

	db_free_result(db_res);

	return(db_stat);
}

db_status db_first(name, attrs, numattrs, res)
	nis_name		name;
	nis_attr		*attrs;
	uint32_t		numattrs;
	struct db_objs		*res;
{
	db_result		*db_res;
	db_status		db_stat;
	char			__buf[NIS_MAXNAMELEN];
	char			*db_name;
	int			i;
	nis_object		*obj, *list;

	bzero((char *)res, sizeof(struct db_objs));
	db_name = __nis_table(name, __buf);
	db_res = db_first_entry(db_name, numattrs, attrs);
	nis_db_sched_update(db_name);

	if (db_res == NULL)
		return(DB_MEMORY_LIMIT);

	db_stat = db_res->status;

	if (db_stat != DB_SUCCESS) {
		db_free_result(db_res);
		return(db_stat);
	}

	list = calloc(db_res->objects.objects_len, sizeof(nis_object));

	if (list == NULL) {
		db_free_result(db_res);
		return(DB_MEMORY_LIMIT);
	}

	for (i = 0; i < db_res->objects.objects_len; i++) {
		if (strcmp(ENTRY_TYPE(db_res), "IN_DIRECTORY") == NULL)
			obj = nis_decode_nis_obj(OBJ_VAL(db_res, i, 1),
						 OBJ_LEN(db_res, i, 1));
		else {
			nis_err("entry isn't a directory object: %s",
						ENTRY_TYPE(db_res));
			continue;
		}
		bcopy(obj, list + i, sizeof(nis_object));
		free(obj);
	}

	res->cookie.n_len = db_res->nextinfo.db_next_desc_len;
	res->cookie.n_bytes = db_res->nextinfo.db_next_desc_val;
	db_res->nextinfo.db_next_desc_len = 0;
	db_res->nextinfo.db_next_desc_val = NULL;
	NIS_RES_NUMOBJ(res) = i;
	NIS_RES_OBJECT(res) = list;

	db_free_result(db_res);

	return(db_stat);
}

db_status db_next(name, nextptr, res)
	nis_name		name;
	netobj			*nextptr;
	struct db_objs		*res;
{
	db_result		*db_res;
	db_status		db_stat;
	db_next_desc		db_next;
	char			__buf[NIS_MAXNAMELEN];
	char			*db_name;
	nis_object		*obj;

	bzero((char *)res, sizeof(struct db_objs));
	db_name = __nis_table(name, __buf);
	db_next.db_next_desc_len = nextptr->n_len;
	db_next.db_next_desc_val = nextptr->n_bytes;

	db_res = db_next_entry(db_name, &db_next);
	nis_db_sched_update(db_name);

	if (db_res == NULL) {
		free(nextptr->n_bytes);
		nextptr->n_bytes = NULL;
		nextptr->n_len = 0;
		return(DB_MEMORY_LIMIT);
	}

	db_stat = db_res->status;
	nextptr->n_len = 0;
	nextptr->n_bytes = NULL;

	if (db_stat != DB_SUCCESS) {
		db_free_result(db_res);
		return(db_stat);
	}

	res->dticks += db_res->ticks;
	res->cookie.n_len = db_res->nextinfo.db_next_desc_len;
	res->cookie.n_bytes = db_res->nextinfo.db_next_desc_val;
	db_res->nextinfo.db_next_desc_len = 0;
	db_res->nextinfo.db_next_desc_val = NULL;

	if (strcmp(ENTRY_TYPE(db_res), "IN_DIRECTORY") == NULL)
		obj = nis_decode_nis_obj(OBJ_VAL(db_res, 0, 1),
					 OBJ_LEN(db_res, 0, 1));
	else {
		nis_err("entry isn't a directory object: %s",
							ENTRY_TYPE(db_res));
		db_free_result(db_res);
		return(DB_BADTABLE);
	}

	NIS_RES_NUMOBJ(res) = 1;
	NIS_RES_OBJECT(res) = obj;

	db_free_result(db_res);

	return(db_stat);
}

db_status db_ibadd(name, attrs, numattrs, obj, table, res)
	nis_name		name;
	nis_attr		*attrs;
	uint32_t		numattrs;
	nis_object		*obj;
	nis_object		*table;
	struct db_objs		*res;
{
	struct entryattrs	*a = NULL;
	entry_obj		tab_ent;
	entry_col		tab_ent_col[2];
	char			*addr;
	uint32_t		s;
	db_result		*db_res;
	db_status		db_stat;
	nis_attr		*atr;
	uint32_t		cnt;
	char			__buf[NIS_MAXNAMELEN];
	char			*db_name;

	if (res != NULL)
		bzero((char *)res, sizeof(struct db_objs));
	if (__type_of(obj) != ENTRY_OBJ || __type_of(table) != TABLE_OBJ)
		return(DB_BADQUERY);

	db_name = __nis_table(name, __buf);

	addr = nis_encode_nis_obj(obj, &s);
	if (addr == NULL)
		return(DB_MEMORY_LIMIT);

	tab_ent_col[0].ec_flags = 0;
	tab_ent_col[0].ec_value.ec_value_val = name;
	tab_ent_col[0].ec_value.ec_value_len = strlen(name);
	tab_ent_col[1].ec_flags = EN_BINARY | EN_XDR;
	tab_ent_col[1].ec_value.ec_value_val = addr;
	tab_ent_col[1].ec_value.ec_value_len = s;

	tab_ent.en_type = "IN_TABLE";
	tab_ent.en_cols.en_cols_len = 2;
	tab_ent.en_cols.en_cols_val = (entry_col *)&tab_ent_col;

	if (numattrs) {
		int			i, j;
		entry_obj		*e = NULL;
		entry_col		*ec = NULL;

		cnt = numattrs;
		atr = attrs;
		/*
		 * Sanity check: if the caller supplies entries,
	 	 * we need to check to make sure that they really
		 * agree with the corresponding entry information.
		 * This is for our benefit. To understand why we
		 * must do this, realize that because of the way
		 * NIS+ entry objects are encapsulated, we are
		 * actually divorcing the supplied entry from the
	 	 * attribute information that the database backend
	 	 * will use in its metadata store (and which will
	 	 * be consulted later for lookups). In order for
		 * things to work right later, the supplied attributes
		 * and the entry have to match: if the attributes
		 * say that the 'name' column has a value of 'foo\0'
		 * which is 4 bytes long, then the entry better damn
		 * well have a 'name' column with a value of 'foo\0'
		 * that's 4 bytes long or else we'll be screwed later.
		 */
		e = &obj->EN_data;
		for (i = 0; i < cnt; i++) {
			for (j = 0; j < e->en_cols.en_cols_len; j++) {
				ec = &e->en_cols.en_cols_val[j];
				if (!bcmp(ec->ec_value.ec_value_val,
					 atr[i].zattr_val.zattr_val_val,
					 ec->ec_value.ec_value_len))
						break;
			}
			if (j == e->en_cols.en_cols_len) {
				free(addr);
				return(DB_BADQUERY);
			}
		}
	} else {
		a = nis_entry_to_attr(&obj->EN_data, &table->TA_data);
		cnt = a->cnt;
		atr = a->attrs;
	}

	db_res = db_add_entry(db_name, cnt, atr, &tab_ent);
	nis_db_sched_update(db_name);

	free(addr);
	if (a != NULL)
		free(a->attrs);

	if (db_res == NULL)
		return(DB_MEMORY_LIMIT);

	db_stat = db_res->status;
	if (res != NULL)
		res->dticks = db_res->ticks;

	db_free_result(db_res);

	return(db_stat);
}

db_status db_ibremove(name, attrs, numattrs, obj, table, res)
	nis_name		name;
	nis_attr		*attrs;
	uint32_t		numattrs;
	nis_object		*obj;
	nis_object		*table;
	struct db_objs		*res;
{
	struct entryattrs	*a = NULL;
	uint32_t		cnt;
	nis_attr		*atr;
	db_result		*db_res;
	db_status		db_stat;
	char			*db_name;
	char			__buf[NIS_MAXNAMELEN];

	if (res != NULL)
		bzero((char *)res, sizeof(struct db_objs));
	if (__type_of(obj) != ENTRY_OBJ || __type_of(table) != TABLE_OBJ)
		return(DB_BADQUERY);

	db_name = __nis_table(name, __buf);

	if (numattrs) {
		atr = attrs;
		cnt = numattrs;
	} else {
		a = nis_entry_to_attr(&obj->EN_data, &table->TA_data);
		atr = a->attrs;
		cnt = a->cnt;
	}

	db_res = db_remove_entry(db_name, cnt, atr);

	if (a != NULL)
		free(a->attrs);

	if (db_res == NULL)
		return(DB_MEMORY_LIMIT);

	db_stat = db_res->status;
	if (res != NULL)
		res->dticks = db_res->ticks;

	db_free_result(db_res);

	return(db_stat);
}

db_status db_cpdir(name, res)
	nis_name		name;
	struct db_objs		*res;
{
	struct db_objs		dres;
	char			__buf[NIS_MAXNAMELEN], fname[NIS_MAXNAMELEN];
	nis_name		db_name = NULL;
	netobj			cookie = { 0, NULL };
	db_status		db_stat;
	uint32_t		xid, stamp;
	DATABASE_CLOCK_DECLS;

	__start_clock(DATABASE_CLOCK);

	if (res != NULL)
		bzero((char *)res, sizeof(struct db_objs));
	if (db_exists(name) != DB_SUCCESS) {
		res->dticks = __stop_clock(DATABASE_CLOCK);
		return(DB_BADTABLE);
	}

	db_stat = db_first(name, NULL, 0, &dres);
	if (db_stat != DB_SUCCESS) {
		res->dticks = __stop_clock(DATABASE_CLOCK);
		return(db_stat);
	}

	while(db_stat == DB_SUCCESS) {
		if (__type_of(NIS_RES_OBJECT(&dres)) == TABLE_OBJ ||
			__type_of(NIS_RES_OBJECT(&dres)) == DIRECTORY_OBJ) {
			snprintf(fname, sizeof(fname), "%s.%s",
				NIS_RES_OBJECT(&dres)->zo_name,
				NIS_RES_OBJECT(&dres)->zo_domain);
			db_name = __nis_table(fname, __buf);
			db_checkpoint(db_name);
		}
		cookie.n_len = dres.cookie.n_len;
		cookie.n_bytes = dres.cookie.n_bytes;
		dres.cookie.n_len = 0;
		dres.cookie.n_bytes = NULL;
		nis_free_dobjs(&dres);
		db_stat = db_next(name, &cookie, &dres);
	}

	free(cookie.n_bytes);
	db_name = __nis_table(name, __buf);
	db_checkpoint(db_name);
	stamp = nis_log_find_update(name, 0, NULL);
	nis_log_new_xid(nis_local_principal(), &xid);
	nis_log_insert_stamp(name, stamp);
	nis_log_commit_stamps();
	nis_log_free_stamps();
	nis_log_commit_xid(xid);

	res->dticks = __stop_clock(DATABASE_CLOCK);

	return(DB_SUCCESS);
}
