/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nis_db_ib.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/cdefs.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nis_db.h>
#include <rpcsvc/nis_tags.h>
#include "nis_extern.h"
#include "nis_common.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nis_db_ib.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $";
#endif

/*
 * The 'information base functions.'
 *
 * These are used for retrieving entries from tables rather than
 * searching the namespace. One particular routine, iblist, is
 * allowed to use callbacks. This is actually handled in the upper
 * level server code.
 *
 * In each case, these functions accept the following arguments:
 * name		: the NIS+ name of the table being searched
 * numattrs	: the number of search attributes
 * attrs	: pointer to an array of nis_attr structs
 * principal	: name of principal making the request
 *		  (NULL means the server is making the request itself
 *		  rather than on behalf of a client)
 *
 * The nis_db_ibmodify() routine may be called with numattrs == 0 and
 * and attrs == NULL. This is because it appears that when a client
 * asks to modify an entry, it may not bother to send search criteria.
 */

/*
 * NIS+ access rights behavior for tables. This information comes from:
 * _SunOS 5.2 Administering NIS+ and DNS_, Sun PN 801-4076-10, revision A,
 * May 1993.
 *
 * o Granting access to read or modify a table
 *
 * The server first checks the table object's rights. If the table
 * grants read access to the group, the principal is allowed to read
 * all of the contents of the table, and the server does not proceed
 * to check rights of columns or entries.
 *
 * If the table object does not grant read access to the group, the
 * server checks the access rights granted by each individual entry.
 * If an entry grants its group read rights, and if the principal
 * belongs to that group (which may be different from the table's
 * group), the server allows the principal to read that entry's
 * contents. Then it checks the rights granted by the next entry.
 *
 * The server then proceeds to check the column rights only if no
 * entry has granted the group read rights. If any entry grants the
 * group read rights, the server does not check column rights.
 *
 * If the server finds that no columns in the table grant their group
 * read rights, the server returns an error message, stating that the
 * principal does not have permission to access the object. However,
 * if any column grants its group read rights, and if the principal
 * belongs to that group, the server displays the body of the table
 * but censors the columns that do not grant their group read rights.
 *
 * Note: this version of the book seems to imply that there are
 * owner/group principals associated with table column rights.
 * In fact, this not the case: the column rights are evaluated
 * based on the table's owner and group principals.
 */
 
/*
 * This function translates data returned from the database library
 * into nis_result form. In the process, it performs access rights
 * checks and filters out data that the calling principal isn't
 * allowed to see. This code is common to the iblist, ibfirst and
 * ibnext routines, so it's separated out here rather than having
 * it repeated three times.
 */
static void __nis_add_obj(res, db_res, principal, table)
	nis_result		*res;
	struct db_objs		*db_res;
	nis_name		principal;
	nis_object		*table;
{
	int			i = 0, j = 0;
	nis_error		t_right = NIS_PERMISSION;
	nis_error		e_right = NIS_PERMISSION;
	nis_error		c_right = NIS_PERMISSION;
	nis_object		*obj = NULL;

	/*
	 * First level of permission checks:
	 * If the principal has read access to the table, then
	 * he has read access to all columns of all entries in
	 * the table too.
	 */

	t_right = e_right = c_right = nis_can_read(principal, table);

	j = NIS_RES_NUMOBJ(res);
	NIS_RES_NUMOBJ(res) += db_res->objects.objects_len;
	if (j == 0)
		NIS_RES_OBJECT(res) = malloc(sizeof(nis_object) *
					NIS_RES_NUMOBJ(db_res));
	else
		NIS_RES_OBJECT(res) = realloc(NIS_RES_OBJECT(res),
						sizeof(nis_object) *
						(NIS_RES_NUMOBJ(db_res) + j));

	for (i = 0; i < NIS_RES_NUMOBJ(db_res); i++) {
		obj = &db_res->objects.objects_val[i];
		/*
		 * Second level of permission checks:
		 * if the principal has access rights for the
		 * entire entry, then he has access to all columns
		 * in the entry (i.e. no mangling is necessary).
		 */
		if (t_right == NIS_PERMISSION)
			e_right = c_right = nis_can_read(principal, obj);
		/*
		 * Third level of permission checks:
		 * if the table's column access rights allow
		 * access to particular columns within the
	 	 * entry, then allow him to see those columns
		 * while coercing all others to the *NP* string.
		 */
		if (e_right == NIS_PERMISSION)
			c_right = nis_mangle_entry(obj, table, principal);

		/*
		 * If the principal wasn't allowed to see any columns,
		 * don't show him the entry at all.
		 */
		if (c_right == NIS_PERMISSION) {
			res->objects.objects_len--; 
			res->objects.objects_val =
				realloc(res->objects.objects_val,
					res->objects.objects_len);
			xdr_free((xdrproc_t)xdr_nis_object, (char *)obj);
		} else {
			bcopy(obj, (char *)&res->objects.objects_val[j],
				sizeof(nis_object));
			j++;
		}
		bzero(obj, sizeof(nis_object));
	}

	free(db_res->objects.objects_val);

	if (!NIS_RES_NUMOBJ(res))
		res->status = NIS_PERMISSION;

	return; 
}

/*
 * This is here to make Sun servers happy. It seems that when
 * doing a dumplog, a Sun NIS+ replica wants a complete set of
 * attributes in lres->le_attrs. Failing to provide them causes
 * the server to corrupt its transaction log and crash.
 */
static void nis_mklog(name, e, t, ltype)
	nis_name		name;
	nis_object		*e, *t;
	log_entry_t		ltype;
{
	struct entryattrs	*a;

	a = nis_entry_to_attr(&e->EN_data, &t->TA_data);
	nis_log_submit(name, e, a->attrs, a->cnt, ltype);
	free(a->attrs);
}

/*
 * The iblist function is a real corker. It's not used just for dumping
 * the entries out of tables: it's also used to generate directory
 * listings. In this case, it may be called without any attributes
 * and we will have to grovel through the entire table, spewing it back
 * to the client a chunk at a time. Callbacks are almost always used
 * for this.
 *
 * Note that this function only does searches with attributes. The other
 * weird cases are handled elsewhere.
 */
nis_error nis_db_iblist(name, flags, numattrs, attrs, result, principal)
	nis_name		name;
	uint32_t		flags;
	uint32_t		numattrs;
	nis_attr		*attrs;
	nis_result		*result;
	nis_name		principal;
{
	db_status		db_stat;
	nis_object		*o;
	nis_error		cerr;
	struct db_objs		dres;

	if (nis_validname(name) != NIS_SUCCESS)
		return(NIS_BADNAME);

	cerr = __nis_ocache_lookup(name, CHECK_SERVER, 0, &o);
	if (cerr != NIS_SUCCESS) {
		if (cerr == NIS_NOTFOUND)
			return(NIS_NOSUCHTABLE);
		return(cerr);
	}

	/*
	 * If the name passed to us is not a table or a directory,
	 * return NIS_PARTIAL. This usually happens if the table
	 * name is actually a link object. The server does not
	 * follow links: the client library does.
	 */
	if (__type_of(o) != TABLE_OBJ &&
		__type_of(o) != DIRECTORY_OBJ) {
		NIS_RES_OBJECT(result) = nis_clone_object(o, NULL);
		NIS_RES_NUMOBJ(result) = 1;
		return(NIS_PARTIAL);
	}

	/*
	 * Do the actual database lookup.
	 */
	db_stat = db_iblookup(name, attrs, numattrs, &dres);

	/*
	 * If we were successful, we have to package up the
	 * entries.
	 */
	if (db_stat == DB_SUCCESS) {
		__nis_add_obj(result, &dres, principal, o);
		result->dticks += dres.dticks;
	} else if (db_stat == DB_NOTFOUND) {
		result->status = NIS_PARTIAL;
		NIS_RES_OBJECT(result) = nis_clone_object(o, NULL);
		NIS_RES_NUMOBJ(result) = 1;
		result->dticks += dres.dticks;
	}

	/*
	 * If we were unable to locate the entries specified by the
	 * search criteria, then we return partial success: we found
	 * the table, but we couldn't find what we wanted in it.
	 * The client library should handle this internally.
	 */
	if (db_stat == DB_NOTFOUND)
		return(NIS_PARTIAL);
	return(db_to_nis_err(db_stat));
}

/*
 * Read first entry in a database.
 */
nis_error nis_db_ibfirst(name, flags, numattrs, attrs, result, principal)
	nis_name		name;
	uint32_t		flags;
	uint32_t		numattrs;
	nis_attr		*attrs;
	nis_result		*result;
	nis_name		principal;
{
	db_status		db_stat;
	nis_object		*o;
	nis_error		r;
	struct db_objs		dres;

	r = __nis_ocache_lookup(name, CHECK_SERVER, 0, &o);

	if (r != NIS_SUCCESS) {
		if (r == NIS_NOTFOUND)
			return(NIS_NOSUCHTABLE);
		return(r);
	}

	if (__type_of(o) != TABLE_OBJ && __type_of(o) != DIRECTORY_OBJ) {
		result->status = NIS_PARTIAL;
		NIS_RES_OBJECT(result) = nis_clone_object(o, NULL);
		NIS_RES_NUMOBJ(result) = 1;
		return(NIS_PARTIAL);
	}

	db_stat = db_ibfirst(name, attrs, numattrs, &dres);

	if (db_stat != NIS_SUCCESS) {
		if (db_stat == DB_NOTFOUND) {
			result->status = NIS_PARTIAL;
			NIS_RES_OBJECT(result) = nis_clone_object(o, NULL);
			NIS_RES_NUMOBJ(result) = 1;
		}
		if (db_stat == DB_NOTFOUND)
			return(NIS_PARTIAL);
		return(db_to_nis_err(db_stat));
	}

	__nis_add_obj(result, &dres, principal, o);

	result->dticks += dres.dticks;

	/*
	 * Alright, we have to return the nextinfo handle to the caller
	 * somehow. The nis_result structure doesn't have anything that
	 * can be used to hold it other than the cookie field, so we use
	 * that.
	 */
	if (db_stat == DB_SUCCESS) {
		result->cookie.n_len = dres.cookie.n_len;
		result->cookie.n_bytes = dres.cookie.n_bytes;
	}

	return(db_to_nis_err(db_stat));
}

/*
 * Note: this routine free()s the nextptr data passed to it.
 * This saves the caller from having to free it.
 */
nis_error nis_db_ibnext(name, flags, nextptr, result, principal)
	nis_name		name;
	uint32_t		flags;
	netobj			*nextptr;
	nis_result		*result;
	nis_name		principal;
{
	db_status		db_stat;
	nis_object		*o;
	struct db_objs		dres;

	if (__nis_ocache_lookup(name, CHECK_SERVER, 0, &o) != NIS_SUCCESS)
		return(NIS_NOSUCHTABLE);

	db_stat = db_ibnext(name, nextptr, &dres);

	if (db_stat != DB_SUCCESS) {
		return(db_to_nis_err(db_stat));
	}

	__nis_add_obj(result, &dres, principal, o);

	result->dticks += dres.dticks;

	if (db_stat == DB_SUCCESS) {
		result->cookie.n_len = dres.cookie.n_len;
		result->cookie.n_bytes = dres.cookie.n_bytes;
	}

	return(db_to_nis_err(db_stat));
}

nis_error nis_db_ibendnext(name, flags, nextptr, principal)
	nis_name		name;
	uint32_t		flags;
	netobj			*nextptr;
	nis_name		principal;
{
	char			*db_name;
	db_next_desc		db_next;
	db_result		*db_res;
	db_status		db_stat;
	char			__buf[NIS_MAXNAMELEN];

	db_next.db_next_desc_len = nextptr->n_len;
	db_next.db_next_desc_val = nextptr->n_bytes;
	
	db_name = __nis_table(name, __buf);

	db_res = db_reset_next_entry(db_name, &db_next);

	if (db_res == NULL)
		return(NIS_NOMEMORY);
	db_stat = db_res->status;

	db_free_result(db_res);
	return(db_to_nis_err(db_stat));
}

static nis_error nis_ib_add(name, flags, nattrs, attrs, obj, t, princ, res)
	nis_name		name;
	uint32_t		flags;
	uint32_t		nattrs;
	nis_attr		*attrs;
	nis_object		*obj;
	nis_object		*t;
	nis_name		princ;
	nis_result		*res;
{
	struct db_objs		dres;
	db_status		db_stat = DB_INTERNAL_ERROR;

	if (nis_can_create(princ, t))
		return(NIS_PERMISSION);

	if (obj == NULL)
		return(NIS_INVALIDOBJ);

	if (nis_dir_cmp(t->zo_domain, obj->zo_domain) != SAME_NAME)
		return(NIS_BADATTRIBUTE);

	db_stat = db_iblookup(name, attrs, nattrs, &dres);
	if (res != NULL)
		res->dticks += dres.dticks;

	/*
	 * If the ADD_OVERWRITE flag is specified, we're
	 * allowed to overwrite an existing entry with the
	 * same attributes. If it is _not_ specified, then
	 * we have to avoid a possible overwrite. The database
	 * backend allows two objects with the same attributes
	 * to exist, but we want to avoid that.
	 */
	if (db_stat == DB_SUCCESS) {
		if (flags & ADD_OVERWRITE) {
			if (nis_can_modify(princ, NIS_RES_OBJECT(&dres))) {
				nis_free_dobjs(&dres);
				return(NIS_PERMISSION);
			}
		} else {
			nis_free_dobjs(&dres);
			return(NIS_NAMEEXISTS);
		}
		nis_free_dobjs(&dres);
	}

	obj->zo_oid.ctime = obj->zo_oid.mtime = time(NULL);

	nis_mklog(name, obj, t, ADD_IBASE);

	db_stat = db_ibadd(name, attrs, nattrs, obj, t, &dres);

	if (db_stat != DB_SUCCESS)
		return(db_to_nis_err(db_stat));

	res->dticks += dres.dticks;

	if (flags & RETURN_RESULT) {
		NIS_RES_OBJECT(res) = nis_clone_object(obj, NULL);
		NIS_RES_NUMOBJ(res) = 1;
	}

	return(NIS_SUCCESS);
}

/*
 * Handler for modifying NIS+ entries.
 * Note: in direct contravention of all logic, it is possible to use
 * the ibmodify routine to _add_ new entries in addition to modifying
 * existing ones. nisaddent(8) is one program that (ab)uses this
 * particular misfeature.
 *
 * In other words, if the caller passes us an object and a set of
 * attributes and we don't find any entries in the table that match
 * the attributes, we add the entry to the table (after turning off
 * all of its EN_MODIFIED flags).
 */

static nis_error nis_merge_entries(old, new, t, princ, e_right)
	nis_object		*old;
	nis_object		*new;
	nis_object		*t;
	nis_name		princ;
	nis_error		e_right;
{
	int			i;
	int			mods = 0;

	/*
	 * The modify operation actually involves merging parts of the
	 * supplied entry (that are marked as EN_MODIFIED) with the
	 * original entry read from the database.
	 */
	for (i = 0; i < t->TA_data.ta_cols.ta_cols_len; i++) {
		if (ENTRY_FLAGS(new, i) & EN_MODIFIED) {
			mods++;
			/* Are we allowed to modify the column? */
			if (e_right == NIS_PERMISSION &&
			    nis_check_access(NIS_MODIFY_ACC,
			     t->TA_data.ta_cols.ta_cols_val[i].tc_rights,
			     princ, t->zo_owner, t->zo_group) == FALSE) {
				/* No: bail */
				return(NIS_PERMISSION);
			}
			/*
			 * Yes: do it up.
			 * XXX: should we leave the EN_MODIFIED flags
			 * turned on so that they'll get saved in the
			 * log entry?
			 */
			ENTRY_FLAGS(new, i) &= ~EN_MODIFIED;
			free(ENTRY_VAL(old, i));
			ENTRY_VAL(old, i) = ENTRY_VAL(new, i);
			ENTRY_LEN(old, i) = ENTRY_LEN(new, i);
			/*
			 * Clear pointers/length in supplied obj
			 * so that svc_freeargs() doesn't generate
			 * a double free.
			 */
			ENTRY_VAL(new, i) = NULL;
			ENTRY_LEN(new, i) = 0;
		}
	}

	if (!mods)
		return(NIS_BADREQUEST);

	return(NIS_SUCCESS);
}

static nis_error nis_ib_modify(name, flags, nattrs, attrs, obj, t, princ, res)
	nis_name		name;
	uint32_t		flags;
	uint32_t		nattrs;
	nis_attr		*attrs;
	nis_object		*obj;
	nis_object		*t;
	nis_name		princ;
	nis_result		*res;
{
	struct db_objs		dres;
	db_status		db_stat = DB_INTERNAL_ERROR;
	nis_error		t_right = NIS_PERMISSION;
	nis_error		e_right = NIS_PERMISSION;
	nis_error		c_right = NIS_PERMISSION;
	nis_object		*o = NULL;

	if (obj == NULL)
		return(NIS_INVALIDOBJ);

	t_right = c_right = e_right = nis_can_modify(princ, t);

	db_stat = db_iblookup(name, attrs, nattrs, &dres);

	res->dticks += dres.dticks;

	/*
	 * If an entry with the same attributes already exists,
	 * then this is a modify operation. If not, it's an add
	 * operation in disguise.
	 *
	 * XXX Careful! There's some trickery involved here because
	 * of the way NIS+ objects are encapsulated within entry
	 * columns of the database. If this modify operation changes
	 * a searchable column in an entry, then we have to remove
	 * the existing entry ourselves first. modify it, then add
	 * it back. If we don't do this, we can get into a situation
	 * where the entry has the new column infomation but is still
	 * indexed in the database under its old search criteria.
	 */
	if (db_stat == DB_SUCCESS) {

		/*
		 * Can't modify more than one entry at a time.
		 */
		if (NIS_RES_NUMOBJ(&dres) > 1) {
			nis_free_dobjs(&dres);
			return(NIS_NOTUNIQUE);
		}

		o = NIS_RES_OBJECT(&dres);

		if (flags & MOD_SAMEOBJ) {
			if (!sameobj(o, obj)) {
				nis_free_dobjs(&dres);
				return(NIS_NOTSAMEOBJ);
			}
		}

		if (t_right == NIS_PERMISSION)
			e_right = nis_can_modify(princ, o);

		/*
	 	 * Check if caller wants to change the non-variant
	 	 * part of the object.
		 */
		if (strcasecmp(o->zo_owner, obj->zo_owner)) {
			if (e_right == NIS_SUCCESS) {
				if (nis_verifyprinc(obj->zo_owner,
				MASTER_ONLY|EXPAND_NAME, NULL) == NIS_SUCCESS) {
					free(o->zo_owner);
					o->zo_owner = obj->zo_owner;
					obj->zo_owner = NULL;
				} else
					return(NIS_BADATTRIBUTE);
			} else
				return(NIS_PERMISSION);
		}

		if (strcasecmp(o->zo_group, obj->zo_group)) {
			if (e_right == NIS_SUCCESS) {
				if (nis_verifygroup_local(obj->zo_group,
			MASTER_ONLY|EXPAND_NAME, NULL, NULL) == NIS_SUCCESS) {
					free(o->zo_group);
					o->zo_group = obj->zo_group;
					obj->zo_group = NULL;
				} else
					return(NIS_BADATTRIBUTE);
			} else
				return(NIS_PERMISSION);
		}

		if (obj->zo_ttl != o->zo_ttl) {
			if (e_right == NIS_SUCCESS)
				o->zo_ttl = obj->zo_ttl;
			else
				return(NIS_PERMISSION);
		}

		if (obj->zo_ttl != o->zo_access) {
			if (e_right == NIS_SUCCESS)
				o->zo_access = obj->zo_access;
			else
				return(NIS_PERMISSION);
		}

		c_right = nis_merge_entries(o, obj, t, princ, e_right);

		if (c_right == NIS_PERMISSION) {
			nis_destroy_object(o);
			return(NIS_PERMISSION);
		}

		o->zo_oid.mtime = time(NULL);
		nis_mklog(name, o, t, ADD_IBASE);
		db_stat = db_ibremove(name, attrs, nattrs, o, t, &dres);
		res->dticks += dres.dticks;
		db_stat = db_ibadd(name, NULL, 0, o, t, &dres);
	} else {
		/*
		 * If no old entry was found, then this operation
		 * will actually be an add. In order to be able to
		 * perform an add, the principal must have create
		 * rights on the table.
		 */
		t_right = e_right = nis_can_create(princ, t);
		if (t_right == NIS_PERMISSION)
			return(NIS_PERMISSION);
		obj->zo_oid.ctime = time(NULL);
		nis_mklog(name, obj, t, ADD_IBASE);
		/*
		 * XXX Sun bug: sometimes nisaddent(8) will try to
		 * perform an add via the nis_modify_entry() routine
		 * and fail to supply a full set of attributes. Work
		 * around this by forcing db_ibadd() to work out the
		 * attributes itself.
		 */
		db_stat = db_ibadd(name, NULL, 0, obj, t, &dres);
	}

	res->dticks += dres.dticks;

	if (db_stat != NIS_SUCCESS) {
		nis_free_dobjs(&dres);
		return(NIS_MODFAIL);
	}

	if (flags & RETURN_RESULT) {
		if (o == NULL) {
			NIS_RES_OBJECT(res) = nis_clone_object(obj, NULL);
			NIS_RES_NUMOBJ(res) = 1;
		} else {
			NIS_RES_OBJECT(res) = nis_clone_object(o, NULL);
			NIS_RES_NUMOBJ(res) = 1;
		}
	}

	nis_free_dobjs(&dres);

	return(NIS_SUCCESS);
}

/*
 * The ibremove function supports some special cases: if the REM_MULTIPLE
 * flag is specified and there are no attributes, then we're supposed to
 * delete _all_ the entries in a table. Permission checking in this case
 * can be confusing. There are two cases where a principal can delete all
 * the entries in a table: either he has destroy rights on the table, or
 * he has destroy rights on all the entries within the table. (Column-level
 * access rights don't enter into it: you can't destroy only a column.)
 * Checking for destroy rights on the table is easy; checking for destroy
 * rights on each entry can be a pain if the table is large.
 */

static nis_error nis_ib_remove(name, flags, nattrs, attrs, obj, t, princ, res)
	nis_name		name;
	uint32_t		flags;
	uint32_t		nattrs;
	nis_attr		*attrs;
	nis_object		*obj;
	nis_object		*t;
	nis_name		princ;
	nis_result		*res;
{
	struct db_objs		dres, dres_rm;
	int			i;
	nis_object		*o;
	db_status		db_stat = DB_INTERNAL_ERROR;
	nis_error		t_right = NIS_PERMISSION;
	netobj			cookie;

	t_right = nis_can_modify(princ, t);

	if (!nattrs && (flags & REM_MULTIPLE)) {
		db_stat = db_ibfirst(name, NULL, 0, &dres);
		res->dticks += dres.dticks;
		if (db_stat != NIS_SUCCESS)
			return(db_to_nis_err(db_stat));
		while(db_stat == DB_SUCCESS) {
			o = NIS_RES_OBJECT(&dres);
			if (t_right != NIS_SUCCESS &&
					nis_can_delete(princ, o)) {
				nis_free_dobjs(&dres);
				return(NIS_PERMISSION);
			}
			nis_mklog(name, o, t, REM_IBASE);
			db_stat = db_ibremove(name, NULL, 0, o, t, &dres_rm);
			res->dticks += dres_rm.dticks;
			if (db_stat != DB_SUCCESS) {
				nis_free_dobjs(&dres);
				return(db_to_nis_err(db_stat));
			}
			nis_free_dobjs(&dres_rm);
			cookie.n_len = dres.cookie.n_len;
			cookie.n_bytes = dres.cookie.n_bytes;
			dres.cookie.n_len = 0;
			dres.cookie.n_bytes = NULL;
			nis_free_dobjs(&dres);
			db_stat = db_ibnext(name, &cookie, &dres);
			res->dticks += dres.dticks;
		}
		/* DB_NOTFOUND means we hit the end of the table */
		if (db_stat == DB_NOTFOUND)
			return(NIS_SUCCESS);
		return(db_to_nis_err(db_stat));
	} else if (!nattrs && !(flags & REM_MULTIPLE))
		return(NIS_BADREQUEST);

	db_stat = db_iblookup(name, attrs, nattrs, &dres);

	res->dticks += dres.dticks;

	if (db_stat != DB_SUCCESS)
		return(NIS_NOTFOUND);

	if (NIS_RES_NUMOBJ(&dres) > 1 && !(flags & REM_MULTIPLE)) {
		nis_free_dobjs(&dres);
		return(NIS_NOTUNIQUE);
	}

	for (i = 0; i < NIS_RES_NUMOBJ(&dres); i++) {
		o = &dres.objects.objects_val[i];
		if (t_right != NIS_SUCCESS && nis_can_delete(princ, o)) {
			nis_free_dobjs(&dres);
			return(NIS_PERMISSION);
		}
		nis_mklog(name, o, t, REM_IBASE);
		db_stat = db_ibremove(name, NULL, 0, o, t, &dres_rm);
		res->dticks += dres_rm.dticks;
		if (db_stat != DB_SUCCESS) {
			nis_free_dobjs(&dres);
			return(db_to_nis_err(db_stat));
		}
		nis_free_dobjs(&dres_rm);
	}

	nis_free_dobjs(&dres);

	return(NIS_SUCCESS);
}


nis_error nis_db_ibop(name, flags, numattrs, attrs,
			obj, principal, op, result)
	nis_name		name;
	uint32_t		flags;
	uint32_t		numattrs;
	nis_attr		*attrs;
	nis_object		*obj;
	nis_name		principal;
	uint32_t		op;
	nis_result		*result;
{
	uint32_t		xid;
	nis_object		*t;
	nis_error		cerr;
	struct entryattrs	*ea = NULL;
	nis_attr		*a = NULL; /* local attr pointer */
	uint32_t		n = 0;  /* local attr count */

	/*
	 * First, do some sanity checks.
	 */
	if (nis_validname(name) != NIS_SUCCESS)
		return(NIS_BADNAME);

	if (obj != NULL && __type_of(obj) != ENTRY_OBJ)
		return(NIS_INVALIDOBJ);

	/*
	 * Dig up the table object.
	 */
	cerr = __nis_ocache_lookup(name, CHECK_MASTER, 0, &t);

	if (cerr != NIS_SUCCESS) {
		if (cerr == NIS_NOTFOUND)
			return(NIS_NOSUCHTABLE);
		return(cerr);
	}

	if (__type_of(t) != TABLE_OBJ)
		return(NIS_NOTSEARCHABLE);

	if (numattrs && numattrs > t->TA_data.ta_cols.ta_cols_len)
		return(NIS_TOOMANYATTRS);

	if (obj != NULL && strcmp(obj->EN_data.en_type, t->TA_data.ta_type))
		return(NIS_TYPEMISMATCH);

	/*
	 * Must have one or the other, _unless_ the operation
	 * is NIS_IBREMOVE. An IBREMOVE with no search criteria
	 * and the REM_MULTIPLE flag set means remove all entries
	 * in a table.
	 */
	if ((obj == NULL && !numattrs) && op != NIS_IBREMOVE)
		return(NIS_BADREQUEST);

	/*
	 * Attributes are typically optional, but if not
	 * supplied by the caller, we need to fabricate them
	 * ourselves (provided the caller provided an object that
	 * we can use as a template).
	 */
	if (numattrs) {
		a = attrs;
		n = numattrs;
	} else {
		if (obj != NULL) {
			ea = nis_entry_to_attr(&obj->EN_data, &t->TA_data);
			a = ea->attrs;
			n = ea->cnt;
		}
	}

	/*
	 * Allocate a transaction log ID.
	 */
	nis_log_new_xid(principal, &xid);

	switch(op) {
	case NIS_IBADD:
		cerr = nis_ib_add(name, flags, n, a, obj, t,
					principal, result);
		break;
	case NIS_IBMODIFY:
		cerr = nis_ib_modify(name, flags, n, a, obj, t,
					principal, result);
		break;
	case NIS_IBREMOVE:
		cerr = nis_ib_remove(name, flags, n, a, obj, t,
					principal, result);
		break;
	default:
		nis_err("invalid information base op: %d", op);
		cerr = NIS_BADREQUEST;
		break;
	}

	if (ea != NULL)
		free(a);

	if (cerr != NIS_SUCCESS) {
		nis_log_abort_xid(xid);
		return(cerr);
	}

	nis_log_commit_xid(xid);
	__nis_flush_one_dir(name);
	nis_master_add_update(name);

	return(NIS_SUCCESS);
}
