/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nis_db_ns.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $
 */

#include <sys/types.h>
#include <sys/cdefs.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nis_db.h>
#include "nis_extern.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nis_db_ns.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $";
#endif

/*
 * The 'namespace' functions.
 *
 * These are used to search and modify the NIS+ 'namespace.' In other
 * words, you use these functions to make new directories, add links
 * between 'objects,' update NIS+ group information, or just generally
 * add/delete/modify NIS+ 'objects.' What you can't do is use these
 * functions to play with entries within tables: for that you need the
 * 'information base' functions, which are in seperate module.
 */

nis_error nis_db_nslookup(name, res, princ)
	nis_name		name;
	nis_result		*res;
	nis_name		princ;
{
	nis_error		cerr;
	db_status		db_stat;
	struct db_objs		dres;
	nis_object		*p, *o;
	DATABASE_CLOCK_DECLS;

	/*
	 * First, do some sanity checks.
	 */
	if (nis_validname(name) != NIS_SUCCESS)
		return(NIS_BADNAME);

	/* We should never see an indexed name. */
	if (strchr(name, '[') != NULL)
		return(NIS_BADNAME);

	/*
	 * If this is the root object we're dealing with,
	 * handle it specially.
	 */
	if (nis_dir_cmp(name, rootname) == SAME_NAME) {
		if (root_obj == NULL)
			return(NIS_NOT_ME);
		__start_clock(DATABASE_CLOCK);
		__nis_ocache_insert(root_obj);
		NIS_RES_OBJECT(res) = nis_clone_object(root_obj, NULL);
		NIS_RES_NUMOBJ(res) = 1;
		res->dticks = __stop_clock(DATABASE_CLOCK);
		return(NIS_SUCCESS);
	}

	/*
	 * If the name is 'above' the root directory, then
	 * the caller is looking for a name outside the namespace
	 * that this server recognizes. We should return the
	 * parent object here.
	 */
	if (nis_dir_cmp(name, rootname) == HIGHER_NAME) {
#ifdef notyet
		NIS_RES_OBJECT(res) = nis_clone_object(parent_obj, NULL);
		NIS_RES_NUMOBJ(res) = 1;
		return(NIS_FOREIGNNS);
#endif
		return(NIS_NOSUCHNAME);
	}

	/* Find the directory object where this name resides. */
	cerr = __nis_ocache_lookup(nis_domain_of(name), 0, 0, &p);

	if (cerr != NIS_SUCCESS)
		return(NIS_NOSUCHNAME);

	/*
	 * Check access rights: caller must have read
	 * rights on the parent directory.
	 */
	if (nis_can_read(princ, p))
		return(NIS_PERMISSION);

	/*
	 * Now find the name. It might be in the cache,
	 * so it wouldn't hurt to look there first. Make
	 * sure not to let the cache code roll over to
	 * nis_lookup() though.
	 */
	cerr = __nis_ocache_lookup(name, CACHE_ONLY|CHECK_SERVER, 0, &o);
	if (cerr != NIS_SUCCESS) {
		db_stat = db_lookup(name, &dres);
		res->dticks = dres.dticks;

		if (db_stat != NIS_SUCCESS)
			return(db_to_nis_err(db_stat));

		NIS_RES_OBJECT(res) = NIS_RES_OBJECT(&dres);
		NIS_RES_NUMOBJ(res) = NIS_RES_NUMOBJ(&dres);

		/* Save the result in the cache */
		__nis_ocache_insert(NIS_RES_OBJECT(&dres));
	} else {
		NIS_RES_OBJECT(res) = nis_clone_object(o, NULL);
		NIS_RES_NUMOBJ(res) = 1;
	}

	return(NIS_SUCCESS);
}

/*
 * Handler for adding a new object to the namespace.
 */
static nis_error nis_ns_add(name, obj, pdir, princ, res)
	nis_name		name;
	nis_object		*obj;
	nis_object		*pdir;
	nis_name		princ;
	nis_result		*res;
{
	db_status		db_stat;
	struct db_objs		dres;
	nis_object		*o;
	nis_error		cerr;

	if (nis_can_create(princ, pdir))
		return(NIS_PERMISSION);

	/* Make sure an object with the same name doesn't already exist. */
	cerr = __nis_ocache_lookup(name, 0, 0, &o);

	if (cerr == NIS_SUCCESS)
		return(NIS_NAMEEXISTS);
	/* If some other unexpected error cropped up, bail. */
/*	if (cerr != NIS_NOTFOUND && cerr != NIS_NOSUCHNAME)
		return(cerr);
*/

	/* Update the OID. */
	obj->zo_oid.ctime = time(NULL);
	obj->zo_oid.mtime = obj->zo_oid.ctime;

	nis_log_submit(name, obj, NULL, 0, ADD_NAME);

	db_stat = db_add(name, obj, &dres);

	res->dticks = dres.dticks;

	if (db_stat != DB_SUCCESS)
		return(db_to_nis_err(db_stat));


	return(NIS_SUCCESS);
}

/*
 * Handler for modifying an object in the namespace.
 */
static nis_error nis_ns_modify(name, obj, pdir, princ, res)
	nis_name		name;
	nis_object		*obj;
	nis_object		*pdir;
	nis_name		princ;
	nis_result		*res;
{
	db_status		db_stat;
	struct db_objs		dres;
	nis_error		cerr;
	nis_object		*o;

	/*
	 * For modification, we have to test the access rights of
	 * the object itself as opposed to the directory in which
	 * the object resides. Hence, we need to grab a copy of it.
	 */
	cerr = __nis_ocache_lookup(name, CHECK_MASTER, MASTER_ONLY, &o);

	if (cerr != NIS_SUCCESS)
		return(cerr);

	if (nis_can_modify(princ, o))
		return(NIS_PERMISSION);

	if (obj->zo_oid.ctime != o->zo_oid.ctime ||
	    obj->zo_oid.mtime != o->zo_oid.mtime)
		return(NIS_NOTSAMEOBJ);

	obj->zo_oid.ctime = o->zo_oid.ctime;
	obj->zo_oid.mtime = time(NULL);

	nis_log_submit(name, o, NULL, 0, MOD_NAME_OLD);
	nis_log_submit(name, obj, NULL, 0, MOD_NAME_NEW);

	db_stat = db_add(name, obj, &dres);

	res->dticks = dres.dticks;

	if (db_stat != DB_SUCCESS)
		return(db_to_nis_err(db_stat));

	if (nis_dir_cmp(name, nis_local_directory()) == SAME_NAME &&
		__type_of(obj) == DIRECTORY_OBJ)
		nis_put_cold_start(&obj->DI_data);

	return(NIS_SUCCESS);
}

/*
 * Handler for removing an object from the namespace.
 */
static nis_error nis_ns_remove(name, obj, pdir, princ, res)
	nis_name		name;
	nis_object		*obj;
	nis_object		*pdir;
	nis_name		princ;
	nis_result		*res;
{
	db_status		db_stat;
	struct db_objs		dres;
	nis_error		cerr;
	nis_object		*o;

	if (nis_can_delete(princ, pdir))
		return(NIS_PERMISSION);

	/*
	 * Need the existing object here for a sanity test, and
	 * for transaction logging in case something goes Horribly
	 * Wrong (tm).
	 */
	cerr = __nis_ocache_lookup(name, CHECK_MASTER, MASTER_ONLY, &o);

	if (cerr != NIS_SUCCESS)
		return(cerr);

	/*
	 * If the object is a table, the table must be empty
	 * otherwise we can't remove it. Now, logically, you'd
	 * think this restriction would apply to directories too,
	 * but nooooooooo! That's not how Sun did it. The test to
	 * avoid the removal of directories containing subdirectories
	 * is done by nis_rmdir_3_svc() instead. Go figure.
	 */
	if (__type_of(o) == TABLE_OBJ) {
		db_stat = db_ibfirst(name, NULL, 0, &dres);
		res->dticks += dres.dticks;
		if (db_stat == DB_SUCCESS) {
			nis_free_dobjs(&dres);
			return(NIS_NOTEMPTY);
		}
	}

	nis_log_submit(name, o, NULL, 0, REM_NAME);

	db_stat = db_remove(name, &dres);

	res->dticks = dres.dticks;

	if (db_stat != DB_SUCCESS)
		return(db_to_nis_err(db_stat));

	return(NIS_SUCCESS);
}

/*
 * Special case handler for operations done on the
 * root object (only applies if we are a root master or,
 * in some cases, a root replica server). Note that we do
 * _not_ perform any transaction logging for operations on
 * the root object. Personally I think that's a mistake, but
 * that's how the Sun server does it.
 */
static nis_error nis_ns_rootop(name, obj, princ, op, res)
	nis_name		name;
	nis_object		*obj;
	nis_name		princ;
	uint32_t		op;
	nis_result		*res;
{

	/* First, see if we actually serve the root directory. */
	if (__is_local_dirname(name) == FALSE)
		return(NIS_NOTMASTER);

	/* Now see if we have a copy of the root object. */
	if (root_obj == NULL)
		return(NIS_NOTMASTER);

	/* Now make sure we really, really are the master server. :) */
	if (__i_am_master(&root_obj->DI_data) == FALSE)
		return(NIS_NOTMASTER);

	switch(op) {
	case NIS_ADD:
		/*
		 * The initial copy of the root object is always created
		 * by nisinit(8) when the NIS+ service is first initialized;
		 * it is never created by the master server. Replica servers
		 * may create a root object after being configured as
		 * root replicas, but this is done through the nis_dump()
		 * or nis_dumplog() facility rather than by an explicit
		 * nis_add(). Thus, a request to add the root object makes
		 * absolutely no sense and is probably a sign that the
		 * caller is insane.
		 */
		return(NIS_NAMEEXISTS);
		break;
	case NIS_MODIFY:
		/*
		 * The root object may be modified on the master when a
		 * root replica server is added (the server list in the
		 * root directory object has to be updated). Permission
		 * checks are still enforced to avoid chaos.
		 */
		if (nis_can_modify(princ, root_obj))
			return(NIS_PERMISSION);
		obj->zo_oid.ctime = root_obj->zo_oid.ctime;
		obj->zo_oid.mtime = time(NULL);
		nis_put_root_obj(obj);
		nis_get_root_obj(&root_obj);
		break;
	case NIS_REMOVE:
		/*
		 * I'm not altogether sure about this, but: if we
		 * are a root replica, it is conceivable that we could
		 * be removed as a replica, in which case we'd no longer
		 * need a copy of the root object around. Note that we
		 * check permissions using the root object itself rather
		 * that the permissions of the parent directory since
		 * the root directory, by definition, can't have a parent.
		 */
		if (nis_can_delete(princ, root_obj))
			return(NIS_PERMISSION);
		if (__i_am_replica(&root_obj->DI_data) == FALSE)
			return(NIS_BADREQUEST);
		nis_destroy_object(root_obj);
		root_obj = NULL;
		unlink(ROOT_OBJ);
		break;
	default:
		nis_err("invalid namespace rootop: %d", op);
		return(NIS_BADREQUEST);
		break;
	}

	__nis_flush_one_dir(name);
	nis_master_add_update(name);

	return(NIS_SUCCESS);
}

/*
 * Do a namespace update operation (add, modify, remove).
 * Note that we have to do special things for operations
 * that concern the root object (if we are a root master) so
 * we attempt to trap these out here.
 */
nis_error nis_db_nsop(name, obj, principal, op, result)
	nis_name		name;
	nis_object		*obj;
	nis_name		principal;
	uint32_t		op;
	nis_result		*result;
{
	nis_error		cerr;
	uint32_t		xid;
	nis_object		*o;

	/*
	 * First, do some sanity checks.
	 */
	if (nis_validname(name) != NIS_SUCCESS)
		return(NIS_BADNAME);

	/* We should never see an entry here. */
	if (obj != NULL && __type_of(obj) == ENTRY_OBJ)
		return(NIS_BADOBJECT);

	/* We should never see an indexed name either. */
	if (strchr(name, '[') != NULL)
		return(NIS_BADNAME);

	/* If zo_name exists, it must be a leaf (i.e. no dots). */
	if (obj != NULL && strchr(obj->zo_name, '.') != NULL)
		return(NIS_INVALIDOBJ);

	/*
	 * If this is the root object we're dealing with,
	 * handle it specially.
	 */
	if (nis_dir_cmp(name, rootname) == SAME_NAME)
		return(nis_ns_rootop(name, obj, principal, op, result));

	/*
	 * Find the parent directory object. We need
	 * this for all the operations that change the
	 * namespace, usually for permissions tests.
	 * And if the parent directory object doesn't
	 * exist, we're in trouble and can't proceed
	 * with the operation anyway. (This also verifies
	 * that we are in fact the master server for the
	 * directory to be updated.)
	 */
	cerr = __nis_ocache_lookup(nis_domain_of(name), CHECK_MASTER,
							MASTER_ONLY, &o);

	if (cerr != NIS_SUCCESS) {
		nis_err("ns_op: failed to find parent object!");
		return(cerr);
	}

	/* Allocate a transaction log ID */
	nis_log_new_xid(principal, &xid);

	switch(op) {
	case NIS_ADD:
		cerr = nis_ns_add(name, obj, o, principal, result);
		break;
	case NIS_REMOVE:
		cerr = nis_ns_remove(name, obj, o, principal, result);
		break;
	case NIS_MODIFY:
		cerr = nis_ns_modify(name, obj, o, principal, result);
		break;
	default:
		nis_err("invalid namespace op: %d", op);
		cerr = NIS_BADREQUEST;
		break;
	}

	/* Operation failed: bail out */
	if (cerr != NIS_SUCCESS) {
		nis_log_abort_xid(xid);
		return(cerr);
	}

	/*
	 * Operation was successful: commit the transaction log,
	 * flush the name from the cache (to refresh it), and frob
	 * the nis_master module to have it propagate changes
	 * to any replicas that may exist.
	 */
	nis_log_commit_xid(xid);
	__nis_flush_one_dir(name);
	nis_master_add_update(name);

	return(NIS_SUCCESS);
}
