/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nis_event.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $
 */

#include <sys/types.h>
#include <sys/cdefs.h>
#include <sys/time.h>
#include <sys/queue.h>

#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nis_extern.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nis_event.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $";
#endif

/*
 * This module contains the server event handler. The event handler
 * is used to schedule tasks that need to be performed after a certain
 * delay, events that need to be performed at regular intervals, or
 * events that need to be triggered when data is pending on a socket.
 * This provides a generic mechanism for doing certain tasks such as
 * standing by tables that haven't been recently used, pinging replica
 * servers, scheduling replica updates in response to pings, scheduling
 * checkpoints, driving the callback module, flushing various caches,
 * and other things.
 *
 * One might consider this a poor man's approach to threads, only
 * without a lot of the hassle normally associated with threads.
 * Maybe once a particular threads package becomes standardized,
 * widespread and stable I might actually start using them; until
 * then I prefer to handle this stuff myself, if you don't mind.
 */

#define NIS_ETYPE_CYCLIC	1	/* event reoccurs at interval */
#define NIS_ETYPE_ONESHOT	2	/* event goes off once */
#define NIS_ETYPE_CONDITIONAL	3	/* event fires if cb returns TRUE */
#define NIS_ETYPE_SOCK		4	/* event depends on socket I/O */


/*
 * Circular queue definitions.
 */
#define CIRCLEQ_HEAD(name, type)					\
struct name {								\
	struct type *cqh_first;		/* first element */		\
	struct type *cqh_last;		/* last element */		\
}

#define CIRCLEQ_ENTRY(type)						\
struct {								\
	struct type *cqe_next;		/* next element */		\
	struct type *cqe_prev;		/* previous element */		\
}

/*
 * Circular queue functions.
 */
#define	CIRCLEQ_INIT(head) {						\
	(head)->cqh_first = (void *)(head);				\
	(head)->cqh_last = (void *)(head);				\
}

#define CIRCLEQ_INSERT_AFTER(head, listelm, elm, field) {		\
	(elm)->field.cqe_next = (listelm)->field.cqe_next;		\
	(elm)->field.cqe_prev = (listelm);				\
	if ((listelm)->field.cqe_next == (void *)(head))		\
		(head)->cqh_last = (elm);				\
	else								\
		(listelm)->field.cqe_next->field.cqe_prev = (elm);	\
	(listelm)->field.cqe_next = (elm);				\
}

#define CIRCLEQ_INSERT_BEFORE(head, listelm, elm, field) {		\
	(elm)->field.cqe_next = (listelm);				\
	(elm)->field.cqe_prev = (listelm)->field.cqe_prev;		\
	if ((listelm)->field.cqe_prev == (void *)(head))		\
		(head)->cqh_first = (elm);				\
	else								\
		(listelm)->field.cqe_prev->field.cqe_next = (elm);	\
	(listelm)->field.cqe_prev = (elm);				\
}

#define CIRCLEQ_INSERT_HEAD(head, elm, field) {				\
	(elm)->field.cqe_next = (head)->cqh_first;			\
	(elm)->field.cqe_prev = (void *)(head);				\
	if ((head)->cqh_last == (void *)(head))				\
		(head)->cqh_last = (elm);				\
	else								\
		(head)->cqh_first->field.cqe_prev = (elm);		\
	(head)->cqh_first = (elm);					\
}

#define CIRCLEQ_INSERT_TAIL(head, elm, field) {				\
	(elm)->field.cqe_next = (void *)(head);				\
	(elm)->field.cqe_prev = (head)->cqh_last;			\
	if ((head)->cqh_first == (void *)(head))			\
		(head)->cqh_first = (elm);				\
	else								\
		(head)->cqh_last->field.cqe_next = (elm);		\
	(head)->cqh_last = (elm);					\
}

#define	CIRCLEQ_REMOVE(head, elm, field) {				\
	if ((elm)->field.cqe_next == (void *)(head))			\
		(head)->cqh_last = (elm)->field.cqe_prev;		\
	else								\
		(elm)->field.cqe_next->field.cqe_prev =			\
		    (elm)->field.cqe_prev;				\
	if ((elm)->field.cqe_prev == (void *)(head))			\
		(head)->cqh_first = (elm)->field.cqe_next;		\
	else								\
		(elm)->field.cqe_prev->field.cqe_next =			\
		    (elm)->field.cqe_next;				\
}
/*
 * The event_time is the actual time that this particular event
 * will next occur. The event_interval is how oftern a cyclic event
 * will occur, in general.
 */
struct nis_event {
	uint32_t		event_id;
	char			*event_desc;
	int			event_type;
	time_t			event_time;
	time_t			event_interval;
	bool_t			(*event_callback)(void *, uint32_t, int);
	void			*event_args;
	void			(*event_freeargs)(void *);
	int			event_sock;
};

static int event_id_seed = 0;

static CIRCLEQ_HEAD(nisq_evhead, nisq_event) nisevhead;

struct nisq_event {
	struct nis_event	*eptr;
	CIRCLEQ_ENTRY(nisq_event) links;
};

void nis_event_init()
{
	CIRCLEQ_INIT(&nisevhead);
	return;
}

static struct nisq_event *nis_event_malloc()
{
	register struct nisq_event	*q;

	q = calloc(1, sizeof(struct nisq_event));
	if (q == NULL) {
		nis_err("out of memory allocating event");
		return(NULL);
	}
	q->eptr = calloc(1, sizeof(struct nis_event));
	if (q->eptr == NULL) {
		free(q);
		nis_err("out of memory allocating event");
		return(NULL);
	}
	return(q);
}

static void nis_event_free(q)
	struct nisq_event	*q;
{
	if (q == NULL)
		return;
	if (q->eptr != NULL) {
		if (verbose)
			nis_err("destroying event #%d", q->eptr->event_id);
		if (q->eptr->event_freeargs != NULL &&
			q->eptr->event_args != NULL)
			(*q->eptr->event_freeargs)(q->eptr->event_args);
		if (q->eptr->event_desc != NULL)
			free(q->eptr->event_desc);
		free(q->eptr);
	}
	free(q);
	return;
}

#define MAXDELAY	0x7FFFFFFF

int nis_event_sched(fds, timeout)
	fd_set			*fds;
	struct timeval		*timeout;
{
	register struct nisq_event	*q, *n;
	register struct nis_event	*e;
	time_t				delay = MAXDELAY;
	time_t				now;
        int                     highfd = 0;

	time(&now);
#ifdef NOISY
	if (verbose)
		nis_err("scheduling events at %s", nis_ctime(now));
#endif
	q = nisevhead.cqh_first;
	while(q != (void *)&nisevhead) {
		n = q->links.cqe_next;
		e = q->eptr;
		switch(e->event_type) {
		case NIS_ETYPE_CYCLIC:
#ifdef NOISY
			if (verbose)
				nis_err("\"%s\" in %d seconds (cyclic)",
					e->event_desc, e->event_interval);
#endif
			if ((e->event_time - now) < delay)
				delay = e->event_time - now;
			break;
		case NIS_ETYPE_ONESHOT:
#ifdef NOISY
			if (verbose)
				nis_err("\"%s\" in %d seconds (oneshot)",
					e->event_desc, e->event_interval);
#endif
			if ((e->event_time - now) < delay)
				delay = e->event_time - now;
			break;
		case NIS_ETYPE_CONDITIONAL:	
#ifdef NOISY
			if (verbose)
				nis_err("\"%s\" in %d seconds (conditional)",
					e->event_desc, e->event_interval);
#endif
			if ((e->event_time - now) < delay)
				delay = e->event_time - now;
			break;
		case NIS_ETYPE_SOCK:
#ifdef NOISY
			if (verbose)
				nis_err("\"%s\" in %d seconds (socket)",
					e->event_desc, e->event_interval);
#endif
			if ((e->event_time - now) < delay)
				delay = e->event_time - now;
			FD_SET(e->event_sock, fds);
			if (e->event_sock > highfd)
				highfd = e->event_sock;
			break;
		default:
			nis_err("bad event type: %d (%s)", e->event_type,
							e->event_desc);
			CIRCLEQ_REMOVE(&nisevhead, q, links);
			nis_event_free(q);
		}
		q = n;
	}

	if (delay == MAXDELAY) {
		timeout->tv_sec = 0;
		timeout->tv_usec = 0;
	} else {
		timeout->tv_sec = delay;
		timeout->tv_usec = 0;
	}

	return(highfd);
}

void nis_event_runq(ev_set)
	fd_set			*ev_set;
{
	register struct nisq_event	*q, *n;
	register struct nis_event	*e;

	time_t			now;

	time(&now);
#ifdef NOISY
	if (verbose)
		nis_err("running scheduled events at %s", nis_ctime(now));
#endif
	q = nisevhead.cqh_first;
	while(q != (void *)&nisevhead) {
		n = q->links.cqe_next;
		e = q->eptr;
		switch(e->event_type) {
		case NIS_ETYPE_CYCLIC:
			if (now >= e->event_time) {
				if (verbose)
					nis_err("\"%s\" (id %d) (cyclic)",
						e->event_desc, e->event_id);
				(*e->event_callback)(e->event_args,
						     e->event_id, 0);
			}
			e->event_time = now + e->event_interval;
			break;
		case NIS_ETYPE_ONESHOT:
			if (now >= e->event_time) {
				if (verbose)
					nis_err("\"%s\" (id %d) (oneshot)",
						e->event_desc, e->event_id);
				(*e->event_callback)(e->event_args,
						     e->event_id, 0);
				CIRCLEQ_REMOVE(&nisevhead, q, links);
				nis_event_free(q);
			}
			break;
		case NIS_ETYPE_CONDITIONAL:
			if (now >= e->event_time) {
				if (verbose)
					nis_err("\"%s\" (id %d) (conditional)",
						e->event_desc, e->event_id);
				if ((*e->event_callback)(e->event_args,
						e->event_id, 0) != TRUE) {
					CIRCLEQ_REMOVE(&nisevhead, q, links);
					nis_event_free(q);
				} else
					e->event_time = now + e->event_interval;
			}
			break;
		case NIS_ETYPE_SOCK:
			if (FD_ISSET(e->event_sock, ev_set)) {
				if (verbose)
					nis_err("\"%s\" (id %d) (socket)",
						e->event_desc, e->event_id);
				if ((*e->event_callback)(e->event_args,
						e->event_id, 1) != TRUE) {
					CIRCLEQ_REMOVE(&nisevhead, q, links);
					nis_event_free(q);
					break;
				} else
					e->event_time = now + e->event_interval;
			}
			if (now >= e->event_time) {
				if ((*e->event_callback)(e->event_args,
						e->event_id, 0) != TRUE) {
					CIRCLEQ_REMOVE(&nisevhead, q, links);
					nis_event_free(q);
				} else
					e->event_time = now + e->event_interval;
			}
			break;
		default:
			nis_err("bad event type: %d", e->event_type);
			CIRCLEQ_REMOVE(&nisevhead, q, links);
			nis_event_free(q);
			break;
		}
		q = n;
	}
}

static uint32_t nis_event_add(cb, args, freeargs,
				sock,interval, type, desc)
	bool_t			(*cb)(void *, uint32_t, int);
	void			*args;
	void			(*freeargs)(void *);
	int			sock;
	int			interval;
	int			type;
	char			*desc;
{
	register struct nisq_event	*q;
	register struct nis_event	*e;

	q = nis_event_malloc();
	if (q == NULL)
		return(0);

	event_id_seed++;
	e = q->eptr;
	e->event_id = event_id_seed;
	if (desc != NULL)
		e->event_desc = strdup(desc);
	else
		e->event_desc = strdup("<unknown event>");
	e->event_type = type;
	e->event_time = time(NULL) + interval;
	e->event_interval = interval;
	e->event_callback = cb;
	e->event_args = args;
	e->event_freeargs = freeargs;
	e->event_sock = sock;

	CIRCLEQ_INSERT_TAIL(&nisevhead, q, links);

	return(e->event_id);
}

uint32_t nis_event_add_cyclic(cb, args, freeargs, interval, desc)
	bool_t			(*cb)(void *, uint32_t, int);
	void			*args;
	void			(*freeargs)(void *);
	int			interval;
	char			*desc;
{
	if (verbose)
		nis_err("adding new cyclic event (%s)", desc);

	return(nis_event_add(cb, args, freeargs, -1, interval,
				NIS_ETYPE_CYCLIC, desc));
}

uint32_t nis_event_add_oneshot(cb, args, freeargs, delay, desc)
	bool_t			(*cb)(void *, uint32_t, int);
	void			*args;
	void			(*freeargs)(void *);
	int			delay;
	char			*desc;
{
	if (verbose)
		nis_err("adding new oneshot event (%s)", desc);

	return(nis_event_add(cb, args, freeargs, -1, delay,
				NIS_ETYPE_ONESHOT, desc));
}

uint32_t nis_event_add_conditional(cb, args, freeargs, delay, desc)
	bool_t			(*cb)(void *, uint32_t, int);
	void			*args;
	void			(*freeargs)(void *);
	int			delay;
	char			*desc;
{
	if (verbose)
		nis_err("adding new conditional event (%s)", desc);

	return(nis_event_add(cb, args, freeargs, -1, delay,
				NIS_ETYPE_CONDITIONAL, desc));
}

uint32_t nis_event_add_sock(cb, args, freeargs, delay, sock, desc)
	bool_t			(*cb)(void *, uint32_t, int);
	void			*args;
	void			(*freeargs)(void *);
	int			delay;
	int			sock;
	char			*desc;
{
	if (verbose)
		nis_err("adding new socket I/O event (%s)", desc);

	return(nis_event_add(cb, args, freeargs, sock, delay,
				NIS_ETYPE_SOCK, desc));
}

uint32_t nis_event_find(desc, args)
	char			*desc;
	void			**args;
{
	register struct nisq_event	*q;

	q = nisevhead.cqh_first;
	while(q != (void *)&nisevhead) {
		if (strcmp(q->eptr->event_desc, desc) == NULL) {
			*args = q->eptr->event_args;
			return(q->eptr->event_id);
		}
		q = q->links.cqe_next;
	}

	return(0);
}

void nis_event_del(id)
	uint32_t		id;
{
	register struct nisq_event	*q, *n;

	q = nisevhead.cqh_first;
	while(q != (void *)&nisevhead) {
		n = q->links.cqe_next;
		if (q->eptr->event_id == id) {
			CIRCLEQ_REMOVE(&nisevhead, q, links);
			nis_event_free(q);
			return;
		}
		q = n;
	}

	return;
}
