/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nis_log.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $
 */

#include <sys/types.h>
#include <sys/cdefs.h>
#include <sys/fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <rpcsvc/nis.h>
#include "nis_extern.h"
#include "nis_log.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nis_log.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $";
#endif

/*
 * This module contains the guts of the transaction logging system.
 * The NIS+ server logs all changes made to the namespace and to "information
 * bases" so that replica servers can replay the log and synchronize
 * themselves with the master.
 * 
 * There are several basic things we need to do:
 * - Create a new empty log if one doesn't exist.
 * - Add a new entry to the log
 * - Dump out log entries so that replicas can copy them
 * - Roll back an operation from the log
 * - Nuke log entries once replicas are all up to date
 *
 * My original notion was to do this with the same stdio file stream
 * stuff I did in libnisdb, but from what I can glean from the (grammatically
 * incorrect and badly punctuated) comments in nis.x we need a more complex
 * mechanism. In the server, the log is used to mark the fact that a
 * transaction has begun, and then later to indicate that it has failed
 * or succeeded. Also, for namespace changes, we have both old and new
 * log entries (where old == previous value and new == updated value).
 * This means that for name changes, one transaction leads to two log
 * entries. It's hard to do that with file streams and stay reentrant.
 *
 * Consequently, we use mmap(). I'm a little leery of this since mmap()
 * has had problems on FreeBSD before, but I can't think of a better way
 * to do it. Note that even though we use XDR to encode log entries into
 * our mmap()ed memory segment, we also include non-XDR'd data (the header
 * and spacer structs). Consequently, the log file is _NOT_ machine
 * independent.
 */

#ifndef _PATH_NIS_LOG
#define _PATH_NIS_LOG "/var/nis/trans.log"
#endif

#ifndef _PATH_NIS_LOG_TMP
#define _PATH_NIS_LOG_TMP "/var/nis/trans.log.tmp"
#endif

static int			log_fd;	/* Log file descriptor */
static struct log_header	*hptr;	/* Pointer to log file header */
static struct log_spacer	*sptr;	/* Pointer to most recent spacer */
static struct log_spacer	*cptr;	/* Current spacer (placeholder) */
static caddr_t			laddr;	/* Address returned by mmap() */
static off_t			lsize;	/* Size of log file */

/*
 * Active XID and principal. If these are defined, there's a
 * transaction in progress. The existence of a single incarnation
 * of this value implies that the logging system is single threaded
 * since only one transaction may be active at a given time.
 * This is fine since the server itself is single threaded: we
 * never have more than one transaction active at any given
 * moment.
 */
#define LOG_INACTIVE		0xFFFFFFFF
static uint32_t		active_xid = LOG_INACTIVE;
static char			active_princ[NIS_MAXNAMELEN];
static struct log_spacer	*active_sptr = NULL;

struct stamps {
	nis_name		dirname;
	uint32_t		stamp;
	struct stamps		*next;
};

static struct stamps		*stlist = NULL;


static nis_error nis_log_mark(state)
	enum log_state		state;
{
	hptr->state = state;
	hptr->upd_time = time(NULL);

	if (msync(laddr, lsize, 0) == -1)
		nis_err("error msyncing log: %s", strerror(errno));

	return(NIS_SUCCESS);
}

nis_error nis_log_new_xid(princ, xid)
	nis_name		princ;
	uint32_t		*xid;
{
	/* see if a transaction is already in progress */
	if (active_xid != LOG_INACTIVE)
		return(NIS_TRYAGAIN);

	hptr->xid++;
	*xid = active_xid = hptr->xid;
	active_sptr = sptr;
	strncpy(active_princ, princ, NIS_MAXNAMELEN);

	nis_log_mark(LOG_UPDATING);

	return(NIS_SUCCESS);
}

nis_error nis_log_current_xid(xid)
	uint32_t		*xid;
{
	if (active_xid != LOG_INACTIVE) {
		*xid = active_xid;
		return(NIS_SUCCESS);
	}
	return(NIS_NOTFOUND);
}

/*
 * Extend the log file by another chunk.
 */
nis_error nis_log_extend()
{

	if (verbose)
		nis_err("extending log by another chunk");
	/*nis_log_mark(LOG_STABLE);*/
	if (munmap(laddr, lsize) == -1)
		nis_err("unmaping log failed: %s", strerror(errno));
	lsize += CHUNKSIZE;
	lseek(log_fd, lsize, SEEK_SET);
	write(log_fd, ".", 1);
	laddr = mmap(0, lsize, PROT_READ|PROT_WRITE, MAP_SHARED, log_fd, 0);
	/*
	 * Now that we've extended the file, we recalculate
	 * our address pointers just in case the file has been
	 * mmap()ed at a different location.
	 */
	hptr = (struct log_header *)laddr;
	sptr = (struct log_spacer *)(laddr + hptr->offset);
	while(sptr->next_offset) {
		caddr_t			ptr;
		if (sptr->magic != SPACER_MAGIC) {
			nis_err("extend: transaction log is corrupt!");
			return(NIS_FAIL);
		}
		ptr = (caddr_t)sptr;
		ptr += sptr->next_offset + sizeof(struct log_spacer);
		sptr = (struct log_spacer *)ptr;
	}

	/*
	 * If a transaction is pending, fix the active_sptr
	 * pointer too or chaos will ensue.
	 */
	if (active_xid != LOG_INACTIVE)
		active_sptr = sptr;

	return(NIS_SUCCESS);
}

/*
 * Write an entry to the log file, extending it by another chunk
 * if necessary.
 */
static nis_error nis_log_add_ent(ent)
	log_entry		*ent;
{
	XDR			xdrs;
	off_t			size, lestimate;
	caddr_t			ptr;
	int			noprinc = 0;


	if (ent->le_princp == NULL) {
		ent->le_princp = active_princ;
		noprinc++;
	}

	size = xdr_sizeof(xdr_log_entry, ent);
	/*
	 * Now that we know how much we'll be adding to the log,
	 * check that it's big enough to hold it all.
	 */
	lestimate = hptr->size + size + sizeof(struct log_spacer);
	if (lestimate >= lsize)
		nis_log_extend();
	sptr->next_offset = size;
	sptr->xid = active_xid;
	ptr = (caddr_t)(sptr) + sizeof(struct log_spacer);
	xdrmem_create(&xdrs, ptr, size, XDR_ENCODE);

	if (!xdr_log_entry(&xdrs, ent)) {
		xdr_destroy(&xdrs);
		nis_err("updating log failed!");
		if (noprinc)
			ent->le_princp = NULL;
		return(NIS_FAIL);
	}

	if (noprinc)
		ent->le_princp = NULL;

	xdr_destroy(&xdrs);
	sptr = (struct log_spacer *)(ptr + size);
	sptr->prev_offset = size;
	sptr->next_offset = 0;
	sptr->magic = SPACER_MAGIC;
	sptr->xid = active_xid;
	hptr->size += sizeof(struct log_spacer) + size;
	hptr->count++;

	if (msync(laddr, lsize, 0) == -1)
		nis_err("error msyncing log: %s", strerror(errno));

	return(NIS_SUCCESS);
}

void nis_log_insert_stamp(dirname, stamp)
	nis_name		dirname;
	time_t			stamp;
{
	struct stamps		*new = stlist;

	while(new != NULL) {
		if (nis_dir_cmp(dirname, new->dirname) == SAME_NAME) {
			new->stamp = stamp;
			return;
		}
		new = new->next;
	}
		
	new = malloc(sizeof(struct stamps));

	new->stamp = stamp;
	new->dirname = strdup(dirname);
	new->next = stlist;
	stlist = new;

	return;
}

void nis_log_free_stamps()
{
	struct stamps		*tmp;

	while(stlist != NULL) {
		tmp = stlist->next;
		free(stlist->dirname);
		free(stlist);
		stlist = tmp;
	}

	return;
}

void nis_log_commit_stamps()
{
	struct stamps		*tmp = stlist;
	log_entry		ent;

	bzero((char *)&ent, sizeof(ent));
	ent.le_type = UPD_STAMP;
	ent.le_princp = nis_local_principal();
	ent.le_object.zo_owner = "";
	ent.le_object.zo_group = "";
	ent.le_object.zo_name = "";
	ent.le_object.zo_oid.ctime = time(NULL);
	ent.le_object.zo_oid.mtime = ent.le_object.zo_oid.ctime;

	__type_of(&ent.le_object) = NO_OBJ;

	while(tmp != NULL) {
		ent.le_name = tmp->dirname;
		ent.le_time = tmp->stamp;
		ent.le_object.zo_domain = tmp->dirname;
		nis_log_add_ent(&ent);
		tmp = tmp->next;
	}

	return;
}

/*
 * Read back next entry in log.
 */
static nis_error nis_log_next_ent(sp, ent, new_sp)
	struct log_spacer	*sp;
	log_entry		**ent;
	struct log_spacer	**new_sp;
{
	caddr_t			ptr;
	XDR			xdrs;
	log_entry		*e;
	int			r;

	e = calloc(1, sizeof(log_entry));

	if (e == NULL)
		return(NIS_NOMEMORY);

	if (sp->magic != SPACER_MAGIC) {
		nis_err("next_ent: transaction log is corrupt!");
		return(NIS_FAIL);
	}
	ptr = (caddr_t)sp;
	ptr += sizeof(struct log_spacer);

	xdrmem_create(&xdrs, ptr, sp->next_offset, XDR_DECODE);
	r = xdr_log_entry(&xdrs, e);
	xdr_destroy(&xdrs);

	if (!r) {
		xdr_free(xdr_log_entry, (char *)e);
		free(e);
		return(NIS_FAIL);
	}

	ptr += sp->next_offset;
	*new_sp = (struct log_spacer *)ptr;
	*ent = e;

	return(NIS_SUCCESS);
}

/*
 * Read back previous entry in log.
 */
static nis_error nis_log_prev_ent(sp, ent, new_sp)
	struct log_spacer	*sp;
	log_entry		**ent;
	struct log_spacer	**new_sp;
{
	caddr_t			ptr;
	XDR			xdrs;
	log_entry		*e;
	int			r;

	e = calloc(1, sizeof(log_entry));

	if (e == NULL)
		return(NIS_NOMEMORY);

	if (sp->magic != SPACER_MAGIC) {
		nis_err("commit: transaction log is corrupt!");
		return(NIS_FAIL);
	}

	ptr = (caddr_t)sp;
	ptr -= sp->prev_offset;
	xdrmem_create(&xdrs, ptr, sp->prev_offset, XDR_DECODE);
	r = xdr_log_entry(&xdrs, e);
	xdr_destroy(&xdrs);

	if (!r) {
		xdr_free(xdr_log_entry, (char *)e);
		free(e);
		return(NIS_FAIL);
	}

	ptr -= sizeof(struct log_spacer);
	*new_sp = (struct log_spacer *)ptr;
	*ent = e;

	return(NIS_SUCCESS);
}

/*
 * Log an update under the currently active transaction ID
 * and principal.
 */
nis_error nis_log_submit(name, obj, attrs, cnt, action)
	nis_name		name;
	nis_object		*obj;
	nis_attr		*attrs;
	uint32_t		cnt;
	log_entry_t		action;
{
	log_entry		ent;
	nis_error		r;

	if (active_xid == LOG_INACTIVE)
		return(NIS_TRYAGAIN);

	bzero((char *)&ent, sizeof(ent));

	ent.le_type = action;
	ent.le_princp = active_princ;
	ent.le_name = name;
	ent.le_time = time(NULL);
	if (obj == NULL) {
		ent.le_object.zo_name = strdup(name);
		ent.le_object.zo_owner = "";
		ent.le_object.zo_group = "";
		ent.le_object.zo_domain = "";
	} else
		bcopy((char *)obj, (char *)&ent.le_object, sizeof(nis_object));
	ent.le_attrs.le_attrs_len = cnt;
	ent.le_attrs.le_attrs_val = attrs;

	r = nis_log_add_ent(&ent);

	if (obj == NULL)
		free(ent.le_object.zo_name);

	return(r);
}

static nis_error nis_log_mktmp(tmpname, h, n, f)
	nis_name		tmpname;
	struct log_header	*h;
	struct log_spacer	*n;
	FILE			**f;
{
	int			r;
	FILE			*fp;

	if (verbose)
		nis_err("creating temporary log file (%s)", tmpname);

	h->magic = HEADER_MAGIC;
	h->state = LOG_STABLE;
	h->count = 0;
	h->xid = 1;
	h->offset = sizeof(struct log_header);
	h->size = sizeof(struct log_header) +
			     sizeof(struct log_spacer);
	h->upd_time = 0;
	bzero((char *)n, sizeof(struct log_spacer));
	n->magic = SPACER_MAGIC;

	fp = fopen(tmpname, "w+");

	if (fp == NULL) {
		nis_err("failed to open temp log file: %s",
					strerror(errno));
		return(NIS_FAIL);
	}

	if ((r = fwrite((char *)h, sizeof(struct log_header), 1, fp)) < 0) {
		nis_err("write failed (%d): %s", r, strerror(errno));
		fclose(fp);
		return(NIS_FAIL);
	}

	if ((r = fwrite((char *)n, sizeof(struct log_spacer), 1, fp)) < 0) {
		nis_err("write failed (%d): %s", r, strerror(errno));
		fclose(fp);
		return(NIS_FAIL);
	}

	fseek(fp, sizeof(struct log_header), SEEK_SET);

	*f = fp;

	return(NIS_SUCCESS);
}

#define MATCH_BYXID		1
#define MATCH_BYNAME		2
#define MATCH_BYTIME		3

static bool_t keep(m, s, p, xid, name, t)
	int			m;
	struct log_spacer	*s;
	caddr_t			p;
	uint32_t		xid;
	nis_name		name;
	time_t			t;
{
	XDR			xdrs;
	log_entry		l;
	bool_t			k = FALSE;

	switch(m) {
	case MATCH_BYXID:
		if (s->xid != xid)
			k = TRUE;
		break;
	case MATCH_BYNAME:
		bzero((char *)&l, sizeof(l));
		xdrmem_create(&xdrs, p, s->next_offset, XDR_DECODE);
		xdr_log_entry(&xdrs, &l);
		xdr_destroy(&xdrs);
		if ((l.le_type == UPD_STAMP &&
		nis_dir_cmp(nis_domain_of(l.le_name), name) != SAME_NAME) &&
			nis_dir_cmp(l.le_name, name) != SAME_NAME)
			k = TRUE;
		xdr_free(xdr_log_entry, (char *)&l);
		break;
	case MATCH_BYTIME:
		bzero((char *)&l, sizeof(l));
		xdrmem_create(&xdrs, p, s->next_offset, XDR_DECODE);
		xdr_log_entry(&xdrs, &l);
		xdr_destroy(&xdrs);
		if (l.le_time != t)
			k = TRUE;
		xdr_free(xdr_log_entry, (char *)&l);
	default:
		break;
	}

	return(k);
}

static nis_error nis_log_copy(m, xid, name, t, h, n, fp)
	int			m;
	uint32_t		xid;
	nis_name		name;
	time_t			t;
	struct log_header	*h;
	struct log_spacer	*n;
	FILE			*fp;
{
	caddr_t			ptr;
	size_t			size = 0;
	struct log_spacer	*s;
	size_t			lpos = 0;

	if (verbose)
		nis_err("pruning unwanted log entries");

	s = (struct log_spacer *)(laddr + hptr->offset);
	lpos += sizeof(struct log_header);

	while(s->next_offset) {
		if (s->magic != SPACER_MAGIC) {
			nis_err("copy: transaction log is corrupt!");
			return(NIS_FAIL);
		}
		ptr = (caddr_t)s;
		ptr += sizeof(struct log_spacer);
		if (keep(m, s, ptr, xid, name, t) == TRUE) {
			h->count++;
			h->xid = s->xid;
			h->upd_time = time(NULL);
			h->size += sizeof(struct log_spacer) + s->next_offset;
			n->magic = SPACER_MAGIC;
			n->next_offset = s->next_offset;
			n->prev_offset = size;
			n->xid = h->xid;
			fwrite((char *)n, sizeof(struct log_spacer), 1, fp);
			fwrite(ptr, s->next_offset, 1, fp);
			size = s->next_offset;
		}
		lpos += s->next_offset + sizeof(struct log_spacer);
		if (lpos > lsize)
			break;
		ptr += s->next_offset;
		s = (struct log_spacer *)ptr;
	}

	n->magic = SPACER_MAGIC;
	n->next_offset = 0;
	n->prev_offset = size;
	n->xid = 0;
	fwrite((char *)n, sizeof(struct log_spacer), 1, fp);
	fseek(fp, (size_t)0, SEEK_SET);
	fwrite((char *)h, sizeof(struct log_header), 1, fp);
	fflush(fp);

	return(NIS_SUCCESS);
}

static nis_error nis_log_reload(tmpname, h, n, fp)
	nis_name		tmpname;
	struct log_header	*h;
	struct log_spacer	*n;
	FILE			*fp;
{
	struct stat		st;

	if (verbose)
		nis_err("replacing old log file and reloading");

	fstat(fileno(fp), &st);
	if (st.st_size < CHUNKSIZE)
		st.st_size = CHUNKSIZE;
	else
		st.st_size += st.st_size % CHUNKSIZE;
	fseek(fp, st.st_size, SEEK_SET);
	fwrite(".", 1, 1, fp);
	fflush(fp);
	fclose(fp);

	if (munmap(laddr, lsize) == -1)
		nis_err("unmaping log failed: %s", strerror(errno));
	close(log_fd);

	if (rename(tmpname, _PATH_NIS_LOG) == -1)
		nis_err("rename failed: %s", strerror(errno));

	nis_init_log();

	return(NIS_SUCCESS);
}

nis_error nis_log_flush_byxid(xid)
	uint32_t		xid;
{
	struct log_spacer	n;
	struct log_header	h;
	FILE			*fp;

	if (verbose)
		nis_err("flushing XID %lu from log", xid);

	if (nis_log_mktmp(_PATH_NIS_LOG_TMP, &h, &n, &fp) != NIS_SUCCESS) {
		nis_err("log flush failed");
		return(NIS_FAIL);
	}

	nis_log_copy(MATCH_BYXID, xid, NULL, 0, &h, &n, fp);

	nis_log_reload(_PATH_NIS_LOG_TMP, &h, &n, fp);

	return(NIS_SUCCESS);
}

nis_error nis_log_flush_byname(dirname)
	nis_name		dirname;
{
	struct log_spacer	n;
	struct log_header	h;
	FILE			*fp;

	if (verbose)
		nis_err("flushing %s from log", dirname);

	if (nis_log_mktmp(_PATH_NIS_LOG_TMP, &h, &n, &fp) != NIS_SUCCESS) {
		nis_err("log flush failed");
		return(NIS_FAIL);
	}

	nis_log_copy(MATCH_BYNAME, 0, dirname, 0, &h, &n, fp);

	nis_log_reload(_PATH_NIS_LOG_TMP, &h, &n, fp);

	return(NIS_SUCCESS);
}

nis_error nis_log_flush_bytime(t)
	time_t			t;
{
	struct log_spacer	n;
	struct log_header	h;
	FILE			*fp;

	if (verbose)
		nis_err("flushing %lu from log", t);

	if (nis_log_mktmp(_PATH_NIS_LOG_TMP, &h, &n, &fp) != NIS_SUCCESS) {
		nis_err("log flush failed");
		return(NIS_FAIL);
	}

	nis_log_copy(MATCH_BYTIME, 0, NULL, t, &h, &n, fp);

	nis_log_reload(_PATH_NIS_LOG_TMP, &h, &n, fp);

	return(NIS_SUCCESS);
}

static nis_error nis_log_commit_ent(l)
	log_entry		*l;
{
#ifdef notyet
	nis_object		*d = NULL;

	if (__type_of(&l->le_object) == TABLE_OBJ)
		__nis_ocache_lookup(l->le_name, CHECK_MASTER,
					MASTER_ONLY, &d);
#endif

	switch(l->le_type) {
	case ADD_NAME:
		if (__type_of(&l->le_object) == TABLE_OBJ)
			db_newtab(l->le_name, &l->le_object);
#ifdef notyet
		if (__type_of(&l->le_object == DIRECTORY_OBJ) &&
			__i_am_server(&l->le_object.DI_data) == TRUE)
			db_newdir(l->le_name);
#endif
		break;
	case REM_NAME:
		if (__type_of(&l->le_object) == TABLE_OBJ)
			db_rm(l->le_name);
#ifdef notyet
		if (__type_of(&l->le_object == DIRECTORY_OBJ) &&
			__i_am_server(&l->le_object.DI_data) == TRUE)
			db_rm(l->le_name);
#endif
		break;
	default:
		/* Nothing to do for other kinds of log entries. */
		break;
	}

	return(NIS_SUCCESS);
}

/*
 * Commit the current transaction id.
 */
nis_error nis_log_commit_xid(xid)
	uint32_t		xid;
{
	struct log_spacer	*sp = active_sptr;
	log_entry		*ent;

	if (active_xid == LOG_INACTIVE)
		return(NIS_TRYAGAIN);

	if (xid != active_xid)
		return(NIS_TRYAGAIN);

	/*
	 * If there were no transactions, roll back the xid count
	 * and do nothing.
	 */
	if (active_sptr == sptr) {
		hptr->xid--;
		nis_log_mark(LOG_STABLE);
		active_xid = LOG_INACTIVE;
		active_sptr = NULL;
		bzero((char *)&active_princ, NIS_MAXNAMELEN);
		if (verbose)
			nis_err("no new log entries; aborting transaction");
		return(NIS_SUCCESS);
	}

	/*
	 * Scan forward through transaction searching for entries
 	 * that result in the creation of new databases and carry
	 * out the creation.
	 */
	while(sp->xid == xid && sp->next_offset) {
		if (nis_log_next_ent(sp, &ent, &sp) != NIS_SUCCESS)
			return(NIS_FAIL);
		/*
		 * If the log action requires the creation/removal
		 * of an underlying database of some kind (table or
		 * directory container), this will do it for us.
	 	 */
		nis_log_commit_ent(ent);

		/*
	 	 * For births and deaths of directory objects,
		 * we also have to add a time stamp. We do _not_
		 * add stamps for modifications to directories.
	 	 */
		if (__type_of(&ent->le_object) == DIRECTORY_OBJ &&
		    (ent->le_type == ADD_NAME || ent->le_type == REM_NAME))
			nis_log_insert_stamp(ent->le_name, ent->le_time);

		xdr_free(xdr_log_entry, (char *)ent);
		free(ent);
	}

	/*
	 * If there are any timestamps pending, flush them out.
	 */
	nis_log_commit_stamps();
	nis_log_free_stamps();

	nis_log_mark(LOG_STABLE);
	active_xid = LOG_INACTIVE;
	active_sptr = NULL;
	bzero((char *)&active_princ, NIS_MAXNAMELEN);

	return(NIS_SUCCESS);
}

/*
 * Reverse the effects of a single log entry.
 */
static nis_error nis_log_rollback_ent(l)
	log_entry		*l;
{
	nis_name		name = l->le_name;
	struct db_objs		res;
	nis_object		*obj = &l->le_object;
	nis_attr		*attrs = l->le_attrs.le_attrs_val;
	uint32_t		numattrs = l->le_attrs.le_attrs_len;
	nis_object		*t = NULL;

	if (__type_of(&l->le_object) == ENTRY_OBJ)
		__nis_ocache_lookup(l->le_name, CHECK_MASTER,
					MASTER_ONLY, &t);
		
	switch(l->le_type) {
	case ADD_NAME:
	case MOD_NAME_NEW:
		db_remove(name, &res);
		break;
	case REM_NAME:
	case MOD_NAME_OLD:
		db_add(name, obj, &res);
		break;
	case ADD_IBASE:
		db_ibremove(name, attrs, numattrs, obj, t, &res);
		break;
	case REM_IBASE:
		db_ibadd(name, attrs, numattrs, obj, t, &res);
		break;
	case MOD_IBASE:
	case UPD_STAMP:
	case LOG_NOP:
	default:
		break;
	}

	return(NIS_SUCCESS);
}

/*
 * We've decided to nullify this transaction. Undo all of the log
 * entries.
 */
nis_error nis_log_abort_xid(xid)
	uint32_t		xid;
{
	struct log_spacer	*sp = sptr;
	log_entry		*ent;

	if (active_xid == LOG_INACTIVE)
		return(NIS_TRYAGAIN);

	if (xid != active_xid)
		return(NIS_TRYAGAIN);

	/*
	 * If there were no transactions, roll back the xid count
	 * and do nothing.
	 */
	if (active_sptr == sptr) {
		hptr->xid--;
		nis_log_mark(LOG_STABLE);
		active_xid = LOG_INACTIVE;
		active_sptr = NULL;
		bzero((char *)&active_princ, NIS_MAXNAMELEN);
		if (verbose)
			nis_err("no new log entries; aborting transaction");
		return(NIS_SUCCESS);
	}

	while(sp->xid == xid) {
		if (nis_log_prev_ent(sp, &ent, &sp) != NIS_SUCCESS)
			return(0);

		nis_log_rollback_ent(ent);

		xdr_free(xdr_log_entry, (char *)ent);
		free(ent);
	}

	nis_log_flush_byxid(xid);

	nis_log_mark(LOG_STABLE);
	active_xid = LOG_INACTIVE;
	active_sptr = NULL;
	bzero((char *)&active_princ, NIS_MAXNAMELEN);

	return(NIS_SUCCESS);
}

nis_error nis_init_log()
{
	struct stat		st;
	int			coldstart = 0;
	size_t			lpos = 0;

	if (stat(_PATH_NIS_LOG, &st) == -1) {
		/* Log doesn't exist yet. Create it. */
		if ((log_fd = open(_PATH_NIS_LOG, O_RDWR|O_CREAT,0600)) == -1) {
			nis_err("couldn't open log: %s", strerror(errno));
			return(NIS_FAIL);
		}
		lseek(log_fd, CHUNKSIZE, SEEK_SET);
		write(log_fd, ".", 1);
		lsize = CHUNKSIZE;
		coldstart++;
	} else {
		if ((log_fd = open(_PATH_NIS_LOG, O_RDWR,0600)) == -1) {
			nis_err("couldn't open log: %s", strerror(errno));
			return(NIS_FAIL);
		}
		lsize = st.st_size;
	}

	laddr = mmap(0, lsize, PROT_READ|PROT_WRITE, MAP_SHARED, log_fd, 0);
	if ((int)laddr == -1) {
		nis_err("mmap failed: %s", strerror(errno));
		close(log_fd);
		return(NIS_FAIL);
	}

	hptr = (struct log_header *)laddr;
	if (coldstart) {
		hptr->magic = HEADER_MAGIC;
		hptr->state = LOG_STABLE;
		hptr->count = 0;
		hptr->xid = 0;
		hptr->offset = sizeof(struct log_header);
		hptr->size = sizeof(struct log_header) +
			     sizeof(struct log_spacer);
		hptr->upd_time = 0;
		sptr = (struct log_spacer *)(laddr + hptr->offset);
		sptr->magic = SPACER_MAGIC;
		nis_log_mark(LOG_STABLE);
	} else {
		sptr = (struct log_spacer *)(laddr + hptr->offset);
		if (hptr->magic != HEADER_MAGIC ||
			sptr->magic != SPACER_MAGIC) {
			nis_err("transaction log corrupt! aborting!");
			munmap(laddr, lsize);
			close(log_fd);
			return(NIS_FAIL);
		}
	}

	lpos += sizeof(struct log_header);

	while(sptr->next_offset) {
		caddr_t			ptr;
		if (sptr->magic != SPACER_MAGIC) {
			nis_err("init: transaction log is corrupt!");
			munmap(laddr, lsize);
			close(log_fd);
			return(NIS_FAIL);
		}
		ptr = (caddr_t)sptr;
		ptr += sptr->next_offset + sizeof(struct log_spacer);
		lpos += sptr->next_offset + sizeof(struct log_spacer);
		if (lpos > lsize)
			break;
		sptr = (struct log_spacer *)ptr;
	}

	return(NIS_SUCCESS);
}

nis_error nis_log_add_entry(l)
	log_entry		*l;
{
	if (active_xid == LOG_INACTIVE)
		return(NIS_TRYAGAIN);

	if (__is_local_dirname(l->le_object.zo_domain) == FALSE)
		return(NIS_SUCCESS);

	nis_log_mark(LOG_UPDATING);
	nis_log_add_ent(l);
	nis_log_mark(LOG_STABLE);

	return(NIS_SUCCESS);
}

static bool_t nis_log_updmatch(l, name)
	log_entry		*l;
	nis_name		name;
{
	switch(l->le_type) {
	case UPD_STAMP:
		if (nis_dir_cmp(l->le_name, name) == SAME_NAME) {
			if (verbose)
				nis_err("(update stamp) name: %s time: %s",
							l->le_name,
							nis_ctime(l->le_time));
			return(TRUE);
		}
		break;
	default:
		if (nis_dir_cmp(l->le_object.zo_domain,
					name) == SAME_NAME) {
			if (verbose)
				nis_err("(entry) name: %s time: %s",
						l->le_name,
						nis_ctime(l->le_time));
			return(TRUE);
		}
		break;
	}

	return(FALSE);
}

static void nis_log_ent_to_stamp(l)
	log_entry		*l;
{
	log_entry		s;

	if (verbose)
		nis_err("making stamp from non stamp");

	bzero((char *)&s, sizeof(s));

	s.le_name = strdup(l->le_object.zo_domain);
	s.le_princp = l->le_princp; l->le_princp = NULL;
	s.le_type = UPD_STAMP;
	s.le_time = l->le_time;
	s.le_object.zo_name = strdup("");
	s.le_object.zo_owner = strdup("");
	s.le_object.zo_group = strdup("");
	s.le_object.zo_domain = strdup(l->le_object.zo_domain);
	s.le_object.zo_oid.ctime = l->le_time;
	s.le_object.zo_oid.mtime = l->le_time;

	__type_of(&s.le_object) = NO_OBJ;

	xdr_free(xdr_log_entry, (char *)l);

	bzero((char *)l, sizeof(log_entry));
	bcopy((char *)&s, (char *)l, sizeof(log_entry));

	return;
}

/*
 * Scan backwards through the log trying to find the last time a
 * given name was updated.
 */
uint32_t nis_log_find_update(name, stamp, res)
	nis_name		name;
	uint32_t		stamp;
	log_entry		**res;
{
	struct log_spacer	*sp = sptr;
	log_entry		*ent;
	time_t			utime = 0;

	if (verbose)
		nis_err("looking for most recent update of %s", name);

	if (strcmp(name, ROOT_OBJ) == NULL || (root_obj != NULL
	    && nis_dir_cmp(name, root_obj->zo_domain) == SAME_NAME))
		name = rootname;

	while(sp->prev_offset) {
		if (nis_log_prev_ent(sp, &ent, &sp) != NIS_SUCCESS)
			return(0);
		if (nis_log_updmatch(ent, name) == TRUE) {
			if (stamp == 0 || stamp == ent->le_time) {
				utime = ent->le_time;
				if (res == NULL) {
					xdr_free(xdr_log_entry, (char *)ent);
					free(ent);
				} else
					*(caddr_t *)res = (caddr_t)ent;
				/* Save our place */
				cptr = sp;
				break;
			}
		}
		xdr_free(xdr_log_entry, (char *)ent);
		free(ent);
	}

	if (res != NULL && utime) {
		if ((*res)->le_type != UPD_STAMP)
			nis_log_ent_to_stamp(*res);
	}
	return(utime);
}

/*
 * Return all log entries starting from the specified time stamp and
 * continuing all the way to the present.
 */
nis_error nis_dump_log(dir, stamp, res)
	nis_name		dir;
	uint32_t		stamp;
	log_result		**res;
{
	struct log_spacer	*sp = cptr;	/* set by find_update() */
	log_result		*lr;
	log_entry		*new = NULL, *list = NULL, *ent = NULL;

	lr = calloc(1, sizeof(log_result));
	if (lr == NULL)
		return(NIS_NOMEMORY);
	list = new = calloc(1, sizeof(log_entry));

	while(sp->next_offset) {
		if (nis_log_next_ent(sp, &ent, &sp) != NIS_SUCCESS)
			return(NIS_FAIL);
		/*
		 * Sanity check: the idea here is to return all of the
		 * log entries that have transpired _after_ the time
		 * specified by the replica. However, nis_log_find_update()
		 * leaves the current position pointer set at the entry
		 * that was logged at the _same_ time as that specified
		 * by the replica. The replica doesn't want this entry
		 * since it already has it, and sending it again would
		 * lead to confusion. So, if this entry's timestamp is
		 * the same as the one specified by the replica, skip it.
		 */
		if (ent->le_time == stamp) {
			xdr_free(xdr_log_entry, (char *)ent);
			free(ent);
			continue;
		}
		if ((nis_dir_cmp(dir, ent->le_object.zo_domain) == SAME_NAME ||
		    nis_dir_cmp(dir, ent->le_name) == SAME_NAME) &&
					ent->le_type != UPD_STAMP) {
			nis_err("selected entry for %s", ent->le_name);
			list = realloc(list,
				(lr->lr_entries.lr_entries_len + 1) *
				sizeof(log_entry));
			if (list == NULL) {
				xdr_free(xdr_log_entry, (char *)ent);
				return(NIS_NOMEMORY);
			}
			new = list + lr->lr_entries.lr_entries_len;
			bcopy((char *)ent, (char *)new, sizeof(log_entry));
			free(ent);
			lr->lr_entries.lr_entries_len++;
			lr->lr_entries.lr_entries_val = list;
		} else {
			xdr_free(xdr_log_entry, (char *)ent);
			free(ent);
		}
	}

	lr->lr_status = NIS_SUCCESS;
	*(caddr_t *)res = (caddr_t)lr;

	return(NIS_SUCCESS);
}

nis_error nis_empty_log()
{
	return(NIS_FAIL);
}

/*
 * This routine is called at start-up once the log system is initialized.
 * It checks to make sure that the log is stable; if it isn't, it tries
 * to bring roll back or roll forward transactions to bring the databases
 * into a consistent state.
 */
nis_error nis_log_recovery()
{

	if (hptr->state == LOG_STABLE)
		return(NIS_SUCCESS);

	if (verbose)
		nis_err("starting database recovery");

	if (hptr->state == LOG_STABLE) {
		if (verbose)
			nis_err("log is stable -- nothing to do");
		return(NIS_SUCCESS);
	}

	if (hptr->state == LOG_UPDATING) {
		if (verbose)
			nis_err("rolling back transaction ID %lu", hptr->xid);
		nis_log_abort_xid(hptr->xid);
		nis_log_flush_byxid(hptr->xid);
	}

	if (verbose)
		nis_err("finishing database recovery");

	return(NIS_FAIL);
}
