/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nis_main.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $
 */

#include <rpcsvc/nis.h>
#include <rpcsvc/nis_db.h>
#if __FreeBSD__ >= 3
#include <rpc/rpc_com.h>
#endif
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <unistd.h>
#include <sys/stat.h>
#include <stdlib.h> /* getenv, exit */
#include <rpc/pmap_clnt.h> /* for pmap_unset */
#include <string.h> /* strcmp */
#include <signal.h>
#ifdef __cplusplus
#include <sysent.h> /* getdtablesize, open */
#endif /* __cplusplus */
#include <memory.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <syslog.h>
#include <sys/wait.h>
#include "nis_extern.h"

#ifndef SIG_PF
#define	SIG_PF void(*)(int)
#endif

#define	_RPCSVC_CLOSEDOWN 120

#ifndef lint
static const char copyright[] = "@(#) Copyright (c) 1996, 1997, 1998\
	Bill Paul. All rights reserved.";
static const char rcsid[] =
	"@(#) $Id: nis_main.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $ (Project Flatbush)";
#endif

int _rpcpmstart;		/* Started by a port monitor ? */
static int _rpcfdtype;
		 /* Whether Stream or Datagram ? */
	/* States a server can be in wrt request */

#define	_IDLE 0
#define	_SERVED 1
#define	_SERVING 2

extern int _rpcsvcstate;	 /* Set when a request is serviced */

char			*progname = "rpc.nisd";
int			securelevel = 2;
time_t			boottime = 0;
int			verbose = 0;
int			readonly = 0;
static int		serving = 0;			
int			__nis_child_must_die = 0;
char			**srvlist = NULL;
int			srvcnt;
nis_name		rootname;
int			max_children = -1;

nis_object		*root_obj = NULL;
int			forked = 0;

static SVCXPRT		*udptransp = NULL;
static SVCXPRT		*tcptransp = NULL;

void __nis_unregister_xprt()
{
	if (verbose)
		nis_err("unregistering RPC service "
				"transport handles (pid=%d)", getpid());
	xprt_unregister(udptransp);
	xprt_unregister(tcptransp);

	return;
}

void __nis_register_xprt()
{
	if (verbose)
		nis_err("re-registering RPC service "
				"transport handles (pid=%d)", getpid());
	xprt_register(udptransp);
	xprt_register(tcptransp);

	return;
}

static
void _msgout(char* msg)
{
	if (debug) {
		if (_rpcpmstart)
			syslog(LOG_ERR, msg);
		else
			(void) fprintf(stderr, "%s\n", msg);
	} else
		syslog(LOG_ERR, msg);
}

/*
 * Note: in order to avoid potential heap corruption, all signals
 * are blocked while any of the main server code is executing. This
 * prevents a signal from being delivered while we're in the middle
 * of a heap allocation (or deallocation).
 */
#if __FreeBSD__ && __FreeBSD__ >= 3
static void
nis_svc_run()
{
	int			pid;
	struct timeval		tv;
	fd_set			*fds;
	int			cb_fds = 0;

	/* Establish the identity of the parent rpc.nisd process. */
	pid = getpid();

	for (;;) {
		if (&svc_fdset != NULL) {
			int bytes = _howmany(FD_SETSIZE, NFDBITS) *
							sizeof(fd_mask);
			fds = calloc(1, bytes);
			bcopy(&svc_fdset, fds, bytes);
		} else
			fds = NULL;

		cb_fds = nis_event_sched(fds, &tv);
		if (cb_fds < svc_maxfd)
			cb_fds = svc_maxfd;

		switch (select(cb_fds + 1, fds, NULL, NULL,
			(tv.tv_sec || tv.tv_usec) ? &tv : NULL)) {
		case -1:
			sigblock(0xFFFFFFFF);
			if (errno == EINTR) {
				if (fds != NULL)
					free(fds);
				sigsetmask(0);
				continue;
			}
			nis_err("svc_run: - select failed: %s",
							strerror(errno));
			if (fds != NULL)
				free(fds);
			sigsetmask(0);
			return;
		case 0:
			sigblock(0xFFFFFFFF);
			nis_event_runq(fds);
			if (fds != NULL)
				free(fds);
			if (readonly) {
				if (kill(pid, 0)) {
					nis_err("help! my parent died!");
					_exit(0);
				}
			}
			sigsetmask(0);
			break;
		default:
			sigblock(0xFFFFFFFF);
			serving = 1;
			nis_event_runq(fds);
			svc_getreqset(fds);
			free(fds);
			serving = 0;
			if (forked && pid != getpid())
				_exit(0);
			sigsetmask(0);
			break;
		}
		if (__nis_child_must_die && pid != getpid()) {
			nis_err("exiting in svc_run");
			_exit(0);
		}
		if (readonly) {
			if (kill(pid, 0)) {
				nis_err("help! my parent died!");
				_exit(0);
			}
			continue;
		}
	}
}
#else
static void
nis_svc_run()
{
#ifdef FD_SETSIZE
	fd_set readfds;
#else
	int readfds;
#endif /* def FD_SETSIZE */
	int pid;
	int fd_setsize = _rpc_dtablesize();
	struct timeval		tv;

	/* Establish the identity of the parent rpc.nisd process. */
	pid = getpid();

	for (;;) {
#ifdef FD_SETSIZE
		readfds = svc_fdset;
#else
		readfds = svc_fds;
#endif /* def FD_SETSIZE */
		nis_event_sched(&readfds, &tv);

		switch (select(fd_setsize, &readfds, NULL, NULL,
			(tv.tv_sec || tv.tv_usec) ? &tv : NULL)) {
		case -1:
			sigblock(0xFFFFFFFF);
			if (errno == EINTR) {
				continue;
			}
			perror("svc_run: - select failed");
			sigsetmask(0);
			return;
		case 0:
			sigblock(0xFFFFFFFF);
			if (readonly) {
				if (kill(pid, 0)) {
					nis_err("help! my parent died!");
					_exit(0);
				}
			}
			sigsetmask(0);
			break;
		default:
			sigblock(0xFFFFFFFF);
			serving = 1;
			svc_getreqset(&readfds);
			serving = 0;
			if (forked && pid != getpid())
				_exit(0);
			sigsetmask(0);
			break;
		}
		sigblock(0xFFFFFFFF);
		nis_event_runq(&readfds);
		sigsetmask(0);
		if (__nis_child_must_die && pid != getpid()) {
			nis_err("exiting in svc_run");
			_exit(0);
		}
		if (readonly) {
			if (kill(pid, 0)) {
				nis_err("help! my parent died!");
				_exit(0);
			}
			continue;
		}
	}
}
#endif

/*
 * SIGUSR1 is our parent's way of telling us that it wants us to
 * die. We have to be careful not to exit immediately though. The
 * parent process will fork() a 'readonly' child whenever it has
 * some work do to, like doing a master/slave resynchronization,
 * which could cause it to block for a long time. We want to avoid
 * blocking, so we spawn a child that can continue serving lookup
 * requests while the parent is busy. But the following can happen:
 *
 * - parent fork()s child and does some work
 * - child receives request and starts handling it
 * - parent finishes its work and signals child to go away _before_
 *   the child has finished servicing its client
 *
 * We don't want the child to die in the middle of a request since
 * that would screw up the client on the other end, so instead we set
 * a flag telling the child do exit after it has finished its work.
 *
 * Callbacks make this tricky. If a callback is in progress, we have
 * finish all of it before the child can exit.
 */
static void reaper(sig)
	int sig;
{
	int			status;
	int			saved_errno;

	saved_errno = errno;
	sigblock(0xFFFFFFFF);

	if (sig == SIGUSR1) {
		__nis_unregister_xprt();
		__nis_child_must_die = 1;
		if (!serving)
			_exit(0);
		errno = saved_errno;
		sigsetmask(0);
		return;
	}

	if (sig == SIGCHLD) {
		while (wait3(&status, WNOHANG, NULL) > 0)
			/* children--*/;
	} else {
		(void) pmap_unset(NIS_PROG, NIS_VERSION);
		exit(0);
	}

	errno = saved_errno;
	sigsetmask(0);

	return;
}

static void usage()
{
	fprintf(stderr, "Usage: %s [-L load] [-S level]\n", progname);
	exit(1);
}

static void
closedown(int sig)
{
	if (_rpcsvcstate == _IDLE) {
		extern fd_set svc_fdset;
		static int size;
		int i, openfd;

		if (_rpcfdtype == SOCK_DGRAM) {
			(void) pmap_unset(NIS_PROG, NIS_VERSION);
			exit(0);
		}
		if (size == 0) {
			size = getdtablesize();
		}
		for (i = 0, openfd = 0; i < size && openfd < 2; i++)
			if (FD_ISSET(i, &svc_fdset))
				openfd++;
		if (openfd <= 1) {
			(void) pmap_unset(NIS_PROG, NIS_VERSION);
			exit(0);
		}
	}
	if (_rpcsvcstate == _SERVED)
		_rpcsvcstate = _IDLE;

	(void) signal(SIGALRM, (SIG_PF) closedown);
	(void) alarm(_RPCSVC_CLOSEDOWN/2);
}

int main(argc, argv)
	int			argc;
	char			*argv[];
{
	int			sock;
	int			proto = 0;
	struct sockaddr_in	saddr;
	int			asize = sizeof (saddr);
	int			ch;
	char			*catalog_file = _NIS_CATALOG;
	int			do_checkpoint = 0;
	struct stat		st;

	while ((ch = getopt(argc, argv, "YBDFvd:L:S:")) != EOF) {
		switch (ch) {
		case 'S':
			securelevel = atoi(optarg);
			break;
		case 'd':
			catalog_file = optarg;
			break;
		case 'v':
			verbose++;
			break;
		case 'D':
			debug++;
			break;
		case 'F':
			do_checkpoint++;
			break;
		case 'L':
			max_children = atoi(optarg);
			break;
		case 'Y':
			nis_err("YP compat mode is not available "
				     "in this implementation of rpc.nisd");
			break;
		case 'B':
			nis_err("DNS forwarding is not available "
				     "in this implementation of rpc.nisd");
			break;
		default:
			usage();
		}
	}

	if (verbose)
		nis_err("NIS+ Server startup.");

	if (getsockname(0, (struct sockaddr *)&saddr, &asize) == 0) {
		int ssize = sizeof (int);

		if (saddr.sin_family != AF_INET)
			exit(1);
		if (getsockopt(0, SOL_SOCKET, SO_TYPE,
				(char *)&_rpcfdtype, &ssize) == -1)
			exit(1);
		sock = 0;
		_rpcpmstart = 1;
		proto = 0;
		openlog(progname, LOG_PID, LOG_DAEMON);
	} else {
		if (!debug) {
			if (daemon(0,0)) {
				err(1, "cannot fork");
			}
			openlog(progname, LOG_PID, LOG_DAEMON);
		}
		sock = RPC_ANYSOCK;
		(void) pmap_unset(NIS_PROG, NIS_VERSION);
	}

	if (verbose)
		nis_err("RPC program registrations ..");

	if ((_rpcfdtype == 0) || (_rpcfdtype == SOCK_DGRAM)) {
		udptransp = svcudp_create(sock);
		if (udptransp == NULL) {
			_msgout("cannot create udp service.");
			exit(1);
		}
		if (!_rpcpmstart)
			proto = IPPROTO_UDP;
		if (!svc_register(udptransp, NIS_PROG, NIS_VERSION, nis_prog_3, proto)) {
			_msgout("unable to register (NIS_PROG, NIS_VERSION, udp).");
			exit(1);
		}
	}

	if ((_rpcfdtype == 0) || (_rpcfdtype == SOCK_STREAM)) {
		tcptransp = svctcp_create(sock, 512, 512);
		if (tcptransp == NULL) {
			_msgout("cannot create tcp service.");
			exit(1);
		}
		if (!_rpcpmstart)
			proto = IPPROTO_TCP;
		if (!svc_register(tcptransp, NIS_PROG, NIS_VERSION, nis_prog_3, proto)) {
			_msgout("unable to register (NIS_PROG, NIS_VERSION, tcp).");
			exit(1);
		}
	}

	if (udptransp == (SVCXPRT *)NULL || tcptransp == (SVCXPRT *)NULL) {
		_msgout("could not create a handle");
		exit(1);
	}
	if (_rpcpmstart) {
		(void) signal(SIGALRM, (SIG_PF) closedown);
		(void) alarm(_RPCSVC_CLOSEDOWN/2);
	}

	if (verbose)
		nis_err(".. rpc registration complete.");

/*
 * Primary server initialization tasks.
 */

	/* Mark time of birth */
	boottime = time(NULL);

	/* Install signal handlers */
	(void) signal(SIGPIPE, SIG_IGN);
	(void) signal(SIGCHLD, (SIG_PF) reaper);
	(void) signal(SIGTERM, (SIG_PF) reaper);
	(void) signal(SIGINT, (SIG_PF) reaper);
	(void) signal(SIGHUP, (SIG_PF) reaper);
	(void) signal(SIGUSR1, (SIG_PF) reaper);

	/* Make sure /var/nis and /var/nis/data exist. */
	if (stat(_PATH_NIS, &st) == -1 || !S_ISDIR(st.st_mode)) {
		_msgout("/var/nis doesn't exist! aborting...");
		(void) pmap_unset(NIS_PROG, NIS_VERSION);
		exit(1);
	}

	/* Set umask */
	umask(077);

	/*
	 * Init the event handler. We need to do this early on.
	 */
	nis_event_init();

	/* Load serving list */
	if (stat(_PATH_NIS_DATA, &st) == -1) {
		if (errno == ENOENT) {
			if (mkdir(_PATH_NIS_DATA, 0750) == -1) {
				_msgout("failed to create /var/nis/data");
				perror(_PATH_NIS_DATA);
				(void) pmap_unset(NIS_PROG, NIS_VERSION);
				exit(1);
			}
		} else {
			perror(_PATH_NIS_DATA);
			(void) pmap_unset(NIS_PROG, NIS_VERSION);
			exit(1);
		}
	} else if (!S_ISDIR(st.st_mode)) {
		_msgout("/var/nis/data is not a directory! aborting...");
		(void) pmap_unset(NIS_PROG, NIS_VERSION);
		exit(1);
	}

	/* Set current directory */
	chdir(_PATH_NIS_DATA);

	nis_load_srvlist();

	/* Load the root.object, if it exists. */
	nis_get_root_obj(&root_obj);

	if (verbose) {
		nis_err("Database initialization ..");
		nis_err("Dictionary is %s", catalog_file);
	}
	/* Initialize the database backend */
	if (db_initialize(catalog_file) == FALSE) {
		_msgout("failed to initialize database");
		(void) pmap_unset(NIS_PROG, NIS_VERSION);
		exit(1);
	}
	if (verbose)
		nis_err(".. database initialization complete.");

	/* Checkpoint our databases if necessary. */
	if (do_checkpoint)
		db_checkpoint("");

	if (root_obj != NULL) {
		/*
		 * Since we are the root server, we can also set the
		 * rootname without having to do an RPC;
		 */
		rootname = strdup(root_obj->DI_data.do_name);
	} else {
		/*
		 * If we aren't a root server, we have to RPC to
		 * learn the name of the root of the namespace.
		 */
		rootname = __nis_find_root();
		if (rootname != NULL)
			rootname = strdup(rootname);
		else
			_msgout("failed to find root of namespace!");
	}

	/*
	 * If we are a root master, we must create the root
	 * object if it doesn't exist.
	 */
	if (root_obj != NULL && __i_am_master(&root_obj->DI_data) == TRUE &&
		db_table_exists("root_dir") != DB_SUCCESS) {
		if (db_newdir(root_obj->DI_data.do_name) != DB_SUCCESS) {
			_msgout("failed to create root directory");
			(void) pmap_unset(NIS_PROG, NIS_VERSION);
			exit(1);
		}
	}

	if (verbose)
		nis_err("Stats counters and"
				" object cache initialization ..");
	/*
	 * Init stats counters.
	 */
	nis_init_stats();

	/*
	 * Init object cache.
	 */
	__nis_init_ocache();

	if (verbose)
		nis_err(".. done.");

#ifdef USE_SECURE_RPC
#ifdef __FreeBSD__
	/* Register the AUTH_DES flavor with the library */
	if (svc_auth_reg(AUTH_DES, _svcauth_des) == -1) {
		_msgout("failed to register AUTH_DES authenticator");
		(void) pmap_unset(NIS_PROG, NIS_VERSION);
		exit(1);
	}
#endif
#endif

	if (verbose)
		nis_err("Callback and cred cache initialization ..");

	/* Init the callback queue */
	nis_init_callbacks();

	/* Init the principal name cache */
	nis_init_cred_cache();

	if (verbose) {
		nis_err(".. done.");
		nis_err("Transaction log initialization ..");
	}

	/* Init the transaction logging system */
	if (nis_init_log() != NIS_SUCCESS) {
		(void) pmap_unset(NIS_PROG, NIS_VERSION);
		exit(1);
	}
	if (verbose)
		nis_err(".. transaction log initialization complete.");

	/* Perform log recovery */
	nis_log_recovery();

	if (verbose)
		nis_err("Checking for master/replica synchronization ..");

	nis_sync();

	if (verbose)
		nis_err(".. sync check complete.");

	/* Should be ready to run now */

	if (verbose)
		nis_err("Service starting.");
	nis_svc_run();
	_msgout("svc_run returned");
	exit(1);
	/* NOTREACHED */
}
