/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nis_master.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/cdefs.h>
#include "nis.h"
#include "nis_extern.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nis_master.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $";
#endif

/*
 * This module contains the master side of the replica update code.
 * Changes are always made on the master server. Once a change is
 * made, the master must ping all the replicas of the directory in
 * which the change was made so that they can re-synchronize themselves.
 * Therefore, we need to be able to:
 *
 * - Ping all replicas of a given directory
 * - Keep a list of pending updates
 *   used to feed update data to the replica.
 */


struct ping_list {
	nis_name		dname;
	directory_obj		*dir;
	uint32_t		stamp;
	struct ping_list	*next;
};

struct ping_list		*plist = NULL;

static nis_error nis_master_insert_ping(name, d)
	nis_name		name;
	directory_obj		*d;
{
	struct ping_list	*tmp, *prev;

	/*
	 * If we are the only server on the server list then
	 * there aren't any replicas to ping.
	 */
	if (d->do_servers.do_servers_len == 1 &&
		nis_dir_cmp(d->do_servers.do_servers_val[0].name,
				nis_local_host()) == SAME_NAME) {
		if (verbose)
			nis_err("no replicas for %s -- not pinging", name);
		xdr_free(xdr_directory_obj, (char *)d);
		free(d);
		return(NIS_TRYAGAIN);
	}

	tmp = plist;

	/*
	 * If we already have a ping scheduled for this directory,
	 * keep it in the queue and update the stamp and directory
	 * object rather than making a second entry.
	 */
	while(tmp) {
		if (nis_dir_cmp(name, tmp->dname) == SAME_NAME) {
			if (verbose)
				nis_err("ping already scheduled for %s",
									name);
			tmp->stamp = nis_log_find_update(name, 0, NULL);
			xdr_free(xdr_directory_obj, (char *)tmp->dir);
			free(tmp->dir);
			tmp->dir = d;
			return(NIS_TRYAGAIN);
		}
		tmp = tmp->next;
	}

	tmp = prev = plist;
	while(tmp != NULL) {
		prev = tmp;
		tmp = tmp->next;
	}

	tmp = malloc(sizeof(struct ping_list));
	if (tmp == NULL)
		return(NIS_NOMEMORY);

	if (verbose)
		nis_err("master queuing update notification for %s", name);

	tmp->dname = strdup(name);
	tmp->dir = d;
	tmp->stamp = nis_log_find_update(name, 0, NULL);
	tmp->next = NULL;

	if (prev == NULL)
		plist = tmp;
	else
		prev->next = tmp;

	return(NIS_SUCCESS);
}

/*
 * Event handler for pinging replicas.
 */
bool_t nis_master_cb(args, id, reason)
	void			*args;
	uint32_t		id;
	int			reason;
{
	if (nis_master_do_updates() == NIS_SUCCESS)
		return(FALSE);
	else
		return(TRUE);
}

nis_error nis_master_add_update(name)
	nis_name		name;
{
	nis_error		r;
	nis_object		*o = NULL;
	directory_obj		*d;

	/* First, find the object */
	__nis_flush_one_dir(name);
	r = __nis_ocache_lookup(name, 0, MASTER_ONLY, &o);

	if (r != NIS_SUCCESS) {
		nis_err("warning: master failed to look up %s!", name);
		/*return(r);*/
	}

	/* Now find the directory in which it resides. */
	if (nis_dir_cmp(name, rootname) == SAME_NAME) {
		if (root_obj != NULL &&
			__i_am_master(&root_obj->DI_data) == TRUE)
			d = __nis_copy_dir(&root_obj->DI_data);
		else {
			nis_err("asked to ping root directory "
				     "but I'm not a root master!");
			return(NIS_NOT_ME);
		}
		r = nis_master_insert_ping(ROOT_OBJ, d);
	} else {
		r = __nis_dcache_lookup(nis_domain_of(name),
					0, MASTER_ONLY, &d);
		if (r != NIS_SUCCESS) {
			nis_err("master failed to find directory %s!",
							o->zo_domain);
			return(r);
		}
		d = __nis_copy_dir(d);
		r = nis_master_insert_ping(nis_domain_of(name), d);
	}

	/*
	 * Now that we're queued a ping for the directory in which
	 * the object resides, we need to check one more thing. If
	 * the object itself is also a directory, then it's possible
 	 * that the replicas serve this directory too, in addition to
 	 * the upper one that we just pinged. If this is the case,
	 * then we need to schedule a ping for this directory too.
	 * This is necessary in the case where you have an existing
	 * directory containing tables and you add a replica for it.
	 * The replica not only needs a copy of the directory object,
	 * it also needs the contents of the directory. However, if
	 * you only ping the parent directory to update the object,
	 * the replica won't be smart enough to realize that it also
	 * needs to transfer the directory contents. Thus, we prod
	 * it ourselves.
	 */

	if (o != NULL && __type_of(o) == DIRECTORY_OBJ) {
		d = __nis_copy_dir(&o->DI_data);
		r = nis_master_insert_ping(name, d);
	}

	if (r == NIS_SUCCESS)
		nis_event_add_conditional(nis_master_cb, NULL, NULL, 120,
					"master pinging replicas");

	return(r);
}

/*
 * Ping all the replicas of a given directory.
 */

static nis_error nis_master_ping_replicas(p)
	struct ping_list	*p;
{
	uint32_t		i;
	directory_obj		*d;
	nis_server		*s;
	int			errs = 0;

	if (verbose)
		nis_err("master pinging servers for %s",
							p->dname);
	d = p->dir;

	/*
	 * Ping all servers, except the first one: we don't want
	 * to ping ourselves.
	 */
	for (i = 1; i < d->do_servers.do_servers_len; i++) {
		s = &d->do_servers.do_servers_val[i];
		if (verbose)
			nis_err("pinging %s...", s->name);
		if (nis_ping_srv(p->dname, s, p->stamp) != NIS_SUCCESS) {
			nis_err("warning: unable to ping replica %s",
					s->name);
			errs++;
			continue;
		}
	}

	if (errs)
		return(NIS_S_SUCCESS);
	return(NIS_SUCCESS);
}

nis_error nis_master_do_updates()
{
	struct ping_list	*tmp = plist, *prev = NULL;
	uint32_t		i, cptime;
	nis_error		r;
	pid_t			cpid;

	if (tmp == NULL)
		return(NIS_SUCCESS);

	nis_create_readonly_child(&cpid);
	if (!cpid)
		return(NIS_SUCCESS);

	/*
	 * Check to see who's naughty and who's nice.
	 */

	while(tmp != NULL) {
		for (i = 1; i < tmp->dir->do_servers.do_servers_len; i++) {
			cptime = 0;
			if (tmp->dir->do_servers.do_servers_val[i].name == NULL)
				continue;
			nis_cptime(tmp->dir->do_name,
				   &tmp->dir->do_servers.do_servers_val[i],
				   &cptime);
			if (verbose)
				nis_err("checking update time"
					" for %s at %s", tmp->dname,
				tmp->dir->do_servers.do_servers_val[i].name);
			if (verbose) {
				nis_err("our time: %s",
						nis_ctime(tmp->stamp));
				nis_err("replica time: %s",
						nis_ctime(cptime));
			}
			if (cptime && cptime == tmp->stamp)
				if (verbose) {
					nis_err("%s is up to date",
				tmp->dir->do_servers.do_servers_val[i]);
				__nis_disqualify_server2(tmp->dir,
				   tmp->dir->do_servers.do_servers_val[i].name);
			}
		}
		if (tmp->dir->do_servers.do_servers_len == 1) {
			xdr_free(xdr_directory_obj, (char *)tmp->dir);
			free(tmp->dname);
			if (prev == NULL) {
				plist = tmp->next;
				free(tmp);
				tmp = plist;
			} else {
				prev->next = tmp->next;
				free(tmp);
				tmp = prev->next;
			}
			continue;
		}
		prev = tmp;
		tmp = tmp->next;
	}

	if (plist == NULL) {
		if (verbose)
			nis_err("pings complete -- "
					"all replicas up to date");
		nis_destroy_readonly_child(cpid);
		return(NIS_SUCCESS);
	}

	tmp = plist;

	/*
	 * Ping all the naughties.
	 */

	while(tmp != NULL) {
		r = nis_master_ping_replicas(tmp);
		if (r != NIS_SUCCESS)
			nis_err("failed to ping replicas for %s: %s",
					tmp->dname, nis_sperrno(r));
		tmp = tmp->next;
	}

	nis_destroy_readonly_child(cpid);

	return(NIS_TRYAGAIN);
}
