/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: nis_nislib.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <syslog.h>
#include <sys/cdefs.h>
#include <sys/fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "nis.h"
#include <rpcsvc/nis_tags.h>
#include "nis_extern.h"

#ifndef lint
static const char rcsid[] = "@(#) $Id: nis_nislib.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $";
#endif

/*
 * This module contains a local version of the __nis_internal_search() and
 * nis_finddirectory() routines from nislib. The idea behind
 * this code is to provide an way for rpc.nisd to safely call nis_lookup()
 * and nis_list() without encountering a deadlock condition where the
 * server might end up trying to make an RPC to itself.
 *
 * If the search code finds itself in a position where it
 * needs to make an RPC to the local server, the RPC call is replaced
 * with a call to a local svc routine.
 */

/*
 * Given a directory_obj, scan the list of servers that serve
 * dir->do_name and choose the one that replies the fastest.
 *
 * The reason we have this here is that the cache code in nislib(3)
 * keeps its own copies of directory objects, and it is these copies
 * that may be passed to this routine.
 */

nis_server *__nis_choose_server(bptr, flags)
	struct _nis_binding	*bptr;
	uint32_t		flags;
{
	nis_error		pstat;
	struct manual_binding	*m = bptr->nb_binding_manual;
	directory_obj		*dir = m->mb_dir;

	if (dir == NULL) {
		if (m->mb_srv == NULL)
			return(NULL);
		else {
			if (nis_dir_cmp(m->mb_srv->name, nis_local_host()) ==
						SAME_NAME)
				return(NULL);
			__nis_ping(bptr);
			return(m->mb_srv);
		}
	} else {
		if (flags & MASTER_ONLY && __i_am_master(dir) == TRUE)
			return(NULL);
		__nis_disqualify_server(dir, nis_local_host());
	}

	if (!dir->do_servers.do_servers_len)
		return(NULL);

	if (flags & MASTER_ONLY || dir->do_servers.do_servers_len == 1) {
		uint32_t		cnt;
		cnt = dir->do_servers.do_servers_len;
		dir->do_servers.do_servers_len = 1;
		__nis_ping(bptr);
		dir->do_servers.do_servers_len = cnt;
		return(m->mb_srv);
	}

	pstat = __nis_ping(bptr);

	return(m->mb_srv);
}

fd_result *nis_finddirectory(dir, name)
	directory_obj		*dir;
	nis_name		name;
{
	fd_result		*fd_res;
	directory_obj		*d;
	nis_error		r;
	struct db_objs		dres;
	uint32_t		size;
	/*
	 * If this is a directory served by us, then fabricate
	 * our own response.
	 */
	if (__is_local_dirname(name) == TRUE) {
		fd_res = calloc(1, sizeof(fd_result));
		fd_res->source = strdup(nis_local_host());
		/* First, check the cache */
		r = __nis_dcache_lookup(name, CACHE_ONLY, 0, &d);
		if (r == NIS_SUCCESS) {
			fd_res->status = r;
			fd_res->dir_data.dir_data_val =
				nis_encode_directory_obj(d, &size);
			fd_res->dir_data.dir_data_len = size;
			return(fd_res);
		}
		/* not in cache: try database */
		bzero((char *)&dres, sizeof(dres));
		r = db_lookup(name, &dres);
		if (r == DB_SUCCESS) {
			/* sanity check: make sure it's a directory */
			if (__type_of(NIS_RES_OBJECT(&dres)) != DIRECTORY_OBJ) {
				xdr_free(xdr_nis_object,
					(char *)NIS_RES_OBJECT(&dres));
				free(NIS_RES_OBJECT(&dres));
				fd_res->status = NIS_BADOBJECT;
				return(fd_res);
			}
			fd_res->status = db_to_nis_err(r);
			fd_res->dir_data.dir_data_val =
		nis_encode_directory_obj(&NIS_RES_OBJECT(&dres)->DI_data,
								&size);
			fd_res->dir_data.dir_data_len = size;
			xdr_free(xdr_nis_object, (char *)NIS_RES_OBJECT(&dres));
			free(NIS_RES_OBJECT(&dres));
			return(fd_res);
		}
	}

	return(nis_finddirectory_core(dir, name));
}

extern nis_result *__nis_search_core __P((nis_name, uint32_t,
					  uint32_t, nis_attr *,
					  uint32_t,
					  int (*)(nis_name,nis_object *,void *),
					  void *));

#define SEARCHOP_LIST	1
#define SEARCHOP_LOOKUP 2

nis_result *__nis_search(name, op, flags, attrs, numattrs,
					callback, userdata)
	nis_name		name;
	uint32_t		op;
	uint32_t		flags;
	nis_attr		*attrs;
	uint32_t		numattrs;
	int (*callback)(nis_name, nis_object *, void *);
	void			*userdata;
{
	bool_t			master_dir = FALSE, local_dir = FALSE;
	db_status		db_stat;
	struct db_objs		dres;
	nis_object		*o = NULL; 
	int			local_search = FALSE;

        if (nis_dir_cmp(name, rootname) == HIGHER_NAME) {
		nis_result		*res = NULL;
		if ((res = calloc(1, sizeof(nis_object))) == NULL)
			return(NULL);
		res->status = NIS_NOSUCHNAME;
		return(res);
	}

	bzero((char *)&dres, sizeof(dres));

	if (__is_local_dirname(name) == TRUE) {
		db_stat = db_lookup(name, &dres);
		if (db_stat == DB_SUCCESS) {
			o = NIS_RES_OBJECT(&dres);
			master_dir = __i_am_master(&o->DI_data);
			local_dir = __i_am_replica(&o->DI_data);
			xdr_free(xdr_nis_object, (char *)o);
			free(o);
		}
	} else if (__is_local_dirname(nis_domain_of(name)) == TRUE) {
		db_stat = db_lookup(nis_domain_of(name), &dres);
		if (db_stat == DB_SUCCESS) {
			o = NIS_RES_OBJECT(&dres);
			master_dir = __i_am_master(&o->DI_data);
			local_dir = __i_am_replica(&o->DI_data);
			xdr_free(xdr_nis_object, (char *)o);
			free(o);
		}
	}

	if (flags & MASTER_ONLY) {
		if (master_dir == TRUE)
			local_search = TRUE;
	} else {
		if (master_dir == TRUE || local_dir == TRUE)
			local_search = TRUE;
	}

	/*
	 * Three things to try for LOOKUP:
	 * 1) search object cache with __nis_ocache_lookup()
	 * 2) search database with db_lookup()
	 * 3) make RPC to other server (making sure to disqualify
	 *    ourselves so we don't deadlock)
	 */
	if (op == SEARCHOP_LOOKUP) {
		nis_result		*res = NULL;
		struct db_objs		dres;

		if (local_search == TRUE) {
			nis_object		*o;
			CACHE_CLOCK_DECLS;

			if ((res = calloc(1, sizeof(nis_object))) == NULL)
				return(NULL);

			/* Try cache first */
			__start_clock(CACHE_CLOCK);
			res->status = __nis_ocache_lookup(name, CACHE_ONLY,
								0, &o);
			res->aticks += __stop_clock(CACHE_CLOCK);
			if (res->status == NIS_SUCCESS) {
				NIS_RES_OBJECT(res) = nis_clone_object(o, NULL);
				NIS_RES_NUMOBJ(res) = 1;
				return(res);
			}

			res->status = db_to_nis_err(db_lookup(name, &dres));
			if (res->status == NIS_SUCCESS) {
				res->dticks += dres.dticks;
				__nis_ocache_insert(NIS_RES_OBJECT(&dres));
				NIS_RES_OBJECT(res) = NIS_RES_OBJECT(&dres);
				NIS_RES_NUMOBJ(res) = NIS_RES_NUMOBJ(&dres);
				return(res);
			} else {
				if (master_dir == FALSE) {
					nis_freeresult(res);
					res = NULL;
					goto do_lookup_rpc;
				}
			}
			return(res);
		} else {
do_lookup_rpc:
			if (verbose)
				nis_err("RPCING FOR %s", name);
			res = __nis_search_core(name, op, flags, attrs,
					numattrs, callback, userdata);

			if (res == NULL)
				return(NULL);
			if (res->status == NIS_SUCCESS)
				__nis_ocache_insert(NIS_RES_OBJECT(res));
			return(res);
		}
	}

	/*
	 * Two things to try for LIST:
	 * 1) search database with db_iblookup()
	 * 2) make RPC to other server (making sure to disqualify
	 *    ourselves so we don't deadlock)
	 */
	if (op == SEARCHOP_LIST) {
		nis_result		*res = NULL;
		struct db_objs		dres;

		if (local_search == TRUE) {
			if ((res = calloc(1, sizeof(nis_object))) == NULL)
				return(NULL);

			res->status = db_to_nis_err(db_iblookup(name,
						attrs, numattrs, &dres));
			if (res->status == DB_SUCCESS) {
				res->status = db_to_nis_err(res->status);
				NIS_RES_OBJECT(res) = NIS_RES_OBJECT(&dres);
				NIS_RES_NUMOBJ(res) = NIS_RES_NUMOBJ(&dres);
				res->dticks = dres.dticks;
				return(res);
			} else {
				if (master_dir == FALSE) {
					nis_freeresult(res);
					res = NULL;
					goto do_list_rpc;
				}
			}
			return(res);
		} else  {
do_list_rpc:
			res = __nis_search_core(name, op, flags, attrs,
					numattrs, callback, userdata);

			return(res);
		}
	}

	return(NULL);
}
