/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nis_replica.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/cdefs.h>
#include "nis.h"
#include "nis_extern.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nis_replica.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $";
#endif

/*
 * This module contains the replica support code. There are two exported
 * routines: nis_replica_add_update() and nis_replica_do_updates(). The
 * former allows us to queue an update that will be processed later when
 * the latter is called from the svc_run() loop.
 *
 * There are two possible ways for a replica to sync itself with its
 * master. In the first case, the replica is only slightly out of date
 * with the master, lacking perhaps only a few table changes. In this
 * case, the replica can ask the master for a copy of the transaction
 * log records needed to bring it up to date. This is done with the
 * nis_dumplog() routine.
 * 
 * In the second case, the replica is so far out of date with the master
 * that no amount of transaction log records can bring it back in line.
 * Here, the replica completely erases the contents of the directory
 * and recopies its entire contents from the master using nis_dump().
 * Since nis_dump() might need to transfer a lot of data, it uses a
 * callback.
 *
 * The replica needs to keep log entries indicating when its last update
 * took place so that it can compare notes with the master. It does this
 * by keeping UPD_STAMP entries in its transaction log with timestamps
 * indicating the directory update times. The problem here is that the
 * replica timestamps have to exactly match the master's update times
 * in order for both to stay in sync: the trick is getting the master to
 * cough up the desired timestamp value and recording it.
 *
 * When using nis_dump(), the master will make life easy for us by
 * providing a complete UPD_STAMP log entry as part of the procedure
 * result. The replica just has to write this entry to its log. (Note:
 * I think that older NIS+ servers circa Solaris 2.3 don't do this.)
 * An exception might be when updating the root object: I've observed
 * that sometimes the master doesn't send an UPD_STAMP entry in this
 * case.
 *
 * Things are harder when using nis_dumplog(). What you have to do
 * here is keep track of all the timestamps in each of the log entries
 * as they go by, saving the most recent one. This timestamp has to
 * be converted into an UPD_STAMP log entry and saved.
 *
 * As a further complication, we have to do special things when we
 * obtain updates concerning directory objects. If we get an ADD_NAME
 * entry for a directory, we have to check to see if we serve the
 * directory or not and, if so, create an underlying database to support
 * it. (This usually happens when someone does an nismkdir that creates
 * a new directory inside of an existing directory that is already
 * being replicated.)
 *
 * Lastly, we have to do certain magic for the root directory. The server
 * may submit ping requests to us for "root.object" which a replica must
 * interpret appropriately.
 */

/*
 * Apply a single update to a directory.
 */
static int nis_dump_cb(dirname, obj, udata)
	nis_name		dirname;
	nis_object		*obj;
	void			*udata;
{
	nis_error		db_err = NIS_FAIL;
	char			nbuf[NIS_MAXNAMELEN + 1];
	nis_object		*o;
	nis_error		cerr;
	ping_args		args;

	snprintf(nbuf, sizeof(nbuf), "%s.%s", obj->zo_name, obj->zo_domain);
	if (verbose)
		nis_err("dump callback applying %s", nbuf);

	switch(__type_of(obj)) {
	case ENTRY_OBJ:
		cerr = __nis_ocache_lookup(nbuf, CHECK_SERVER, MASTER_ONLY, &o);
		if (cerr == NIS_SUCCESS)
			db_err = db_to_nis_err(db_ibadd(nbuf, NULL,
						     0, obj, o, NULL));
		break;
	case DIRECTORY_OBJ:
	case LINK_OBJ:
	case TABLE_OBJ:
	case GROUP_OBJ:
	case PRIVATE_OBJ:
		db_remove(nbuf, NULL);
		db_err = db_to_nis_err(db_add(nbuf, obj, NULL));
		if (__type_of(obj) == TABLE_OBJ)
			db_newtab(nbuf, obj);
		if (__type_of(obj) == DIRECTORY_OBJ &&
			__i_am_replica(&obj->DI_data) == TRUE) {
			db_newdir(nbuf);
			args.stamp = 0;
			args.dir = nbuf;
			nis_replica_add_update(&args);
		}
		break;
	default:
		nis_err("trying to add invalid object");
		break;
	}

	if (db_err != NIS_SUCCESS)
		nis_err("warning: dump callback failed to add "
			     "object: %s", nis_sperrno(db_err));
	return(0);
}

static nis_error nis_replica_replay_log(lres)
	log_result		*lres;
{
	uint32_t		i;
	log_entry		*l;
	nis_error		r = NIS_FAIL;
	int			errs = 0;
	nis_object		*o;
	ping_args		args;
	uint32_t		xid;

	/*
	 * From what I can tell, the Sun NIS+ server treats
	 * each log entry read from the master as a separate
	 * transaction, hence each entry gets its own transaction ID.
	 * I don't agree with this: we obtain all entries in one
	 * transaction (a single call to nis_dumplog()), so logically
	 * they should all be applied in a single transaction.
	 */
	nis_log_new_xid(nis_local_principal(), &xid);

	if (verbose)
		nis_err("starting log update, %d entries",
			lres->lr_entries.lr_entries_len);

	for (i = 0; i < lres->lr_entries.lr_entries_len; i++) {
		l = &lres->lr_entries.lr_entries_val[i];
		if (verbose)
			nis_err("processing log entry for %s (%s)",
				 l->le_name, nis_ctime(l->le_time));

		switch(l->le_type) {
		case ADD_NAME:
			r = db_to_nis_err(db_add(l->le_name, &l->le_object,
									NULL));
			if (__type_of(&l->le_object) == DIRECTORY_OBJ &&
				__i_am_replica(&l->le_object.DI_data) == TRUE) {
				db_newdir(l->le_name);
				args.stamp = l->le_time;
				args.dir = l->le_name;
				nis_replica_add_update(&args);
			}
			if (__type_of(&l->le_object) == TABLE_OBJ)
				db_newtab(l->le_name, &l->le_object);
			nis_log_add_entry(l);
			break;
		case MOD_NAME_NEW:
			r = db_to_nis_err(db_add(l->le_name, &l->le_object,
									NULL));
			nis_log_add_entry(l);
			if (__type_of(&l->le_object) == DIRECTORY_OBJ &&
				__i_am_replica(&l->le_object.DI_data) == TRUE) {
				args.stamp = l->le_time;
				args.dir = l->le_name;
				nis_replica_add_update(&args);
			}
			break;
		case REM_NAME:
			r = db_to_nis_err(db_remove(l->le_name, NULL));
			if (__type_of(&l->le_object) == DIRECTORY_OBJ ||
				__type_of(&l->le_object) == TABLE_OBJ)
				db_rm(l->le_name);
			nis_log_add_entry(l);
			break;
		case ADD_IBASE:
			__nis_ocache_lookup(l->le_name, CHECK_SERVER,
						MASTER_ONLY, &o);
			r = db_to_nis_err(db_ibadd(l->le_name,
					   l->le_attrs.le_attrs_val,
					   l->le_attrs.le_attrs_len,
					   &l->le_object,
					   o, NULL));
			nis_log_add_entry(l);
			break;
		case REM_IBASE:
			__nis_ocache_lookup(l->le_name, CHECK_SERVER,
						MASTER_ONLY, &o);
			r = db_to_nis_err(db_ibremove(l->le_name,
					      l->le_attrs.le_attrs_val,
					      l->le_attrs.le_attrs_len,
					      &l->le_object,
					      o, NULL));
			nis_log_add_entry(l);
			break;
		case UPD_STAMP:
			nis_log_add_entry(l);
			r = NIS_SUCCESS;
			break;
		case MOD_IBASE:	/* XXX never used */
			break;
		case MOD_NAME_OLD:
			nis_log_add_entry(l);
		case LOG_NOP:
		default:
			r = NIS_SUCCESS;
			break;
		}
		if (r != NIS_SUCCESS) {
			nis_err("applying log entries: %s",
					 nis_sperrno(r));
			errs++;
		}
	}

	/*
	 * Now timestamp any modified directories and save
	 * log entries.
	 */
	nis_log_commit_xid(xid);

	if (errs)
		return(NIS_S_SUCCESS);
	return(NIS_SUCCESS);
}

static nis_error nis_replica_dumplog(dirname, master)
	nis_name		dirname;
	nis_server		*master;
{
	uint32_t		utime = 0;
	log_result		*res;
	nis_error		s;

	utime = nis_log_find_update(dirname, 0, NULL);

	if (verbose)
		nis_err("dumplog: our last update "
			     "time is: %s", nis_ctime(utime));

	/*
	 * This saves an RPC. If we already know that our last
	 * update time is 'never,' then we should proceed directly
	 * to nis_dump().
	 */
	if (!utime) {
		if (verbose)
			nis_err("our last update is 'never' -- "
				     "forcing a resync");
		return(NIS_RESYNC);
	}

	res = nis_dumplog(master, dirname, utime);

	if (res == NULL)
		return(NIS_NOMEMORY);

	s = res->lr_status;

	/*
	 * If the master doesn't have enough log entries to
	 * satisfy us, it will tell us to resync.
	 */
	if (res->lr_status != NIS_SUCCESS) {
		xdr_free(xdr_log_result, (char *)res);
		free(res);
		if (verbose)
			nis_err("dumplog: %s", nis_sperrno(s));
		return(s);
	}

	/*
	 * Master gave us a bunch of entries to apply to our
	 * databases; roll them in.
	 */
	nis_replica_replay_log(res);

	xdr_free(xdr_log_result, (char *)res);
	free(res);

	return(NIS_SUCCESS);
}

static nis_error nis_replica_dump(dirname, master)
	nis_name		dirname;
	nis_server		*master;
{
	log_result		*res;
	nis_error		s;
	uint32_t		xid;

	/*
	 * First step in resynchronizing is to throw out
	 * all the existing stuff in a directory.
	 */
	if (db_exists(dirname)) {
		if (nis_empty_directory(dirname) != NIS_SUCCESS) {
			nis_err("dumplog: failed to empty %s", dirname);
			return(NIS_NOTEMPTY);
		}
	}

	/* Now recreate the directory from scratch. */
	db_newdir(dirname);
	if (nis_is_in_srvlist(dirname) == FALSE)
		nis_add_to_srvlist(dirname);

	/*
	 * Now repopulate it from the master.
 	 * The nis_dump_cb() callback handler will be invoked
	 * called to add in each new object as it arrives.
	 */
	res = nis_dump(master, dirname, nis_dump_cb);

	if (res == NULL)
		return(NIS_NOMEMORY);

	s = res->lr_status;

	nis_log_new_xid(nis_local_principal(), &xid);
	if (s == NIS_SUCCESS || s == NIS_CBRESULTS) {
		/*
		 * If a log entry was supplied here, use it to update
		 * the log. If not, then there are perhaps one of two
		 * things happening:
		 *
		 * 1) This is the root object we're updating. I think
		 *    Sun NIS+ servers sometimes don't supply log entries
		 *    for changes made to the root object.
		 * 2) The master for this directory is a very old Sun NIS+
		 *    server, which doesn't support the practice of
		 *    returning an UPD_STAMP log entry in the nis_dump()
		 *    result.
		 *
		 * For both these cases, we need to manually check
		 * the update time of the directory object and
		 * fabricate an UPD_STAMP.
		 */
		if (!res->lr_entries.lr_entries_len ||
		     res->lr_entries.lr_entries_val == NULL) {
				uint32_t		u;

				if (nis_dir_cmp(dirname, rootname)
							 == SAME_NAME)
					nis_err("warning: ancient "
						"version of rpc.nisd "
						"running as root master "
						"for directory %s", dirname);
				if (nis_cptime(dirname,
					master, &u) == NIS_SUCCESS)
					nis_log_insert_stamp(dirname, u);
				else
					nis_err("warning: couldn't "
						"create update stamp for "
						"directory %s", dirname);
		} else {
			nis_log_add_entry(res->lr_entries.lr_entries_val);
		}
	} else
		nis_err("nis_dump failed: %s", nis_sperrno(s));

	xdr_free(xdr_log_result, (char *)res);
	free(res);
	nis_log_commit_xid(xid);
	if (s == NIS_SUCCESS || s == NIS_CBRESULTS)
		return(NIS_SUCCESS);
	return(s);
}

static nis_error nis_replica_update(dirname, since)
	nis_name		dirname;
	uint32_t		since;
{
	nis_server		*master;
	directory_obj		*d;
	nis_error		cerr;

	if (verbose)
		nis_err("processing update of %s", dirname);

	__nis_flush_one_dir(dirname);
	cerr = __nis_dcache_lookup(dirname, CHECK_SERVER|LOOKUP_ONLY,
					    MASTER_ONLY|NO_CACHE, &d);
	if (cerr != NIS_SUCCESS)
		return(cerr);

	/* Sanity check */
	if (__i_am_replica(d) == FALSE) {
		nis_err("update_replica: asked to replicate directory "
			     "%s which I do not serve!", dirname);
		return(NIS_NOT_ME);
	}

	/*
	 * First of all, this could be a directory for which we have
	 * just been made a replica. If this is the case, we don't have
	 * an underlying table that describes it yet, so we should
	 * make one.
	 *
	 * If this is the root directory, then we should create a
	 * table file called 'root_dir.'
	 */
	cerr = db_to_nis_err(db_newdir(dirname));
	if (cerr != NIS_SUCCESS && cerr != NIS_NOTUNIQUE) {
		nis_err("replica_update: failed to create new directory "
			     "%s: %s", dirname, nis_sperrno(cerr));
		return(cerr);
	}

	/*
	 * Also add to serving list if necessary.
	 */
	if (nis_is_in_srvlist(dirname) == FALSE)
		nis_add_to_srvlist(dirname);

	/*
	 * If the directory is also the local directory,
	 * update the cold start object too.
	 */
	if (nis_dir_cmp(dirname, nis_local_directory()) == SAME_NAME)
		nis_update_cold();

	/*
	 * Updating the cold start object may have forced a cache
	 * flush. Read the directory object again to be safe.
	 */
	cerr = __nis_dcache_lookup(dirname, CHECK_SERVER|LOOKUP_ONLY,
					    MASTER_ONLY|NO_CACHE, &d);
	if (cerr != NIS_SUCCESS)
		return(cerr);

	master = &d->do_servers.do_servers_val[0];

	/*
	 * Easy case: now try to pull over pending log entries from
	 * the master.
	 */
	cerr = nis_replica_dumplog(dirname, master);

	if (cerr == NIS_SUCCESS || cerr == NIS_DUMPLATER)
		return(cerr);

	/*
	 * Difficult case: we are so totally out of date that we
	 * have to completely tear down and rebuild the directory
	 * from scratch.
	 */
	if (cerr == NIS_RESYNC)
		cerr = nis_replica_dump(dirname, master);

	return(cerr);
}

nis_error nis_replica_update_root()
{
	nis_object		*o;
	nis_name		newroot;
	nis_error		r;

	newroot = __nis_find_root();
	free(rootname);
	rootname = newroot;

	if (rootname == NULL) {
		nis_err("failed to learn new name of namespace root!");
		return(NIS_FAIL);
	}

	r = __nis_ocache_lookup(rootname, LOOKUP_ONLY,
				MASTER_ONLY|NO_CACHE, &o);

	if (r != NIS_SUCCESS) {
		nis_err("failed to update root object! (%s)", rootname);
		return(NIS_FAIL);
	}

	if (__type_of(o) != DIRECTORY_OBJ)
		return(NIS_INVALIDOBJ);

	if (__is_local_dir(&o->DI_data) == TRUE) {
		nis_put_root_obj(o);
		nis_get_root_obj(&root_obj);
        	if (db_table_exists("root_dir") != DB_SUCCESS) {
			if (db_newdir(rootname) != DB_SUCCESS)
			nis_err("failed to create root directory");
		}
		nis_add_to_srvlist(rootname);
	} else {
		nis_destroy_object(root_obj);
		unlink(_PATH_NIS_ROOT_OBJ);
		root_obj = NULL;
		nis_remove_from_srvlist(rootname);
	}

	if (nis_dir_cmp(rootname, nis_local_directory()) == SAME_NAME)
		nis_update_cold();

	return(NIS_SUCCESS);
}

struct update_list {
	nis_name		dir;
	uint32_t		stamp;
	nis_error		stat;
	struct update_list	*next;
};

static struct update_list 	*ulist = NULL;

/*
 * Event callback for replica updates.
 */
bool_t nis_replica_cb(args, id, reason)
	void			*args;
	uint32_t		id;
	int			reason;
{
	if (nis_replica_do_updates() == NIS_SUCCESS)
		return(FALSE);
	else
		return(TRUE);
}

/*
 * We've been pinged; schedule a rendez-vous with the master.
 */
nis_error nis_replica_add_update(p_args)
	ping_args		*p_args;
{
	struct update_list	*tmp, *prev;

	if (verbose)
		nis_err("replica scheduling update for: %s", p_args->dir);

	tmp = prev = ulist;
	while(tmp != NULL) {
		if (nis_dir_cmp(p_args->dir, tmp->dir) == SAME_NAME) {
			if (verbose) 
				nis_err("update for %s already queued",
								tmp->dir);
			return(NIS_SUCCESS);
		}
		prev = tmp;
		tmp = tmp->next;
	}

	tmp = malloc(sizeof(struct update_list));
	if (tmp == NULL)
		return(NIS_NOMEMORY);

	tmp->dir = strdup(p_args->dir);
	tmp->stamp = p_args->stamp;
	tmp->stat = NIS_FAIL;
	tmp->next = NULL;

	if (prev == NULL)
		ulist = tmp;
	else
		prev->next = tmp;

	nis_event_add_conditional(nis_replica_cb, NULL, NULL,
					20, "replica update");

	return(NIS_SUCCESS);
}

/*
 * Process pending updates.
 */
nis_error nis_replica_do_updates()
{
	struct update_list	*tmp = ulist, *ptr;
	int			tries = 0, errs = 0;
	int			childpid = 0;

	if (tmp == NULL)
		return(NIS_SUCCESS);

	nis_create_readonly_child(&childpid);
	if (!childpid)
		return(NIS_SUCCESS);

	while(tmp != NULL) {
		if (tmp->stat != NIS_SUCCESS) {
			tries++;
			if (nis_dir_cmp(tmp->dir, ROOT_OBJ) == SAME_NAME ||
			    nis_dir_cmp(tmp->dir, rootname) == SAME_NAME) {
				tmp->stat = nis_replica_update(rootname,
								tmp->stamp);
				nis_replica_update_root();
			} else {
				tmp->stat = nis_replica_update(tmp->dir,
								tmp->stamp);
			}
			if (tmp->stat != NIS_SUCCESS && tmp->stat != NIS_NOT_ME)
				errs++;
		}
		tmp = tmp->next;
	}

	/*
	 * If there are no updates pending, free
	 * the list.
	 */
	if (tries == 0 || errs == 0) {
		tmp = ulist;
		while(tmp != NULL) {
			free(tmp->dir);
			ptr = tmp->next;
			free(tmp);
			tmp = ptr;
		}
		ulist = NULL;
	}

	nis_destroy_readonly_child(childpid);

	if (ulist != NULL)
		return(NIS_TRYAGAIN);

	return(NIS_SUCCESS);
}
