/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id$
 */

#include <sys/types.h>
#include <sys/cdefs.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef USE_SECURE_RPC
#include <rpc/des_crypt.h>
#endif
#include <rpcsvc/nis_db.h>

#include "nis_extern.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nis_server.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $";
#endif /* not lint */

/*
 * The NIS_LOOKUP procedure checks for and retrieves an NIS+ object.
 * The specified name must be fully qualified and can not be an indexed
 * name, which means you can not use NIS_LOOKUP to retrieve entry objects
 * (you should use NIS_IBLIST) instead).
 *
 * This procedure has the follwing behavior:
 *
 * - A principal name is generated for the purpose of access rigths
 *   checks. A calling principal must have read access to an object
 *   before being allowed to retrieve a copy.
 *
 * NIS_LOOKUP calls nis_db_nslookup() to do the actual object retrieval.
 * The nslookup routine will perform the actual db search and do the required
 * access rights checks.
 */
nis_result *
nis_lookup_3_svc(ns_request *argp, struct svc_req *rqstp)
{
	nis_result		result;
	nis_name		principal;
	char			_pbuf[NIS_MAXNAMELEN];
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	__start_clock(SERVER_CLOCK);

	bzero((char *)&result, sizeof(result));

	if (verbose)
		nis_err("nis_lookup for %s from %s", argp->ns_name,
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	principal = nis_principal(rqstp, _pbuf);

	result.status = nis_db_nslookup(argp->ns_name, &result, principal);
	result.zticks = __stop_clock(SERVER_CLOCK);

	nis_update_stats(NIS_LOOKUP,
			 result.status == NIS_SUCCESS ? TRUE : FALSE,
			 result.zticks);

	if (!svc_sendreply(rqstp->rq_xprt, xdr_nis_result, (char *)&result))
		nis_err("sendreply in nis_lookup failed");

	if (verbose)
		nis_err("result: %s", nis_sperrno(result.status));

	xdr_free(xdr_nis_result, (char *)&result);

	return(NULL);
}

/*
 * The nis_add() service is used to add new entities to the NIS
 * hierarchy. This is a bit confusing because nis_add() can do
 * different things depending on what sort of entity we're creating.
 * For directories, we add an nis_object struct to a parent directory
 * that basically describes the directory to be created and lets the
 * system know that it exists. The nis_mkdir() procedure is what actually
 * creates the 'table' file representing the directory. For NIS tables,
 * we create the table ourselves. For links, we don't have to create any
 * new table files, we just add new nis_object structs into existing ones.
 *
 * This is actually a little strange. It seems odd that there should be
 * a seperate function for creating a directory but not for creating a
 * table (ibcreate/ibdestroy, for example). It's possible I'll stumble on
 * the reason for this as things progress, but for now, I'm a little
 * baffled.
 */

nis_result *
nis_add_3_svc(ns_request *argp, struct svc_req *rqstp)
{
	static nis_result	result;
	register nis_name	principal;
	char			_pbuf[NIS_MAXNAMELEN];
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	__start_clock(SERVER_CLOCK);
	bzero((char *)&result, sizeof(result));

	if (verbose)
		nis_err("nis_add for %s from %s", argp->ns_name,
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	if (readonly) {
		result.status = NIS_TRYAGAIN;
		result.zticks = __stop_clock(SERVER_CLOCK);
		if (verbose)
			nis_err("result: %s", nis_sperrno(result.status));
		return(&result);
	}

	principal = nis_principal(rqstp, _pbuf);

	result.status = nis_db_nsop(argp->ns_name,
					argp->ns_object.ns_object_val,
					principal, NIS_ADD, &result);

	result.zticks = __stop_clock(SERVER_CLOCK);
	nis_update_stats(NIS_ADD,
			 result.status == NIS_SUCCESS ? TRUE : FALSE,
			 result.zticks);

	if (verbose)
		nis_err("result: %s", nis_sperrno(result.status));

	return(&result);
}

/*
 * Modify the nis_object structure that describes an NIS object of
 * some kind (directory, table, whatever).
 */

nis_result *
nis_modify_3_svc(ns_request *argp, struct svc_req *rqstp)
{
	static nis_result	result;
	register nis_name	principal;
	char			_pbuf[NIS_MAXNAMELEN];
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	__start_clock(SERVER_CLOCK);

	bzero((char *)&result, sizeof(result));

	if (verbose)
		nis_err("nis_modify for %s from %s", argp->ns_name,
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	if (readonly) {
		result.status = NIS_TRYAGAIN;
		result.zticks = __stop_clock(SERVER_CLOCK);
		if (verbose)
			nis_err("result: %s", nis_sperrno(result.status));
		return(&result);
	}

	principal = nis_principal(rqstp, _pbuf);

	result.status = nis_db_nsop(argp->ns_name,
				    argp->ns_object.ns_object_val,
				    principal, NIS_MODIFY, &result);

	result.zticks = __stop_clock(SERVER_CLOCK);
	nis_update_stats(NIS_MODIFY,
			 result.status == NIS_SUCCESS ? TRUE : FALSE,
			 result.zticks);
	if (verbose)
		nis_err("result: %s", nis_sperrno(result.status));

	return(&result);
}

/*
 * Nuke an nis_object that describes an NIS entity of some sort
 * (table, directory, whatever).
 */

nis_result *
nis_remove_3_svc(ns_request *argp, struct svc_req *rqstp)
{
	static nis_result	result;
	register nis_name	principal;
	char			_pbuf[NIS_MAXNAMELEN];
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	__start_clock(SERVER_CLOCK);
	bzero((char *)&result, sizeof(result));

	if (verbose)
		nis_err("nis_remove for %s from %s", argp->ns_name,
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	if (readonly) {
		result.status = NIS_TRYAGAIN;
		result.zticks = __stop_clock(SERVER_CLOCK);
		if (verbose)
			nis_err("result: %s", nis_sperrno(result.status));
		return(&result);
	}

	principal = nis_principal(rqstp, _pbuf);

	result.status = nis_db_nsop(argp->ns_name,
				    argp->ns_object.ns_object_val,
				    principal, NIS_REMOVE, &result);

	result.zticks = __stop_clock(SERVER_CLOCK);
	nis_update_stats(NIS_REMOVE,
			 result.status == NIS_SUCCESS ? TRUE : FALSE,
			 result.zticks);
	if (verbose)
		nis_err("result: %s", nis_sperrno(result.status));
	return(&result);
}

/*
 * The iblist routine. This function can be used to lookup explicit
 * entries in a table or dump the whole table. It can be called in
 * the following ways:
 * - With an object name and search criteria. In this case, the
 *   object is a table and the search criteria indicate entries
 *   we want to find. A callback endpoint may be specified.
 * - With an object name and no search criteria. The object is a
 *   table name; the lack of search criteria indicates that the
 *   caller wants to dump the whole table. A callback may be specified.
 * - With an object name and no search criteria. The object is a
 *   directory name. Here, the directory contents are dumped (nisls(1)
 *   does this). Callbacks are not supported, so an NIS_NOCALLBACK error
 *   is returned if the caller specified a callback endpoint.
 *
 * Handling returns with callbacks is tricky: if the caller specifies
 * a callback endpoint but the iblist would result in an NIS_PARTIAL
 * error, we have to send back the table object in the nis_result struct
 * and the NIS_PARTIAL error via callback. The Solaris NIS+ client code
 * expects this behavior and tends to SEVG if it doesn't get it.
 *
 * Note: the Sun NIS+ server only supports callbacks in the
 * nis_iblist() procedure.
 */
nis_result *
nis_iblist_3_svc(ib_request *argp, struct svc_req *rqstp)
{
	nis_result		result;
	nis_result		cb_result;
	nis_object		*obj;
	nis_object		*head = NULL, *list = NULL;
	int			i = 0;
	nis_name		principal;
	char			_pbuf[NIS_MAXNAMELEN];
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	__start_clock(SERVER_CLOCK);

	bzero((char *)&result, sizeof(nis_result));
	bzero((char *)&cb_result, sizeof(nis_result));

	if (verbose)
		nis_err("nis_iblist for %s from %s", argp->ibr_name,
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));
	/*
	 * Client wants a callback; prepare some useful things.
	 */
	if (argp->ibr_cbhost.ibr_cbhost_len) {
		nis_make_cookie(rqstp, &cb_result.cookie);
		result.cookie.n_len = cb_result.cookie.n_len;
		result.cookie.n_bytes = cb_result.cookie.n_bytes;
		cb_result.status = NIS_CBRESULTS;
	}

	principal = nis_principal(rqstp, _pbuf);

	/*
	 * Check the callback host information and see if a
	 * public key is included. If the caller used DES
	 * authentication and didn't supply a public key,
	 * find it ourselves.
	 */
	if (argp->ibr_cbhost.ibr_cbhost_len) {
		nis_server		*s;
		char			*p;

		s = &argp->ibr_cbhost.ibr_cbhost_val[0];
		if (rqstp->rq_cred.oa_flavor == AUTH_DES && !s->pkey.n_len) {
			p = __nis_getpkey(principal, NULL);
			if (p != NULL) {
				s->pkey.n_len = strlen(p) + 1;
				s->pkey.n_bytes = p;
				s->key_type = NIS_PK_DH;
			}	
		}
	}
	/*
	 * Special case: caller asked for a callback and wants
	 * to dump the whole table. We have a special callback routine
	 * to deal with that.
	 */
	if (argp->ibr_cbhost.ibr_cbhost_len &&
		!argp->ibr_srch.ibr_srch_len) {

		/* Note: callback routine free()s the cookie. */
		if (nis_callback_all(argp->ibr_cbhost.ibr_cbhost_val,
			argp->ibr_name, principal, argp->ibr_flags,
			rqstp, &cb_result.cookie, SERVER_CLOCK) != NIS_SUCCESS)
			nis_err("warning: callback failed");

		if (verbose)
			nis_err("replying via callback");
		nis_update_stats(NIS_IBLIST, TRUE, __stop_clock(SERVER_CLOCK));
		return(NULL);
	}

	if (argp->ibr_srch.ibr_srch_len) {
	/*
	 * Retrieve objects matching specified search criteria.
	 */
		result.status = nis_db_iblist(argp->ibr_name, argp->ibr_flags,
					      argp->ibr_srch.ibr_srch_len,
					      argp->ibr_srch.ibr_srch_val,
					      &result, principal);

	} else {
		netobj			cookie;
		netobj			nextp = { 0, NULL };

		cookie.n_len = result.cookie.n_len;
		cookie.n_bytes = result.cookie.n_bytes;
	/*
	 * Dump the whole table.
	 */
		result.status = nis_db_ibfirst(argp->ibr_name, argp->ibr_flags,
					       argp->ibr_srch.ibr_srch_len,
					       argp->ibr_srch.ibr_srch_val,
					       &result, principal);

		if (result.status == NIS_SUCCESS) {
			head = list = (nis_object *)malloc(sizeof(nis_object));
			while (result.status == NIS_SUCCESS) {
				head = realloc(head, (i+1) *
						 sizeof(nis_object));
				list = head + i;
				obj = NIS_RES_OBJECT(&result);
				bcopy(obj, list, sizeof(nis_object));
				free(obj);
				i++;
				NIS_RES_OBJECT(&result) = NULL;
				NIS_RES_NUMOBJ(&result) = 0;
				nextp.n_len = result.cookie.n_len;
				nextp.n_bytes = result.cookie.n_bytes;
				result.cookie.n_len = 0;
				result.cookie.n_bytes = NULL;
				xdr_free(xdr_nis_result, (char *)&result);
				result.status = nis_db_ibnext(argp->ibr_name,
							      argp->ibr_flags,
							      &nextp, &result,
							      principal);
			}
			if (result.status == NIS_NOTFOUND)
				result.status = NIS_SUCCESS;
			result.cookie.n_len = cookie.n_len;
			result.cookie.n_bytes = cookie.n_bytes;
		}
	}

	if (list != NULL) {
		result.objects.objects_len = i;
		result.objects.objects_val = head;
	}

	if (verbose)
		nis_err("result: %s", nis_sperrno(result.status));

	if (cb_result.status == NIS_CBRESULTS) {
		if (result.status != NIS_SUCCESS) {
			nis_error		lstat = result.status;
			result.status = NIS_CBRESULTS;
			result.zticks = __stop_clock(SERVER_CLOCK);
			svc_sendreply(rqstp->rq_xprt, xdr_nis_result,
						(char *)&result);
			nis_update_stats(NIS_IBLIST,
				result.status == NIS_SUCCESS ? TRUE : FALSE,
				result.zticks);
			xdr_free(xdr_nis_result, (char *)&result);
			bzero((char *)&result, sizeof(result));
			result.status = lstat;
		} else {
			cb_result.dticks = result.dticks;
			cb_result.zticks = __stop_clock(SERVER_CLOCK);
			svc_sendreply(rqstp->rq_xprt, xdr_nis_result,
						(char *)&cb_result);
			nis_update_stats(NIS_IBLIST,
				result.status == NIS_SUCCESS ? TRUE : FALSE,
				result.zticks);
		}
		nis_callback_batch(result.objects.objects_val,
					result.objects.objects_len,
					argp->ibr_cbhost.ibr_cbhost_val,
					result.status, rqstp,
					&cb_result.cookie);
		cb_result.cookie.n_len = 0;
		cb_result.cookie.n_bytes = NULL;
		result.cookie.n_len = 0;
		result.cookie.n_bytes = NULL;
		result.objects.objects_val = NULL;
		result.objects.objects_len = 0;
		xdr_free(xdr_nis_result, (char *)&result);
	} else {
		result.zticks = __stop_clock(SERVER_CLOCK);
		if (!svc_sendreply(rqstp->rq_xprt, xdr_nis_result,
							 (char *)&result))
			nis_err("sendreply in nis_iblist failed");
		nis_update_stats(NIS_IBLIST,
			 result.status == NIS_SUCCESS ? TRUE : FALSE,
			 result.zticks);
		xdr_free(xdr_nis_result, (char *)&result);
	}

	/* Should free all the memory we allocated. */
	xdr_free(xdr_nis_result, (char *)&cb_result);
	
	return(NULL);
}

nis_result *
nis_ibadd_3_svc(ib_request *argp, struct svc_req *rqstp)
{
	static nis_result	result;
	nis_name		principal;
	char			_pbuf[NIS_MAXNAMELEN];
	SERVER_CLOCK_DECLS;


	if (__nis_child_must_die)
		return(NULL);

	__start_clock(SERVER_CLOCK);

	bzero((char *)&result, sizeof(nis_result));

	if (verbose)
		nis_err("nis_ibadd for %s from %s", argp->ibr_name,
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	if (readonly) {
		result.status = NIS_TRYAGAIN;
		result.zticks = __stop_clock(SERVER_CLOCK);
		if (verbose)
			nis_err("result: %s", nis_sperrno(result.status));
		return(&result);
	}

	if (argp->ibr_cbhost.ibr_cbhost_len) {
		result.status = NIS_NOCALLBACK;
		result.zticks = __stop_clock(SERVER_CLOCK);
		if (verbose)
			nis_err("result: %s", nis_sperrno(result.status));
		return(&result);
	}

	principal = nis_principal(rqstp, _pbuf);

	result.status = nis_db_ibop(argp->ibr_name, argp->ibr_flags,
					argp->ibr_srch.ibr_srch_len,
					argp->ibr_srch.ibr_srch_val,
					argp->ibr_obj.ibr_obj_val,
					principal, NIS_IBADD, &result);

	result.zticks = __stop_clock(SERVER_CLOCK);
	if (!svc_sendreply(rqstp->rq_xprt, xdr_nis_result, (char *)&result))
		nis_err("sendreply failed in nis_ibadd");

	nis_update_stats(NIS_IBADD,
			 result.status == NIS_SUCCESS ? TRUE : FALSE,
			 result.zticks);

	if (verbose)
		nis_err("result: %s", nis_sperrno(result.status));

	xdr_free(xdr_nis_result, (char *)&result);

	return(NULL);
}

nis_result *
nis_ibmodify_3_svc(ib_request *argp, struct svc_req *rqstp)
{
	static nis_result	result;
	nis_name		principal;
	char			_pbuf[NIS_MAXNAMELEN];
	SERVER_CLOCK_DECLS;


	if (__nis_child_must_die)
		return(NULL);

	__start_clock(SERVER_CLOCK);

	bzero((char *)&result, sizeof(nis_result));

	if (verbose)
		nis_err("nis_ibmodify for %s from %s", argp->ibr_name,
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	if (readonly) {
		result.status = NIS_TRYAGAIN;
		result.zticks = __stop_clock(SERVER_CLOCK);
		if (verbose)
			nis_err("result: %s", nis_sperrno(result.status));
		return(&result);
	}

	if (argp->ibr_cbhost.ibr_cbhost_len) {
		result.status = NIS_NOCALLBACK;
		result.zticks = __stop_clock(SERVER_CLOCK);
		if (verbose)
			nis_err("result: %s", nis_sperrno(result.status));
		return(&result);
	}

	principal = nis_principal(rqstp, _pbuf);

	bzero((char *)&result, sizeof(nis_result));
	result.status = nis_db_ibop(argp->ibr_name, argp->ibr_flags,
					argp->ibr_srch.ibr_srch_len,
					argp->ibr_srch.ibr_srch_val,
					argp->ibr_obj.ibr_obj_val,
					principal, NIS_IBMODIFY, &result);

	result.zticks = __stop_clock(SERVER_CLOCK);
	if (!svc_sendreply(rqstp->rq_xprt, xdr_nis_result, (char *)&result))
		nis_err("sendreply failed in nis_ibmodify");

	nis_update_stats(NIS_IBMODIFY,
			 result.status == NIS_SUCCESS ? TRUE : FALSE,
			 result.zticks);

	if (verbose)
		nis_err("result: %s", nis_sperrno(result.status));

	xdr_free(xdr_nis_result, (char *)&result);

	return(NULL);
}

nis_result *
nis_ibremove_3_svc(ib_request *argp, struct svc_req *rqstp)
{
	static nis_result	result;
	nis_name		principal;
	char			_pbuf[NIS_MAXNAMELEN];
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	__start_clock(SERVER_CLOCK);

	bzero((char *)&result, sizeof(nis_result));

	if (verbose)
		nis_err("nis_ibremove for %s from %s", argp->ibr_name,
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	if (readonly) {
		result.status = NIS_TRYAGAIN;
		result.zticks = __stop_clock(SERVER_CLOCK);
		if (verbose)
			nis_err("result: %s", nis_sperrno(result.status));
		return(&result);
	}

	if (argp->ibr_cbhost.ibr_cbhost_len) {
		result.status = NIS_NOCALLBACK;
		result.zticks = __stop_clock(SERVER_CLOCK);
		if (verbose)
			nis_err("result: %s", nis_sperrno(result.status));
		return(&result);
	}

	principal = nis_principal(rqstp, _pbuf);

	bzero((char *)&result, sizeof(nis_result));
	result.status = nis_db_ibop(argp->ibr_name, argp->ibr_flags,
					argp->ibr_srch.ibr_srch_len,
					argp->ibr_srch.ibr_srch_val,
					argp->ibr_obj.ibr_obj_val,
					principal, NIS_IBREMOVE, &result);

	result.zticks = __stop_clock(SERVER_CLOCK);
	if (!svc_sendreply(rqstp->rq_xprt, xdr_nis_result, (char *)&result))
		nis_err("sendreply failed in nis_ibremove");

	nis_update_stats(NIS_IBREMOVE,
			 result.status == NIS_SUCCESS ? TRUE : FALSE,
			 result.zticks);

	if (verbose)
		nis_err("result: %s", nis_sperrno(result.status));

	xdr_free(xdr_nis_result, (char *)&result);

	return(NULL);
}

nis_result *
nis_ibfirst_3_svc(ib_request *argp, struct svc_req *rqstp)
{
	static nis_result	result;
	nis_name		principal;
	char			_pbuf[NIS_MAXNAMELEN];
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	__start_clock(SERVER_CLOCK);

	bzero((char *)&result, sizeof(nis_result));

	if (verbose)
		nis_err("nis_ibfirst for %s from %s", argp->ibr_name,
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	if (argp->ibr_cbhost.ibr_cbhost_len) {
		result.status = NIS_NOCALLBACK;
		result.zticks = __stop_clock(SERVER_CLOCK);
		if (verbose)
			nis_err("result: %s", nis_sperrno(result.status));
		return(&result);
	}

	principal = nis_principal(rqstp, _pbuf);

	bzero((char *)&result, sizeof(nis_result));
	result.status = nis_db_ibfirst(argp->ibr_name, argp->ibr_flags,
					argp->ibr_srch.ibr_srch_len,
					argp->ibr_srch.ibr_srch_val,
					&result, 
					principal);

	result.zticks = __stop_clock(SERVER_CLOCK);
	if (!svc_sendreply(rqstp->rq_xprt, xdr_nis_result, (char *)&result))
		nis_err("sendreply failed in nis_ibfirst");

	nis_update_stats(NIS_IBFIRST,
			 result.status == NIS_SUCCESS ? TRUE : FALSE,
			 result.zticks);

	if (verbose)
		nis_err("result: %s", nis_sperrno(result.status));

	xdr_free(xdr_nis_result, (char *)&result);

	return(NULL);
}

nis_result *
nis_ibnext_3_svc(ib_request *argp, struct svc_req *rqstp)
{
	static nis_result  result;
	nis_name		principal;
	char			_pbuf[NIS_MAXNAMELEN];
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	__start_clock(SERVER_CLOCK);

	bzero((char *)&result, sizeof(nis_result));

	if (verbose)
		nis_err("nis_ibnext for %s from %s", argp->ibr_name,
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	if (argp->ibr_cbhost.ibr_cbhost_len) {
		result.status = NIS_NOCALLBACK;
		result.zticks = __stop_clock(SERVER_CLOCK);
		if (verbose)
			nis_err("result: %s", nis_sperrno(result.status));
		return(&result);
	}

	principal = nis_principal(rqstp, _pbuf);

	bzero((char *)&result, sizeof(nis_result));
	result.status = nis_db_ibnext(argp->ibr_name, argp->ibr_flags,
					&argp->ibr_cookie,
					&result, 
					principal);

	result.zticks = __stop_clock(SERVER_CLOCK);
	if (!svc_sendreply(rqstp->rq_xprt, xdr_nis_result, (char *)&result))
		nis_err("sendreply failed in nis_ibnext");

	nis_update_stats(NIS_IBNEXT,
			 result.status == NIS_SUCCESS ? TRUE : FALSE,
			 result.zticks);

	if (verbose)
		nis_err("result: %s", nis_sperrno(result.status));

	xdr_free(xdr_nis_result, (char *)&result);

	return(NULL);
}

/*
 * Hunt down and return the directory_obj structure that describes
 * the named directory.
 */

fd_result *
nis_finddirectory_3_svc(fd_args *argp, struct svc_req *rqstp)
{
	fd_result		result;
	char			*addr;
	uint32_t		s;
	struct signature	*sig;
	char			*pkey;
/*	nis_name		principal; */
	directory_obj		*d;
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	__start_clock(SERVER_CLOCK);

	bzero((char *)&result, sizeof(fd_result));

	if (verbose)
		nis_err("nis_finddirectory for %s from %s",
				argp->dir_name,
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	result.source = strdup(nis_local_host());

	/* Any principal can call nis_finddirectory(). */
/*	principal = nis_principal(rqstp);
*/

	result.status = __nis_dcache_lookup(argp->dir_name,
					CHECK_SERVER, 0, &d);

	if (result.status != NIS_SUCCESS) {
		if (result.status == NIS_NAMEUNREACHABLE)
			result.status = NIS_NOSUCHNAME;
		if (result.status == NIS_BADNAME)
			result.status = NIS_BADOBJECT;
		nis_update_stats(NIS_FINDDIRECTORY,
		 	result.status == NIS_SUCCESS ? TRUE : FALSE,
		 	__stop_clock(SERVER_CLOCK));
		if (verbose)
			nis_err("result: %s", nis_sperrno(result.status));
		if (!svc_sendreply(rqstp->rq_xprt, xdr_fd_result,
							 (char *)&result))
			nis_err("sendreply failed in nis_finddirectory");
		xdr_free(xdr_fd_result, (char *)&result);
		return(NULL);
	}

	addr = nis_encode_directory_obj(d, &s);

	result.dir_data.dir_data_len = s;
	result.dir_data.dir_data_val = (char *)addr;

	/*
	 * If the sender gave us a requester name, sign the 'object'
	 * using our secret key and their public key.
	 *
	 * NOTE: The Sun rpc.nisd does _NOT_ sign directory objects
	 * even though it's supposed to! (This is supposed to be fixed
	 * in Solaris 2.6; assuming it is, _and_ assuming Sun doesn't
	 * break or change the checksum algorithm, FreeBSD and Solaris
	 * signatures should be identical.)
	 */
#ifdef USE_SECURE_RPC
	if ((pkey = __nis_getpkey(argp->requester, NULL)) != NULL) {
		sig = __nis_calculate_encrypted_cksum((unsigned char *)addr, s,
						pkey, DES_ENCRYPT | DES_HW );
		if (sig != NULL) {
			result.signature.signature_val =
					malloc(sig->signature_len);
			bcopy(sig->signature_val,
				result.signature.signature_val,
				sig->signature_len);
			result.signature.signature_len = sig->signature_len;
		}
		free(pkey);
	}
#endif
	if (verbose)
		nis_err("result: %s", nis_sperrno(result.status));

	if (!svc_sendreply(rqstp->rq_xprt, xdr_fd_result, (char *)&result))
		nis_err("sendreply failed in nis_finddirectory");

	nis_update_stats(NIS_FINDDIRECTORY,
		 result.status == NIS_SUCCESS ? TRUE : FALSE,
		 __stop_clock(SERVER_CLOCK));

	xdr_free(xdr_fd_result, (char *)&result);

	return(NULL);
}

nis_taglist *
nis_status_3_svc(nis_taglist *argp, struct svc_req *rqstp)
{
	static nis_taglist  *result;
	nis_name		princ;
	char			_pbuf[NIS_MAXNAMELEN];
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	if (verbose)
		nis_err("nis_status called from %s",
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	__start_clock(SERVER_CLOCK);
	princ = nis_principal(rqstp, _pbuf);
	result = nis_getstate(argp, princ);
	if (!svc_sendreply(rqstp->rq_xprt, xdr_nis_taglist, (char *)result))
		nis_err("sendreply failed in nis_status");

	nis_freetags(result->tags.tags_val, result->tags.tags_len);
	free(result);

	nis_update_stats(NIS_STATUS, TRUE, __stop_clock(SERVER_CLOCK));

	return (NULL);
}

log_result *
nis_dumplog_3_svc(dump_args *argp, struct svc_req *rqstp)
{
	log_result		*lptr;
	static log_result	result;
	nis_error		lrstat;
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	if (verbose) {
		nis_err("nis_dumplog called from %s for directory %s",
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr),
				argp->da_dir);
		nis_err("nis_dumplog args: %s %s", argp->da_dir,
						nis_ctime(argp->da_time));
	}

	__start_clock(SERVER_CLOCK);
/*
	if (readonly) {
		result.lr_status = NIS_DUMPLATER;
		return(&result);
	}
*/

	if (argp->da_cbhost.da_cbhost_len) {
		result.lr_status = NIS_NOCALLBACK;
		nis_update_stats(NIS_DUMPLOG, FALSE,
				__stop_clock(SERVER_CLOCK));
		return(&result);
	}

	lrstat = nis_log_find_update(argp->da_dir, argp->da_time, NULL);

	if (lrstat)
		lrstat = nis_dump_log(argp->da_dir, argp->da_time, &lptr);
	else
		lrstat = NIS_RESYNC;

	if (lrstat == NIS_SUCCESS) {
		if (!svc_sendreply(rqstp->rq_xprt, xdr_log_result,
							 (char *)lptr))
			nis_err("sendreply failed in nis_dumplog");
		xdr_free(xdr_log_result, (char *)lptr);
		free(lptr);
	} else {
		log_result		lr;
		bzero((char *)&lr, sizeof(lr));

		lr.lr_status = lrstat;
		if (!svc_sendreply(rqstp->rq_xprt, xdr_log_result, (char *)&lr))
			nis_err("sendreply failed in nis_dumplog");
	}

	nis_update_stats(NIS_DUMPLOG, TRUE, __stop_clock(SERVER_CLOCK));

	return (NULL);
}

log_result *
nis_dump_3_svc(dump_args *argp, struct svc_req *rqstp)
{
	static log_result	result;
	log_entry		*ent;
	nis_name		principal;
	char			_pbuf[NIS_MAXNAMELEN];
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	if (verbose)
		nis_err("nis_dump called from %s for directory %s",
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr),
				argp->da_dir);

	__start_clock(SERVER_CLOCK);

	bzero((char *)&result, sizeof(log_result));
	if (!argp->da_cbhost.da_cbhost_len) {
		result.lr_status = NIS_BADREQUEST;
		nis_update_stats(NIS_DUMP, FALSE, __stop_clock(SERVER_CLOCK));
		return(&result);
	}

	result.lr_status = nis_log_find_update(argp->da_dir, 0, &ent);
	if (!result.lr_status) {
		result.lr_status = NIS_NOTFOUND;
		nis_update_stats(NIS_DUMP, FALSE, __stop_clock(SERVER_CLOCK));
		return(&result);
	}

	principal = nis_principal(rqstp, _pbuf);

	nis_make_cookie(rqstp, &result.lr_cookie);
	result.lr_status = NIS_CBRESULTS;
	result.lr_entries.lr_entries_len = 1;
	result.lr_entries.lr_entries_val = ent;
	if (!svc_sendreply(rqstp->rq_xprt, xdr_log_result, (char *)&result))
		nis_err("sendreply failed in nis_dump");

	nis_callback_dump(argp->da_cbhost.da_cbhost_val,
				argp->da_dir, principal,
				rqstp, &result.lr_cookie);

	/* save the cookie */
	result.lr_cookie.n_len = 0;
	result.lr_cookie.n_bytes = NULL;
	xdr_free(xdr_log_result, (char *)&result);

	nis_update_stats(NIS_DUMP, FALSE, __stop_clock(SERVER_CLOCK));
	return (NULL);
}

bool_t *
nis_callback_3_svc(netobj *argp, struct svc_req *rqstp)
{
	static bool_t  result;
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	if (verbose)
		nis_err("nis_callback called from %s",
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	__start_clock(SERVER_CLOCK);
	/*
	 * Okay: should we be enforcing access restrictions
	 * here or not? I can't figure out what to check against
	 * anyway: we don't assign ownership for a callback 'thread.'
	 * (They aren't really threads anyway.)
	 */
	if (nis_callback_status(argp) == NIS_NOTFOUND)
		result = FALSE;
	else
		result = TRUE;

	nis_update_stats(NIS_CALLBACK, result, __stop_clock(SERVER_CLOCK));

	return (&result);
}

uint32_t *
nis_cptime_3_svc(nis_name *argp, struct svc_req *rqstp)
{
	static uint32_t  result;
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	if (verbose)
		nis_err("nis_cptime called from %s",
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	__start_clock(SERVER_CLOCK);

	result = nis_log_find_update(*argp, 0, NULL);

	nis_update_stats(NIS_CPTIME, result ? TRUE : FALSE,
				__stop_clock(SERVER_CLOCK));
	return (&result);
}

/*
 * Event callback for checkpoints. This callback is a bit tricky
 * because it creates a readonly child to do the work because
 * checkpointing may take a while. Note that the event manager
 * frees the arguments for us.
 */
static bool_t nis_checkpoint_cb(arg, id, reason)
	void			*arg;
	uint32_t		id;
	int			reason;
{
	char			*argp;
	pid_t			childpid;
	struct db_objs		dres;
	nis_name		dir = NULL;

	argp = arg;

	nis_create_readonly_child(&childpid);
	if (!childpid)
		return(FALSE);

	if (argp == NULL) {
		while((dir = nis_enum_srvlist(dir)) != NULL)
			db_cpdir(dir, &dres);
	} else
		db_cpdir(argp, &dres);

	nis_destroy_readonly_child(childpid);

	return(FALSE);
}

cp_result *
nis_checkpoint_3_svc(nis_name *argp, struct svc_req *rqstp)
{
	static cp_result	result;
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	if (verbose)
		nis_err("nis_checkpoint called from %s",
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	__start_clock(SERVER_CLOCK);

	bzero((char *)&result, sizeof(cp_result));

	if (readonly) {
		result.cp_status = NIS_TRYAGAIN;
		result.cp_zticks = __stop_clock(SERVER_CLOCK);
		nis_update_stats(NIS_CHECKPOINT, FALSE, result.cp_zticks);
		return(&result);
	}

	if (strlen(*argp) && db_exists(*argp) != NIS_SUCCESS) {
		result.cp_zticks = __stop_clock(SERVER_CLOCK);
		result.cp_status = NIS_NOSUCHNAME;
		nis_update_stats(NIS_CHECKPOINT, FALSE, result.cp_zticks);
		return(&result);
	}

	nis_event_add_oneshot(nis_checkpoint_cb,
		strlen(*argp) ? strdup(*argp) : NULL, free,
			10, "checkpoint dir");

	result.cp_status = NIS_SUCCESS;
	result.cp_zticks = __stop_clock(SERVER_CLOCK);

	nis_update_stats(NIS_CHECKPOINT, TRUE, result.cp_zticks);

	return (&result);
}

void *
nis_ping_3_svc(ping_args *argp, struct svc_req *rqstp)
{
	static char * result;
	nis_error		r;
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	if (verbose) {
		nis_err("nis_ping called from %s",
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));
		nis_err("nis_ping args: %s %s", argp->dir,
							nis_ctime(argp->stamp));
	}

	if (readonly)
		return((void *) &result);

	__start_clock(SERVER_CLOCK);
	if ((r = nis_replica_add_update(argp)) != NIS_SUCCESS) {
		nis_err("warning: failed to queue update for %s: %s",
						argp->dir, nis_sperrno(r));
		nis_update_stats(NIS_PING, FALSE, __stop_clock(SERVER_CLOCK));
	} else
		nis_update_stats(NIS_PING, TRUE, __stop_clock(SERVER_CLOCK));

	return((void *) &result);
}

nis_taglist *
nis_servstate_3_svc(nis_taglist *argp, struct svc_req *rqstp)
{
	static nis_taglist  *result;
	nis_name		princ;
	char			_pbuf[NIS_MAXNAMELEN];
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	if (verbose)
		nis_err("nis_servstate called from %s",
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));
	if (readonly)
		return(result);

	__start_clock(SERVER_CLOCK);

	princ = nis_principal(rqstp, _pbuf);
	result = nis_setstate(argp, princ);
	if (!svc_sendreply(rqstp->rq_xprt, xdr_nis_taglist, (char *)result))
		nis_err("sendreply failed in nis_status");

	nis_freetags(result->tags.tags_val, result->tags.tags_len);
	free(result);

	nis_update_stats(NIS_SERVSTATE, TRUE, __stop_clock(SERVER_CLOCK));

	return (NULL);
}

nis_error *
nis_mkdir_3_svc(nis_name *argp, struct svc_req *rqstp)
{
	static nis_error	result;
	directory_obj		*d;
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	if (verbose)
		nis_err("nis_mkdir for %s called from %s", (char *)*argp,
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	if (readonly) {
		result = NIS_TRYAGAIN;
		return(&result);
	}

	__start_clock(SERVER_CLOCK);

	/*
	 * Can only create the directory if we serve it.
	 */
	result = __nis_dcache_lookup((char *)*argp, LOOKUP_ONLY|CHECK_SERVER,
					MASTER_ONLY, &d);

	if (result != NIS_SUCCESS) {
		nis_update_stats(NIS_MKDIR, FALSE, __stop_clock(SERVER_CLOCK));
		return(&result);
	}

	result = db_to_nis_err(db_newdir((char *)*argp));

	if (result == NIS_SUCCESS) {
		nis_add_to_srvlist((char *)*argp);
		nis_update_stats(NIS_MKDIR, TRUE, __stop_clock(SERVER_CLOCK));
	} else
		nis_update_stats(NIS_MKDIR, FALSE, __stop_clock(SERVER_CLOCK));

	return (&result);
}

/*
 * The nis_rmdir routine is rather gross. The Sun NIS+ server seems to
 * handle directory removals like this:
 * - nisrmdir(8) calls nis_remove() to remove the directory object
 * - nisrmdir(8) then calls nis_rmdir() to remove the directory
 * - the nis_rmdir_3_svc() routine in rpc.nisd traverses the leftover
 *   directory table (using db_first/db_next), removing all objects it
 *   encounters as it goes. If at some point another directrory is
 *   encountered (i.e. the directory being removed itself contained other
 *   subdirectories), then the removal is aborted.
 *
 *   Presumeably, the aborting of the directory removal is handled by
 *   the transaction logging code somehow. Whatever the case, that's
 *   how I do it here.
 *
 *   NOTE: the Solaris 2.5.1 server appears to botch the abort: if there
 *   are tables in the directory, they are not correctly restored. Only
 *   the first entry in the table is put back.
 *
 */

nis_error *
nis_rmdir_3_svc(nis_name *argp, struct svc_req *rqstp)
{
	static nis_error	result;
	db_status		db_stat;
	struct db_objs		dres;
	char			__buf[NIS_MAXNAMELEN];
	char			name[NIS_MAXNAMELEN];
	directory_obj		*d;
	nis_object		*obj;
	uint32_t		xid;
	nis_name		principal;
	netobj			cookie;
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	if (verbose)
		nis_err("nis_rmdir for %s called from %s", (char *)*argp,
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	if (readonly) {
		result = NIS_TRYAGAIN;
		return(&result);
	}

	__start_clock(SERVER_CLOCK);

	result = __nis_dcache_lookup((char *)*argp, LOOKUP_ONLY,
					MASTER_ONLY, &d);
	/*
	 * If the directory doesn't exist, then that's because
	 * the directory object was removed. This is okay. If
	 * some other error occurs, then we should signal it.
	 */
	if (result != NIS_SUCCESS && result != NIS_NOTFOUND &&
		result != NIS_NOSUCHNAME && result != NIS_NAMEUNREACHABLE) {
		nis_update_stats(NIS_RMDIR, FALSE, __stop_clock(SERVER_CLOCK));
		return(&result);
	}
	/*
	 * If the directory object does still exist, check that we
	 * no longer serve it.
	 */
	if (result == NIS_SUCCESS) {
		if (__is_local_dir(d) == TRUE) {
			nis_err("rmdir called for directory %s"
					" which we still serve!",
						(char *)*argp);
			result = NIS_NOT_ME;
			nis_update_stats(NIS_RMDIR, FALSE,
					__stop_clock(SERVER_CLOCK));
			return(&result);
		}
	}

	/*
	 * For a master server, it's a mistake to remove a directory
	 * that contains subdirectories since doing so would leave
	 * the subdirectories orphaned. For a replica however, it's
	 * alright to allow this.
	 */
	if (result == NIS_SUCCESS) {
		if (__i_am_master(d) == FALSE)
			goto skipcheck;
	}


	/*
	 * We only get the principal here to make the transaction
	 * logging module happy.
	 */
	principal = nis_principal(rqstp, __buf);

	db_stat = db_first((char *)*argp, NULL, 0, &dres);
	result = db_to_nis_err(db_stat);
	if (result != NIS_SUCCESS && result != NIS_NOTFOUND) {
		nis_update_stats(NIS_RMDIR, FALSE, __stop_clock(SERVER_CLOCK));
		return(&result);
	}

	nis_log_new_xid(principal, &xid);
	while(db_stat == DB_SUCCESS) {
		cookie.n_len = dres.cookie.n_len;
		cookie.n_bytes = dres.cookie.n_bytes;
		obj = NIS_RES_OBJECT(&dres);
		if (__type_of(obj) == DIRECTORY_OBJ) {
			nis_log_abort_xid(xid);
			result = NIS_NOTEMPTY;
			if (verbose)
				nis_err("directory %s not empty",
							(char *)*argp);
			nis_update_stats(NIS_RMDIR, FALSE,
					__stop_clock(SERVER_CLOCK));
			return(&result);
		}
		snprintf(name, sizeof(name), "%s.%s", obj->zo_name,
						      obj->zo_domain);
		nis_log_submit(name, obj, NULL, 0, REM_NAME);
		if ((db_stat = db_remove(name, NULL)) != DB_SUCCESS) {
			nis_log_abort_xid(xid);
			result = db_to_nis_err(db_stat);
			nis_update_stats(NIS_RMDIR, FALSE,
					__stop_clock(SERVER_CLOCK));
			return(&result);
		}
		__nis_flush_one_dir(name);
		nis_destroy_object(obj);
		db_stat = db_next((char *)*argp, &cookie, &dres);
	}
	nis_log_commit_xid(xid);

skipcheck:

	result = db_rm((char *)*argp);

	__nis_flush_one_dir((char *)*argp);

	if (d != NULL || __i_am_master(d) != TRUE)
		nis_log_flush_byname(*argp);

	if (result == NIS_SUCCESS) {
		nis_remove_from_srvlist((char *)*argp);
		if (nis_dir_cmp((char *)*argp, rootname) == SAME_NAME)
			nis_replica_update_root();
		nis_update_stats(NIS_RMDIR, TRUE, __stop_clock(SERVER_CLOCK));
	} else
		nis_update_stats(NIS_RMDIR, FALSE, __stop_clock(SERVER_CLOCK));

	return (&result);
}

/*
 * It should be noted that the Sun NIS+ server does not implement
 * the UDPKEYS procedure even though it's defined in the protocol.
 * The Sun nisupdkeys(8) command uses nis_modify(3) instead.
 */

nis_error *
nis_updkeys_3_svc(nis_name *argp, struct svc_req *rqstp)
{
	static nis_error	result;
	char			*pkey;
	uint32_t		i;
	nis_object		*o, *n;
	directory_obj		*d;
	nis_server		*srv;
	nis_name		principal;
	char			_pbuf[NIS_MAXNAMELEN];
	SERVER_CLOCK_DECLS;

	if (__nis_child_must_die)
		return(NULL);

	if (verbose)
		nis_err("nis_updkeys for %s called from %s", (char *)*argp,
				inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr));

	if (readonly) {
		result = NIS_TRYAGAIN;
		return(&result);
	}

	__start_clock(SERVER_CLOCK);

	result = __nis_ocache_lookup(*argp, CHECK_SERVER, 0, &o);

	if (result != NIS_SUCCESS) {
		nis_update_stats(NIS_UPDKEYS, FALSE,
				__stop_clock(SERVER_CLOCK));
		return(&result);
	}

	n = nis_clone_object(o, NULL);
	d = &n->DI_data;

	for (i = 0; d->do_servers.do_servers_len; i++) {
		srv = &d->do_servers.do_servers_val[i];
		if ((pkey = __nis_getpkey(srv->name, NULL)) != NULL) {
			srv->key_type = NIS_PK_DH;
			srv->pkey.n_len = strlen(pkey) + 1;
			if (srv->pkey.n_bytes != NULL)
				free(srv->pkey.n_bytes);
			srv->pkey.n_bytes = pkey;
		}
	}

	principal = nis_principal(rqstp, _pbuf);
	result = nis_db_nsop(*argp, n, principal, NIS_MODIFY, NULL);
	nis_destroy_object(n);

	nis_update_stats(NIS_UPDKEYS, result == NIS_SUCCESS ? TRUE : FALSE,
					__stop_clock(SERVER_CLOCK));
	return (&result);
}
