/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: nis_state.c,v 1.2 1998/07/18 21:22:25 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/cdefs.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nis_tags.h>
#include "nis_extern.h"

#ifndef lint
static const char rcsid[] =
	"@(#) $Id: nis_state.c,v 1.2 1998/07/18 21:22:25 wpaul Exp wpaul $";
#endif

extern void __nis_get_grpstats __P(( uint32_t *, uint32_t *,
					uint32_t * ));

/*
 * In Solaris 2.5.x, Sun added a new wrinkle to rpc.nisd: the serving
 * list. A file called /var/nis/data/serving_list is created by nisinit(8)
 * which contains a list of all the directories served by this particular
 * instance of rpc.nisd. A tag was also added to allow this list to be
 * viewed with nisstat(8).
 *
 * This helps cut a little overhead off the server operation since it lets
 * us to quickly see if we are a server for a directory without having to
 * do an nis_lookup() on ourselves or a replica.
 */

nis_error nis_load_srvlist()
{
	FILE			*fp;
	char			buf[NIS_MAXNAMELEN + 1];
	char			**new;

	srvlist = calloc(1, sizeof(char *));
	if (srvlist == NULL)
		return(NIS_NOMEMORY);

	srvcnt = 0;
	srvlist[srvcnt] = NULL;

	if ((fp = fopen(_PATH_NIS_SRVLIST, "r")) == NULL) {
		if (errno == ENOENT)
			return(NIS_NOTFOUND);
		else
			return(NIS_SYSTEMERROR);
	}

	while(fgets(buf, NIS_MAXNAMELEN, fp)) {
		if (strchr(buf, '\n') != NULL)
			*strchr(buf, '\n') = '\0';
		new = realloc(srvlist, (srvcnt + 2) * sizeof(char *));
		if (new == NULL) {
			fclose(fp);
			return(NIS_NOMEMORY);
		}
		new[srvcnt] = strdup(buf);
		srvcnt++;
		new[srvcnt] = NULL;
		srvlist = new;
	}

	fclose(fp);

	return(NIS_SUCCESS);
}

nis_error nis_dump_srvlist()
{
	FILE			*fp;
	char			buf[NIS_MAXNAMELEN + 1];
	char			**tmp = srvlist;

	if ((fp = fopen(_PATH_NIS_SRVLIST, "w+")) == NULL) {
		nis_err("failed to dump serving list: %s",
				 strerror(errno));
		return(NIS_SYSTEMERROR);
	}

	while (*tmp != NULL) {
		snprintf(buf, sizeof(buf), "%s\n", *tmp);
		fputs(buf, fp);
		tmp++;
	}

	fclose(fp);

	return(NIS_SUCCESS);
}

bool_t nis_is_in_srvlist(dir)
	nis_name		dir;
{
	char			**tmp = srvlist;

	while (*tmp != NULL) {
		if (nis_dir_cmp(dir, *tmp) == SAME_NAME)
			return(TRUE);
		tmp++;
	}

	return(FALSE);
}

nis_error nis_add_to_srvlist(dir)
	nis_name		dir;
{
	char			**tmp;

	if (nis_is_in_srvlist(dir) == TRUE)
		return(NIS_NOTUNIQUE);

	tmp = realloc(srvlist, (srvcnt + 2) * sizeof(char *));
	if (tmp == NULL) {
		nis_err("out of memory allocating srvlist entry");
		return(NIS_NOMEMORY);
	}
	tmp[srvcnt] = strdup(dir);
	srvcnt++;
	tmp[srvcnt] = NULL;
	srvlist = tmp;

	nis_dump_srvlist();

	return(NIS_SUCCESS);
}

nis_error nis_remove_from_srvlist(dir)
	nis_name		dir;
{
	char			**tmp = srvlist, **new;
	int			tmpcnt = 0;

	if (nis_is_in_srvlist(dir) == FALSE)
		return(NIS_NOTFOUND);

	new = malloc(sizeof(char *) * srvcnt - 1);

	while(*tmp != NULL) {
		if (nis_dir_cmp(dir, *tmp) != SAME_NAME) {
			new[tmpcnt] = *tmp;
			tmpcnt++;
		} else
			free(*tmp);
		tmp++;
	}
	srvcnt--;
	free(srvlist);
	srvlist = new;
	srvlist[srvcnt] = NULL;

	nis_dump_srvlist();

	return(NIS_SUCCESS);
}

nis_name nis_enum_srvlist(dir)
	nis_name		dir;
{
	char			**tmp = srvlist;

	if (dir == NULL)
		return(*tmp);

	while (*tmp != NULL) {
		if (nis_dir_cmp(dir, *tmp) == SAME_NAME)
			break;
		tmp++;
	}

	tmp++;

	return(*tmp);
}

/*
 * Operations statistics. These are part of the server state, so they
 * are kept here. We keep track of things like how many times our procedures
 * are called and their success/failure rates.
 */

struct s_info {
	uint32_t		call;
	uint32_t		fail;
	uint32_t		tick_tot;
	uint32_t		tick_avg;
	int			tick_cnt;
};

static struct s_info	nis_stat[NIS_UPDKEYS];
static int nis_stats_enabled = 1;

#define S_ALL	0

void nis_init_stats()
{
	bzero((char *)&nis_stat, sizeof(nis_stat));

	return;
}

#define NSTAT	nis_stat[op]

void nis_update_stats(op, result, ticks)
	uint32_t		op;
	bool_t			result;
	uint32_t		ticks;
{
	if (!nis_stats_enabled)
		return;

	if (op > NIS_UPDKEYS)
		return;

	NSTAT.call++;
	if (NSTAT.tick_cnt == 16) {
		NSTAT.tick_cnt = 0;
		NSTAT.tick_tot = 0;
		NSTAT.tick_avg = 0;
	}
	NSTAT.tick_cnt++;
	NSTAT.tick_tot += ticks;
	NSTAT.tick_avg = NSTAT.tick_tot / NSTAT.tick_cnt;
	nis_stat[S_ALL].call++;
	if (result == FALSE) {
		NSTAT.fail++;
		nis_stat[S_ALL].fail++;
	}

	return;
}

static char *matrix[] = { "NULLPROC", "LOOKUP", "ADD", "MODIFY", "REMOVE",
			  "LIST", "ADDENTRY", "MODENTRY", "REMENTRY",
			  "FIRSTENTRY", "NEXTENTRY", "RESRVD1", "FINDDIR",
			  "RESRVD2", "STATUS",
			  "DUMPLOG", "DUMP", "CALLBACK", "CPTIME",
			  "CHECKPOINT", "PING", "SERVSTATE", "MKDIR",
			  "RMDIR", "UPDKEYS" };

nis_taglist *nis_getstate(tags, princ)
	nis_taglist		*tags;
	nis_name		princ;
{
	uint32_t		i, j;
	nis_tag			*tag = NULL;
	nis_taglist		*res;
	char			buf[1024], tbuf[1024];
	struct rusage		rusage;
	uint32_t		days = 0, hrs = 0, mins = 0, secs = 0;
	time_t			uptime;
	uint32_t		calls, hits, misses;
	char			**slist = srvlist;
	bool_t			perm_ok;

	res = calloc(1, sizeof(nis_taglist));
	res->tags.tags_val = calloc(tags->tags.tags_len, sizeof(nis_tag));
	res->tags.tags_len = tags->tags.tags_len;

	snprintf(buf, sizeof(buf), "admin.%s", nis_local_directory());
	perm_ok = nis_check_access(NIS_READ_ACC, NIS_STATE_ACCESS,
				princ, nis_local_principal(), buf);

	for (i = 0; i < tags->tags.tags_len; i++) {
		tag = &tags->tags.tags_val[i];
		if (perm_ok == FALSE) {
			res->tags.tags_val[i].tag_type = tag->tag_type;
			res->tags.tags_val[i].tag_val =
					strdup("permission denied");
			continue;
		}
		switch(tag->tag_type) {
		case TAG_UPTIME:
			uptime = time(NULL) - boottime;
			days = uptime / 86400;
			hrs = (uptime - (days * 86400)) / 3600;
			mins = (uptime - (days * 86400) - (hrs * 3600)) / 60;
			secs = uptime - (days * 86400) - (hrs * 3600) -
						(mins * 60);
			snprintf(buf, 1024, "up %uD, %02u:%02u:%02u",
				days, hrs, mins, secs);
			res->tags.tags_val[i].tag_type = TAG_UPTIME;
			res->tags.tags_val[i].tag_val = strdup(buf);
			break;
		case TAG_SECURITY_LEVEL:
			snprintf(buf, 1024, "%d", securelevel);
			res->tags.tags_val[i].tag_type = TAG_SECURITY_LEVEL;
			res->tags.tags_val[i].tag_val = strdup(buf);
			break;
		case TAG_DNSFORWARDING:
			res->tags.tags_val[i].tag_type = TAG_DNSFORWARDING;
			res->tags.tags_val[i].tag_val =
				strdup("DNS forwarding not supported");
			break;
		case TAG_NISCOMPAT:
			res->tags.tags_val[i].tag_type = TAG_NISCOMPAT;
			res->tags.tags_val[i].tag_val =
				strdup("YP compat mode not supported");
			break;
		case TAG_S_STORAGE:
			getrusage(RUSAGE_SELF, &rusage);
			snprintf(buf, 1024, "%ld", rusage.ru_idrss);
			res->tags.tags_val[i].tag_type = TAG_S_STORAGE;
			res->tags.tags_val[i].tag_val = strdup(buf);
			break;
		case TAG_S_DCACHE:
			__nis_get_dcache_stats(&hits, &misses, &calls);
			snprintf(buf, 1024, "C=%u:H=%u:M=%u:HR=%u%%",
					calls, hits, misses,
					calls == 0 ? 0 :
					((calls - misses)*100/calls));
			res->tags.tags_val[i].tag_type = TAG_S_DCACHE;
			res->tags.tags_val[i].tag_val = strdup(buf);
			break;
		case TAG_S_GCACHE:
			__nis_get_grpstats(&hits, &misses, &calls);
			snprintf(buf, 1024, "C=%u:H=%u:M=%u:HR=%u%%",
					calls, hits, misses,
					calls == 0 ? 0 :
					((calls - misses)*100/calls));
			res->tags.tags_val[i].tag_type = TAG_S_GCACHE;
			res->tags.tags_val[i].tag_val = strdup(buf);
			break;
		case TAG_DEBUG:
			snprintf(buf, 1024, "%s", debug ? "ON" : "OFF");
			res->tags.tags_val[i].tag_type = TAG_DEBUG;
			res->tags.tags_val[i].tag_val = strdup(buf);
			break;
		case TAG_DIRLIST:
			bzero(buf, sizeof(buf));
			while(*slist != NULL) {
				strncat(buf, *slist, sizeof(buf) - 1);
				slist++;
				if (*slist != NULL)
					strncat(buf, " ", sizeof(buf) - 1);
			}
			res->tags.tags_val[i].tag_type = TAG_DIRLIST;
			res->tags.tags_val[i].tag_val = strdup(buf);
			break;
		case TAG_ROOTSERVER:
			snprintf(buf, 1024, "%s", root_obj != NULL
							 ? "ON" : "OFF");
			res->tags.tags_val[i].tag_type = TAG_ROOTSERVER;
			res->tags.tags_val[i].tag_val = strdup(buf);
			break;
		case TAG_OPSTATS:
			if (!nis_stats_enabled) {
				res->tags.tags_val[i].tag_type = TAG_OPSTATS;
				res->tags.tags_val[i].tag_val =
						strdup("disabled");
				break;
			}
			bzero(buf, sizeof(buf));
			for (j = 1; j < NIS_UPDKEYS; j++) {
				if (nis_stat[j].call) {
					snprintf(tbuf, sizeof(tbuf),
						 "OP=%s:C=%u:E=%u:T=%u",
						 matrix[j],
						 nis_stat[j].call,
						 nis_stat[j].fail,
						 nis_stat[j].tick_avg);
					strncat(buf, tbuf, sizeof(buf) - 1);
					if (j != NIS_UPDKEYS)
						strncat(buf, "\n",
						 sizeof(buf) - 1);
				}
			}
			res->tags.tags_val[i].tag_type = TAG_OPSTATS;
			res->tags.tags_val[i].tag_val = strdup(buf);
			break;
		case TAG_HEAP:
		default:
			res->tags.tags_val[i].tag_type = tag->tag_type;
			res->tags.tags_val[i].tag_val =
					strdup("unknown");
			break;
		}
	}

	return(res);
}

nis_taglist *nis_setstate(tags, princ)
	nis_taglist		*tags;
	nis_name		princ;
{
	nis_taglist		*res;
	nis_tag			*tag = NULL;
	register unsigned int	i;
	char			buf[NIS_MAXNAMELEN];
	bool_t			perm_ok;

	res = calloc(1, sizeof(nis_taglist));
	if (!tags->tags.tags_len)
		return(res);
	res->tags.tags_val = calloc(tags->tags.tags_len, sizeof(nis_tag));
	res->tags.tags_len = tags->tags.tags_len;

	snprintf(buf, sizeof(buf), "admin.%s", nis_local_directory());
	perm_ok = nis_check_access(NIS_MODIFY_ACC, NIS_STATE_ACCESS,
				princ, nis_local_principal(), buf);

	for (i = 0; i < tags->tags.tags_len; i++) {
		tag = &tags->tags.tags_val[i];
		res->tags.tags_val[i].tag_type = tag->tag_type;
		if (perm_ok == FALSE) {
			res->tags.tags_val[i].tag_val =
				strdup("permission denied");
			continue;
		} else
			res->tags.tags_val[i].tag_val = strdup(tag->tag_val);
		switch(tag->tag_type) {
		case TAG_DEBUG:
			if (strcmp(tag->tag_val, "on") == 0) {
				if (verbose)
					nis_err("enabling verbose mode");
				verbose = 1;
			} else if (strcmp(tag->tag_val, "off") == 0) {
				if (verbose)
					nis_err("disabling verbose mode");
				verbose = 0;
			} else {
				free(res->tags.tags_val[i].tag_val);
				res->tags.tags_val[i].tag_val =
						strdup("unchanged");
			}
			break;
		case TAG_DCACHE_ONE_REFRESH:
			if (verbose)
				nis_err("flushing dir: %s", tag->tag_val);
			__nis_flush_one_dir(tag->tag_val);
			break;
		case TAG_DCACHE:
			if (verbose)
				nis_err("flushing dircache");
			__nis_flush_dcache();
			break;
		case TAG_GCACHE:
			if (verbose)
				nis_err("flushing group cache");
			nis_flushgroups();
			break;
		case TAG_GCACHE_ONE:
			if (verbose)
				nis_err("flushing %s from group cache",
							tag->tag_val);
			__nis_flush_one_group(tag->tag_val);
			break;
		case TAG_HEAP:
			free(res->tags.tags_val[i].tag_val);
			res->tags.tags_val[i].tag_val =
						strdup("<Not enabled>");
			break;
		case TAG_STATS:
			free(res->tags.tags_val[i].tag_val);
			if (nis_stats_enabled) {
				if (verbose)
					nis_err("disabling stats counters");
				nis_stats_enabled = 0;
				nis_init_stats();
				res->tags.tags_val[i].tag_val = strdup("off");
			} else {
				if (verbose)
					nis_err("enabling stats counters");
				nis_stats_enabled = 1;
				res->tags.tags_val[i].tag_val = strdup("on");
			}
			break;
		case TAG_OCACHE:
			if (verbose)
				nis_err("flushing object cache");
			__nis_flush_ocache();
			break;
		default:
			break;
		}
	}

	return(res);
}
