/*
 * Copyright (c) 1996, 1997
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: nis_subr.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $
 */

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/cdefs.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nis_db.h>
#include "nis_extern.h"

#ifndef lint
static const char rcsid[] = "@(#) $Id: nis_subr.c,v 1.1 1998/07/08 19:23:20 wpaul Exp wpaul $";
#endif

/*
 * This routine fork()s a child process that is used to handle lookups
 * while the parent rpc.nisd process does some other time consuming work.
 * The child will handle lookup, iblist, ibfirst, ibnext and finddirectory
 * request _only_, since the parent will probably be performing a checkpoint
 * or some other operation on the databases that should not be interrupted.
 * The parent process will unregister its RPC transport handles so that
 * it will not bothered by incoming requests.
 */
nis_error nis_create_readonly_child(cpid)
	int			*cpid;
{
	if (readonly)
		return(NIS_TRYAGAIN);


	switch((*cpid = fork())) {
	case -1:
		return(NIS_SYSTEMERROR);
		break;
	case 0:
		readonly = 1;
		progname = "rpc.nisd-readonly";
		return(NIS_SUCCESS);
	default:
		break;
	}

	/*
	 * We only want the child to handle incoming
	 * requests, so we tell ourselves to shut up.
	 */
	__nis_unregister_xprt();

	return(NIS_SUCCESS);
}

/*
 * This routine terminates a child process that was spawned to handle
 * lookups while the parent was busy. It also reactivates the parent's
 * RPC transport handles so it can receive RPC requests again.
 */
nis_error nis_destroy_readonly_child(cpid)
	int			cpid;
{
	if (readonly)
		return(NIS_TRYAGAIN);

	kill(cpid, SIGUSR1);
	__nis_register_xprt();

	return(NIS_SUCCESS);
}

/*
 * Generate a unique cookie. The cookie is used to identify a callback
 * 'thread' that has been set in motion. The caller can use this cookie
 * and the NIS_CALLBACK procedure to check the status of a callback
 * 'thread.'
 */
nis_error nis_make_cookie(rqstp, cookie)
	struct svc_req		*rqstp;
	netobj			*cookie;
{
	char *ptr;

	cookie->n_len = sizeof(struct svc_req) + sizeof(SVCXPRT);
	if ((cookie->n_bytes = calloc(1, cookie->n_len)) == NULL) {
		nis_err("no memory for cookie: %s", strerror(errno));
		return(NIS_NOMEMORY);
	}

	ptr = cookie->n_bytes;

	bcopy((char *)rqstp, ptr, sizeof(struct svc_req));
	ptr += sizeof(struct svc_req);
	bcopy((char *)rqstp->rq_xprt, ptr, sizeof(SVCXPRT));

	return(NIS_SUCCESS);
}

/*
 * This routine encapsulates an NIS+ object as an entry object. This is
 * used by the nis_db_iblist/first/next() routines when asked to list a
 * directory. The variant part of the object is destroyed and replaced
 * by an entry with a single column containing a 32-bit int that represents
 * the object type (zotypes). The value is in network byte order.
 *
 * Note: getting this routine right is tricky. You have to remember to
 * free the variant part of the object before replacing it with the
 * faked-up entry or you will leak memory.
 */
nis_object *nis_encap_obj(db_res, i)
	db_result		*db_res;
	uint32_t		i;
{
	nis_object		*obj;
	entry_col		*col;
	zotypes			otype;

	obj = nis_decode_nis_obj(OBJ_VAL(db_res, i, 1), OBJ_LEN(db_res, i, 1));
	otype = __type_of(obj);
	xdr_free(xdr_objdata, (char *)&obj->zo_data);
	bzero((char *)&obj->zo_data, sizeof(objdata));
	__type_of(obj) = ENTRY_OBJ;
	obj->EN_data.en_type = strdup("NIS object");
	obj->EN_data.en_cols.en_cols_len = 1;
	col = malloc(sizeof(entry_col));
	col->ec_flags = OBJ_FLAGS(db_res, 0, 1);
	col->ec_value.ec_value_val = malloc(sizeof(zotypes));
	*(uint32_t *)col->ec_value.ec_value_val = htonl(otype);
	col->ec_value.ec_value_len = sizeof(zotypes);
	obj->EN_data.en_cols.en_cols_val = col;

	return(obj);
}

/*
 * Block out columns of an entry that principal doesn't
 * have the right to see. If the principal doesn't have
 * the right to see any columns, return NIS_PERMISSION.
 */
nis_error nis_mangle_entry(ent, obj, principal)
	nis_object		*ent;
	nis_object		*obj;
	nis_name		principal;
{
	table_col		*tcol;
	table_obj		*t = &obj->TA_data;
	nis_name		owner = obj->zo_owner;
	nis_name		group = obj->zo_group;
	register uint32_t	i, cnt = 0;

	for (i = 0; i < t->ta_cols.ta_cols_len; i++) {
		tcol = &t->ta_cols.ta_cols_val[i];
		if (nis_check_access(NIS_READ_ACC, tcol->tc_rights,
				     principal, owner, group) == FALSE) {
			free(ENTRY_VAL(ent, i));
			ENTRY_VAL(ent, i) = strdup("*NP*");
			ENTRY_LEN(ent, i) = strlen("*NP*") + 1;
			cnt++;
		}
	}

	if (cnt == t->ta_cols.ta_cols_len)
		return(NIS_PERMISSION);

	return(NIS_SUCCESS);
}

nis_error db_to_nis_err(db_error)
	db_status db_error;
{
	switch(db_error) {
	case (DB_SUCCESS):
		return(NIS_SUCCESS);
		break;
	case (DB_NOTUNIQUE):
		return(NIS_NOTUNIQUE);
		break;
	case (DB_BADQUERY):
		return(NIS_BADREQUEST);
		break;
	case (DB_BADTABLE):
		return(NIS_INVALIDOBJ);
		break;
	case (DB_MEMORY_LIMIT):
		return(NIS_NOMEMORY);
		break;
	case (DB_STORAGE_LIMIT):
		return(NIS_NOFILESPACE);
		break;
	case (DB_NOTFOUND):
		return(NIS_NOTFOUND);
		break;
	case (DB_INTERNAL_ERROR):
		return(NIS_SYSTEMERROR);
		break;
	case (DB_BADOBJECT):
		return(NIS_BADOBJECT);
		break;
	default:
		return(NIS_SYSTEMERROR);
		break;
	}
}

nis_error nis_validname(str)
	char *str;
{
	if (str == NULL)	/* string pointer shouldn't be NULL */
		return(NIS_BADNAME);

	switch(*str) {
	case ' ':
	case '\t':	/* tab -- whitespace */
	case '.':
	case '[':	/* XXX */
	case ']':
	case ',':
	case '=':
	case '@':		
	case '+':		/* 'The initial character may not be a */
	case '-':		/* terminal character or the characters */
		return(NIS_BADNAME);	/* '@', '+' or '-'. (No terminals?) */
		break;
	default:
		break;
	}

	if (strchr(str, '/'))
		return(NIS_BADNAME);	/* No slashes allowed anywhere in */
				/* an NIS+ name. */

	/* Name must be fully qualified. */
	if (*(str + strlen(str) - 1) != '.')
		return(NIS_BADNAME);

	return(NIS_SUCCESS);
}

nis_error nis_put_root_obj(root_obj)
	nis_object		*root_obj;
{
	FILE			*fp;
	XDR			xdrs;

	if ((fp = fopen(_PATH_NIS_ROOT_OBJ, "w+")) == NULL) {
		return(NIS_SYSTEMERROR);
	}

	xdrstdio_create(&xdrs, fp, XDR_ENCODE);

	if (!xdr_nis_object(&xdrs, root_obj)) {
		fclose(fp);
		xdr_destroy(&xdrs);
		return(NIS_SYSTEMERROR);
	}


	fclose(fp);
	xdr_destroy(&xdrs);

	return(NIS_SUCCESS);
}

nis_error nis_get_root_obj(root_obj)
	nis_object		**root_obj;
{
	nis_object		*obj;
	FILE			*fp;
	XDR			xdrs;

	if (*root_obj != NULL) {
		xdr_free(xdr_nis_object, (char *)*root_obj);
		free(*root_obj);
	}

	obj = calloc(1, sizeof(nis_object));

	if (obj == NULL)
		return(NIS_NOMEMORY);

	if ((fp = fopen(_PATH_NIS_ROOT_OBJ, "r")) == NULL) {
		free(obj);
		return(NIS_SYSTEMERROR);
	}

	xdrstdio_create(&xdrs, fp, XDR_DECODE);

	if (!xdr_nis_object(&xdrs, obj)) {
		free(obj);
		fclose(fp);
		xdr_destroy(&xdrs);
		return(NIS_SYSTEMERROR);
	}


	fclose(fp);
	xdr_destroy(&xdrs);
	*(caddr_t *)root_obj = (caddr_t)obj;

	return(NIS_SUCCESS);
}

struct entryattrs *nis_entry_to_attr(ent, table)
	entry_obj *ent;
	table_obj *table;
{
	int i, cnt = 0;
	nis_attr *attrs, *p;
	static struct entryattrs eattrs;

	p = attrs = calloc(1, sizeof(nis_attr));

	for (i = 0; i < table->ta_cols.ta_cols_len; i++) {
		if (!(table->ta_cols.ta_cols_val[i].tc_flags & TA_SEARCHABLE))
			continue;
		if (!ent->en_cols.en_cols_val[i].ec_value.ec_value_len)
			continue;
		p->zattr_ndx = table->ta_cols.ta_cols_val[i].tc_name;
		p->zattr_val.zattr_val_len =
			ent->en_cols.en_cols_val[i].ec_value.ec_value_len;
		p->zattr_val.zattr_val_val =
			ent->en_cols.en_cols_val[i].ec_value.ec_value_val;
		if (i < table->ta_cols.ta_cols_len) {
			attrs = (nis_attr *)realloc(attrs,
						  (cnt+2) * sizeof(nis_attr));
			p = attrs + (cnt + 1);
		}
		cnt++;
	}

	eattrs.cnt = cnt;
	if (cnt)
		eattrs.attrs = attrs;
	else {
		free(attrs);
		eattrs.attrs = NULL;
	}

	return(&eattrs);
}

char *nis_encode_directory_obj(obj, s)
	directory_obj *obj;
	uint32_t *s;
{
	uint32_t size = sizeof(directory_obj);
	char *addr = NULL;
	XDR xdrs;

	size = xdr_sizeof(xdr_directory_obj, obj);

	if ((addr = calloc(1, size)) == NULL)
			return(NULL);
	xdrmem_create(&xdrs, addr, size, XDR_ENCODE);
	if (!xdr_directory_obj(&xdrs, obj)) {
		xdr_destroy(&xdrs);
		free(addr);
		return(NULL);
	}

	xdr_destroy(&xdrs);
	*s = size;

	/* Note: must be free()'ed later! */
	return(addr);
}

char *nis_encode_nis_obj(obj, s)
	nis_object *obj;
	uint32_t *s;
{
	uint32_t size = sizeof(nis_object);
	char *addr = NULL;
	XDR xdrs;

	size = xdr_sizeof(xdr_nis_object, obj);
	if ((addr = calloc(1, size)) == NULL)
		return(NULL);
	xdrmem_create(&xdrs, addr, size, XDR_ENCODE);
	if (!xdr_nis_object(&xdrs, obj)) {
		xdr_destroy(&xdrs);
		free(addr);
		return(NULL);
	}

	xdr_destroy(&xdrs);
	*s = size;

	/* Note: must be free()'ed later! */
	return(addr);
}

/*
 * Decodes an XDR'ed buffer into nis_object form.
 */
nis_object *nis_decode_nis_obj(addr, size)
	char			*addr;
	uint32_t		size;
{
	XDR xdrs;
	nis_object *obj;

	if ((obj = calloc(1, sizeof(nis_object))) == NULL)
		return(NULL);
	xdrmem_create(&xdrs, addr, size, XDR_DECODE);
	if (!xdr_nis_object(&xdrs, obj)) {
		xdr_destroy(&xdrs);
		free(obj);
		return(NULL);
	}
	xdr_destroy(&xdrs);

	/* Note: must be free()'ed later! */
	return(obj);
}

bool_t __is_local_dirname(dirname)
	nis_name		dirname;
{
	return(nis_is_in_srvlist(dirname));
}

bool_t __is_local_dir(dir)
	directory_obj		*dir;
{
	uint32_t		i;
	nis_name		l = nis_local_host();

	if (dir == NULL)
		return(FALSE);
	if (!dir->do_servers.do_servers_len)
		return(FALSE);

	for (i = 0; i < dir->do_servers.do_servers_len; i++) {
		if (!strcasecmp(dir->do_servers.do_servers_val[i].name, l))
			return(TRUE);
	}

	return(FALSE);
}

bool_t __i_am_replica(dir)
	directory_obj		*dir;
{
	uint32_t		i;
	nis_name		l = nis_local_host();

	if (dir == NULL)
		return(FALSE);
	if (!dir->do_servers.do_servers_len)
		return(FALSE);

	if (dir->do_servers.do_servers_len == 1)
		return(FALSE);

	for (i = 1; i < dir->do_servers.do_servers_len; i++) {
		if (!strcasecmp(dir->do_servers.do_servers_val[i].name, l))
			return(TRUE);
	}

	return(FALSE);
}

bool_t __i_am_master(dir)
	directory_obj		*dir;
{
	nis_name		l = nis_local_host();

	if (dir == NULL)
		return(FALSE);
	if (!dir->do_servers.do_servers_len)
		return(FALSE);

	if (nis_dir_cmp(dir->do_servers.do_servers_val[0].name, l) == SAME_NAME)
		return(TRUE);
	return(FALSE);
}

nis_error __nis_check_master(name)
	nis_name		name;
{
	directory_obj			*d;
	nis_error			derr;

	derr = __nis_dcache_lookup(nis_domain_of(name), CHECK_MASTER, 0, &d);

	return(derr);
}

/*
 * This routine must be reentrant. 'buf' is assumed to point to a
 * buffer of NIS_MAXNAMELEN bytes.
 */
nis_name __nis_table(name, buf)
	nis_name		name;
	char			*buf;
{
	nis_name		ptr;

	ptr = nis_name_of(name);

	if (ptr == NULL) {
		if (nis_dir_cmp(name, rootname) == SAME_NAME)
			return("root_dir");
		else
			return(NULL);
	}

	strncpy(buf, ptr, NIS_MAXNAMELEN);

	return(buf);
}

/*
 * If we are a root replica server, we need to keep our root.object
 * up to date. However, we can't just ask the master server for "root.object";
 * we need to request a copy of the root directory object. But to do that,
 * we need to know the name of the root directory. How do we find it?
 * Well, there isn't really an easy way; all we can do is climb up the
 * namespace looking for directories until we 'bang our heads against the
 * ceiling,' i.e. we get an NIS_NOTFOUND error.
 *
 * We start searching using our default directory. This function should
 * be called as few times as possible since it can waste many cycles.
 */

nis_name __nis_find_root()
{
	nis_name		cur = nis_local_directory();
	int			done = 0;
	nis_object		*o;
	nis_error		cerr;

	while(!done) {
		cerr = __nis_ocache_lookup(nis_domain_of(cur), 0, 0, &o);
		switch(cerr) {
			case NIS_NOTFOUND:
			case NIS_NOSUCHNAME:
			case NIS_FOREIGNNS:
			case NIS_NOT_ME:
			case NIS_NAMEUNREACHABLE:
			case NIS_NOMEMORY:
				done++;
				break;
			case NIS_SUCCESS:
			case NIS_S_SUCCESS:
				cur = nis_domain_of(cur);
				break;
			default:
				if (verbose)
					nis_err("__nis_find_root() "
						"returns NULL");
				return(NULL);
		}
	}

	if (verbose)
		nis_err("__nis_find_root() returns %s", cur);
	return(strdup(cur));
}

/*
 * If we are forced to do an nis_dump() to completely re-sync a directory,
 * then we need to empty it out first. This means destroying all tables
 * within the directory and then all objects within the directory itself.
 * The exception is directory objects: we can't delete directory objects
 * since that would leave orphaned subdirectories around.
 */

nis_error nis_empty_directory(dirname)
	nis_name		dirname;
{
	nis_error		err;
	char			nbuf[NIS_MAXNAMELEN];
	nis_object		*obj;
	db_status		db_stat;
	struct db_objs		dres;

	if (verbose)
		nis_err("emptying %s", dirname);

	__nis_flush_one_dir(dirname);

	/*
	 * Start enumerating the directory contents.
	 */

	db_stat = db_first(dirname, NULL, 0, &dres);

	if (db_stat != DB_SUCCESS)
		return(db_to_nis_err(db_stat));

	while (1) {
		obj = NIS_RES_OBJECT(&dres);
		snprintf(nbuf, sizeof(nbuf), "%s.%s", obj->zo_name,
							obj->zo_domain);
		switch(__type_of(obj)) {
		case TABLE_OBJ:
			db_rm(nbuf);
			__nis_flush_one_dir(nbuf);
			break;
		case GROUP_OBJ:
		case LINK_OBJ:
		case PRIVATE_OBJ:
		case DIRECTORY_OBJ:
		case ENTRY_OBJ:
			__nis_flush_one_dir(nbuf);
			/* Do nothing. */
			break;
		default:
			nis_err("empty_dir: trying to delete "
				     "unknown obj type");
			break;
		}
		nis_destroy_object(obj);
		db_stat = db_next(dirname, &dres.cookie, &dres);
		if (db_stat != DB_SUCCESS) {
			err = db_to_nis_err(db_stat);
			break;
		}
	}

	/* NIS_NOTFOUND means we hit the end of the directory */
	if (err == NIS_NOTFOUND)
		return(NIS_SUCCESS);
	return(err);
}

/*
 * Convert time_t to string _without_ embedded newline.
 * Bah.
 */
char *nis_ctime(t)
	time_t			t;
{
	static char		buf[64];

	strcpy(buf, ctime(&t));
	*strchr(buf, '\n')  = '\0';

	return(buf);
}

nis_error nis_update_cold()
{
	directory_obj		*d;
	nis_error		cerr;

	if (verbose)
		nis_err("updating coldstart file");

	cerr = __nis_dcache_lookup(nis_local_directory(),
					LOOKUP_ONLY,
					MASTER_ONLY|NO_CACHE, &d);

	if (cerr != NIS_SUCCESS)
		return(cerr);

	nis_put_cold_start(d);

	return(NIS_SUCCESS);
}

void nis_free_dobjs(dres)
	struct db_objs		*dres;
{
	nis_result		r;

	bzero((char *)&r, sizeof(r));
	NIS_RES_OBJECT(&r) = NIS_RES_OBJECT(dres);
	NIS_RES_NUMOBJ(&r) = NIS_RES_NUMOBJ(dres);
	xdr_free(xdr_nis_result, (char *)&r);
	NIS_RES_NUMOBJ(dres) = 0;
	NIS_RES_OBJECT(dres) = NULL;

	if (dres->cookie.n_len) {
		free(dres->cookie.n_bytes);
		dres->cookie.n_len = 0;
		dres->cookie.n_bytes = NULL;
	}

	return;
}
