/*
 * Copyright (c) 1996, 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: nis_sync.c,v 1.1 1998/07/08 19:23:20 wpaul Exp $
 */

#include <sys/types.h>
#include <sys/cdefs.h>
#include <string.h>
#include <stdlib.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/nislib.h>
#include "nis_extern.h"

#ifndef lint
static const char rcsid[] =
	 "@(#) $Id: nis_sync.c,v 1.1 1998/07/08 19:23:20 wpaul Exp $";
#endif


/*
 * Get latest update time of a directory from a remote
 * server.
 */
nis_error nis_cptime(dir, srv, utime)
	nis_name		dir;
	nis_server		*srv;
	uint32_t		*utime;
{
	struct _nis_binding	*bptr = NULL;
	uint32_t		*cp_res;
	struct rpc_err		rerr;
	nis_error		r;

	r = __nis_create_binding(NULL, srv,
				USE_DGRAM|NO_AUTHINFO, &bptr);

	if (r != NIS_SUCCESS)
		return(r);

	if (__nis_connect_server(bptr) != NIS_SUCCESS) {
		__nis_destroy_binding(bptr);
		return(NIS_FAIL);
	}

	cp_res = nis_cptime_3(&dir, bptr->nb_clnt);

	if (cp_res == NULL) {
		clnt_geterr(bptr->nb_clnt, &rerr);
		nis_err("nis_cptime failed: %s",
				clnt_sperrno(rerr.re_status));
	}

	__nis_destroy_binding(bptr);

	if (cp_res == NULL)
		return(NIS_RPCERROR);

	*utime = *cp_res;

	return(NIS_SUCCESS);
}

/*
 * Ping a remote server so that it knows to synchronize
 * with its master.
 */
nis_error nis_ping_srv(dir, srv, utime)
	nis_name		dir;
	nis_server		*srv;
	uint32_t		utime;
{
	ping_args		args;
	struct timeval		tv = { 0, 0 };
	struct _nis_binding	*bptr = NULL;
	nis_error		r;

	args.dir = dir;
	args.stamp = utime;

	r = __nis_create_binding(NULL, srv,
				USE_DGRAM|NO_AUTHINFO, &bptr);

	if (r != NIS_SUCCESS)
		return(r);

	if (__nis_connect_server(bptr) != NIS_SUCCESS) {
		__nis_destroy_binding(bptr);
		return(NIS_FAIL);
	}

	clnt_control(bptr->nb_clnt, CLSET_TIMEOUT, &tv);
	nis_ping_3(&args, bptr->nb_clnt);

	__nis_destroy_binding(bptr);

	return(NIS_SUCCESS);
}

/*
 * Find out if we need to synchronize ourselves with
 * our master server.
 */
static void nis_sync_with_master(d)
	directory_obj		*d;
{
	nis_server		*s;
	uint32_t		my_utime;
	uint32_t		master_utime;
	ping_args		args;

	s = &d->do_servers.do_servers_val[0];

	my_utime = nis_log_find_update(d->do_name, 0, NULL);

	if (nis_cptime(d->do_name, s, &master_utime) != NIS_SUCCESS) {
		nis_err("sync warning: unable to obtain update "
			     "time for directory %s from master %s -- "
			     "server may be down", d->do_name, s->name);
		return;
	}

	if (master_utime != my_utime) {
		args.dir = d->do_name;
		args.stamp = master_utime;
		if (nis_replica_add_update(&args) != NIS_SUCCESS)
			nis_err("sync warning: replica unable to "
				     "schedule update of %s", d->do_name);
	}

	return;
}

/*
 * Find out of our replicas are up to date and ping
 * them if they aren't.
 */
static void nis_sync_replicas(d)
	directory_obj		*d;
{
	int			i;
	nis_server		*s;
	uint32_t		my_utime;
	uint32_t		replica_utime;

	/*
	 * If only one server (us) then there's nothing
	 * to do. Just return.
	 */
	if (d->do_servers.do_servers_len == 1)
		return;

	my_utime = nis_log_find_update(d->do_name, 0, NULL);

	for (i = 1; i < d->do_servers.do_servers_len; i++) {
		s = &d->do_servers.do_servers_val[i];
		if (nis_cptime(d->do_name, s, &replica_utime) != NIS_SUCCESS) {
			nis_err("sync warning: unable to obtain update "
				     "time for directory %s from replica %s -- "
				     "server may be down", d->do_name, s->name);
			continue;
		}
		if (replica_utime != my_utime) {
			nis_err("replica time is: %s our time "
				     "is %s", nis_ctime(replica_utime),
					      nis_ctime(my_utime));
			nis_master_add_update(d->do_name);
		}
	}

	return;
}

/*
 * Check for synchronization between masters and replicas. Go through
 * list of directories we serve. If we are a master, verify all our
 * replicas are alive and in up to date. If we are a replica, verify
 * that master is alive and that we are in sync with it.
 * 
 * Careful! The directory returned by __nis_dcache_lookup() might be
 * invalidated at some point as part of the master/replica update procedures.
 * We have to make a local copy to avoid having stale pointers.
 */
void nis_sync()
{
	nis_name		dir = NULL;
	directory_obj		*d, *local;
	nis_error		r;

	while((dir = nis_enum_srvlist(dir)) != NULL) {
		r = __nis_dcache_lookup(dir, LOOKUP_ONLY, MASTER_ONLY, &d);
		if (r != NIS_SUCCESS) {
			nis_err("warning: failed to locate object "
				     "for directory %s which we serve", dir);
			continue;
		}
		if (__i_am_master(d) == TRUE) {
			local = __nis_copy_dir(d);
			nis_sync_replicas(local);
			xdr_free(xdr_directory_obj, (char *)local);
			free(local);
			continue;
		}
		if (__i_am_replica(d) == TRUE) {
			local = __nis_copy_dir(d);
			nis_sync_with_master(d);
			xdr_free(xdr_directory_obj, (char *)local);
			free(local);
			continue;
		}
	}

	return;
}
